/*!
  @file     Converter_PageNoIterator.hpp
  @author   Henrik
  @ingroup  Converter
  @brief    
*/

/* ========== licence begin  GPL
  Copyright (c) 2000-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
*/


#ifndef CONVERTER_PAGENOITERATOR_HPP
#define CONVERTER_PAGENOITERATOR_HPP

#include "Converter/Converter_TempMapControl.hpp"
#include "Converter/Converter_PermMapControl.hpp"
#include "Converter/Converter_PermStaticMapControl.hpp"

class Converter_PageNoIterator
{

public:


    Converter_PageNoIterator(IOMan_DeviceNo volNo,
                                Converter_TempMapControl *tempMapControl,
                                Converter_PermMapControl *permMapControl,
                                Converter_PermStaticMapControl *permStaticMapControl,
                                SAPDB_Bool iterEnd = SAPDB_FALSE)
    :
    m_VolNo(volNo),
    m_PermMapControl(permMapControl),
    m_PermStaticMapControl(permStaticMapControl),
    m_MapType(Map_Temp),
    m_IntraMapIter(tempMapControl->BeginBlockIterator(volNo)),
    m_IntraMapEndIter(tempMapControl->EndBlockIterator(volNo))
    {
        if (iterEnd)
        {
            m_MapType = Map_INVALID;
        }
        else
        {
            while ((m_IntraMapIter == m_IntraMapEndIter) && (m_MapType != Map_INVALID))
            {
                ++(*this);
            }
        }
    }

    Converter_PageNoIterator()
    {}

    /*!
      @brief Dereferences the iterator
      @return (IOMan_Blockaddress) current block
    */

    void GetPageNo(Data_PageNo &pageNo,
                   Data_PageRecoveryMode &recMode) const
    {
        pageNo = *m_IntraMapIter;
        
        switch(m_MapType)
        {
         case Map_Temp:
            recMode = Data_PageRecoveryMode( Data_Dynamic, Data_NotRecoverable );
            break;
         case Map_Perm:
            recMode = Data_PageRecoveryMode( Data_Dynamic, Data_Recoverable );
            break;
         case Map_PermStatic:
            recMode = Data_PageRecoveryMode( Data_Static, Data_Recoverable );
            break;
         default:
            SAPDBERR_ASSERT_STATE( SAPDB_FALSE );
           ;
        }
    }

    /*!
      @brief Prefix increment; advances iterator
      @return (Converter_BlockAddressIterator&amp;) this instance
    */
    Converter_PageNoIterator& operator++()
    {
        ++m_IntraMapIter;
        while ((m_MapType != Map_INVALID) && (m_IntraMapIter == m_IntraMapEndIter) )
        {
            // next map control
            if (m_MapType == Map_Temp)
            {
                m_MapType = Map_Perm;
            } else if (m_MapType == Map_Perm)
            {
                m_MapType = Map_PermStatic;
            }
            else
            {
                m_MapType = Map_INVALID;
            }

            switch(m_MapType)
            {
             case Map_Perm:
                m_IntraMapIter = m_PermMapControl->BeginBlockIterator(m_VolNo);
                m_IntraMapEndIter = m_PermMapControl->EndBlockIterator(m_VolNo);
                break;
             case Map_PermStatic:
                m_IntraMapIter = m_PermStaticMapControl->BeginBlockIterator(m_VolNo);
                m_IntraMapEndIter = m_PermStaticMapControl->EndBlockIterator(m_VolNo);
                break;
            }
        }

        return *this;
    }

    SAPDB_Bool operator==(const Converter_PageNoIterator &comp) const
    {
        return ((m_MapType == Map_INVALID) && (m_MapType == comp.m_MapType)) || (
            (m_MapType == comp.m_MapType) && (m_IntraMapIter == comp.m_IntraMapIter));
    }

    SAPDB_Bool operator!=(const Converter_PageNoIterator &comp) const
    {
        return ! ((*this)==comp);

    }


    void operator=(Converter_PageNoIterator const &source)
    {
        m_MapType              = source.m_MapType;
        m_VolNo                = source.m_VolNo;
        m_IntraMapIter         = source.m_IntraMapIter;
        m_IntraMapEndIter      = source.m_IntraMapEndIter;
        m_PermMapControl       = source.m_PermMapControl;
        m_PermStaticMapControl = source.m_PermStaticMapControl;
    }


private:

    IOMan_DeviceNo m_VolNo;

    typedef enum {
       Map_Temp,
       Map_Perm,
       Map_PermStatic,
       Map_INVALID}
    MapType;

    Converter_MapPageNoIterator m_IntraMapIter;
    Converter_MapPageNoIterator m_IntraMapEndIter;

    Converter_PermMapControl *m_PermMapControl;
    Converter_PermStaticMapControl *m_PermStaticMapControl;

    MapType m_MapType;

};

#endif // CONVERTER_PAGENOITERATOR_HPP
