/*!
  @file           SQL_Connection.cpp
  @author         ThomasA
  @brief          Provides a 'connection' for SQLDBC calls.

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


\endif
*/

#include "geo00.h"
#include "ggg01.h"
#include "Oms/OMS_DbpError.hpp"
#include "Oms/OMS_Defines.h"
#include "Oms/OMS_Globals.hpp"
#include "CppSQL/SQL_SQLDBCConnection.hpp"
#include "CppSQL/SQL_SQLDBCRuntime.hpp"
#include "liveCache/LVC_IliveCacheSink.hpp"
#include "liveCache/LVC_KernelInterface.hpp"
#include "SAPDB/Interfaces/SQLDBC/SQLDBC.h"

using namespace SQLDBC;

/*!
   @brief implementation class
 */
class SQL_ConnectionImplementation
{
public :
    /*!
       @brief constructor, create the runtime object used by SQLDBC
     */
    SQL_ConnectionImplementation(IliveCacheSink& kernelSink, SAPDBMem_IRawAllocator& allocator, int packetSize)
        : m_allocator(allocator)
        , m_runTime  (kernelSink, allocator, packetSize)
        , m_environment(&m_runTime)
        , m_connection(0)
    {
    }
    /*!
       @brief destructor, releases connection
     */
    ~SQL_ConnectionImplementation();
    /*!
       @brief establishes the connection
     */
    SQLDBC_Connection* Connect();
    /*!
       @brief returns the allocator used
     */
    SAPDBMem_IRawAllocator& GetAllocator()
    {
        return m_allocator;
    }
private :
    SAPDBMem_IRawAllocator& m_allocator;         //!< reference to the allocator given to an used by SQLDBC
    SQL_SQLDBCRuntime       m_runTime;           //!< instance implementing the run time interface used by SQLDBC
    SQLDBC_Environment      m_environment;       //!< SQLDBC environment
    SQLDBC_Connection*      m_connection;        //!< connection pointer
};

//-----------------------------------------------------------------------------

SQLDBC_Connection* SQL_ConnectionImplementation::Connect()
{
    SQLDBC_Connection* m_connection = m_environment.createConnection(m_allocator); 
    SQLDBC_ConnectProperties properties;
    //properties.setProperty ("APPLICATION", "KNL");
    properties.setProperty ("STATEMENTCACHESIZE", "0");
    //properties.setProperty ("SQLMODE", "INTERNAL");
    if (SQLDBC_NOT_OK == m_connection->connect(properties))
    {
        return 0;
    }
    else
    {
        return m_connection;;
    }
}

//-----------------------------------------------------------------------------

SQL_ConnectionImplementation::~SQL_ConnectionImplementation()
{
    if (m_connection)
    {
       m_connection->disconnect();                      
       m_environment.releaseConnection(m_connection);
    }
}

//-----------------------------------------------------------------------------

SQL_SQLDBCConnection::SQL_SQLDBCConnection(int packetSize)
: m_connection(0)
, m_implementation(0)
{
    IliveCacheSink* pKernelSink = OMS_Globals::GetKernelInterface()->GetSinkPtr();
    SAPDBMem_IRawAllocator* pAllocator; 
    pKernelSink->GetSessionAllocator(pAllocator);
    m_implementation = new(*pAllocator) SQL_ConnectionImplementation(*pKernelSink, *pAllocator, packetSize);
    m_connection     = m_implementation->Connect();
    if (0 == m_connection)
    {
        this->ThrowInvalidConnection();
    }
}

//-----------------------------------------------------------------------------

SQL_SQLDBCConnection::~SQL_SQLDBCConnection()
{
    destroy (m_implementation, m_implementation->GetAllocator());
}

//-----------------------------------------------------------------------------

void SQL_SQLDBCConnection::ThrowInvalidConnection()
{
    throw DbpError (e_no_sqldbc_connection, "SQLDBC connection not established");
}


