/*!
  @file           SQL_Connection.hpp
  @author         ThomasA
  @brief          Provides a 'connection' for SQLDBC calls.

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


\endif
*/

#ifndef SQL_DBC_HPP
#define SQL_DBC_HPP

#include "SAPDB/Interfaces/SQLDBC/SQLDBC.h"

class SQL_SQLDBCRuntime;

class SQL_ConnectionImplementation; //!< forward declaration

namespace SQLDBC {
    class SQLDBC_Connection;
    class SQLDBC_Statement;
    class SQLDBC_PreparedStatement;
}

using namespace SQLDBC; 

/*!
   @brief provides a SQLDBC connection inside a cpp db-procedure embedded in the liboms.
          This connection can then be used to run methods of the SQLDBC interface.   
 */
class AFX_EXT_CLASS SQL_SQLDBCConnection
{
public :
    /*!
       @brief constructor, establishes the connection.
       @param packetSize defines the size of the packet for the communication with the kernel
     */
    SQL_SQLDBCConnection(int packetSize = 4 * 4096);
    /*!
       @brief destructor, closes the connection
     */
    virtual ~SQL_SQLDBCConnection();

    /*!
     * @brief Creates an SQLDBC_Statement object for sending SQL statements
     * to the database.
     
     * SQL statements without parameters are normally executed using 
     * SQLDBC_Statement objects.
     * Applications executing SQL statements several times or 
     * binding parameters for input and/or output must
     * use SQLDBC_PreparedStatement objects.
     * @return The SQLDBC_Statement object created by this method.
     * @see SQLDBC_Statement
     */
    inline SQLDBC::SQLDBC_Statement *createStatement();
  
     /*!
      * @brief Creates an SQLDBC_PreparedStatement object for sending SQL statements
      * to the database.
      
      * An SQLDBC_PreparedStatement object is 'prepared' in the database server and can use binding variables for input/output parameters. 
      * @return The SQLDBC_PreparedStatement object created by this method.
      * @see SQLDBC_PreparedStatement
      */
     inline SQLDBC::SQLDBC_PreparedStatement *createPreparedStatement();
    

     /*!
      * @brief Releases an SQLDBC_Statement.
      * @param stmt A reference to an SQLDBC_Statement object to be released. 
      */
     inline void releaseStatement(SQLDBC::SQLDBC_Statement *stmt);
  
     /*!
      * @brief Releases an SQLDBC_PreparedStatement.
      * @param stmt A reference to an SQLDBC_PreparedStatement object to be released. 
      */
     inline void releaseStatement(SQLDBC::SQLDBC_PreparedStatement *stmt);

     /*!
      * @brief Returns the kernel version.
      *
      * It will be returned in the following 
      * format:  
      *   \<major_release\>\<minor_releaser\>\<correction_level\>
      *   - mayor_release (1 digit)
      *   - minor_release (2 digit)
      *   - correction_level (2 digits)
      * @return The kernel version, which is a computed \<major_release\> * 10000 +
      * \<minor_release\> * 100 + \<correction_level\>.  Example: For version 7.4.04 version number 70404 is returned.
      */
     inline SQLDBC_Int4 getKernelVersion();

     /*!
      * @brief Returns whether the database is an unicode database or
      * not.
      * @return #SQLDBC_TRUE if the database is an unicode database.
      */
     inline SQLDBC_Bool isUnicodeDatabase();

private :
    void ThrowInvalidConnection();                  //!< throws an error if the connection could not be established
    SQLDBC::SQLDBC_Connection*    m_connection;     //!< SQLDBC connection
    SQL_ConnectionImplementation* m_implementation; //!< pointer to the class instance implementing SQL_SQLDBCConnection  
};

//-----------------------------------------------------------------------------

inline SQLDBC::SQLDBC_Statement *SQL_SQLDBCConnection::createStatement()
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->createStatement(); 
};

inline SQLDBC::SQLDBC_PreparedStatement *SQL_SQLDBCConnection::createPreparedStatement()
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->createPreparedStatement(); 
};

inline void SQL_SQLDBCConnection::releaseStatement(SQLDBC::SQLDBC_Statement *stmt)
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->releaseStatement(stmt); 
};
  
inline void SQL_SQLDBCConnection::releaseStatement(SQLDBC::SQLDBC_PreparedStatement *stmt)
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->releaseStatement(stmt); 
};

inline SQLDBC_Int4 SQL_SQLDBCConnection::getKernelVersion()
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->getKernelVersion(); 
};

inline SQLDBC_Bool SQL_SQLDBCConnection::isUnicodeDatabase()
{
  if (0 == m_connection) {
    this->ThrowInvalidConnection();
  }
  return m_connection->isUnicodeDatabase(); 
};

#endif
