/*!
  \file    DBMSrvCmd_SilentDispatcherUsingCommand.cpp
  \author  MarcW
  \ingroup DBMServer commands
  \brief   definition of classes for handling DBMServer
           dealing with the autoextender

    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/

#include "hcn42.h"
#include "hcn50.h"
#include "hcn51.h"

#include "DBM/Cli/Eventing/Dispatcher/DBMCliEvtDisp_Properties.hpp"
#include "DBM/Srv/Commands/Auto/DBMSrvCmdAuto_AutoExtend.hpp"
#include "DBM/Srv/Commands/Auto/DBMSrvCmdAuto_AutoUpdateStatistics.hpp"
#include "DBM/Srv/Commands/DBMSrvCmd_EventAdministrationCommand.hpp"
#include "DBM/Srv/Commands/DBMSrvCmd_SilentDispatcherUsingCommand.hpp"
#include "DBM/Srv/Parameters/DBMSrvPar_Iterator.hpp"
#include "DBM/Srv/Process/DBMSrvProc_ServerProcess.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_InternalResource.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_LineAccessorEventingSilent.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_LineAccessorResourceLock.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_LineIteratorEventing.hpp"

#include "RunTime/System/RTESys_ProcessState.hpp"

DBMSrvCmd_SilentDispatcherUsingCommand::DBMSrvCmd_SilentDispatcherUsingCommand(
        CommandKey commandKey,
        const char* longHelp )
        : DBMSrvCmd_ShmUsingCommand( commandKey, false, longHelp) {

    cn42GetFileName(cn00DBMServerData::vcontrol()->dbname, FGET_SEVTDISPCFG_CN42, m_Configfile);
}

tcn00_Error DBMSrvCmd_SilentDispatcherUsingCommand::runUsingSharedMemory(
    VControlDataT* vcontrol,
    CommandT* command,
    DBMSrv_Reply& theReply) {

    tcn00_Error rc(OK_CN00);
    theReply.startWithOK();

    // lock dispatcher communication
    DBMSrvShM_LineAccessorResourceLock communicationLock(
        DBMSrvShM_InternalResource::SilentEventDispatcherCommunication);

    if( !communicationLock.lock() ) {
        // error someone else is communicating...
        return theReply.startWithMessageList(
            DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SHMRESLOCK,
            Msg_Arg("DBMSRV_COMP", m_Configfile.asCharp() )));
    }

    DBMSrvMsg_Error neverSeen;
    m_SilentDispatcherProps.setFileName(m_Configfile.asCharp());
    m_SilentDispatcherProps.load(neverSeen);

    // let derived class do the work
    rc = runLockProtected( vcontrol, command, theReply);

    // release resource lock
    communicationLock.unlock();
    return rc;
}

tcn00_Error DBMSrvCmd_SilentDispatcherUsingCommand::runLockProtected(
    VControlDataT* vcontrol,
    CommandT* command,
    DBMSrv_Reply& theReply) {

    DBMSrvMsg_Error errOut;
    theReply.startWithOK();
  
    if( command->oArguments.getCount() != 1 )
        // return silently
        return OK_CN00;

    if( 0 == strcmp(command->oArguments.getValue(1).CharPtr(), Start) ) {
        // we're called because kernel will leave offline state
        if( !isDispatcherUp() ) {
            bool mustStart(migrateDispatcherConfiguration(vcontrol, theReply, errOut));
            if( mustStart && errOut.IsEmpty() )
                startDispatcher(errOut);
        }
    }
    else if( 0 == strcmp(command->oArguments.getValue(1).CharPtr(), StopAndDelete) ) {
        // we're called from db_drop command
        if( stopDispatcher(errOut) ) {
            // the dispatcher is stopped successfully
            // we can delete the configuration file
            deleteConfigurationFile(errOut);
        }
    }
    else if( 0 == strcmp(command->oArguments.getValue(1).CharPtr(), StopAll) ) {
        // kernel went offline
        stopDispatcher(errOut);
    }

    if( errOut.IsEmpty() )
        return OK_CN00;
    else
        return theReply.startWithMessageList(errOut);
}

bool DBMSrvCmd_SilentDispatcherUsingCommand::migrateDispatcherConfiguration(
        VControlDataT* vcontrol,
        DBMSrv_Reply& theReply,
        DBMSrvMsg_Error& errOut) {

    // check if event dispatcher must be started
    cn51DBMConfigValue  cfgRunDispatcher (vcontrol->dbname, CFG_RUNDISPATCHERFLAG_CN51);

    bool runEvtDisp(false);
    runEvtDisp = stricmp(((tsp00_Namec)cfgRunDispatcher).asCharp(), "yes") == 0;
    
    // backward compatibility
    cn51DBMConfigValue  cfgAutoExt (vcontrol->dbname, CFG_AUTOEXTFLAG_CN51);
    if( stricmp(((tsp00_Namec)cfgAutoExt).asCharp(), "yes")== 0 ) {
        tsp00_C512c newval;
        newval.rawAssign("no");
        cfgAutoExt = newval;
        newval.rawAssign("yes");
        cfgRunDispatcher = newval;
        runEvtDisp = true;
    }

    if( !runEvtDisp )
        return false;

    tni34_AVirtualFile newConfigFile(
            m_Configfile.asCharp(),
            tni34_AVirtualFile::FromDisk,
            tni34_AVirtualFile::DoNotOpenNow);

    if( newConfigFile.IsGood() && !newConfigFile.Exists() ) {
        // a new file does not exist yes

        DBMCliEvtDisp_Properties newProps(Tools_DynamicUTF8String(m_Configfile.asCharp()));

        tsp00_Pathc oldConfigfileName;
        cn42GetFileName(cn00DBMServerData::vcontrol()->dbname, FGET_EVTDISPCONF_CN42, oldConfigfileName);
        tni34_AVirtualFile oldConfigFile(
                oldConfigfileName.asCharp(),
                tni34_AVirtualFile::FromDisk,
                tni34_AVirtualFile::DoNotOpenNow);
        Tools_Properties oldProps(Tools_DynamicUTF8String(oldConfigfileName.asCharp()));
        if( oldConfigFile.Exists() )
            oldProps.load(errOut);
        if( oldProps.isEmpty() ) {
            tsp00_C512c no;
            no.rawAssign("no");
            cfgRunDispatcher = no;
            runEvtDisp = false;
        }
        else {
            // auto extend
            Tools_DynamicUTF8String autoExtProp(
                    oldProps.getProperty(Tools_DynamicUTF8String("DB_ABOVE_LIMIT")));
            if( autoExtProp.Length() > 0 ) {
                // transfer entry
                Tools_DynamicUTF8String newExtendEvent;
                Tools_DynamicUTF8String autoExtendCondition;
                Tools_DynamicUTF8String autoExtendHandler;
                getEventCategoryNameFromID( sp31ei_db_filling_above_limit, theReply, newExtendEvent);
                DBMSrvCmdAuto_AutoExtend::buildConfigLine(
                        newExtendEvent,
                        DBMSrvCmdAuto_AutoExtend::m_PercentageDefault,
                        autoExtendCondition,
                        autoExtendHandler);

                newProps.setProperty(autoExtendCondition, autoExtendHandler);
            }
            // update statistics
            Tools_DynamicUTF8String updStatProp(
                    oldProps.getProperty(Tools_DynamicUTF8String("UPDATE_STATISTICS")));
            if( updStatProp.Length() > 0 ) {
                // transfer entry
                Tools_DynamicUTF8String newUpdStatEvent;
                Tools_DynamicUTF8String updStatCondition;
                Tools_DynamicUTF8String updStatHandler;
                getEventCategoryNameFromID( sp31ei_upd_stat_wanted, theReply, newUpdStatEvent);
                DBMSrvCmdAuto_AutoUpdateStatistics::buildConfigLine(
                        newUpdStatEvent,
                        updStatCondition,
                        updStatHandler);

                newProps.setProperty(updStatCondition, updStatHandler);
            }
            
            // add version number
            bool mustStore(false);
            if( !newProps.isEmpty() ) {
                
                newProps.setProperty(
                        Tools_DynamicUTF8String(DBMCliEvtDisp_Properties::FileVersionKey),
                        Tools_DynamicUTF8String("7.6.00.04"));
                mustStore = true;
            }            
            
            // store the transferred lines if necessary
            if( !mustStore || !newProps.store(errOut) ) {
                tsp00_C512c no;
                no.rawAssign("no");
                cfgRunDispatcher = no;
                runEvtDisp = false;
            }
        }        
        
        // delete old file
        oldConfigFile.Remove();
    }

    return runEvtDisp;
}

void DBMSrvCmd_SilentDispatcherUsingCommand::getEventHandler(
    const Tools_DynamicUTF8String& aEventName,
    Tools_DynamicUTF8String& outKey,
    Tools_DynamicUTF8String& outValue) {
    
    outKey.Erase();
    outValue.Erase();

    Tools_List<Tools_Property*>* allLines = m_SilentDispatcherProps.getAllProperties();
    if( allLines != NULL ) {
        Tools_List<Tools_Property*>::iterator it(allLines->begin());
        while( it != allLines->end() ) {
            Tools_DynamicUTF8String key((*it)->getKey());
            if( key.Find(aEventName) != Tools_DynamicUTF8String::NPos ) {
                outKey = key;
                outValue = (*it)->getValue();
                delete *it;
            }
            ++it;
        }
        delete allLines;
    }
}

bool DBMSrvCmd_SilentDispatcherUsingCommand::isDispatcherUp() {
    DBMSrvShM_LineIteratorEventing internalDispatchers(true);
    if( !internalDispatchers.isValid() )
        return false;

    int linesFound(0);
    int dispatcherPid(0);
    int instanceID(0);
    while( internalDispatchers.hasNext() ) {
        ++linesFound;
        dispatcherPid = internalDispatchers.getDispatcherProcessID();
        instanceID = internalDispatchers.getInstanceID();
        internalDispatchers.next();
    }

    if( linesFound != 1 )
        return false;

    RTE_OSPid dispatcherOSPid(dispatcherPid);
    RTESys_ProcessState procState(dispatcherOSPid);
    bool isUp(procState.State() == RTESys_ProcessState::Alive);
    if( !isUp ) {
        // there is an internal dispatcher line, but the dispatcher is not up, so we remove the line
        DBMSrvShM_LineAccessorEventing internalDispatcher(instanceID);
        internalDispatcher.removeLineEventing();
    }
    return isUp;
}

bool DBMSrvCmd_SilentDispatcherUsingCommand::stopDispatcher(DBMSrvMsg_Error& theErrList) {

    if( !isDispatcherUp() )
        return true;

    DBMSrvShM_LineIteratorEventing internalDispatchers(true);
    if( !internalDispatchers.isValid() ) {
        theErrList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODISPATCHERSTOP));
        return false;
    }

    int linesFound(0);
    int instanceID(-1);
    while( internalDispatchers.hasNext() ) {
        ++linesFound;
        instanceID = internalDispatchers.getInstanceID();
        internalDispatchers.next();
    }

    if( linesFound > 1 ) {
        theErrList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODISPATCHERSTOP));
        return false;
    }

    if( linesFound == 0 )
        return true;

    DBMSrvShM_LineAccessorEventing theInternalDispatcher(instanceID);
    theInternalDispatcher.doCancel();

    // now wait until the line is removed...
    int waitCount(300); 
    while( waitCount > 0 && theInternalDispatcher.exists() ) {
        RTE_ISystem::DoSleep(1000);
        --waitCount;
    }

    if( theInternalDispatcher.exists() ) {
        theErrList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODISPATCHERSTOP));
        return false;
    }

    if( m_SilentDispatcherProps.getCount()==1 )
        return deleteConfigurationFile( theErrList );
    else
        return true;
}

bool DBMSrvCmd_SilentDispatcherUsingCommand::deleteConfigurationFile(DBMSrvMsg_Error& errOut) {
    // we delete the configuration file
    if( m_SilentDispatcherProps.deleteStorage(errOut) ) {
        // write dbm configuration
        cn51DBMConfigValue cfgRunDispatcher(
                cn00DBMServerData::vcontrol()->dbname, 
                CFG_RUNDISPATCHERFLAG_CN51);
        tsp00_C512c no;
        no.rawAssign("no");
        cfgRunDispatcher = no;
        return true;
    }
    else
        return false;
}

bool DBMSrvCmd_SilentDispatcherUsingCommand::startDispatcher(DBMSrvMsg_Error& theErrList) {

    if( isDispatcherUp() )
        return true;

    char upwdBuffer[256];
    DBMSrvProc_ArgumentList theArgList;

    // find the dbm super user
    cn50DBMUser superUser(
        cn00DBMServerData::vcontrol()->dbname,
        cn50DBMUser::getColdUser(cn00DBMServerData::vcontrol()->dbname));

    // build command line
    theArgList.AddArgument("start");
    theArgList.AddArgument("-internal");
    theArgList.AddArgument(m_Configfile.asCharp());
    theArgList.AddArgument("-l");
    tsp00_Pathc logFile;
    cn42GetFileName(cn00DBMServerData::vcontrol()->dbname, FGET_EVTDISPPRT_CN42, logFile);
    theArgList.AddArgument(logFile);
    theArgList.AddArgument("-d");
    theArgList.AddArgument(cn00DBMServerData::vcontrol()->dbname.asCharp());
    theArgList.AddArgument("-u");
    sprintf(upwdBuffer, "%s,%s",
            superUser.getUserName().asCharp(),
            superUser.getClearMasterPwd().asCharp());
    theArgList.AddArgument(upwdBuffer);

    DBMSrvProc_ServerProcess dispatcherProc(
        DBMSrvProc_ServerProcessCommandLineConverter::KeyEventDispatcherProgram,
        theArgList);

    if( dispatcherProc.started() ) {
        int count(1000);
        while( count>0 && !isDispatcherUp() ) {
            RTE_ISystem::DoSleep(300);
            --count;
        }

        if( isDispatcherUp() ) {
            // write dbm configuration
            cn51DBMConfigValue cfgRunDispatcher(
                    cn00DBMServerData::vcontrol()->dbname, 
                    CFG_RUNDISPATCHERFLAG_CN51);
            tsp00_C512c yes;
            yes.rawAssign("yes");
            cfgRunDispatcher = yes;
            return true;
        }
        else {
            dispatcherProc.getMessagesInto(theErrList);
            theErrList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODISPATCHERSTART));
            return false;
        }
    }
    else {
        dispatcherProc.getMessagesInto(theErrList);
        theErrList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_EXECUTE));
        return false;
    }
}

tcn00_Error DBMSrvCmd_SilentDispatcherUsingCommand::getEventCategoryNameFromID(
        const tsp31_event_ident_Enum eventID,
        DBMSrv_Reply& theReply,
        Tools_DynamicUTF8String& eventCategoryName ) {

    eventCategoryName = DBMSrvCmd_EventAdministrationCommand::m_EventCategoryUnknown;
            
    DBMSrvPar_Parameters* kernelParameters =
        DBMSrvPar_Parameters::GetInstance(cn00DBMServerData::vcontrol()->dbname, true, theReply);

    if( kernelParameters == NULL ) {
        return theReply.ReturnCode();
    }

    DBMSrvPar_ClassIterator classEventsIterator(kernelParameters, DBMSrvPar_Parameter::XPClassValueEvents);
    while( *classEventsIterator != NULL ) {
        DBMSrvPar_XPLine* helpLine((*classEventsIterator)->pHelp);
        if( helpLine != NULL && helpLine->Line() != NULL ) {
            tsp31_event_ident_Enum currentID = (tsp31_event_ident_Enum) atoi(helpLine->Line());
            if( currentID == eventID ) {
                eventCategoryName = (*classEventsIterator)->szParameterName.asCharp();
                break;
            }
        }
        ++classEventsIterator;
    }

    delete kernelParameters;
    return OK_CN00;
}

const char* const DBMSrvCmd_SilentDispatcherUsingCommand::StopAndDelete("stopAndDelete");
const char* const DBMSrvCmd_SilentDispatcherUsingCommand::StopAll("stopAll");
const char* const DBMSrvCmd_SilentDispatcherUsingCommand::Start("start");
