/*!
  \file    DBMSrvCmdDb_DbAddVolume.cpp
  \author  MarcW
  \ingroup DBMServer commands
  \brief   definition of class for handling DBMServer
           command db_addvolume

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (C) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

\endif
*/

#include <stdlib.h>

#include "hcn40.h"
#include "hcn50.h" // cn50DBMUser
#include "hcn90.h"

#include "DBM/Srv/Commands/Db/DBMSrvCmdDb_DbAddVolume.hpp"
#include "DBM/Srv/KernelAccess/DBMSrvKnl_Session.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_InternalResource.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_LineAccessorResourceLock.hpp"

DBMSrvCmdDb_DbAddVolume::DBMSrvCmdDb_DbAddVolume()
: DBMSrvCmd_ShmUsingCommand( KeyDbAddVolume, false )
{
    m_longHelp=
    "@command db_addvolume With this command, you define additional volumes "
        "to extend the disk space for the database instance. If only parameter <vol_mode> "
        "is given, location and size are determined as follows: "
        "If DBMServer parameter AutoExtDir (for data volumes) resp. AutoExtDirLog (for log "
        "volumes) contains the name of an existing directory, the new volume is created there. "
        "In case of log mirroring, the mirror volume is created in AutoExtDirLogMirror. "
        "If the DBMServer parameter for the requested volume type is not set, the volume is "
        "created in the rundirectoy of the database instance."
        "The new volume's name is created as follows: The name of the preceding volume of the "
        "same type is searched for the last occurence of it's index in it's name. If found, "
        "the name of the new volume is the name of the old volume, in which the old index is "
        "replaced with the new index. If not found, the new volume's name is the old volume's name "
        "with a dot and the new index appended. This applies to data, log and mirror volumes. "
        "The new volume's size in pages is calculated as follows: "
        "If DBMServer parameter AutoExtSize (for data volumes) resp. AutoExtSizeLog (for log "
        "volumes) is set to a number value, this is taken as the new volume's size. If the "
        "parameter is a percent value (ends with a '%' character), the new volume's size "
        "is calculated as the given percentage of the sum of all existing data resp. "
        "log volumes. If the parameter is not set, the new volume's size is the size of the "
        "preceding volume of the same type."
    "@preconditions You have the server authorization ParamFull or AccessUtility."
    "@syntax db_addvolume <vol_mode> [-fd <filldegree>]|[<vol_name> <vol_type> <vol_size> [<mvol_name> <mvol_type>]]"
    "@param <vol_mode> Mode of volume that is to be added. Possible values are:"
        "@value DATA \"\" Data Volume"
        "@value LOG \"\" Log Volume"
        "@param <filldegree> The actual filling degree is compared to <filldegree>. "
            "If it is bigger, new volumes are added, until the actual filling degree "
            "is smaller that <filldegree>. To be used only together with "
            "parameter <vol_mode> == DATA. No other parameters besides these two "
            "may be set. 1 <= <filldegree> <= 100 must be true. If this parameter is given, "
            "the number of added volumes is replied."
    "@param <vol_name> Name of volume that is to be added. If it's a relative path, the new volume "
        "is created relative to the database instance's rundirectory."
    "@param <vol_type> Type of volume that is to be added."
        "@value F \"File\" File"
        "@value L \"Link\" Link"
        "@value R \"Raw Device\" Raw Device"
    "@param <size> Size of the volume to be added in pages."
    "@param <mvol_name> Name of mirror volume that is to be added. Use only "
        "with <vol_mode> LOG and log mirroring activated. If it's a relative path, the new volume "
        "is created relative to the database instance's rundirectory."
    "@param <mvol_type> Type of mirror volume that is to be added. Use only "
        "with <vol_mode> LOG and log mirroring activated."
        "@value F \"File\" File"
        "@value L \"Link\" Link"
        "@value R \"Raw Device\" Raw Device"
    "@reply OK<NL>"
        "[<num_added>]"; 
}

tcn00_Error DBMSrvCmdDb_DbAddVolume::runUsingSharedMemory(
    VControlDataT* vcontrol,
    CommandT* command,
    DBMSrv_Reply& theReply) {

    int volsAdded(0);
        
    // check number of arguments
    SAPDB_Int argCount(command->oArguments.getCount());
    int fillDegree(0);
    bool useFillDegree(argCount==3);
    bool needMore(true);
    tcn00_Error  retVal(OK_CN00);

    if( argCount != 1 && argCount != 3 && argCount != 4 && argCount != 6 ) {
        retVal =  theReply.startWithError(ERR_PARAM_CN00);
        getUsageText(theReply);
        return retVal;
    }
    
    if( useFillDegree ) {
        char szMode[PARAMETER_MAXLEN_CN90];
        char szSwitch[PARAMETER_MAXLEN_CN90];
        char szFillDegree[PARAMETER_MAXLEN_CN90];

        cn90GetToken(command->args, szMode, 1, PARAMETER_MAXLEN_CN90);
        cn90GetToken(command->args, szSwitch, 2, PARAMETER_MAXLEN_CN90);
        cn90GetToken(command->args, szFillDegree, 3, PARAMETER_MAXLEN_CN90);

        if( stricmp(szMode, XP_DEV_DAT_CN00 ) != 0 ) {
            Msg_List errVolMode(
                   DBMSrv_DBMError(ADDVOLWRONGMODE), 1, szMode);
            retVal = theReply.startWithMessageList(errVolMode);
        } else if( stricmp(szSwitch, m_FillDegreeParameterName) != 0 ) {
            retVal = theReply.startWithError(ERR_PARAM_CN00);
        } else {
            fillDegree=atoi(szFillDegree);
            if( fillDegree <= 0 || fillDegree > 100 ) {
                Msg_List errFillDegree(
                    DBMSrv_DBMError(ADDVOLERRFDEGRANGE), 1, szFillDegree);
                retVal =  theReply.startWithMessageList(errFillDegree);
            }
        }

        if( retVal != OK_CN00 ) {
            getUsageText(theReply);
            return retVal;
        }
    }

    DBMSrvShM_LineAccessorResourceLock addVolumeLock(DBMSrvShM_InternalResource::AddVolume);
    if( !addVolumeLock.lock() ) {
        Msg_List errLockAddVol(
            DBMSrv_DBMError(SHMRESLOCK), 1, addVolumeLock.getResourceName());
        return theReply.startWithMessageList(errLockAddVol);
    }
    do {
        if( useFillDegree ) {
            cn50DBMUser       usrDBM    ( vcontrol->dbname, cn50DBMUser::getUser(vcontrol->dbname, FIRSTDBM_CN50));
            DBMSrvKnl_Session oSession;

            if (!oSession.NormalConnect(vcontrol->dbname.asCharp(),
                                        usrDBM.getUserName().asCharp(),
                                        usrDBM.getClearMasterPwd().asCharp())) {
                retVal = theReply.startWithMessageList(oSession.LastMessage());
            } // end if
            if( retVal != OK_CN00 )
                break;

            DBMSrvKnl_Session::ResultType resType;
            if( oSession.Execute(
                m_sqlStmtGetDataFillingPercentage,
                resType) ) {

                DBMSrv_SQLRecord & oRecord = oSession.GetRecordObject();
                SAPDB_Int4 realFillDegree(oRecord.fieldAsInt(1));

                needMore = (retVal == OK_CN00) && (realFillDegree > fillDegree);
            }
            else {
                // error
                needMore = false;
                retVal = theReply.startWithMessageList(oSession.LastMessage());
                break;
            }
        }

        if( needMore ) {
            CommandT newCommand;
            char newArguments[1024];
            SAPDB_strcpy(newArguments, command->args);
            newCommand.nIndex=CommandT::InvalidIndex;
            newCommand.setArgs(newArguments);
            retVal = cn40DbAddDevice(
                                    vcontrol,
                                    &newCommand, // must do this , because it's changed...
                                    theReply.giveData(),
                                    theReply.giveLength(),
                                    theReply.giveMaximalLength());
            theReply.adaptLength();
            if( retVal == OK_CN00)
                ++volsAdded;
        }
    } while ( retVal == OK_CN00 && useFillDegree && needMore );

    addVolumeLock.unlock();

    if( useFillDegree ) {
        // put number of added volumes into reply
        char volsAddedBuf[128];
        sprintf(volsAddedBuf, "%d", volsAdded);
        theReply.appendLine(volsAddedBuf);
    }

    return retVal;
}

const char* const DBMSrvCmdDb_DbAddVolume::m_FillDegreeParameterName = "-fd";

const char* const DBMSrvCmdDb_DbAddVolume::m_sqlStmtGetDataFillingPercentage =
    "select \"Data            (%)\" from SYSDBM.INFO_STATE";
