/*!
  @file           DBMSrvKnl_Session.cpp
  @author         Bernd Vorsprach - bernd.vorsprach@sap.com
  @brief          DBMServer to Kernel Communication - Implementation

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


\endif
*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include <stdlib.h>
#include <ctype.h>

#include "SAPDB/DBM/Srv/KernelAccess/DBMSrvKnl_Session.hpp"
#include "SAPDB/ToolsCommon/Tools_DynamicUTF8String.hpp"
#include "ToolsCommon/Tools_System.hpp"

#include "heo02.h"
#include "heo03.h"
#include "hsp02.h"

#include "SAPDB/PacketInterface/PIn_Packet.h"

#define CMD_CONNECT_P1  "CONNECT \""
#define CMD_CONNECT_P2  "\" IDENTIFIED by :PW TIMEOUT 0"
#define CMD_CONNECT_P3  " NO TEMP FILE"
#define CMD_RELEASE     "COMMIT WORK RELEASE"

/*-----------------------------------------------------------------------------*/
/*! @brief constructor */
  /*! @ brief constructor */
DBMSrv_SQLField::DBMSrv_SQLField
  ( tsp1_param_info * aInfo,
    DataType          aType )
{
  SAPDB_memcpy(&oInfo, aInfo, sizeof(oInfo));
  oType  = aType;
  iValue = 0;
  sValue = "";
  bNull  = false;
}  // end DBMSrv_SQLField::DBMSrv_SQLField

/*-----------------------------------------------------------------------------*/
/*! @brief constructor */
DBMSrvKnl_Session :: DBMSrvKnl_Session
    ( )
{
  this->Reset();
} // end DBMSrvKnl_Session :: DBMSrvKnl_Session

/*-----------------------------------------------------------------------------*/
/*! @brief destructor */
DBMSrvKnl_Session :: ~DBMSrvKnl_Session
    ( )
{
  Release();
} // end DBMSrvKnl_Session :: ~DBMSrvKnl_Session

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: NormalConnect
    ( const char * szDB,
      const char * szUserIn,
      const char * szPassword,
      bool         bWithNoTempFiles )
{
  this->oMessageList.ClearMessageList();

  // check user and password
  if ((szUserIn == NULL) || (szPassword == NULL) || (strlen(szUserIn) == 0) || (strlen(szPassword) == 0)) {
    this->oMessageList.Overrule(Msg_List(DBMSRV_ERR_MISSDBUSRPWD));
    this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_PARAM));
    return false;
  } // end if

  oState.SetDatabase(szDB);

  this->bNoTempFiles = bWithNoTempFiles;

  if (!bWithNoTempFiles) {
    bool         bLogFull       = false;
    bool         bDataFull      = false;

    oState.IsFull(bLogFull, bDataFull);
    if (!oState.DBMMsg().IsEmpty()) {
      this->oMessageList = oState.DBMMsg();
      return false;
    } else if (bDataFull) {
      // database full error
      this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_DATAFULL);
      return false;
    } // end if
  } // end if
  this->bAdminConnect = (oState.State() != DBMSrvKnl_State::StateOnline);

  if (this->Connect(DBMSrvKnl_Session::taskNormal, szDB )) {
    if(!this->ConnectUser(szUserIn, szPassword, bWithNoTempFiles)) {
      this->Release();
    } else {
      this->nConnected   = DBMSrvKnl_Session::taskNormal;
      szDatabase.rawAssign(szDB);
    } // end if
  } // end if

  return this->oMessageList.IsEmpty();
} // end DBMSrvKnl_Session :: NormalConnect

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: EventConnect
    ( const char * szDB )
{
  this->oMessageList.ClearMessageList();

  if (this->Connect(DBMSrvKnl_Session::taskEvent, szDB )) {
    this->nConnected = DBMSrvKnl_Session::taskEvent;
    szDatabase.rawAssign(szDB);
  } // end if

  return this->oMessageList.IsEmpty();
} // end DBMSrvKnl_Session :: EventConnect

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Release
    ( )
{
  bool bReturn = true;
  DBMSrvMsg_Error oLocalMsg;

  if (this->IsConnected()) {
    if (!this->bAdminConnect) {
      // send release command
      // create request packet
      PIn_RequestPacket  oPacket(this->pPacket, this->nPacketSize, this->nUnicode);
      // initialise request packet
      oPacket.InitVersion(csp_comp_db_manager, NULL);
      // create segement
      PIn_RequestSegment oSegment = oPacket.AddSegment(sp1m_dbs, sp1sm_session_sqlmode);
      oSegment.SetProducer(sp1pr_internal_cmd);
      // create command part
      PIn_Part oCommandPart = oSegment.AddPart(sp1pk_command);
      // fill commadn part
      oCommandPart.AddArgument(CMD_RELEASE, (int) strlen(CMD_RELEASE));
      // close part
      oSegment.ClosePart(oCommandPart);
      // close segment
      oPacket.CloseSegment(oSegment);
      // send request to kernel
      if (!Request(oPacket.Length())) {
        if (bReturn) oLocalMsg = this->LastMessage();
        bReturn = false;
      } else {
        // receive answer from kernel and create answer packet
        PIn_ReplyPacket  oReplyPacket(Receive());
        // check answer packet
        if (!oReplyPacket.IsValid()) {
          if (bReturn) oLocalMsg = this->LastMessage();
          bReturn = false;
        } // end if
      } // end if
    } // end if

    sqlarelease (this->nSessionHandle);

    if (!oLocalMsg.IsEmpty()) this->oMessageList = oLocalMsg;
  } // end if

  this->Reset();
  return bReturn;
} // end DBMSrvKnl_Session :: Release

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: BreakSession
    ( )
{
  this->Cancel();
  sqlarelease (this->nSessionHandle);
  this->Reset();
  return true;
} // end DBMSrvKnl_Session :: BreakSession

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Request
  ( const char * szCmd )
{
  oMessageList.ClearMessageList();

  this->bAllowedInDbFull = AllowedInDbFull(szCmd);

  if (!this->IsConnected()) {
    this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODBSESSION);
    return false;
  } // end if

  // create request packet
  PIn_RequestPacket  oPacket(this->pPacket, this->nPacketSize, this->nUnicode);
  // initialise request packet
  oPacket.InitVersion(csp_comp_db_manager, NULL);
  // create segement
  PIn_RequestSegment oSegment = oPacket.AddSegment(sp1m_dbs, sp1sm_session_sqlmode);
  oSegment.SetProducer(sp1pr_internal_cmd);

  oSegment.GetRawHeader()->sp1c_with_info          () = true;
  oSegment.GetRawHeader()->sp1c_commit_immediately () = this->bAutoCommit;

  // create command part
  PIn_Part oCommandPart = oSegment.AddPart(sp1pk_command);
  // fill command part
  oCommandPart.AddArgument();

  if (this->nUnicode) {
    oCommandPart.AddDataUTF8ToUCS2(szCmd, (int) strlen(szCmd));
  } else {
    oCommandPart.AddData(szCmd, (int) strlen(szCmd));
  } // end if

  // close part
  oSegment.ClosePart(oCommandPart);

  // mass cmd handling in case of fetch
  if (strncmp("FETCH ", szCmd, 6) == 0) {
    oSegment.GetRawHeader()->sp1c_mass_cmd           () = true;
    // create result count part
    PIn_Part oResultCountPart = oSegment.AddPart(sp1pk_resultcount);

    // add number
    unsigned char oFixed[7] = {0x00,0xC5,0x32,0x76,0x70,0x00,0x00};
    oResultCountPart.AddArgument(&oFixed[0], sizeof(oFixed));

    // close part
    oSegment.ClosePart(oResultCountPart);
  } // end if

  // close segment
  oPacket.CloseSegment(oSegment);

  // send request to kernel
  if (!Request(oPacket.Length())) {
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: Request

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: EventRequest
  ( )
{
  oMessageList.ClearMessageList();

  if (!this->IsConnected()) {
    this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODBSESSION);
    return false;
  } // end if

  // create request packet
  PIn_RequestPacket  oPacket(this->pPacket, this->nPacketSize, this->nUnicode);
  // initialise request packet
  oPacket.InitVersion(csp_comp_db_manager, NULL);
  // create segement
  PIn_RequestSegment oSegment = oPacket.AddSegment(sp1m_wait_for_event, sp1sm_session_sqlmode);

//  oSegment.SetProducer(sp1pr_internal_cmd);

//  oSegment.GetRawHeader()->sp1c_with_info          () = true;
//  oSegment.GetRawHeader()->sp1c_commit_immediately () = this->bAutoCommit;

  // create command part
  PIn_Part oCommandPart = oSegment.AddPart(sp1pk_command);

  // close part
  oSegment.ClosePart(oCommandPart);

  // close segment
  oPacket.CloseSegment(oSegment);

  // send request to kernel
  if (!Request(oPacket.Length())) {
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: EventRequest

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session ::  Receive
  ( ResultType & oResultType )
{
  oMessageList.ClearMessageList();

  if (!this->IsConnected()) {
    this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODBSESSION);
    return false;
  } // end if

  oResultType = DBMSrvKnl_Session::resultUnknown;

  // receive answer from kernel and create answer packet
  PIn_ReplyPacket  oPacket(Receive());
  // check answer packet
  if (!oPacket.IsValid()) {
    return false;
  } // end if

  // read data
  PIn_ReplySegment oSegment     = oPacket.GetFirstSegment ();

  if (oSegment.ErrorCode() == 0) {

    // decide on answer the next steps
    if (oSegment.GetRawHeader()->sp1s_no_of_parts() != 0) {
      PIn_Part oData        = oSegment.FindPart(sp1pk_data);
      PIn_Part oShortInfo   = oSegment.FindPart(sp1pk_shortinfo);
      PIn_Part oResultCount = oSegment.FindPart(sp1pk_resultcount);
      PIn_Part oResultTable = oSegment.FindPart(sp1pk_resulttablename);
      PIn_Part oMsgList     = oSegment.FindPart(sp1pk_message_list);

      if (oShortInfo.IsValid() && oResultCount.IsValid() && oResultTable.IsValid()) {
        bool bEnd;
        if (!this->oRecord.getFirstRecord(bEnd)) {
          return false;
        } // end if
        oResultType = DBMSrvKnl_Session::resultSQL;
      } else if (oShortInfo.IsValid() && oData.IsValid()) {
        oResultType = DBMSrvKnl_Session::resultUtility;
      } else if (oMsgList.IsValid()) {
        oResultType = DBMSrvKnl_Session::resultMessage;
      } else if (oResultCount.IsValid()) {
        // result count only (answer for insert or update)
        oResultType = DBMSrvKnl_Session::resultNothing;
      } // end if

    } else {
      oResultType = DBMSrvKnl_Session::resultNothing;
    } // end if

  } else {
    this->ReadErrorText(oSegment);
    oResultType = DBMSrvKnl_Session::resultError;
    if (oSegment.ErrorCode() == 100) {
      DBMSrvMsg_Error tmpMsg = this->oMessageList;
      this->Execute("CLOSE");
      this->oMessageList = tmpMsg;
    } // end if
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: Receive

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session ::  EventReceive
  ( void * const pEventBuffer, 
    size_t nBufferSize )
{
  oMessageList.ClearMessageList();

  if (!this->IsConnected()) {
    this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_NODBSESSION);
    return false;
  } // end if

  // receive answer from kernel and create answer packet
  PIn_ReplyPacket  oPacket(Receive());
  // check answer packet
  if (!oPacket.IsValid()) {
    return false;
  } // end if

  // read data
  PIn_ReplySegment oSegment     = oPacket.GetFirstSegment ();

  if (oSegment.ErrorCode() == 0) {

    // decide on answer the next steps
    if (oSegment.GetRawHeader()->sp1s_no_of_parts() != 0) {
      PIn_Part oData        = oSegment.FindPart(sp1pk_data);
      if (oData.IsValid()) {
        SAPDB_memcpy(pEventBuffer, oData.GetReadData(1), nBufferSize);
      } // end if
    } // end if

  } else {
    this->ReadErrorText(oSegment);
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: Receive

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Execute
  ( const char * szCmd )
{
  ResultType  oResultType;
  return this->Execute(szCmd, oResultType);
} // end DBMSrvKnl_Session :: Execute

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: ExecuteWithoutCommit
  ( const char * szCmd )
{
  bool bResult = false;
  bool bCommit = this->bAutoCommit;
  this->bAutoCommit  = false;
  ResultType  oResultType;
  bResult = this->Execute(szCmd, oResultType);
  this->bAutoCommit = bCommit;
  return bResult;
} // end DBMSrvKnl_Session :: ExecuteWithoutCommit

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Execute
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError)
{
  return Execute(szDB, szUser, szPassword, szCommand, oError, false);
} // end DBMSrvKnl_Session :: Execute

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: ExecuteWithNoTempFiles
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError)
{
  return Execute(szDB, szUser, szPassword, szCommand, oError, true);
} // end DBMSrvKnl_Session :: ExecuteWithNoTempFiles

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Execute
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError,
      bool              bWithNoTempFiles)
{
  DBMSrvKnl_Session oSession;

  if (!oSession.NormalConnect(szDB, szUser, szPassword, bWithNoTempFiles)) {
    oError = oSession.LastMessage();
  } else {
    if (!oSession.Execute(szCommand)) {
      oError = oSession.LastMessage();
    } // end if
    oSession.Release();
  } // end if

  return oError.IsEmpty();
} // end DBMSrvKnl_Session :: Execute


/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: Execute
  ( const char * szCmd,
    ResultType & oResultType )
{
  if (Request(szCmd)) {
    return Receive(oResultType);
  } // end if
  return false;
} // end DBMSrvKnl_Session :: Execute

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
const Msg_List DBMSrvKnl_Session :: KernelMessage
  ( ) const
{
  Msg_List oKernelMsg;

  oKernelMsg.ClearMessageList();

  PIn_ReplyPacket  oPacket(this->pReceivePacket);
  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oPart   = oSegment.FindPart(sp1pk_message_list);
    if (oPart.IsValid()) {
      PIn_MessageListPart oMsgListPart(oPart);
      oKernelMsg = oMsgListPart.MessageList();
    } // end if
  } // end if

  return oKernelMsg;
} // end DBMSrvKnl_Session :: KernelMessage

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: Connect
    ( DBMSrvKnl_Session::TaskType   oTaskTypeIn,
      const char                     * szDB )
{
  this->oMessageList.ClearMessageList();

  // check user and password
  if ((szDB == NULL) || (strlen(szDB) == 0)) {
    this->oMessageList.Overrule(Msg_List(DBMSRV_ERR_MISSDBUSRPWD));
    this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_PARAM));
    return false;
  } // end if

  tsp01_Service       nService;
  tsp00_TaskId        nTaskId;
  tsp00_ErrText       pasErrtext;
  tsp01_CommErr       nCommRc;

  // fill argument nTaskId for sqlaconnect
  sqluid (&nTaskId);

  // fill argument nService for sqlaconnect
  switch (oTaskTypeIn) {
    case DBMSrvKnl_Session::taskNormal:
      nService.becomes(srvUser_esp01);
      break;
    case DBMSrvKnl_Session::taskEvent:
      nService.becomes(srvEvent_esp01);
      break;
    default:
      {
        Msg_List oMsg(DBMSRV_ERR_UNKNOWNTASK);
        this->oMessageList.Overrule(oMsg);
        return false;
      }
      break;
  } // end switch

  // connect to kernel
  tsp00_NodeId tspNode;
  tsp00_DbName tspDB;

  sqlaconnect( nTaskId,
               tspNode.c2p(""),
               tspDB.c2p(szDB),
               nService,
               1,
               this->nSessionHandle,
               this->nPacketSize,
               (void **) &this->pPacket,
               pasErrtext,
               nCommRc );

  if (nCommRc != commErrOk_esp01) {
    tsp00_ErrTextc       szErrtext;

    szErrtext.p2c(pasErrtext);

    this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(nCommRc), szErrtext.asCharp(), 0));
    this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_RTE));
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: Connect

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: ConnectUser
    ( const char * szUsr,
      const char * szPassword,
      bool         bNoTempFile)
{
  this->oMessageList.ClearMessageList();

  {
    size_t i = 0;
    size_t j = 0;
    bool bUpper = true;

    // username
    bUpper = true;
    this->szUser.rawAssign(szUsr);
    j = 0;
    for (i = 0; i <= strlen(szUsr); ++i) {
      if (((this->szUser[i] == '"') || (this->szUser[i] == '\'')) && ((i == 0) || (i == (strlen(this->szUser)-1)))) {
        bUpper = false;
      } else {
        if (bUpper) {
          this->szUser[j] = toupper (this->szUser[i]);
        } else {
          this->szUser[j] = this->szUser[i];
        } // end if
        ++j;
      } // end if
    } // end for

    // password
    bUpper = true;
    this->szPwd.rawAssign(szPassword);
    j = 0;
    for (i = 0; i <= strlen(this->szPwd); ++i) {
      if (((this->szPwd[i] == '"') || (this->szPwd[i] == '\'')) && ((i == 0) || (i == (strlen(this->szPwd)-1)))) {
        bUpper = false;
      } else {
        if (bUpper) {
          this->szPwd[j] = toupper (this->szPwd[i]);
        } else {
          this->szPwd[j] = this->szPwd[i];
        } // end if
        ++j;
      } // end if
    } // end for
  } // end block

  {
    // create request packet
    PIn_RequestPacket  oPacket(this->pPacket, this->nPacketSize, this->nUnicode);
    // initialise request packet
    oPacket.InitVersion(csp_comp_db_manager, NULL);
    // create segement
    PIn_RequestSegment oSegment = oPacket.AddSegment(sp1m_dbs, sp1sm_session_sqlmode);

    oSegment.SetProducer(sp1pr_internal_cmd);

    // create command part
    PIn_Part oCommandPart = oSegment.AddPart(sp1pk_command);
    // fill commadn part
    oCommandPart.AddArgument(CMD_CONNECT_P1, (int) strlen(CMD_CONNECT_P1));
    oCommandPart.AddData    (this->szUser,   (int) strlen(this->szUser));
    oCommandPart.AddData    (CMD_CONNECT_P2, (int) strlen(CMD_CONNECT_P2));
    if (bNoTempFile) {
      oCommandPart.AddData    (CMD_CONNECT_P3, (int) strlen(CMD_CONNECT_P3));
    } // end if
    // close command part
    oSegment.ClosePart(oCommandPart);
    // create data part for crypted password
    PIn_Part oDataPart = oSegment.AddPart(sp1pk_data);

    // fill data part with crypted password
    tsp00_Byte    cByte = csp_defined_byte;
    tsp00_CryptPw cryptPwd;
    tsp00_Name    pasPwd;
    tsp00_TermId  oTermid;

    pasPwd.c2p(this->szPwd);
    s02applencrypt (pasPwd, cryptPwd);

    sqltermid (oTermid);

    oDataPart.AddArgument(&cByte,   sizeof(cByte));
    oDataPart.AddData    (cryptPwd, sizeof(cryptPwd));
    oDataPart.AddData    (&cByte,   sizeof(cByte));
    oDataPart.AddData    (oTermid,  sizeof(oTermid));

    // close part
    oSegment.ClosePart(oDataPart);
    // close segment
    oPacket.CloseSegment(oSegment);

    // send request to kernel
    if (!Request(oPacket.Length())) {
      return false;
    } // end if
  } // end block

  {
    // receive answer from kernel and create answer packet
    PIn_ReplyPacket  oPacket(Receive());
    // check answer packet
    if (!oPacket.IsValid()) {
      return false;
    } // end if

    // read data
    PIn_ReplySegment oSegment     = oPacket.GetFirstSegment ();

    if (oSegment.ErrorCode() == 0) {

      PIn_Part         oPart = oSegment.FindPart(sp1pk_session_info_returned);

      if (oPart.IsValid()) {
        PIn_SessionInfoPart oSessionInfoPart(oPart);
        this->nUnicode = oSessionInfoPart.IsUnicode();
        this->bAutoCommit = true;
      } // end if

      oPart = oSegment.FindPart(sp1pk_user_info_returned);
      if (oPart.IsValid()) {
        PIn_UserInfoPart oUserInfoPart(oPart);
      } // end if

    } else {
      this->ReadErrorText(oSegment);
      return false;
    } // end if

  } // end block

  return true;
} // DBMSrvKnl_Session :: ConnectUser

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: FetchRecordBlock
    ( bool       & bEnd,
      bool         bFirst )
{
  Tools_DynamicUTF8String sCommand;
  tsp00_Int4              nParams = this->ParamCount();
  tsp00_Int4              i = 0;

  bEnd = false;

  sCommand = bFirst ? "FETCH FIRST INTO ?" : "FETCH NEXT INTO ?";

  for (i = 1; i < nParams; ++i) {
    sCommand.Append(", ?");
  } // end for

  if (!this->Execute(sCommand.CharPtr())) {
    if (!bFirst && (this->oMessageList.NextMessage() != NULL) && (this->oMessageList.NextMessage()->ID() == 100)) {
        this->Execute("CLOSE");
        this->oMessageList.ClearMessageList();
        bEnd = true;
        return true;
    } // end if
    return false;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: FetchRecordBlock

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: Request
    ( SAPDB_Int4 nRequestLength )
{
  tsp00_ErrText  pasErrtext;
  tsp01_CommErr  nCommRc;

  sqlarequest(this->nSessionHandle,
              this->pPacket,
              nRequestLength,
              pasErrtext,
              nCommRc);

  if (nCommRc != commErrOk_esp01) {
    tsp00_ErrTextc       szErrtext;

    szErrtext.p2c(pasErrtext);

    this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(nCommRc), szErrtext.asCharp(), 0));
    this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_RTE));
    return false;
  } else {
    this->bRequestPending = true;
  } // end if

  return true;
} // end DBMSrvKnl_Session :: Request

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
tsp00_Int4 DBMSrvKnl_Session :: ParamCount
  ( )
{
  PIn_ReplyPacket  oPacket(this->pReceivePacket);
  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oShortInfo   = oSegment.FindPart(sp1pk_shortinfo);
    if (oShortInfo.IsValid()) {
      return oShortInfo.GetRawPart()->sp1p_arg_count();
    } // end if
  } // end if

  return 0;
} // end DBMSrvKnl_Session :: ParamCount

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
tsp1_param_info * DBMSrvKnl_Session :: ParamInfo
  ( )
{
  PIn_ReplyPacket  oPacket(this->pReceivePacket);
  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oShortInfo = oSegment.FindPart(sp1pk_shortinfo);
    if (oShortInfo.IsValid()) {
      return  (tsp1_param_info *) oShortInfo.GetReadData();
    } // end if
  } // end if

  return NULL;
} // DBMSrvKnl_Session :: ParamInfo

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
tsp00_Int4 DBMSrvKnl_Session :: RecordSize
  ( )
{
  tsp00_Int4        nRecordSize = 0;
  tsp00_Int4        nParamCount = this->ParamCount();
  tsp00_Int4        i           = 0;
  tsp1_param_info * pParamInfo  = this->ParamInfo();

  for (i = 0; i < nParamCount; ++i) {
    nRecordSize = nRecordSize + pParamInfo[i].sp1i_in_out_len;
  } // end for

  return nRecordSize;
} // DBMSrvKnl_Session :: RecordSize

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
tsp00_Int4 DBMSrvKnl_Session :: RecordCount
  ( )
{
  tsp00_Int4        RecordCount = 0;

  PIn_ReplyPacket  oPacket(this->pReceivePacket);
  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oPart = oSegment.FindPart(sp1pk_data);
    if (oPart.IsValid()) {
      PIn_DataPart oDataPart(oPart);
      RecordCount = oDataPart.GetRawPart()->sp1p_arg_count();
    } // end if
  } // end if

  return RecordCount;
} // DBMSrvKnl_Session :: ResultCount

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
DBMSrv_SQLRecord & DBMSrvKnl_Session :: GetRecordObject
  ( )
{
  this->oRecord.RefreshRecordData(DBMSrv_SQLRecord::refreshWithoutRecordCounter);
  return this->oRecord;
} // end DBMSrvKnl_Session :: GetRecordObject

  /*-----------------------------------------------------------------------------*/
/*! @brief public member */
Tools_DynamicUTF8String DBMSrvKnl_Session :: FieldName
  ( const tsp00_Int4     nField )
{
  Tools_DynamicUTF8String sField;

  PIn_ReplyPacket  oPacket(this->pReceivePacket);

  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oColumnNames = oSegment.FindPart(sp1pk_columnnames);
    int                     nParamCount = this->ParamCount();

    if (oColumnNames.IsValid()) {
      const tsp00_Byte * pNames      = oColumnNames.GetReadData();
      int                i           = 0;
      int                nLen        = 0;

      // get columns
      for (i = 1; i <= nParamCount; i++) {
        nLen = *pNames++;
        if (i == nField) {
          if (this->nUnicode) {
            sField.ConvertFromUCS2_Unaligned_Swap((Tools_DynamicUTF8String::UCS2Char *) pNames, (Tools_DynamicUTF8String::UCS2Char *) (pNames + nLen));
          } else {
            sField.ConvertFromASCII_Latin1((Tools_DynamicUTF8String::ASCIIConstPointer) pNames, (Tools_DynamicUTF8String::ASCIIConstPointer) (pNames + nLen));
          } // end if
          break;
        } // end if
        pNames += nLen;
      } // end for

    } else {
      int i = 0;
      Tools_DynamicUTF8String sNumber;
      for (i = 1; i <= nParamCount; i++) {
        sNumber.ConvertFromInt(i);
        sField.Assign("Field").Append(sNumber);
      } // end for
    }  // end if
  } // end if

  return sField;
} // end DBMSrvKnl_Session :: FieldName

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: GetValue
    ( const tsp00_Int4     nField,
      DBMSrv_SQLField    & oField )
{
  PIn_ReplyPacket  oPacket(this->pReceivePacket);
  // check answer packet
  if (oPacket.IsValid()) {
    PIn_ReplySegment oSegment = oPacket.GetFirstSegment ();
    PIn_Part oPart   = oSegment.FindPart(sp1pk_data);

    if (oPart.IsValid()) {
      PIn_DataPart                   oDataPart(oPart);
      PIn_DataPart::ConversionNeeded oConversion = PIn_DataPart::ConvertUnknown;
      const char *                   pValue      = NULL;
      tsp00_Int4                     nLength     = 0;

      switch (oField.oType) {
        case DBMSrv_SQLField::typeInt:
          oField.iValue = 0;
          oField.bNull = oDataPart.IsNull(&(oField.oInfo));
          if (!oField.bNull) {
            if (!oDataPart.GetValueAsInt(&(oField.oInfo), oField.iValue)) {
              return false;
            } // end if
          } // end if
          break;
        case DBMSrv_SQLField::typeString:
          oField.sValue = "";
          oField.bNull = oDataPart.IsNull(&(oField.oInfo));
          if (!oDataPart.GetValueAsChar(&(oField.oInfo), pValue, nLength, oConversion)) {
            return false;
          } // end if

          switch (oSegment.GetRawHeader()->sp1r_function_code()) {
            case csp1_administrator_fc:
            case csp1_diagnose_fc:
              oField.sValue.ConvertFromASCII_Latin1((Tools_DynamicUTF8String::ASCIIConstPointer) pValue, (Tools_DynamicUTF8String::ASCIIConstPointer) (pValue + nLength));
              break;
            default:
              switch (oConversion) {
                case PIn_DataPart::ConvertAlways:
                  oField.sValue.ConvertFromUCS2_Unaligned_Swap((Tools_DynamicUTF8String::UCS2Char *) pValue, (Tools_DynamicUTF8String::UCS2Char *) (pValue + (nLength*2)));
                  break;
                case PIn_DataPart::ConvertNever:
                  oField.sValue.ConvertFromASCII_Latin1((Tools_DynamicUTF8String::ASCIIConstPointer) pValue, (Tools_DynamicUTF8String::ASCIIConstPointer) (pValue + nLength));
                  break;
                case PIn_DataPart::ConvertSometimes:
                  if (this->nUnicode) {
                    oField.sValue.ConvertFromUCS2_Unaligned_Swap((Tools_DynamicUTF8String::UCS2Char *) pValue, (Tools_DynamicUTF8String::UCS2Char *) (pValue + (nLength*2)));
                  } else {
                    oField.sValue.ConvertFromASCII_Latin1((Tools_DynamicUTF8String::ASCIIConstPointer) pValue, (Tools_DynamicUTF8String::ASCIIConstPointer) (pValue + nLength));
                  } // end if
                  break;
                default:
                  return false;
                  break;
              } // end switch
              break;
          } // end swicth

          oField.sValue.TrimTrailing();
          break;
        default:
          return false;
          break;
      } // end switch

    } // end if
  } // end if

  return true;
} // end DBMSrvKnl_Session :: GetValue

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
tsp1_packet * DBMSrvKnl_Session :: Receive
    (  )
{
  tsp00_ErrText  pasErrtext;
  tsp01_CommErr  nCommRc;
  SAPDB_Int4     nLength;

  this->pReceivePacket = NULL;

  // poll here for answer
  if ((!this->bAllowedInDbFull) && (!this->bNoTempFiles) && (this->nConnected == DBMSrvKnl_Session::taskNormal)) {
    SAPDB_UInt4  nMilliSeconds  = 1;
    bool         bLogFull       = false;
    bool         bDataFull      = false;

    while (!this->ReplyAvailable()) {
      RTE_ISystem::DoSleep(nMilliSeconds);
      if (nMilliSeconds > 100) {
        oState.IsFull(bLogFull, bDataFull);
        if (!oState.DBMMsg().IsEmpty()) {
          this->oMessageList = oState.DBMMsg();
          this->BreakSession();
          return NULL;
        } else if (bDataFull) {
          // database full error
          this->oMessageList = DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_DATAFULL);
          this->BreakSession();
          return NULL;
        } // end if
      } else {
        nMilliSeconds *= 2;
      } // end if
    } // end while
  } // end if

  sqlareceive (this->nSessionHandle,
               (void**)&this->pReceivePacket,
               nLength,
               pasErrtext,
               nCommRc);

  this->bRequestPending = false;

  if (nCommRc != commErrOk_esp01) {
    tsp00_ErrTextc       szErrtext;

    szErrtext.p2c(pasErrtext);

    this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(nCommRc), szErrtext.asCharp(), 0));
    this->oMessageList.Overrule(DBMSrvMsg_Error((nCommRc == commErrShutdown_esp01)?DBMSrvMsg_Error::DBMSrv_SHUTDOWN:DBMSrvMsg_Error::DBMSrv_RTE));
    return NULL;
  } // end if

  return pReceivePacket;
} // end DBMSrvKnl_Session :: Receive

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
bool DBMSrvKnl_Session :: ReplyAvailable
  ( )
{
  tsp00_ErrText  pasErrtext;
  tsp01_CommErr  nCommRc;
  sqlareplyavailable (this->nSessionHandle, pasErrtext, nCommRc);
  return (nCommRc == commErrOk_esp01);
} // end DBMSrvKnl_Session :: ReplyAvailable

/*-----------------------------------------------------------------------------*/
/*! @brief public member */
void DBMSrvKnl_Session :: Cancel
  ( )
{
  sqlacancel (this->nSessionHandle);
} // end DBMSrvKnl_Session :: Cancel

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
void DBMSrvKnl_Session :: ReadErrorText
  ( PIn_ReplySegment & oSegment )
{
    PIn_Part oPart     = oSegment.FindPart(sp1pk_errortext);
    PIn_Part oPart2    = oSegment.FindPart(sp1pk_message_list);

    if (oPart.IsValid()) {
      PIn_ErrorTextPart oErrorTextPart(oPart);
      Tools_DynamicUTF8String sError;
      if (this->nUnicode) {
        sError.ConvertFromUCS2_Unaligned_Swap((Tools_DynamicUTF8String::UCS2Char *) oErrorTextPart.ErrorText(), (Tools_DynamicUTF8String::UCS2Char *) (oErrorTextPart.ErrorText() + oErrorTextPart.ErrorLen()));
      } else {
        sError.ConvertFromASCII_Latin1((Tools_DynamicUTF8String::ASCIIConstPointer) oErrorTextPart.ErrorText(), (Tools_DynamicUTF8String::ASCIIConstPointer) (oErrorTextPart.ErrorText() + oErrorTextPart.ErrorLen()));
      } // end if

      if (oPart2.IsValid()) {
        PIn_MessageListPart oMsgListPart(oPart2);
        this->oMessageList.Overrule(oMsgListPart.MessageList());
      } // end if
      this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(oSegment.ErrorCode()), sError.CharPtr(), 0));
      this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SQL));
    } else {
      // may be it was a backup/recovery without error text part but an errortext
      // inside the answer record in the column "TAPE ERRORTEXT"
      Tools_DynamicUTF8String sError;
      DBMSrv_SQLRecord & oLocalRecord = this->GetRecordObject();
      sError = oRecord.fieldAsChar("TAPE ERRORTEXT");
      if (sError.Length() > 0) {
        this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(oSegment.ErrorCode()), sError.CharPtr(), 0));
      } else {
        this->oMessageList.Overrule(Msg_List(DBMSrv_ExtendedError(oSegment.ErrorCode()), "(no error text available from database kernel)", 0));
      } // end if
      this->oMessageList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SQL));
    } // end if

} // end DBMSrvKnl_Session :: ReadErrorText

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
bool DBMSrvKnl_Session :: AllowedInDbFull
  ( const char * szCmd )
{
  // check for
  //     add data
  //     save log
  //     auto save
  //     takeover

  const char * szAllowed [] = {"ADD DATA",
                               "ADD LOG",
                               "SAVE LOG",
                               "AUTOSAVE",
                               "TAKEOVER",
                               NULL
                              };

  Tools_DynamicUTF8String szString(_Tools_UTF8Ptr(szCmd));
  Tools_DynamicUTF8String sz1stToken;
  Tools_DynamicUTF8String sz2ndToken;

  szString.ToUpper();
  szString.Trim();

  Tools_DynamicUTF8String::BasisElementIndex nSpace = szString.Find(" ");

  // extract first two tokens
  if (nSpace != Tools_DynamicUTF8String::NPos) {
    sz1stToken = szString.SubStr(0, nSpace);
    szString   = szString.SubStr(nSpace);
    szString.Trim();
    nSpace = szString.Find(" ");
    if (nSpace != Tools_DynamicUTF8String::NPos) {
      sz2ndToken = szString.SubStr(0, nSpace);
    } else {
      sz2ndToken = szString;
    } // end if
  } else {
    sz1stToken = szString;
  } // end if

  // build string
  szString.Assign(sz1stToken).Append(" ").Append(sz2ndToken);

  // compare with allowed list
  int nAllowed = 0;
  while (szAllowed[nAllowed] != NULL) {
    if (szString.Compare(szAllowed[nAllowed]) == 0) {
      return true;
    } // end if
    ++nAllowed;
  } // end while

  return false;
} // end DBMSrvKnl_Session :: AllowedInDbFull

/*-----------------------------------------------------------------------------*/
/*! @brief private member */
void DBMSrvKnl_Session :: Reset
    ( )
{
  this->nSessionHandle   = 0;
  this->nPacketSize      = 0;
  this->pPacket          = NULL;
  this->pReceivePacket   = NULL;
  this->nUnicode         = 0;
  this->szUser.Init();
  this->szPwd.Init();
  this->szDatabase.Init();
  this->nConnected       = DBMSrvKnl_Session::taskUnknown;
  this->bAutoCommit      = false;
  this->bNoTempFiles     = false;
  this->bAdminConnect    = false;
  this->bAllowedInDbFull = false;
  this->bRequestPending  = false;
  this->oUserType        = DBMSrvKnl_Session::userUnknown;
  oRecord.SetSession(this);

} // end DBMSrvKnl_Session :: Reset

/*-----------------------------------------------------------------------------*/
/*******************************************************************************/
/*-----------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------*/
/*! @brief private constructor */
DBMSrv_SQLRecord :: DBMSrv_SQLRecord
    ( )
{
  this->RefreshRecordData(DBMSrv_SQLRecord::resetAll);
} // end DBMSrv_SQLRecord :: DBMSrv_SQLRecord

/*-----------------------------------------------------------------------------*/
/*! @brief destructor */
DBMSrv_SQLRecord :: ~DBMSrv_SQLRecord
    (  )
{
} // end DBMSrv_SQLRecord :: ~DBMSrv_SQLRecord

/*! @brief private member */
void DBMSrv_SQLRecord :: SetSession
    ( DBMSrvKnl_Session * pSession )
{
  this->pKernelSession = pSession;
} // end DBMSrv_SQLRecord :: SetSession

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_Int4 DBMSrv_SQLRecord :: fieldAsInt ( const tsp00_Int4 nField  ) const
{
  if (nField <= this->nParamCount && nField > 0) {
    DBMSrv_SQLField oField(&this->pParamInfo[nField-1], DBMSrv_SQLField::typeInt);
    oField.oInfo.sp1i_bufpos() = oField.oInfo.sp1i_bufpos() + ((this->nCurrentRecord-1)*this->nRecordSize);

    if (this->pKernelSession->GetValue(nField, oField)) {
      return oField.iValue;
    } // end if
  } // end if

  return 0;
} // end DBMSrv_SQLRecord :: fieldAsInt

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_Int4 DBMSrv_SQLRecord :: fieldAsInt ( const Tools_DynamicUTF8String & szField ) const
{
  return fieldAsInt(fieldNumber(szField));
} // end DBMSrv_SQLRecord :: fieldAsInt

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
Tools_DynamicUTF8String DBMSrv_SQLRecord :: fieldAsChar
      ( const tsp00_Int4 nField,
        const char     cStringQuote) const
{
  if (nField <= this->nParamCount && nField > 0) {
    DBMSrv_SQLField oField(&this->pParamInfo[nField-1], DBMSrv_SQLField::typeString);
    oField.oInfo.sp1i_bufpos() = oField.oInfo.sp1i_bufpos() + ((this->nCurrentRecord-1)*this->nRecordSize);

    if (!this->IsLong(oField.oInfo.sp1i_data_type)) {
      if (this->pKernelSession->GetValue(nField, oField)) {
        if (cStringQuote != 0 && !oField.bNull && this->IsString(oField.oInfo.sp1i_data_type)) {
          if (oField.sValue.Length() == 0) {
            oField.sValue.Append(cStringQuote);
          } else {
            oField.sValue.Insert(0, cStringQuote);
          } // end if
          oField.sValue.Append(cStringQuote);
        } // end if
      } // end if
      return oField.sValue;
    } else {
      return "(long)";
    } // end if
  } // end if

  return "";
} // end DBMSrv_SQLRecord :: fieldAsChar

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
Tools_DynamicUTF8String DBMSrv_SQLRecord :: fieldAsChar
      ( const Tools_DynamicUTF8String & szField,
        const char                      cStringQuote) const
{
  return fieldAsChar(fieldNumber(szField));
} // end DBMSrv_SQLRecord :: fieldAsChar

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_DataType DBMSrv_SQLRecord :: fieldType ( const tsp00_Int4 nField  ) const
{
  tsp00_DataType oDataType;

  if ((nField <= this->nParamCount) && (nField > 0) && (this->pParamInfo != NULL)) {
    oDataType = this->pParamInfo[nField-1].sp1i_data_type;
  } else {
    oDataType.becomes(dunknown);
  } // end if

  return oDataType;
} // end DBMSrv_SQLRecord :: fieldType

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_DataType DBMSrv_SQLRecord :: fieldType ( const Tools_DynamicUTF8String & szField  ) const
{
  return fieldType(fieldNumber(szField));
} // end DBMSrv_SQLRecord :: fieldType

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
bool DBMSrv_SQLRecord :: isNull ( const tsp00_Int4 nField  ) const
{
  if (nField <= this->nParamCount && nField > 0) {
    DBMSrv_SQLField oField(&this->pParamInfo[nField-1], DBMSrv_SQLField::typeString);
    oField.oInfo.sp1i_bufpos() = oField.oInfo.sp1i_bufpos() + ((this->nCurrentRecord-1)*this->nRecordSize);

    if (this->pKernelSession->GetValue(nField, oField)) {
      return oField.bNull;
    } // end if
  } // end if

  return false;
} // end DBMSrv_SQLRecord :: isNull

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
bool DBMSrv_SQLRecord :: isNull ( const Tools_DynamicUTF8String &  szField ) const
{
  return this->isNull(this->fieldNumber(szField));
} // end DBMSrv_SQLRecord :: isNull

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
size_t DBMSrv_SQLRecord :: size ( const tsp00_Int4 nField  ) const
{
  return this->pParamInfo[nField-1].sp1i_length;
} // end DBMSrv_SQLRecord :: size

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
size_t DBMSrv_SQLRecord :: size ( const Tools_DynamicUTF8String &  szField  ) const
{
  return this->size(this->fieldNumber(szField));
} // end DBMSrv_SQLRecord :: size

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
size_t DBMSrv_SQLRecord :: maxSize ( const tsp00_Int4 nField  ) const
{
  return (this->size(nField) > this->fieldName(nField).Length() ? this->size(nField) : this->fieldName(nField).Length());
} // end DBMSrv_SQLRecord :: maxSize

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
size_t DBMSrv_SQLRecord :: maxSize ( const Tools_DynamicUTF8String &  szField  ) const
{
  return this->maxSize(this->fieldNumber(szField));
} // end DBMSrv_SQLRecord :: maxSize

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
bool DBMSrv_SQLRecord :: exists ( const tsp00_Int4 nField  ) const
{
  return (nField <= this->nParamCount && nField > 0);
} // end DBMSrv_SQLRecord :: exists

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
bool DBMSrv_SQLRecord :: exists ( const Tools_DynamicUTF8String &  szField  ) const
{
  return this->exists(this->fieldNumber(szField));
} // end DBMSrv_SQLRecord :: exists

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_Int4 DBMSrv_SQLRecord :: fieldNumber ( const Tools_DynamicUTF8String &  szField  ) const
{
  tsp00_Int4  nField = 0;
  int         i = 1;

  while (nField == 0 && i <= this->nParamCount) {
    nField = (szField == this->fieldName(i)) ? i : 0;
    ++i;
  } // end if

  return nField;
} // end DBMSrv_SQLRecord :: fieldNumber

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
Tools_DynamicUTF8String DBMSrv_SQLRecord :: fieldName ( const tsp00_Int4 nField  ) const
{
  return this->pKernelSession->FieldName(nField);
} // end DBMSrv_SQLRecord :: fieldName

/*-----------------------------------------------------------------------------*/
/*! @brief get data */
tsp00_Int4 DBMSrv_SQLRecord :: fields ( )
{
  return this->nParamCount;
} // end DBMSrv_SQLRecord :: fields

/*-----------------------------------------------------------------------------*/
/*! @brief check for string */
bool DBMSrv_SQLRecord :: IsString
  ( tsp00_DataType aDataType ) const
{
  bool bRc = false;

  switch(aDataType) {
    // Had I hit all data types???
    case dcha:
    case dunicode:
    case dvarchara:
    case ddate:
    case dtime:
    case dtimestamp:
    case dstra:
      bRc = true;
  } // end switch

  return bRc;

} // DBMSrv_SQLRecord :: IsString

/*-----------------------------------------------------------------------------*/
/*! @brief check for string */
bool DBMSrv_SQLRecord :: IsLong
  ( tsp00_DataType aDataType ) const
{
  bool bRc = false;

  switch(aDataType) {
    // Had I hit all data types???
    case dstra:
    case dstrb:
      bRc = true;
  } // end switch

  return bRc;

} // DBMSrv_SQLRecord :: IsLong

/*-----------------------------------------------------------------------------*/
/*! @brief get record */
bool DBMSrv_SQLRecord :: getNextRecord
  ( bool & bEnd )
{
  return getRecord(bEnd, false);
} // end DBMSrv_SQLRecord :: getNextRecord

/*-----------------------------------------------------------------------------*/
/*! @brief get record */
bool DBMSrv_SQLRecord :: getFirstRecord
  ( bool & bEnd )
{
  return getRecord(bEnd, true);
} // end DBMSrv_SQLRecord :: getFirstRecord

/*-----------------------------------------------------------------------------*/
/*! @brief get record */
bool DBMSrv_SQLRecord :: getRecord
  ( bool & bEnd,
    bool   bFirst )
{
  bool bReturn = true;

  if (this->nCurrentRecord < this->nMaxRecords) {
    ++nCurrentRecord;
  } else {

    bReturn = this->pKernelSession->FetchRecordBlock(bEnd, bFirst );

    if (bReturn && !bEnd) {
      this->RefreshRecordData(DBMSrv_SQLRecord::refreshAll);
    } else {
      this->RefreshRecordData(DBMSrv_SQLRecord::resetAll);
    } // end if

  } // end if

  return bReturn;
} // end DBMSrv_SQLRecord :: getRecord

/*-----------------------------------------------------------------------------*/
/*! @brief refresh record */
void DBMSrv_SQLRecord :: RefreshRecordData
  ( RefreshMode oRefreshMode )
{
  switch (oRefreshMode) {
    case refreshAll:
      this->nCurrentRecord = 1;
      // no break, fall trough!!!
    case refreshWithoutRecordCounter:
      this->nParamCount    = this->pKernelSession->ParamCount();
      this->pParamInfo     = this->pKernelSession->ParamInfo();
      this->nMaxRecords    = this->pKernelSession->RecordCount();
      this->nRecordSize    = this->pKernelSession->RecordSize();
      break;
    case resetAll:
      this->nParamCount    = 0;
      this->pParamInfo     = NULL;
      this->nMaxRecords    = 0;
      this->nRecordSize    = 0;
      this->nCurrentRecord = 1;
      break;
  } // end switch
} // end DBMSrv_SQLRecord :: RefreshRecordData
