/*!
  @file           DBMSrvKnl_Session.hpp
  @author         Bernd Vorsprach - bernd.vorsprach@sap.com
  @brief          DBMServer to Kernel Communication - Specification

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


\endif
*/

#ifndef _DBMSrvKnl_Session_HPP_
#define _DBMSrvKnl_Session_HPP_

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */
#include "SAPDB/DBM/Srv/DBMSrv_Common.hpp"
#include "SAPDB/ToolsCommon/Tools_DynamicUTF8String.hpp"
#include "SAPDB/DBM/Srv/KernelAccess/DBMSrvKnl_State.hpp"

#include "SAPDB/PacketInterface/PIn_SpecialParts.h"

/*
  -----------------------------------------------------------------------------
  forward declarations
  -----------------------------------------------------------------------------
 */
class  PIn_ReplySegment;
class  DBMSrvKnl_Session;

/*! @brief field value */
class DBMSrv_SQLField {
private:
  friend class DBMSrv_SQLRecord;
  friend class DBMSrvKnl_Session;

  /*! @brief   data type */
  typedef enum {
    typeInt,
    typeString
  } DataType;

  /*! @ brief constructor */
  DBMSrv_SQLField
    ( tsp1_param_info * aInfo,
      DataType          aType);

  /*! @brief member for paraminfo */
  tsp1_param_info         oInfo;
  /*! @ brief member fro requested type */
  DataType                oType;
  /*! @ brief member for integer */
  SAPDB_Int4              iValue;
  /*! @ brief member for string */
  Tools_DynamicUTF8String sValue;
  /*! @ brief null flag */
  bool                    bNull;

}; // end class DBMSrv_SQLField

/*! @brief SQL Record Object */
class DBMSrv_SQLRecord {

  public:
    /*! @brief   RefreshMode */
    typedef enum {
      refreshAll,
      refreshWithoutRecordCounter,
      resetAll
    } RefreshMode;

    /*! @brief destructor */
    ~DBMSrv_SQLRecord  ( );

    /*! @brief      returns a field contents

                    If there no valid conversion from DB datatype to SAPDB_Int4
                    or any other error occured a value of 0 will be returned.
                    Check also isNull() and fieldType() if you won't trust this
                    behavior.

        @param      nField  [IN] - number of field

        @return     field value
     */
    SAPDB_Int4  fieldAsInt ( const SAPDB_Int4 nField  ) const;

    /*! @brief      returns a field contents

                    If there no valid conversion from DB datatype to SAPDB_Int4
                    or any other error occured a value of 0 will be returned.
                    Check also isNull() and fieldType() if you won't trust this
                    behavior.

        @param      szField [IN] - name of field

        @return     field value
     */
    SAPDB_Int4  fieldAsInt ( const Tools_DynamicUTF8String & szField ) const;

    /*! @brief      returns a field contents

                    If there no valid conversion from DB datatype to _TCHAR *
                    or any other error occured a empty String will be returned.
                    Check also isNull() and fieldType() if you won't trust this
                    behavior.

                    The return value is a temporarely allocated buffer. It will
                    only valid until the next call of fieldAsChar or until the
                    destruction of the object.

      @param        nField       [IN] - number of field
      @param        cStringQuote [in] - character for quoting strings
                                        0 means no quoting
      @return       field value
     */
    Tools_DynamicUTF8String fieldAsChar ( const SAPDB_Int4 nField,
                                          const char       cStringQuote = 0) const;

    /*! @brief      returns a field contents

                    If there no valid conversion from DB datatype to _TCHAR *
                    or any other error occured a empty String will be returned.
                    Check also isNull() and fieldType() if you won't trust this
                    behavior.

                    The return value is a temporarely allocated buffer. It will
                    only valid until the next call of fieldAsChar or until the
                    destruction of the object.

      @param        szField      [IN] - name of field
      @param        cStringQuote [in] - character for quoting strings
                                        0 means no quoting
      @return       field value
     */
    Tools_DynamicUTF8String fieldAsChar ( const Tools_DynamicUTF8String & szField,
                                          const char                      cStringQuote = 0) const;

    /*! @brief      returns a field type
        @param      nField       [IN] - number of field
        @return     field type
     */
    tsp00_DataType fieldType ( const SAPDB_Int4  nField  ) const;

    /*! @brief      returns a field type
        @param      szField      [IN] - name of field
        @return     field type
     */
    tsp00_DataType fieldType ( const Tools_DynamicUTF8String &  szField ) const;

    /*! @brief      returns a the nullvalue flag
        @param      nField       [IN] - number of field
        @return     true if null, otherwise false
     */
    bool isNull ( const SAPDB_Int4 nField  ) const;

    /*! @brief      returns a the nullvalue flag
        @param      szField      [IN] - name of field
        @return     true if null, otherwise false
     */
    bool isNull ( const Tools_DynamicUTF8String &  szField ) const;

    /*! @brief      return the maximal size of a field value
        @param      nField       [IN] - number of field
        @return     size of field
     */
    size_t size ( const SAPDB_Int4 nField  ) const;

    /*! @brief      return the maximal size of a field value
        @param      szField      [IN] - name of field
        @return     size of field
     */
    size_t size ( const Tools_DynamicUTF8String &  szField  ) const;

    /*! @brief      return the size of a column regarding the columname and the value size
        @param      nField       [IN] - number of field
        @return     size of field
     */
    size_t maxSize ( const SAPDB_Int4 nField  ) const;

    /*! @brief      return the size of a column regarding the columname and the value size
        @param      szField      [IN] - name of field
        @return     size of field
     */
    size_t maxSize ( const Tools_DynamicUTF8String &  szField  ) const;

    /*! @brief      check the existence of a field
        @param      nField       [IN] - number of field
        @return     true if null, otherwise false
     */
    bool exists ( const SAPDB_Int4 nField  ) const;

    /*! @brief      check the existence of a field
        @param      szField      [IN] - name of field
        @return     true if null, otherwise false
     */
    bool exists ( const Tools_DynamicUTF8String &  szField ) const;

    /*! @brief      returns a field name

        @param      nField  [IN] - number of field, enumeration of fileds is
                                   starting with 1

                    This is a temporarely allocated buffer. It will only valid
                    until the next call of fieldName or until the destruction
                    of the object.

        @return     field name
     */
    Tools_DynamicUTF8String fieldName ( const SAPDB_Int4 nField  ) const;

    /*! @brief      returns a field number
        @param      szField  [IN] - name of field
        @return     field number
     */
    SAPDB_Int4 fieldNumber ( const Tools_DynamicUTF8String & szField  ) const;

    /*! @brief      gets the nex record
        @param      bEnd [OUT] - no record available
        @return     success or not
     */
    bool getNextRecord ( bool & bEnd );

    /*! @brief      returns the number of fields
        @return     number of fields
     */
    SAPDB_Int4 fields ( );

  private:
    /*! @brief constructor
     */
    DBMSrv_SQLRecord  ( );
    /*! @brief set kernel session
        @param  pSession [IN]  - kernel session object
     */
    void SetSession ( DBMSrvKnl_Session * pSession );
    /*! @brief      gets the first record
        @param      bEnd [OUT] - no record available
        @return     success or not
     */
    bool getFirstRecord ( bool & bEnd );
    /*! @brief      gets the next/first record
        @param      bEnd   [OUT] - no record available
        @param      bFirst [IN]  - first or next record?
        @return     success or not
     */
    bool getRecord ( bool & bEnd, bool bFirst);
    /*! @brief      refresh the record data
        @param      bResetRecordCounter  [IN] - reset the current record or not
     */
    void RefreshRecordData
      ( RefreshMode oRefreshMode );
    /*! @brief check for string */
    bool IsString
      ( tsp00_DataType aDataType ) const;
    /*! @brief check for long */
    bool IsLong
      ( tsp00_DataType aDataType ) const;

    /*! @brief private member */
    DBMSrvKnl_Session *    pKernelSession;
    /*! @brief private member */
    SAPDB_Int4             nParamCount;
    /*! @brief private member */
    tsp1_param_info *      pParamInfo;
    /*! @brief private member */
    SAPDB_Int4             nCurrentRecord;
    /*! @brief private member */
    SAPDB_Int4             nMaxRecords;
    /*! @brief private member */
    SAPDB_Int4             nRecordSize;

    friend class DBMSrvKnl_Session;
}; // end class DBMSrv_SQLRecord

/*!
  @brief     Kernel session object

  One object of this class represents a kernel session

 */
class DBMSrvKnl_Session {

public:
  /*! @brief   user type */
  typedef enum {
    userUnknown,
    userDBM,
    userSYSDBA,
    userSQL
  } UserType;

  /*! @brief   task type */
  typedef enum {
    taskUnknown,
    taskNormal,
    taskEvent
  } TaskType;

  /*! @brief   result type */
  typedef enum {
    resultUnknown,
    resultNothing,
    resultError,
    resultMessage,
    resultUtility,
    resultSQL
  } ResultType;

  /*! @brief   Constructor */
  DBMSrvKnl_Session
    (  );

  /*! @brief   Destructor */
  ~DBMSrvKnl_Session
    (  );

  /*! @brief connect to user task */
  bool NormalConnect
    ( const char * szDB,
      const char * szUser,
      const char * szPassword,
      bool         bWithNoTempFiles = false );

  /*! @brief   connect to  event task */
  bool EventConnect
    ( const char * szDB );

  /*! @brief   execute a statement */
  bool Execute
    ( const char * szCommand,
      ResultType & oResultType );

  /*! @brief   execute a statement */
  bool Execute
    ( const char * szCommand );

  /*! @brief   execute a statement */
  bool ExecuteWithoutCommit
    ( const char * szCommand );

  /*! @brief   execute a statement */
  static bool Execute
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError );

  /*! @brief   execute a statement */
  static bool ExecuteWithNoTempFiles
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError );

  /*! @brief  execute a request with statement
      @param  szCommand [IN] -null terminated statement
      @return true if successfull, in case of false use LastMessage()
   */
  bool Request
    ( const char * szCommand );

  /*! @brief   execute a request to the event task
      @return true if successfull, in case of false use LastMessage()
   */
  bool EventRequest
    ( );

  /*! @brief  receive a answer after request
      @param  oResultType [OUT ] - the type of the kernel answer
      @return true if successfull, in case of false use LastMessage()
   */
  bool Receive
    ( ResultType & oResultType );

  /*! @brief  receive a answer after request from the eventtask
      @param  pEventBuffer [IN] - pointer to buffer for the eventdescription
      qparam  nBufferSize  [IN] - size of buffer
      @return true if successfull, in case of false use LastMessage()
   */
  bool EventReceive
    ( void * const pEventBuffer, 
      size_t nBufferSize  );

  /*! @brief  check whether a answer is available or not
      @return true if answer is available
   */
  bool ReplyAvailable
    (  );

  /*! @brief   set cancel flag for kernel task */
  void Cancel
    (  );

  /*! @brief   release a connection */
  bool Release
    ( );

  /*! @brief  Returns the last message
      @return A reference to the internal message object.
   */
  const DBMSrvMsg_Error & LastMessage
    ( ) const
    {return this->oMessageList;}

  /*! @brief  Returns the kernel message
      @return A Msg_List object with the message from the database kernel.
   */
  const Msg_List KernelMessage
    ( ) const;

  /*! @brief get connect flag
      @return true if connected
   */
  bool IsConnected
    ( ) const
    {return (this->nConnected != taskUnknown);}

  /*! @brief get record object */
  DBMSrv_SQLRecord & GetRecordObject
    ( );

  /*! @brief  get the database name of the connection
      @return A const pointer to a null terminated string. The
              string will be empty if there is no cionnection to
              a database.
   */
  const char * GetDatabase() const { return this->szDatabase.asCharp();}

  /*! @brief  get the request pending state of the session
      @return true if there is a reuest pending otherwise false.
   */
  bool IsRequestPending() const { return this->bRequestPending;}

private:
  /*! @brief parameter count of answer */
  SAPDB_Int4 ParamCount
    ();

  /*! @brief parameter info of answer */
  tsp1_param_info * ParamInfo
    ( );

  /*! @brief size of one record */
  SAPDB_Int4 RecordSize
    ();

  /*! @brief count of records in result */
  SAPDB_Int4 RecordCount
    ();

  /*! @brief name of a field */
  Tools_DynamicUTF8String FieldName
    ( const SAPDB_Int4     nField );

  /*! @brief get a field value */
  bool GetValue
    ( const SAPDB_Int4     nField,
      DBMSrv_SQLField    & oField );

  /*! @brief   connect to  a task */
  bool Connect
    (TaskType     oTaskType,
     const char * szDB );

  /*! @brief   connect connect user to session */
  bool ConnectUser
    ( const char * szUser,
      const char * szPassword,
      bool         bNoTempFile);

  /*! @brief   reset all members of the object */
  void Reset();

  /*! @brief   fetch record*/
  bool FetchRecordBlock
    ( bool & bEnd,
      bool   bFirst = false );

  /*! @brief   execute a statement */
  static bool Execute
    ( const char      * szDB,
      const char      * szUser,
      const char      * szPassword,
      const char      * szCommand,
      DBMSrvMsg_Error & oError,
      bool              bWithNoTempFiles);

  /*! @brief   send request to kernel */
  bool Request
    ( SAPDB_Int4 nRequestLength );

  /*! @brief   receive answer from kernel */
  tsp1_packet * Receive
    ( );

  /*! @brief   cancel a request */
  bool BreakSession
    (  );

  void ReadErrorText
    ( PIn_ReplySegment & oSegment );

  bool AllowedInDbFull
    ( const char * szCommand );

  /*! @brief Messagelist object */
  DBMSrvMsg_Error        oMessageList;
  /*! @brief   kernel session handle */
  SAPDB_Int4             nSessionHandle;
  /*! @brief   size of paket */
  SAPDB_Int4             nPacketSize;
  /*! @brief   request packet */
  tsp1_packet         *  pPacket;
  /*! @brief   receive packet */
  tsp1_packet         *  pReceivePacket;
  /*! @brief   unicode flag */
  SAPDB_Int4             nUnicode;
  /*! @brief   username     */
  tsp00_KnlIdentifierc   szUser;
  /*! @brief   password     */
  tsp00_Namec            szPwd;
  /*! @brief   database name */
  tsp00_DbNamec          szDatabase;
  /*! @brief   connect flag */
  TaskType               nConnected;
  /*! @brief   auto commit flag */
  bool                   bAutoCommit;
  /*! @brief user type information */
  UserType               oUserType;
  /*! @brief   no temporary files flag */
  bool                    bNoTempFiles;
  /*! @brief   flag if connected in admin */
  bool                    bAdminConnect;
  /*! @brief   command allowed in db full */
  bool                    bAllowedInDbFull;
  /*! @brief data base state object */
  DBMSrvKnl_State        oState;
  /*! @brief record object */
  DBMSrv_SQLRecord       oRecord;
  /*! @brief   command allowed in db full */
  bool                    bRequestPending;

  friend class DBMSrv_SQLRecord;
}; // end class DBMSrvKnl_Session

#endif /* _DBMSrvKnl_Session_HPP_ */
