/*!
    @file           DBMSrvShM_Base.cpp
    @author         MarcW
    @brief          Base class for shared memory - implementation

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

\endif
*/

#include <stdio.h>
#include <time.h>

#include "heo06.h"
#include "heo41.h"
#include "hsp100.h"

#include "SAPDB/DBM/Srv/DBMSrv_Common.hpp"
#include "SAPDB/DBM/Logging/DBMLog_Logger.hpp"
#include "SAPDB/DBM/Srv/Message/DBMSrvMsg_Error.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_Base.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_InitialisationSynchroniser.hpp"
#include "SAPDB/RunTime/RTE_ISystem.hpp"
#include "SAPDB/RunTime/System/RTESys_LockArea.hpp"
#include "SAPDB/RunTime/System/RTESys_ProcessState.hpp"
#include "SAPDBCommon/SAPDB_string.h"
#include "ToolsCommon/MemoryManagement/ToolsMeMa_GuardedOperation.hpp"

#ifdef _WIN32
const char* const DBMSrvShM_Base::SHM_NAME("\\dbm.shm");
const char* const DBMSrvShM_Base::SHM_SIZEINFO("\\dbm.shi");
#else
const char* const DBMSrvShM_Base::SHM_NAME("/dbm.shm");
const char* const DBMSrvShM_Base::SHM_SIZEINFO("/dbm.shi");
#endif

const size_t DBMSrvShM_Base::SHM_SIZE_MAX(1024*1024);
const int DBMSrvShM_Base::NUMBER_OF_LINETYPES(4);
const SAPDB_UInt4 DBMSrvShM_Base::SLEEP_TIME(30000);

DBMSrvShM_Base::DBMSrvShM_Base(const char* shmDir, const bool limitedAccess)
        : m_IsValid(false),
          m_LimitedAccess(limitedAccess),
          m_Persistence(shmDir, SHM_SIZEINFO, SHM_NAME),
          m_pShmVersion(NULL),
          m_LengthOperationOpen(false),
          m_pSharedMemory(NULL),
          m_pInfoShm(NULL),
          m_SizeInfoShm(0),
          m_pSpinlock(NULL),
          m_pBackinglock(NULL),
          m_pLockingPid(NULL),
          m_pLockCounter(NULL),
          m_pSaveHalfSwitcher(NULL),
          m_pPlayers(NULL),
          m_pPlayersLocal(NULL),
          m_pData(NULL) {

    // code version
    tsp100_VersionID1  VersionID1;
    sp100_GetVersionID ( VersionIDType1_esp100, s100buildnumberEx, &VersionID1 );
    sprintf(m_CodeVersion, "%02d.%02d.%02d.%03d"DBMSrvShM_Base_SHM_AUTHORS,
                    (int) VersionID1.MajorVersion_sp100,
                    (int) VersionID1.MinorVersion_sp100,
                    (int) VersionID1.CorrLevel_sp100,
                    (int) VersionID1.BuildNumberPrefix_sp100);

    // initialize random number generation for sleeping when allocating shared memory
    // see method openSharedMemory
    srand((unsigned int)time(NULL));

    SAPDB_strcpy(m_ShmIDInfo, m_Persistence.getInfoFile());
    SAPDB_strcpy(m_ShmIDData, m_Persistence.getDataFile());

    bool shmExists(false);
    Msg_List chkLst;
    if( !m_Persistence.exists(shmExists, chkLst) ) {
        // checking persistence failed: panic
        DBMLog_Logger::getInstance().addMessageList(chkLst, true);
    }

    bool initializingProcess(
        DBMSrvShM_InitialisationSynchroniser::permissionToInitialize(m_ShmIDInfo));

    m_SizeInfoShm =
        2 * sizeof(RTESys_LockArea) + 
        2 * sizeof(RTESys_LockArea) / sizeof(void*) + sizeof(void*) + 
        sizeof(*m_pLockingPid) +
        sizeof(*m_pLockCounter) +
        sizeof(*m_pShmVersion) + 
        sizeof(*m_pSaveHalfSwitcher) + 
        sizeof(*m_pPlayers);

    m_pInfoShm = openSharedMemory(m_ShmIDInfo, m_SizeInfoShm);
    if( m_pInfoShm != NULL ) {
        // define spinlocks
        m_pSpinlock = new RTESync_Spinlock(
                        ((RTESys_LockArea *)m_pInfoShm)->GetLockValue(),
                        false,
                        false);

        m_pBackinglock = new RTESync_Spinlock(
                        ((RTESys_LockArea *)(m_pInfoShm + sizeof(RTESys_LockArea)))->GetLockValue(),
                        false,
                        false);

        // locking pid (points to an aligned address)
        m_pLockingPid = (RTE_OSPid*)(
                                        ((size_t)(m_pInfoShm + 2 * sizeof(RTESys_LockArea)) /
                                        sizeof(void*)) * sizeof(void*) +
                                        sizeof(void*)
                                    );

        // lock counter (points to an aligned address)
        m_pLockCounter = (int*)(
            ((size_t)((SAPDB_Byte*)m_pLockingPid + sizeof(*m_pLockingPid)) /
                                    sizeof(void*)) * sizeof(void*) + 
                                    sizeof(void*)
                            );

        m_pShmVersion = (DBMSrvShM_SaveType<char[DBMSrvShM_Base_SHM_VERSION_LENGTH]>*)
                                    ((SAPDB_Byte*)m_pLockCounter + 
                                        sizeof(*m_pLockCounter));

        m_pSaveHalfSwitcher = (DBMSrvShM_SaveType<DBMSrvShM_HalfSwitcher>*)
                                    ((SAPDB_Byte*)m_pShmVersion + 
                                        sizeof(*m_pShmVersion));

        m_pPlayers = (DBMSrvShM_ParticipantSet*)
                                        ((SAPDB_Byte*)m_pSaveHalfSwitcher + 
                                        sizeof(*m_pSaveHalfSwitcher));

        if( initializingProcess ) {
            // we're creating the shared memory
            m_pSpinlock->Unlock();
            m_pBackinglock->Unlock();
            *m_pLockingPid=0;
            m_pShmVersion->setValue(m_CodeVersion);
            m_pPlayers->clear();
            m_pPlayers->insert(RTESys_ProcessID());
            // default shm size
            DBMSrvShM_HalfSwitcher tmpHS;
            // for each line type, there are 2 save ints and one save size_t
            size_t initialSize(0);
            initialSize =  NUMBER_OF_LINETYPES * (sizeof(DBMSrvShM_SaveType<size_t>) +
                                                2 * sizeof(DBMSrvShM_SaveType<int>));

            tmpHS.setMemSize(initialSize);
            tmpHS.setDataSize(initialSize);
            tmpHS.setOffset(0);
            m_pSaveHalfSwitcher->setValue(tmpHS);
            // now the losers can continue...
            ++(*m_pLockCounter);
        }
        else {
            // shared memory existed earlier (we're reusing it)
            // wait until lockcounter is != 0 (otherwise someone else is still
            // initializing shared memory)
            while( (*m_pLockCounter) == 0 )
                RTE_ISystem::DoSleep(100);

            // here we have a fully initialized shared memory
            bool limitedAccess(m_LimitedAccess);
            m_LimitedAccess = false;
            while( !lockMemory() )
                RTE_ISystem::DoSleep(100);

            char *bufFileVersion(NULL);
            char *bufCodeVersion(NULL);
            bufFileVersion = getVersion(true);
            bufCodeVersion = getVersion(false);

            int versionMatch(strcmp( bufFileVersion, bufCodeVersion));
            if( versionMatch < 0 ) {
                // code version is newer than the file version
                // migrate the shm file
                // we update the file version in the shi
                m_pShmVersion->setValue(m_CodeVersion);
                // existing shared memory is migrated
                // now we can work with the existing shared memory
            }
            else if( versionMatch == 0 ) {
                // the versions are identical
                // no migration necessary
            }
            else {
                // versionMatch > 0
                // code version is older than the file version
                // we cannot work with the exisiting shared memory

                DBMLog_Logger::getInstance().addMessageList(
                    DBMSrvMsg_Error(
                        DBMSrvMsg_Error::DBMSrv_SHMVERSIONMISMATCH,
                        Msg_Arg("DBMSRV_SHMVERSCURR", bufFileVersion),
                        Msg_Arg("DBMSRV_SHMVERSREQ", bufCodeVersion)), true);
                delete[] bufFileVersion;
                delete[] bufCodeVersion;
                unlockMemory();
                m_LimitedAccess = limitedAccess;
                return;
            }
            delete[] bufFileVersion;
            delete[] bufCodeVersion;
            m_pPlayers->insert(RTESys_ProcessID());
            unlockMemory();
            m_LimitedAccess = limitedAccess;
        }
        // successful end of constructor...
        m_IsValid = true;
    }
}

SAPDB_Byte* DBMSrvShM_Base::openSharedMemory(const char* const aShmId, const unsigned int aSize) {
    int i(10);

    SAPDB_Byte* theShm = (SAPDB_Byte*) sqlAllocSharedMem(aShmId, aSize );
    while( i>0 && theShm==NULL ) {
        // sleep for 50ms maximum
        RTE_ISystem::DoSleep((unsigned int)((double)((50*rand()))/(double)RAND_MAX));
        theShm = (SAPDB_Byte*) sqlAllocSharedMem(aShmId, aSize);
        --i;
    }

    if( theShm == NULL ) {
        char size[32];
        sprintf(size, "%u", aSize);
        DBMLog_Logger::getInstance().addMessageList(
            DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SHMALLOCFAILED, Msg_Arg("DBMSRV_COMP1", aShmId), Msg_Arg("DBMSRV_COMP2", size)), true);
    }

    return theShm;
}

bool DBMSrvShM_Base::attachMemory() {
    DBMSrvShM_HalfSwitcher tmpHS;
    if( m_LengthOperationOpen ) {
        tmpHS = m_HalfSwitcher;
    }
    else {
        m_pSaveHalfSwitcher->getValue(tmpHS);
    }

    unsigned int memSize = (unsigned int)tmpHS.getMemSize();

    m_pSharedMemory = openSharedMemory(m_ShmIDData, memSize);
    m_pData = m_pSharedMemory;
    
    if( m_pSharedMemory == NULL ) {
        m_pLastValidAddress = NULL;
        return false;
    }
    else {
        m_pLastValidAddress = (SAPDB_Byte*)(m_pSharedMemory + (memSize-1));
        return true;
    }
}

void DBMSrvShM_Base::detachMemory() {
    if( m_pSharedMemory != NULL ) {
        unsigned int memSize = (unsigned int)m_HalfSwitcher.getMemSize();

        sqlFreeSharedMem(m_pSharedMemory, memSize);
        m_pSharedMemory = NULL;
        m_pLastValidAddress = NULL;
    }
}

DBMSrvShM_Base::~DBMSrvShM_Base() {
    if( !m_IsValid )
        return;

    if( !m_LimitedAccess )
        detachMemory();

    // quit the game
    bool limitedAccess(m_LimitedAccess);
    m_LimitedAccess = false;
    while( !lockMemory() )
        RTE_ISystem::DoSleep(100);

    m_pPlayers->remove(RTESys_ProcessID());
    unlockMemory();
    m_LimitedAccess = limitedAccess;

    sqlFreeSharedMem(m_pInfoShm, m_SizeInfoShm);

    if( m_pPlayersLocal != NULL )
        delete m_pPlayersLocal;
}

void DBMSrvShM_Base::getPlayerPIDsInto(Tools_Vector<RTE_OSPid>& resultVector) {
    m_pPlayers->getPIDsInto(resultVector);
}

// precondition: the process that runs me has locked it

bool DBMSrvShM_Base::extendMemory(SAPDB_Byte* firstByte, int newBytes) {
    if( m_LimitedAccess )
        return false;

    if( newBytes == 0 ) {
        return true;
    }
    else {
        if( m_LengthOperationOpen ) {
            m_pSaveHalfSwitcher->setValue(m_HalfSwitcher);
            m_LengthOperationOpen = false;
        }

        // hier haben wir also keine length operation offen,
        // d.h. der switcher im shm und der eigene zeigen dasselbe wahre
        size_t bytes((size_t)abs(newBytes));
        size_t copyOffset(firstByte - (m_pData + m_HalfSwitcher.getOffset()));
        size_t oldDataSize(m_HalfSwitcher.getDataSize());
        DBMSrvShM_HalfSwitcher tmpHS;
        tmpHS = m_HalfSwitcher;

        if( !tmpHS.isUpperHalf() ) {
            // copy data to upper half
            ToolsMeMa_GuardedOperation::MemMove(
                    m_pData,
                    m_pData + tmpHS.getOffset(),
                    tmpHS.getDataSize(),
                    __FILE__,
                    __LINE__);
            tmpHS.setMemSize(tmpHS.getDataSize());
            tmpHS.setOffset(0);
            m_pSaveHalfSwitcher->setValue(tmpHS);
        }

        // hier sind wir wieder in der oberen Hlfte und haben keine Lngenoperation offen
        // alles ist sicher und korrekt im shm
        size_t tmpMemSize;
        size_t tmpDataSize;
        size_t tmpOffset;
        if( newBytes > 0 ) {
            tmpMemSize = 2 * tmpHS.getMemSize() + 2 * bytes;
            tmpDataSize = tmpHS.getDataSize() + bytes;
            tmpOffset = tmpHS.getDataSize() + bytes;
        }
        else {
            tmpMemSize = 2 * tmpHS.getMemSize() - bytes;
            tmpDataSize = tmpHS.getDataSize() - bytes;
            tmpOffset = tmpHS.getDataSize();
        }

        tmpHS.setMemSize(tmpMemSize);
        tmpHS.setDataSize(tmpDataSize);
        tmpHS.setOffset(tmpOffset);

        detachMemory(); __shm_crashtest_divideByZero(0, __FILE__);
        m_HalfSwitcher = tmpHS;

        m_LengthOperationOpen = true;
        if( !attachMemory() )
            return false;
        // now shm is open with new size
        // copy everything into lower half
        ToolsMeMa_GuardedOperation::MemMove(
            m_pData + tmpOffset,
            m_pData,
            oldDataSize,
            __FILE__,
            __LINE__);
        // either make room or shrink in lower half
        if( newBytes > 0 ) {
            void* theDest = m_pData + tmpOffset + copyOffset + bytes;
            void* theSource = m_pData + tmpOffset + copyOffset;
            SAPDB_Long theLength = oldDataSize - copyOffset;
            ToolsMeMa_GuardedOperation::MemMove(
                theDest, theSource,  theLength, __FILE__, __LINE__);
            // if we extend the memory, we initialize the new bytes to 0, because otherwise
            // member variables of line layouts have undefined values. This can cause trouble
            // when trying to find e.g. the next line of a type.
            ToolsMeMa_GuardedOperation::MemFill(
                theSource, (SAPDB_Byte)0, (SAPDB_Long)newBytes, __FILE__, __LINE__);
        }
        else {
            ToolsMeMa_GuardedOperation::MemMove(
                m_pData + tmpOffset + copyOffset,
                m_pData + tmpOffset + copyOffset + bytes,
                oldDataSize - copyOffset,
                __FILE__,
                __LINE__);
        }
        return true;
    }
}

bool DBMSrvShM_Base::lockMemory() {
    if( m_LimitedAccess )
        return false;

    bool viaBackingLock(false);
    bool takeOver(false);
    int currentCount(0);

    if( m_pSpinlock->TryLock() ) {
        // we could not get lock
        if( *m_pLockingPid == 0 ) {
            currentCount = *m_pLockCounter;
            RTE_ISystem::DoSleep(SLEEP_TIME);
            if( m_pSpinlock->TryLock() ) {
                // even after waiting, we could not get the lock
                takeOver = (*m_pLockingPid == 0 && currentCount == *m_pLockCounter);
                // takeOver true means the locking process still could not
                // write its pid to shm
            }
            else {
                // now we could get the lock, okay
            }
        }
        // check if locker is still alive
        RTESys_ProcessState procState(*m_pLockingPid);
        if( !takeOver && procState.State() == RTESys_ProcessState::Alive ) {
            // locker is alive, so we wait for the lock to be unlocked
            m_pSpinlock->Lock();
            __shm_crashtest_divideByZero(2, __FILE__);
        }
        else {
            // locker is not alive or stuck while entering its pid
            // so we try to take over the session
            currentCount = *m_pLockCounter;
            m_pBackinglock->Lock();
///////////// crashing between here...
            __shm_crashtest_divideByZero(10, __FILE__);
            if( currentCount != *m_pLockCounter ) {
                // someone else got the lock before me, I give up
                m_pBackinglock->Unlock();
                return false;
            }
            else {
                // okay, nobody else managed to get the lock, we take the shm
                viaBackingLock = true;
            }
        }
    }

    // here I am the only one

    // increment lock counter, never let it be zero (initialization sync would fail)
    (*m_pLockCounter) += (((*m_pLockCounter)+1) == 0)?2:1;

    // copy our pid to shm
    __shm_crashtest_divideByZero(9, __FILE__);
    *m_pLockingPid = RTESys_ProcessID();
    if( viaBackingLock ) {
        m_pBackinglock->Unlock();
///////////////// ... and here can cause the shm to remain locked!!
    }
    __shm_crashtest_divideByZero(1, __FILE__);
    m_pSaveHalfSwitcher->getValue(m_HalfSwitcher);
    return attachMemory();
}

void DBMSrvShM_Base::unlockMemory() {
    if( m_LimitedAccess )
        return;
    
    if( *m_pLockingPid != RTESys_ProcessID() ) {
        char line[32];
        sprintf(line, "%d", __LINE__ + 1);
        Msg_List wrongPid(DBMSrv_DBMError(SHMPANIC), 2, __FILE__, line);
        DBMLog_Logger::getInstance().addMessageList(wrongPid, true);
        return;
    }

    // if size changed, persist switcher
    if( m_LengthOperationOpen ) {
        m_pSaveHalfSwitcher->setValue(m_HalfSwitcher);
        m_LengthOperationOpen = false;
    }

    // detach from memory
    detachMemory();

    // remove my pid
    *m_pLockingPid = 0;
    __shm_crashtest_divideByZero(4, __FILE__);
    m_pSpinlock->Unlock();
}

bool DBMSrvShM_Base::versionOkay() {
    char* currentVersion = this->getActualVersion();
    char* requiredVersion = this->getRequiredVersion();
    bool isOkay(SAPDB_strcmp(currentVersion, requiredVersion) == 0);
    delete [] currentVersion;
    delete [] requiredVersion;
    return isOkay;
}

char* DBMSrvShM_Base::getActualVersion() {
    return getVersion(true);
}

char* DBMSrvShM_Base::getRequiredVersion() {
    return getVersion(false);
}

char* DBMSrvShM_Base::getVersion(bool current) {
    size_t versLen = DBMSrvShM_Base_SHM_VERSION_LENGTH - // contains a byte for the 0
                     strlen(DBMSrvShM_Base_SHM_AUTHORS);
    char* versBuf = new char[versLen];
    if( current ) {
        char buf[DBMSrvShM_Base_SHM_VERSION_LENGTH];
        m_pShmVersion->getValue(buf);
        SAPDB_strncpy(versBuf, buf, versLen-1);
    }
    else
        SAPDB_strncpy(versBuf, m_CodeVersion, versLen-1);
    
    versBuf[versLen-1] = '\0';
    return versBuf;
}

SAPDB_Byte* DBMSrvShM_Base::getStartOf(int lineType) {
    if( m_LimitedAccess )
        return NULL;

    // pointer to place where start is stored (as offset)
    DBMSrvShM_SaveType<size_t>* tmpSaveSO((DBMSrvShM_SaveType<size_t>*)
       (m_pData +
        m_HalfSwitcher.getOffset() + 
        lineType * sizeof(DBMSrvShM_SaveType<size_t>) +
        lineType * 2 * sizeof(DBMSrvShM_SaveType<int>)
        ));

    size_t startOf(0);
    tmpSaveSO->getValue(startOf);
    // return the absolute position in shm plus stored offset
    return m_pData +
           m_HalfSwitcher.getOffset() + 
           NUMBER_OF_LINETYPES * sizeof(DBMSrvShM_SaveType<size_t>) +
           NUMBER_OF_LINETYPES * 2 * sizeof(DBMSrvShM_SaveType<int>) +
           startOf;
}

void DBMSrvShM_Base::shiftStartOf(int lineType, int offset) {
    if( m_LimitedAccess )
        return;

    // shift all line types with higher!! index
    // (stored after!! the lineType)
    for( int i = lineType + 1; i < NUMBER_OF_LINETYPES; i++ ) {
        DBMSrvShM_SaveType<size_t>* tmpSaveSO((DBMSrvShM_SaveType<size_t>*)
        (m_pData +
            m_HalfSwitcher.getOffset() + 
            i * sizeof(DBMSrvShM_SaveType<size_t>) +
            i * 2 * sizeof(DBMSrvShM_SaveType<int>)
        ));
        
        size_t tmpSO;
        tmpSaveSO->getValue(tmpSO);
        tmpSO = (size_t)(((int)tmpSO) + offset);
        tmpSaveSO->setValue(tmpSO);
    }
}

int DBMSrvShM_Base::getNextIndexOf(int lineType) {
    if( m_LimitedAccess )
        return -1;

    int nextIndexOf;
    // pointer to place where index is stored
    DBMSrvShM_SaveType<int>* saveNI((DBMSrvShM_SaveType<int>*)
       (m_pData +
        m_HalfSwitcher.getOffset() + 
        lineType * sizeof(DBMSrvShM_SaveType<size_t>) +
        lineType * 2 * sizeof(DBMSrvShM_SaveType<int>) + 
        sizeof(DBMSrvShM_SaveType<size_t>) + 
        sizeof(DBMSrvShM_SaveType<int>)
        ));

    saveNI->getValue(nextIndexOf);
    return nextIndexOf;
}

void DBMSrvShM_Base::setNextIndexOf(int lineType, int nextIndex) {
    if( m_LimitedAccess )
        return;

    // pointer to place where index is stored
    DBMSrvShM_SaveType<int>* saveNI((DBMSrvShM_SaveType<int>*)
       (m_pData +
        m_HalfSwitcher.getOffset() + 
        lineType * sizeof(DBMSrvShM_SaveType<size_t>) +
        lineType * 2 * sizeof(DBMSrvShM_SaveType<int>) + 
        sizeof(DBMSrvShM_SaveType<size_t>) + 
        sizeof(DBMSrvShM_SaveType<int>)
        ));

    saveNI->setValue(nextIndex);
}

int DBMSrvShM_Base::getExistingLinesOf(int lineType) {
    if( m_LimitedAccess )
        return -1;

    int exLines;
    // pointer to place where number is stored
    DBMSrvShM_SaveType<int>* saveEL((DBMSrvShM_SaveType<int>*)
       (m_pData +
        m_HalfSwitcher.getOffset() + 
        lineType * sizeof(DBMSrvShM_SaveType<size_t>) +
        lineType * 2 * sizeof(DBMSrvShM_SaveType<int>) + 
        sizeof(DBMSrvShM_SaveType<size_t>)
        ));

    saveEL->getValue(exLines);
    return exLines;
}

void DBMSrvShM_Base::setExistingLinesOf(int lineType, int existingLines) {
    if( m_LimitedAccess )
        return;

    // pointer to place where number is stored
    DBMSrvShM_SaveType<int>* saveEL((DBMSrvShM_SaveType<int>*)
       (m_pData +
        m_HalfSwitcher.getOffset() + 
        lineType * sizeof(DBMSrvShM_SaveType<size_t>) +
        lineType * 2 * sizeof(DBMSrvShM_SaveType<int>) + 
        sizeof(DBMSrvShM_SaveType<size_t>)
        ));

    saveEL->setValue(existingLines);
}
