/*!
    @file           DBMSrvShM_LineAccessorEventing.cpp
    @author         MarcW
    @brief          

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/

#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_LineAccessorEventing.hpp"

DBMSrvShM_LineAccessorEventing::DBMSrvShM_LineAccessorEventing(
        const int instanceID)
        : DBMSrvShM_LineAccessorBase(),
          m_InstanceID(instanceID) {

    if( m_ShmAdmin.lock() ) {
        DBMSrvShM_LineLayoutEventing* le =
                m_ShmAdmin.getLineLayoutEventing(instanceID);
        m_Exists = (le != NULL);
        if( m_Exists ) {
            setIndex(le->getIndex());
            m_Validity = true;
        }
        m_ShmAdmin.unlock();
    }
}

bool DBMSrvShM_LineAccessorEventing::exists() {
    if( m_ShmAdmin.lock() ) {
        DBMSrvShM_LineLayoutEventing* le =
                m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
        m_Exists = (le != NULL);
        m_ShmAdmin.unlock();
        m_Validity = true;
    }
    else {
        m_Exists = false;
        m_Validity = false;
    }

    return DBMSrvShM_LineAccessorBase::exists();
}

DBMSrvShM_LineAccessorEventing::DBMSrvShM_LineAccessorEventing(
        const char* dispatcherPID,
        const char* visibilityStatusString,    
        const char* commStatusString,
        const char* dbmUser,
        const char* nodeName,
        const char* configFilePath,
        const char* logFilePath)
      : DBMSrvShM_LineAccessorBase(),
        m_VisibiltyCheck(false) {

    SAPDB_Byte visibilityStatus(atoi(visibilityStatusString));
    SAPDB_Byte commStatus(atoi(commStatusString));

    m_VisibiltyCheck = (visibilityStatus == INTERNAL);

    if( m_ShmAdmin.lock() ) {
        DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(
                m_VisibiltyCheck,
                atoi(dispatcherPID),
                commStatus,
                dbmUser,
                nodeName,
                configFilePath,
                logFilePath);
        m_Exists = (le != NULL);
        if( m_Exists ) {
            setIndex(le->getIndex());
            m_InstanceID = le->getInstanceID();
            m_Validity = true;
        }
        m_ShmAdmin.unlock();    
    }
}

bool DBMSrvShM_LineAccessorEventing::getCommunicationStatus(
        InternalCommStatus& commStatus) {

    DBMSrvShM_LineLayoutEventing* le(NULL);
    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists )
        commStatus = (InternalCommStatus)le->getInternalCommStatus();

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setCommunicationStatus(
        const InternalCommStatus newCommStatus) {
    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists )
        le->setInternalCommStatus((SAPDB_Byte)newCommStatus);
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::getVisibilityStatus(
        VisibilityStatus& visibilityStatus) {

    DBMSrvShM_LineLayoutEventing* le(NULL);
    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists )
        visibilityStatus = (VisibilityStatus)(le->isInternal()?0:1);

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setVisibilityStatus(
        const VisibilityStatus newVisibilityStatus) {

    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists )
        le->setAccessibleFlag((SAPDB_Byte)newVisibilityStatus);
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::getDBMUser(char*& dbmUser) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists ) {
        size_t len = le->getDBMUserLen();
        dbmUser = new char[len+1];
        if( dbmUser != NULL ) {
            SAPDB_memcpy(dbmUser, le->getDBMUser(), len);
            dbmUser[len] = '\0';
        }
    }

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setDBMUser(const char* newDbmUser) {
    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        SAPDB_Byte* start = le->getDBMUser();
        size_t oldLen = le->getDBMUserLen();
        size_t newLen = strlen(newDbmUser);
        m_ShmAdmin.shift(
                DBMSrvShM_ShmAdmin::LineTypeEventingUser,
                start,
                (int)newLen - (int)oldLen);
        le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
        le->setDBMUser(newDbmUser);
    }
    m_ShmAdmin.unlock();
    return true;
}


bool DBMSrvShM_LineAccessorEventing::getNodeName(char*& outNodeName) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists ) {
        size_t len = le->getNodeNameLen();
        outNodeName = new char[len+1];
        if( outNodeName != NULL ) {
            SAPDB_memcpy(outNodeName, le->getNodeName(), len);
            outNodeName[len] = '\0';
        }
    }

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setNodeName(const char* newNodeName) {
    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        SAPDB_Byte* start = le->getNodeName();
        size_t oldLen = le->getNodeNameLen();
        size_t newLen = strlen(newNodeName);
        m_ShmAdmin.shift(
                DBMSrvShM_ShmAdmin::LineTypeEventingUser,
                start,
                (int)newLen - (int)oldLen);
        le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
        le->setNodeName(newNodeName);
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::getConfigFilePath(char*& configFilePath) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists ) {
        size_t len = le->getConfigFilePathLen();
        configFilePath = new char[len+1];
        if( configFilePath != NULL ) {
            SAPDB_memcpy(configFilePath, le->getConfigFilePath(), len);
            configFilePath[len] = '\0';
        }
    }

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setConfigFilePath(const char* newConfigFilePath) {
    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        SAPDB_Byte* start = le->getConfigFilePath();
        size_t oldLen = le->getConfigFilePathLen();
        size_t newLen = strlen(newConfigFilePath);
        m_ShmAdmin.shift(
                DBMSrvShM_ShmAdmin::LineTypeEventingUser,
                start,
                (int)newLen - (int)oldLen);
        le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
        le->setConfigFilePath(newConfigFilePath);
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::getLogFilePath(char*& logFilePath) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    if( !m_ShmAdmin.lock() )
        return false;
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists ) {
        size_t len = le->getLogFilePathLen();
        logFilePath = new char[len+1];
        if( logFilePath != NULL ) {
            SAPDB_memcpy(logFilePath, le->getLogFilePath(), len);
            logFilePath[len] = '\0';
        }
    }

    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setLogFilePath(const char* newLogFilePath) {
    if( !isValid() )
        return false;

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        SAPDB_Byte* start = le->getLogFilePath();
        size_t oldLen = le->getLogFilePathLen();
        size_t newLen = strlen(newLogFilePath);
        m_ShmAdmin.shift(
                DBMSrvShM_ShmAdmin::LineTypeEventingUser,
                start,
                (int)newLen - (int)oldLen);
        le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
        le->setLogFilePath(newLogFilePath);
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::removeBGSRVPid() {

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        le->removeBGSRVPid();
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::setBGSRVPid() {
   
    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);
    if( m_Exists ) {
        le->setBGSRVPid(RTESys_ProcessID());
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::getBGSRVPid( RTE_OSPid& pid) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    if( !m_ShmAdmin.lock() )
        return false;

    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    if( m_Exists )
        pid = le->getBGSRVPid();

    m_ShmAdmin.unlock();
    return m_Exists;
}

bool DBMSrvShM_LineAccessorEventing::doCancel() {

    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);    
    if( m_Exists ) {
        le->doCancel();
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::unCancel() {
    
    if( !m_ShmAdmin.lock() )
        return false;

    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);    
    if( m_Exists ) {
        le->unCancel();
    }
    m_ShmAdmin.unlock();
    return true;
}

bool DBMSrvShM_LineAccessorEventing::isCancelled(bool& iscanc) {
    DBMSrvShM_LineLayoutEventing* le(NULL);

    m_ShmAdmin.lock();
    le = m_ShmAdmin.getLineLayoutEventing(m_InstanceID);
    m_Exists = (le != NULL);

    iscanc = !m_Exists || le->isCancelled();

    m_ShmAdmin.unlock();
    return m_Exists;
}

bool DBMSrvShM_LineAccessorEventing::removeLineEventing() {
 
    if( !m_ShmAdmin.lock() )
        return false;

    m_ShmAdmin.removeLineLayoutEventing(m_InstanceID);
    DBMSrvShM_LineLayoutEventing* le = m_ShmAdmin.getLineLayoutEventing(getIndex());
    m_Exists = (le != NULL);    
    m_ShmAdmin.unlock();
    return true;
}