/*!
    @file           DBMSrvShM_Persistence.cpp
    @author         MarcW
    @brief          description of the shared memory, of maintenance "from outside"

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
*/

#include "heo06.h"

#include "DBM/Srv/Message/DBMSrvMsg_Error.hpp"
#include "DBM/Srv/SharedMemory/DBMSrvShM_Persistence.hpp"

#include "RunTime/System/RTESys_ProcessID.hpp"

DBMSrvShM_Persistence::DBMSrvShM_Persistence(
            const char* const rootDir,
            const char* const infoFile,
            const char* const dataFile) {

    setRootDir(rootDir);
    setInfoFile(infoFile);
    setDataFile(dataFile);
}

void DBMSrvShM_Persistence::setRootDir(const char* const rootDir) {
    m_RootDir = rootDir; //null pointer is allowed
    pathsBuilt = false;
}

void DBMSrvShM_Persistence::setInfoFile(const char* const infoFile) {
    m_InfoFile = infoFile; //null pointer is allowed
    pathsBuilt = false;
}

void DBMSrvShM_Persistence::setDataFile(const char* const dataFile) {
    m_DataFile = dataFile; //null pointer is allowed
    pathsBuilt = false;
}

void DBMSrvShM_Persistence::buildPaths() {
    if( pathsBuilt )
        return;

    m_InfoFileAbs = m_RootDir;
    m_InfoFileAbs += m_InfoFile;
    m_DataFileAbs = m_RootDir;
    m_DataFileAbs += m_DataFile;
    pathsBuilt = true;
}

bool DBMSrvShM_Persistence::exists(bool& doesExist, Msg_List& errList) {
    buildPaths();

    bool errorFound(false);
    tsp05_RteFileInfo info_Info;
    tsp05_RteFileError err_Info;
    tsp05_RteFileInfo info_Data;
    tsp05_RteFileError err_Data;
    char errtxtBuf[41];
    errtxtBuf[40] = '\0';

    sqlfinfoc( m_InfoFileAbs.CharPtr(), &info_Info, &err_Info );
    sqlfinfoc( m_DataFileAbs.CharPtr(), &info_Data, &err_Data );

    errorFound = (err_Info.sp5fe_result == vf_notok);
    if( errorFound ) {
        // report: error checking info file
        errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_FILECHK,
                                         Msg_Arg("DBMSRV_COMP1", m_InfoFileAbs.CharPtr()),
                                         Msg_Arg("DBMSRV_COMP2", err_Info.sp5fe_text.asCharp())));
    }

    errorFound = errorFound || (err_Data.sp5fe_result == vf_notok);
    if( err_Data.sp5fe_result == vf_notok ) {
        // report: error checking data file
        errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_FILECHK,
                                         Msg_Arg("DBMSRV_COMP1", m_DataFileAbs.CharPtr()),
                                         Msg_Arg("DBMSRV_COMP2", err_Info.sp5fe_text.asCharp())));
    }

    if( !errorFound ) {
        /*
        we cannot check this since if someone else has opened the shared memory,
        it is not writable for us...
        if( info_Info.sp5fi_exists &&
            info_Data.sp5fi_exists &&
            (!info_Info.sp5fi_readable ||
            !info_Info.sp5fi_writeable ||
            !info_Data.sp5fi_readable ||
            !info_Data.sp5fi_writeable) ) {
            // shm cannot be accessed correctly
            errorFound = true;
            errList = errList +
                DBMSrvMsg_Error(
                    DBMSrvMsg_ Error::SHMFILEPERM,
                    m_InfoFileAbs.CharPtr(),
                    m_DataFileAbs.CharPtr());
        }
        else
        */
        if( info_Info.sp5fi_exists && !info_Data.sp5fi_exists ) {
            // report: only info file exists
            errorFound = true;
            errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SHMNODATAFILE));
        }
        else if( !info_Info.sp5fi_exists && info_Data.sp5fi_exists ) {
            // report: only data file exists
            errorFound = true;
            errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_SHMNOINFOFILE));
        }
        else
             doesExist = info_Info.sp5fi_exists != 0;
    }

    return !errorFound;
}

bool DBMSrvShM_Persistence::remove(Msg_List& errList) {
    buildPaths();
    
    bool errorFound(false);
    tsp05_RteFileError err_Info;
    tsp05_RteFileError err_Data;

    sqlferasec( m_InfoFileAbs.CharPtr(), &err_Info );
    sqlferasec( m_DataFileAbs.CharPtr(), &err_Data );

    errorFound = err_Info.sp5fe_result == vf_notok;
    if( errorFound ) {
        // report: could not remove info file
        errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_FILEDEL,
                                         Msg_Arg("DBMSRV_COMP1", m_InfoFileAbs.CharPtr()),
                                         Msg_Arg("DBMSRV_COMP2", err_Info.sp5fe_text.asCharp())));
    }

    errorFound = errorFound || err_Data.sp5fe_result == vf_notok;
    if( err_Data.sp5fe_result == vf_notok ) {
        // report: cound not remove data file
        errList.Overrule(DBMSrvMsg_Error(DBMSrvMsg_Error::DBMSrv_FILEDEL,
                                         Msg_Arg("DBMSRV_COMP1",m_DataFileAbs.CharPtr()),
                                         Msg_Arg("DBMSRV_COMP2", err_Data.sp5fe_text.asCharp())));
    }

    return !errorFound;
}
