/*!
    @file           DBMSrvShM_Shmadmin.cpp
    @author         MarcW
    @brief          

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*/
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_ShmAdmin.hpp"
#include "SAPDBCommon/SAPDB_string.h"
#include "ToolsCommon/MemoryManagement/ToolsMeMa_GuardedOperation.hpp"

char*               DBMSrvShM_ShmAdmin::shmDir(NULL);
DBMSrvShM_ShmAdmin* DBMSrvShM_ShmAdmin::pInstance(NULL);
const int           DBMSrvShM_ShmAdmin::maxAllowedLinesEventingUser(7);
const int           DBMSrvShM_ShmAdmin::maxAllowedLinesEventingSilent(1);
const int           DBMSrvShM_ShmAdmin::maxAllowedLinesResourceLock(100);
const int           DBMSrvShM_ShmAdmin::maxAllowedLinesScheduler(1);

DBMSrvShM_Persistence* DBMSrvShM_ShmAdmin::getShmPersistence() {
    return new DBMSrvShM_Persistence("", DBMSrvShM_Base::SHM_SIZEINFO, DBMSrvShM_Base::SHM_NAME);
}

DBMSrvShM_ShmAdmin::DBMSrvShM_ShmAdmin(const bool isInfoInstance )
        : shm(NULL),
          m_ScopeMemberCheckOnly(LineTypeUninitialized) {
    if( shmDir != NULL )
        shm = new DBMSrvShM_Base(shmDir, isInfoInstance);
}

bool DBMSrvShM_ShmAdmin::shmVersionOkay() {
    return shm->versionOkay();
}

void DBMSrvShM_ShmAdmin::setShmDir(const char* aDir) {
    shmDir = new char[strlen(aDir)+1];
    SAPDB_strcpy(shmDir, aDir);
}

DBMSrvShM_ShmAdmin& DBMSrvShM_ShmAdmin::Instance() {
    if( pInstance == NULL )
        pInstance = new DBMSrvShM_ShmAdmin(false);

    return *pInstance;
}

DBMSrvShM_ShmAdmin& DBMSrvShM_ShmAdmin::InfoInstance() {
    if( pInstance == NULL )
        pInstance = new DBMSrvShM_ShmAdmin(true);

    return *pInstance;
}

bool DBMSrvShM_ShmAdmin::getLockStatus(
            bool& normalLockLocked,
            bool& backingLockLocked,
            RTE_OSPid& lockingPid ) {
    return shm->getLockStatus(normalLockLocked, backingLockLocked, lockingPid);
}

bool DBMSrvShM_ShmAdmin::lock() {
    if( !shm->isValid() )
        return false;

    int attempts(10);
    while ( attempts > 0) {
        if( shm->lockMemory() )
            return true;
        else
            --attempts;
    }
    return false;
}

void DBMSrvShM_ShmAdmin::unlock() {
    shm->unlockMemory();
}

bool DBMSrvShM_ShmAdmin::resolveDeadlock() {
    return shm->resolveDeadlockDirty();
}

void DBMSrvShM_ShmAdmin::shift(DBMSrvShM_ShmAdmin::LineType lineType, SAPDB_Byte* firstByte, int offset){

    // extend memory (or shrink)
    shm->extendMemory(firstByte, offset);
    shm->shiftStartOf(lineType, offset);

    m_ScopeMemberCheckOnly = lineType;
}

DBMSrvShM_LineLayoutResourceLock* DBMSrvShM_ShmAdmin::getLineLayoutResourceLock
                (const char* resourceName) {

    int existingLines(shm->getExistingLinesOf(LineTypeResourceLock));
    if( existingLines > maxAllowedLinesResourceLock )
        // should never happen
        return NULL;
    DBMSrvShM_LineLayoutResourceLock* lyt(NULL);
    int nextIndex(shm->getNextIndexOf(LineTypeResourceLock));
    int count = existingLines;
    lyt = (DBMSrvShM_LineLayoutResourceLock*)shm->getStartOf(LineTypeResourceLock);
    
    while(
        count>0 &&
        isInScope(lyt) &&
        !(strlen(resourceName) == lyt->getResourceNameLen() &&
          0 == strncmp(resourceName, (const char*)lyt->getResourceName(), lyt->getResourceNameLen())) ) {

        lyt = lyt->getNext();
        --count;
    }
          
    if( count > 0 ) {
        // a line for this resource exists already
        if( isInScope(lyt) )
            return lyt;
        else
            return NULL;
    }
    else {
        // this line does not exist, so create a new one
        // make room for new line
        shift(  LineTypeResourceLock,
                (SAPDB_Byte*)lyt,
                (int)(sizeof(DBMSrvShM_LineLayoutResourceLock) + strlen(resourceName)));
        // update admin information
        shm->setNextIndexOf(LineTypeResourceLock, nextIndex+1);
        shm->setExistingLinesOf(LineTypeResourceLock, existingLines+1);

        count = existingLines;
        lyt = (DBMSrvShM_LineLayoutResourceLock*)shm->getStartOf(LineTypeResourceLock);
        
        for( ; count>0 && isInScope(lyt); --count)
            lyt = lyt->getNext();

        if( count > 0 || !isInScope(lyt) )
            // some existing layout on the way is outof scope
            return NULL;
        else {
            lyt->setIndex(nextIndex);
            lyt->setResourceName(resourceName);
            lyt->setBGSRVPid(0);
            return lyt;
        }
    }
}

void DBMSrvShM_ShmAdmin::removeLineLayoutResourceLock
                (const char* resourceName) {

    int existingLines(shm->getExistingLinesOf(LineTypeResourceLock));
    DBMSrvShM_LineLayoutResourceLock* lyt(NULL);

    int count = existingLines;
    lyt = (DBMSrvShM_LineLayoutResourceLock*)shm->getStartOf(LineTypeResourceLock);

    while(
        count>0 &&
        isInScope(lyt) &&
        !(strlen(resourceName) == lyt->getResourceNameLen() &&
          0 == strncmp( resourceName, (const char*)lyt->getResourceName(), lyt->getResourceNameLen())) ) {
        lyt = lyt->getNext();
        count--;
    }
    
    if( count > 0 && isInScope(lyt) ) {
        // a line for this resource exists, we must remove it
        shift(LineTypeResourceLock, (SAPDB_Byte*)lyt, -1 * (int)((SAPDB_Byte*)lyt->getNext()-(SAPDB_Byte*)lyt) );
        existingLines--;
        shm->setExistingLinesOf(LineTypeResourceLock, existingLines);
    }
}


DBMSrvShM_LineLayoutEventing* DBMSrvShM_ShmAdmin::getLineLayoutEventing(int instanceID ) {
    int existingLines(shm->getExistingLinesOf(LineTypeEventingUser));
    if( existingLines == 0 )
        return NULL;
    
    DBMSrvShM_LineLayoutEventing* lyt(NULL);
    lyt = (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingUser);
    for( ; existingLines>0 && isInScope(lyt); --existingLines) {
        if( lyt->getInstanceID() == instanceID )
            return lyt;
        lyt = lyt->getNext();
    }
    return NULL;
}

void DBMSrvShM_ShmAdmin::removeLineLayoutEventing(int instanceID) {
    DBMSrvShM_LineLayoutEventing* lyt(getLineLayoutEventing(instanceID));

    if( !isInScope(lyt) )
        return;

    // line exists, we must remove it
    shift(LineTypeEventingUser, (SAPDB_Byte*)lyt, -1 * (int)((SAPDB_Byte*)lyt->getNext()-(SAPDB_Byte*)lyt) );
    int existingLines(shm->getExistingLinesOf(LineTypeEventingUser));
    --existingLines;
    shm->setExistingLinesOf(LineTypeEventingUser, existingLines);
}

DBMSrvShM_LineLayoutEventing* DBMSrvShM_ShmAdmin::getLineLayoutEventing(
        bool& visibilityCheck,
        int dispatcherPID,
        SAPDB_Byte commStatus,
        const char* dbmUser,
        const char* nodeName,
        const char* configFilePath,
        const char* logFilePath) {

    int existingLines(shm->getExistingLinesOf(LineTypeEventingUser));

    if( existingLines > maxAllowedLinesEventingUser )
        // should never happen
        return NULL;

    bool makeInternal(visibilityCheck);

    DBMSrvShM_LineLayoutEventing* lyt(NULL);
    int nextIndex(shm->getNextIndexOf(LineTypeEventingUser));

    int lastInstanceID(-1); // this will be the ID of the new instance
    int currInstanceID(-1);
    bool slotFound(false);

    int count = existingLines;
    lyt = (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingUser);
    for( ; count>0 && isInScope(lyt); count--) {
        if( visibilityCheck && lyt->isInternal() )
            return NULL;

        currInstanceID = lyt->getInstanceID();
        if( !slotFound ) {
            if( currInstanceID > lastInstanceID+1 ) {
                // found a slot
                slotFound = true;
                ++lastInstanceID;
            }
            else
                lastInstanceID = currInstanceID;
        }
        lyt = lyt->getNext();
    }
    
    visibilityCheck = false;

    if( !slotFound )
        ++lastInstanceID;

    if( count>0 )
        return NULL;

    // now lyt points behind last eventing eventing line layout

    // make room for new line
    shift(
            LineTypeEventingUser,
            (SAPDB_Byte*)lyt,
            (int)(sizeof(DBMSrvShM_LineLayoutEventing) +
                    strlen(dbmUser) + strlen(nodeName) + strlen(configFilePath) + strlen(logFilePath)));
    
    // update admin information
    shm->setNextIndexOf(LineTypeEventingUser, nextIndex+1);
    shm->setExistingLinesOf(LineTypeEventingUser, existingLines+1);

    count = existingLines;
    lyt = (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingUser);
    for( ; count>0 && isInScope(lyt); count--)
        lyt = lyt->getNext();

    if( count>0 )
        return NULL;

    lyt->setIndex(nextIndex);
    lyt->setAccessibleFlag(makeInternal?0:1);
    lyt->setInstanceID(lastInstanceID);
    lyt->setBGSRVPid((RTE_OSPid)dispatcherPID);
    lyt->setInternalCommStatus(commStatus);
    lyt->unCancel();
    lyt->setDBMUser(dbmUser);
    lyt->setNodeName( nodeName );
    lyt->setLogFilePath( logFilePath );
    lyt->setConfigFilePath( configFilePath );

    return lyt;
}

DBMSrvShM_LineLayoutEventing* DBMSrvShM_ShmAdmin::getFirstLineLayoutEventing() {
    int existingLines(shm->getExistingLinesOf(LineTypeEventingUser));
    if( existingLines == 0 )
        return NULL;
    else
        return (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingUser);
}

DBMSrvShM_LineLayoutEventing* DBMSrvShM_ShmAdmin::getNextLineLayoutEventing(
        const DBMSrvShM_LineLayoutEventing* aLayout) {

    DBMSrvShM_LineLayoutEventing* next = aLayout->getNext();
    if( (SAPDB_Byte*)next >= shm->getStartOf(LineTypeEventingSilent) ) 
        return NULL;
    else
        return aLayout->getNext();
}


DBMSrvShM_LineLayoutEventing* DBMSrvShM_ShmAdmin::getLineLayoutEventingSilent( bool createIfNecessary ) {
    int existingLines(shm->getExistingLinesOf(LineTypeEventingSilent));

    if( !createIfNecessary && (existingLines == 0) )
        return NULL;

    if( existingLines > maxAllowedLinesEventingSilent )
        // should never happen
        return NULL;

    DBMSrvShM_LineLayoutEventing* lyt(NULL);
    if( existingLines == maxAllowedLinesEventingSilent ) {
        // it exists, so return it
        lyt = (DBMSrvShM_LineLayoutEventing*) shm->getStartOf(LineTypeEventingSilent);
        if( isInScope (lyt) )
            return lyt;
        else
            return NULL;
    }

    int nextIndex(shm->getNextIndexOf(LineTypeEventingSilent));

    int count = existingLines;
    lyt = (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingSilent);
    for( ; count>0 && isInScope(lyt); count--)
        lyt = lyt->getNext();

    if( count>0 )
        return NULL;

    // now lyt points behind last eventing line layout

    // make room for new line
    shift(LineTypeEventingSilent, (SAPDB_Byte*)lyt, sizeof(DBMSrvShM_LineLayoutEventing));
    // update admin information
    shm->setNextIndexOf(LineTypeEventingSilent, nextIndex+1);
    shm->setExistingLinesOf(LineTypeEventingSilent, existingLines+1);

    count = existingLines;
    lyt = (DBMSrvShM_LineLayoutEventing*)shm->getStartOf(LineTypeEventingSilent);
    for( ; count>0 && isInScope(lyt); count--)
        lyt = lyt->getNext();

    if( count>0 )
        return NULL;

    lyt->setIndex(nextIndex);
    lyt->unCancel();
    lyt->setNodeName( "" );
    lyt->setLogFilePath( "" );
    lyt->setConfigFilePath( "" );
    return lyt;
}

void DBMSrvShM_ShmAdmin::removeLineLayoutEventingSilent() {
    DBMSrvShM_LineLayoutEventing* lyt(getLineLayoutEventingSilent(false));
    if( !isInScope(lyt) )
        return;

    // line exists, we must remove it
    shift(LineTypeEventingSilent, (SAPDB_Byte*)lyt, -1 * (int)((SAPDB_Byte*)lyt->getNext()-(SAPDB_Byte*)lyt) );
    int existingLines(shm->getExistingLinesOf(LineTypeEventingSilent));
    existingLines--;
    shm->setExistingLinesOf(LineTypeEventingSilent, existingLines);
}

DBMSrvShM_LineLayoutScheduler* DBMSrvShM_ShmAdmin::getLineLayoutScheduler( bool createIfNecessary ) {
    int existingLines(shm->getExistingLinesOf(LineTypeScheduler));

    if( !createIfNecessary && (existingLines == 0) )
        return NULL;

    if( existingLines > maxAllowedLinesScheduler )
        // should never happen
        return NULL;

    DBMSrvShM_LineLayoutScheduler* lyt(NULL);
    if( existingLines == maxAllowedLinesScheduler ) {
        // it exists, so return it
        lyt = (DBMSrvShM_LineLayoutScheduler*) shm->getStartOf(LineTypeScheduler);
        if( isInScope(lyt) )
            return lyt;
        else
            return NULL;
    }

    int nextIndex(shm->getNextIndexOf(LineTypeScheduler));

    int count = existingLines;
    lyt = (DBMSrvShM_LineLayoutScheduler*)shm->getStartOf(LineTypeScheduler);
    for( ; count>0 && isInScope(lyt); count--)
        lyt = lyt->getNext();

    if( count>0 )
        return NULL;

    // now lyt points behind last line layout of preceding type

    // make room for new line
    shift(LineTypeScheduler, (SAPDB_Byte*)lyt, sizeof(DBMSrvShM_LineLayoutScheduler));
    // update admin information
    shm->setNextIndexOf(LineTypeScheduler, nextIndex+1);
    shm->setExistingLinesOf(LineTypeScheduler, existingLines+1);

    count = existingLines;
    lyt = (DBMSrvShM_LineLayoutScheduler*)shm->getStartOf(LineTypeScheduler);
    for( ; count>0 && isInScope(lyt); count--)
        lyt = lyt->getNext();

    if( count>0 )
        return NULL;

    lyt->setIndex(nextIndex);
    lyt->setBGSRVPid(0);
    lyt->setInfoType(0);
    lyt->setRecipient(0);
    lyt->setVariablePart("");
    lyt->unCancel();
    return lyt;
}

void DBMSrvShM_ShmAdmin::removeLineLayoutScheduler() {
    DBMSrvShM_LineLayoutScheduler* lyt(getLineLayoutScheduler(false));
    if( lyt == NULL )
        return;
    
    if( !isInScope(lyt) )
        return;

    // line exists, we must remove it
    shift(LineTypeScheduler, (SAPDB_Byte*)lyt, -1 * (int)((SAPDB_Byte*)lyt->getNext()-(SAPDB_Byte*)lyt) );
    int existingLines(shm->getExistingLinesOf(LineTypeScheduler));
    --existingLines;
    shm->setExistingLinesOf(LineTypeScheduler, existingLines);
}

bool DBMSrvShM_ShmAdmin::isInScope( DBMSrvShM_LineLayoutResourceLock * aLayout) {

    do {
        const SAPDB_Byte* lastByte(NULL);

        // common base check
        if( !scopeBaseCheck(aLayout, lastByte) )
            break;

        // check if the member variables are in scope
        if( (SAPDB_Byte*)((SAPDB_Byte*)aLayout + (sizeof(*aLayout)-1)) > lastByte )
            break;

        if( m_ScopeMemberCheckOnly == DBMSrvShM_ShmAdmin::LineTypeResourceLock ) {
            m_ScopeMemberCheckOnly = DBMSrvShM_ShmAdmin::LineTypeUninitialized;
            return true;
        }

        // now we know we can access the members
        if( (SAPDB_Byte*)((SAPDB_Byte*)aLayout + sizeof(*aLayout)-1 + aLayout->getResourceNameLen()) > lastByte )
            break;

        return true;
    } while(true);

    return false;
}

bool DBMSrvShM_ShmAdmin::isInScope( DBMSrvShM_LineLayoutEventing * aLayout) {
    
    do {
        const SAPDB_Byte* lastByte(NULL);

        // common base check
        if( !scopeBaseCheck(aLayout, lastByte) )
            break;

        // check if the member variables are in scope
        if( (SAPDB_Byte*)((SAPDB_Byte*)aLayout + (sizeof(*aLayout)-1)) > lastByte )
            break;

        if( m_ScopeMemberCheckOnly == DBMSrvShM_ShmAdmin::LineTypeEventingSilent ||
            m_ScopeMemberCheckOnly == DBMSrvShM_ShmAdmin::LineTypeEventingUser) {

            m_ScopeMemberCheckOnly = DBMSrvShM_ShmAdmin::LineTypeUninitialized;
            return true;
        }

        // now we know we can access the members
            if( (SAPDB_Byte*)(
                    (SAPDB_Byte*)aLayout +
                    sizeof(*aLayout)-1 +
                    aLayout->getDBMUserLen() +
                    aLayout->getNodeNameLen() +
                    aLayout->getConfigFilePathLen() +
                    aLayout->getLogFilePathLen()) > lastByte )
            break;
        
        // all checks passed
        return true;
    } while(true);

    return false;
}

bool DBMSrvShM_ShmAdmin::isInScope( DBMSrvShM_LineLayoutScheduler * aLayout) {
    
    do {
        const SAPDB_Byte* lastByte(NULL);

        // common base check
        if( !scopeBaseCheck(aLayout, lastByte) )
            break;

        // check if the member variables are in scope
        if( (SAPDB_Byte*)((SAPDB_Byte*)aLayout + (sizeof(*aLayout)-1)) > lastByte )
            break;

        if( m_ScopeMemberCheckOnly == DBMSrvShM_ShmAdmin::LineTypeScheduler ) {
            m_ScopeMemberCheckOnly = DBMSrvShM_ShmAdmin::LineTypeUninitialized;
            return true;
        }

        // now we know we can access the members
        if( (SAPDB_Byte*)((SAPDB_Byte*)aLayout + sizeof(*aLayout)-1 + aLayout->getVariablePartLen() + aLayout->getUserNameLen()) > lastByte )
            break;
        
        // all checks passed
        return true;
    } while(true);

    return false;
}

bool DBMSrvShM_ShmAdmin::scopeBaseCheck(
            DBMSrvShM_LineLayoutBase * aLayout,
            const SAPDB_Byte*& shmLastByte) {
    // check layout and shm pointers themselves
    if( aLayout == NULL || shm == NULL )
        return false;

    const SAPDB_Byte* shmFirstByte(NULL);
    // get the current memory borders
    shm->getMemoryBorders(shmFirstByte, shmLastByte);

    // check if there is shm allocated
    if( shmFirstByte == NULL || shmLastByte == NULL )
        return false;

    // check if the start of the layout is in scope   
    return ((SAPDB_Byte*)aLayout >= shmFirstByte) && ((SAPDB_Byte*)aLayout <= shmLastByte);
}