/*!
    @file           DBMSrvShM_ShmAdmin.hpp
    @author         MarcW
    @brief          shared memory administration class

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

\endif
*/

#if !defined(DBMSrvShM_ShmAdmin_HPP)
#define DBMSrvShM_ShmAdmin_HPP

#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_Base.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_LineLayoutEventing.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_LineLayoutResourceLock.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_LineLayoutScheduler.hpp"
#include "SAPDB/DBM/Srv/SharedMemory/DBMSrvShM_Persistence.hpp"

/*!
    @brief DBMServer shm communication administration class

    Each process has a singleton of this class. Access to shm from higher level classes
    is only though this singleton. Creation, deletion and modifying of lines of each type
    is supported by according methods for each type of line.
    @see DBMSrvShM_Base
*/
class DBMSrvShM_ShmAdmin {
public:
    static DBMSrvShM_Persistence* getShmPersistence();

    /*!
        @brief status check
        @return true if there is an instance, false otherwise
    */
    static bool isAttached() {return pInstance != NULL;}

    // all different line types, must be consistent with max number of lines types!!
    enum LineType {
        LineTypeEventingUser,
        LineTypeEventingSilent,
        LineTypeResourceLock,
        LineTypeScheduler,

        LineTypeUninitialized
    };


    /*!
        @brief set folder in which shm file is to be created

        @param aDir [IN] folder in which shared memory file is to be created
    */
    static void setShmDir(const char* aDir);

    /*!
        @brief free memory for the shmDir path

        at the moment I don't know when to call this...
    */
    static void removeShmDir() {
        delete[] shmDir;
    }

    /*! 
        @brief get a reference to the only instance

        @return reference to the only instance
    */
    static DBMSrvShM_ShmAdmin& Instance();

    static DBMSrvShM_ShmAdmin& InfoInstance();

    static void deleteInstance() {
        if(pInstance!=NULL) {
            delete pInstance;
            pInstance = NULL;
        }
    }

    /*!
        @brief version check 
        
        @return true, if the current program can work with the shm that's used
    */
    bool shmVersionOkay();

    char* getActualVersion() {
        return shm->getActualVersion();
    }

    char* getRequiredVersion() {
        return shm->getRequiredVersion();
    }

    bool isValid() {
        return shm != NULL && shm->isValid();
    }

    bool getLockStatus(
            bool& normalLockLocked,
            bool& backingLockLocked,
            RTE_OSPid& lockingPid );

    void getParticipantPIDsInto(Tools_Vector<RTE_OSPid>& resultVector) {
        shm->getPlayerPIDsInto(resultVector);
    }

    /*!
        @brief lock shm

        @return true, if exclusive access to the shm is guaranteed, else false
    */
    bool lock();

    /*!
        @brief unlock shm

        Release exclusive access to the shm.
    */
    void unlock();

    bool resolveDeadlock();

    /*!
        @brief change shm size

        @see DBMSrvShM_Base.extendMemory()
    */
    void shift(DBMSrvShM_ShmAdmin::LineType lineType, SAPDB_Byte* firstByte, int offset);

    /*! @brief destructor */
    ~DBMSrvShM_ShmAdmin() {
        delete shm;
    }

    /*
     *************** line type specifics
     */

    /*  schduler, one line at max */
    DBMSrvShM_LineLayoutScheduler* getLineLayoutScheduler(bool createIfNecessary);
    void removeLineLayoutScheduler();    

    /* resource locks */
    DBMSrvShM_LineLayoutResourceLock* getLineLayoutResourceLock(const char* resourceName);
    void removeLineLayoutResourceLock(const char* resourceName);

    /*  silent eventing, one line at max */
    DBMSrvShM_LineLayoutEventing* getLineLayoutEventingSilent(bool createIfNecessary);
    void removeLineLayoutEventingSilent();    

    /* eventing */

    /*!
        @brief return pointer to specific line in shm of type eventing

        A pointer to an existing line is returned.
        @param index [IN] id of line to return
        @return pointer to line of type eventing in shm with id index
    */
    DBMSrvShM_LineLayoutEventing* getLineLayoutEventing(int InstanceID);

    /*!
        @brief get a new line in shm of type eventing

        A new line with the passed filepath is created in shm and a pointer to it is returned.
        @param filePath [IN] filepath to store in the new line
        @return pointer to the new line
    */
    DBMSrvShM_LineLayoutEventing* getLineLayoutEventing(
            bool& visibilityCheck,
            int dispatcherPID,
            SAPDB_Byte commStatus,
            const char* dbmUser,
            const char* nodeName,
            const char* configFilePath,
            const char* logFilePath);
    /*!
        @brief remove specific line in shm of type eventing

        The line of type eventing is removed from shm if it exists.
        @param index [IN] id of line to remove
    */
    void removeLineLayoutEventing(int index);

    /*!
        @brief get pointer to first line of type eventing in shm

        If no line of type eventing exists, this pointer is not valid!
        @return pointer to first line of type eventing in shm
    */
    DBMSrvShM_LineLayoutEventing* getFirstLineLayoutEventing();

    DBMSrvShM_LineLayoutEventing* getNextLineLayoutEventing( const DBMSrvShM_LineLayoutEventing* aLayout);

    /*!
        @brief get number of existing of type eventing

        @return number of existing of type eventing
    */
    int getExistingLinesEventing() { return shm->getExistingLinesOf(LineTypeEventingUser); }

private:
    static char* shmDir;
    /*! @brief the one and only instance */
    static DBMSrvShM_ShmAdmin* pInstance;

    DBMSrvShM_ShmAdmin(const bool isInfoInstance);

    DBMSrvShM_Base* shm;

    /*
     *************** line type specifics
     */

    /* eventing */
    static const int maxAllowedLinesEventingUser;

    /* "silent" eventing (eventing hidden from user, used for other tasks) */
    static const int maxAllowedLinesEventingSilent;

    /* exclusive access to resources */
    static const int maxAllowedLinesResourceLock;

    /* communication with the scheduler */
    static const int maxAllowedLinesScheduler;

    LineType m_ScopeMemberCheckOnly;

    // check where a layout is completely in shm (including variable part!!)
    // for each line layout type we need an extra method
    bool isInScope( DBMSrvShM_LineLayoutScheduler * aLayout);
    bool isInScope( DBMSrvShM_LineLayoutResourceLock * aLayout);
    bool isInScope( DBMSrvShM_LineLayoutEventing * aLayout);

    // called from the isInScope methods
    bool scopeBaseCheck(
        DBMSrvShM_LineLayoutBase * aLayout,
        const SAPDB_Byte*& shmLastByte);
};

#endif //DBMSrvShM_ShmAdmin_HPP
