/*****************************************************************************/
/*!
  @file         Data_PrimTreeStatisticsCoordinator.cpp
  @author       MartinKi
  @ingroup

  @brief        Implementation file for class
                Data_PrimTreeStatisticsCoordinator.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2003 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"
#include "SQLManager/SQLMan_UpdateStatistics.hpp"

#include "DataAccess/Data_PrimTreeStatisticsCoordinator.hpp"
#include "DataAccess/Data_UpdStatRequestFile.hpp"
#include "DataAccess/Data_Types.hpp"
#include "SAPDBCommon/Messages/SDBMsg_Data.h"

#include "ServerTasks/SrvTasks_JobCollectTableStatistics.hpp"
#include "ServerTasks/SrvTasks_Job.hpp"
#include "ServerTasks/SrvTasks_JobList.hpp"
#include "ServerTasks/SrvTasks_ICoordinator.hpp"

#include "FileDirectory/FileDir_IDirectory.hpp"
#include "FileDirectory/FileDir_Table.hpp"
#include "FileDirectory/FileDir_FileNo.hpp"

#include "Container/Container_Vector.hpp"

// Content of include files
#include "hbd998.h"
#include "hbd01.h"  // PASCAL: Configuration_Parameter

#include "Trace/Trace_Entry.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDB/SAPDBCommon/SAPDB_RangeCode.hpp" // Kernel_move_and_fill

#include "Messages/Msg_List.hpp"

#include "hgg04.h" // g04build_temp_tree_id
#include "KernelCommon/Kernel_VTrace.hpp"

#include "RunTime/RTE_Message.hpp"
#include "RunTime/Tasking/RTETask_ITask.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

typedef SrvTasks_JobList<SrvTasks_JobCollectTableStatistics> Srv_StatJobList;

/*========================================================================*
 *  EXTERNAL VARIABLES                                                    *
 *========================================================================*/
/*========================================================================*
 *  GLOBAL VARIABLES                                                      *
 *========================================================================*/
/*========================================================================*
 *  LOCAL VARIABLES                                                       *
 *========================================================================*/
class JobQueue;

class JobInfo {
    SrvTasks_JobCollectTableStatistics* m_job;
    SAPDB_Bool     m_isLastRun;
    SAPDB_Bool     m_isRunning;
    SAPDB_Int2     m_auxFileIndex;

    friend class JobQueue;

public:
    JobInfo(
        SrvTasks_JobCollectTableStatistics& job,
        const SAPDB_Bool                    isLastRun)
        : m_job( &job ),
          m_isLastRun( isLastRun ),
          m_isRunning( false ),
          m_auxFileIndex( 0 )
    {}

    /****************************************************************/

    SAPDB_Bool IsValid() const
    {
        return m_job->GetTableKey().IsValid();
    }

    /****************************************************************/

    const Data_PrimTreeStatisticsCoordinator::TableKey& GetTableKey() const
    {
        return m_job->GetTableKey();
    }

    /****************************************************************/

    SrvTasks_JobCollectTableStatistics& GetJob()
    {
        return *m_job;
    }

    /****************************************************************/

    const FileDir_FileNo& GetAuxFileNo() const
    {
        return m_job->GetAuxFileNo();
    }

    /****************************************************************/

    SAPDB_Int2 GetAuxFileIndex() const
    {
        return m_auxFileIndex;
    }

    /****************************************************************/

    SAPDB_Int4 GetJobId()
    {
        return m_job->GetId();
    }

    /****************************************************************/

    void SetIsRunning(SAPDB_Bool isRunning)
    {
        m_isRunning = isRunning;
    }

    /****************************************************************/

    SAPDB_Bool IsRunning() const
    {
        return m_isRunning;
    }

    /****************************************************************/

    const JobInfo& operator=(const JobInfo& ji)
    {
        this->m_job          = ji.m_job;
        this->m_isLastRun    = ji.m_isLastRun;
        this->m_isRunning    = ji.m_isRunning;
        this->m_auxFileIndex = ji.m_auxFileIndex;

        return *this;
    }
};

/*************************************************************************/

class JobQueue
{
    typedef Data_UpdStatRequestFile::StatisticsRequestInfo RequestInfo;
    typedef Container_Vector<JobInfo> RunQueue;

    Data_UpdStatRequestFile& m_requestFile;
    Data_UpdStatRequestFile::Iterator m_iterator;
    Data_PrimTreeStatisticsCoordinator::TableKey m_tableKey;
    FileDir_FileNo m_curFileNo;
    RunQueue m_runQueue;
    SAPDB_Int4 m_taskIdCounter;
    SAPDB_Int2 m_maxRunQueueLength;
    tgg00_TransContext& m_trans;

public:
    JobQueue(
        tgg00_TransContext&      trans,
        Data_UpdStatRequestFile& requestFile,
        SAPDB_Int2               maxServerTasks,
        Msg_List&                errMsg)
        : m_trans( trans ),
          m_requestFile( requestFile ),
          m_runQueue( *reinterpret_cast<SAPDBMem_IRawAllocator*>
                      ( trans.trAllocator_gg00 ) ),
          m_taskIdCounter( 0 ),
          m_maxRunQueueLength( maxServerTasks ),
          m_iterator ( requestFile.GetIterator( trans ) )
    {
        Data_UpdStatRequestFile::Iterator::ReturnCode itRetCode;
        itRetCode = m_iterator.Next(trans, errMsg );

        if ( itRetCode == Data_UpdStatRequestFile::Iterator::Okay ) {
            RequestInfo& curRequest =
                m_iterator.GetStatisticsRequest( errMsg );
            m_curFileNo = curRequest.tableFileNo;
            m_tableKey = m_iterator.GetTableKey();
        }
        m_runQueue.Reserve( m_maxRunQueueLength );
    }

    /****************************************************************/

    ~JobQueue()
    {
        tgg00_Surrogate surr;
        tgg00_BasisError auxErr = m_trans.trError_gg00;
        RunQueue::Iterator it = m_runQueue.Begin();
        for ( RunQueue::Iterator itEnd = m_runQueue.End();
              it != itEnd; ++it ) {
            // FIXME: this does not handle multiple entries for the same table
            surr.rawAssign( it->GetAuxFileNo().GetCharArray() );
            bd998DestroyTempFile( m_trans, surr );
            m_trans.trError_gg00 = e_ok;
        }
        m_trans.trError_gg00 = auxErr;
    }

    /****************************************************************/

    SAPDB_Bool PrepareNextJob(
        tgg00_TransContext&   trans,
        SrvTasks_WaitingTask& waitingTask,
        Msg_List&             errMsg)
    {
        SAPDBTRACE_METHOD_DEBUG(
            "Data_PrimTreeStatisticCoordinator::JobQueue::PrepareNextJob",
            DataTree_Trace, 5 );

        // FIXME: check if task in run queue is runnable instead of
        // simply preparing next table

        if ( ! m_tableKey.IsValid() ) {
            SAPDBTRACE_WRITELN( DataTree_Trace, 5, "no more tables" );
            return false;
        }

        SAPDBTRACE_WRITELN(
            DataTree_Trace, 5, "preparing current table: " << m_curFileNo );

        // FIXME: recycle temp files if possible
        tgg00_FileId fileId;
        this->BuildTempFileId( trans, fileId, m_taskIdCounter );

        bd998CreateTempFile( trans, fileId );

        if ( trans.trError_gg00 != e_ok ) {
            // FIXME: add error message
            return false;
        }

        FileDir_FileNo auxFileNo( fileId.fileTabId_gg00() );

        Data_UpdStatRequestFile::StatisticsRequestInfo&
            requestInfo = m_iterator.GetStatisticsRequest( errMsg );

        SrvTasks_Job::Attributes jobAttributes( m_taskIdCounter );

        SrvTasks_JobCollectTableStatistics* pJob =
            new ( *reinterpret_cast<SAPDBMem_IRawAllocator*>
                  ( trans.trAllocator_gg00 ) )
            SrvTasks_JobCollectTableStatistics(
                jobAttributes,
                m_requestFile.GetFileNo(),
                m_tableKey,
                requestInfo.tableFileNo,
                auxFileNo,
                m_iterator.GetNoOfPagesToSample(),
                waitingTask);

        if ( pJob == 0 ) {
            // FIXME: 'out of memory'
            return false;
        }

        m_runQueue.InsertFront( JobInfo( *pJob, true ) );
        ++m_taskIdCounter;

        // FIXME: this should be done in SetIsRunning but then we
        // can't advance the iterator here:
        m_iterator.SetBeginTimeStamp( trans, errMsg );

        // advance iterator
        Data_UpdStatRequestFile::Iterator::ReturnCode itRetCode;
        itRetCode = m_iterator.Next(trans, errMsg );

        if ( itRetCode == Data_UpdStatRequestFile::Iterator::Okay ) {
            RequestInfo& curRequest =
                m_iterator.GetStatisticsRequest( errMsg );
            m_curFileNo = curRequest.tableFileNo;
            SAPDBTRACE_WRITELN(
                DataTree_Trace, 5, "next table will be: " << m_curFileNo );
            SAPDBERR_ASSERT_STATE( curRequest.columnCount > 0 );
        } else {
            m_curFileNo.Invalidate();
        }
        m_tableKey = m_iterator.GetTableKey();

        return true;
    }

    /****************************************************************/

    JobInfo& GetPreparedJob()
    {
        return m_runQueue[0];
    }

    /****************************************************************/

    SAPDB_Bool JobsAvailable() const
    {
        return m_tableKey.IsValid();
    }

    /****************************************************************/

    SAPDB_Bool AllJobsDone() const
    {
        return (! JobsAvailable()) && m_runQueue.IsEmpty();
    }

    /****************************************************************/

    SAPDB_Bool MarkJobDone(
        tgg00_TransContext trans,
        SAPDB_Int          jobId,
        Msg_List&          errMsg)
    {
        RunQueue::Iterator it = m_runQueue.Begin();
        for ( RunQueue::Iterator itEnd = m_runQueue.End();
              it != itEnd; ++it ) {
            if ( jobId == it->GetJobId() ) {
                SAPDB_Bool ok = true;

                Data_UpdStatRequestFile::Iterator::ReturnCode itRetCode;
                Data_UpdStatRequestFile::Iterator finishedFile =
                    m_requestFile.GetIterator( trans );

                itRetCode = finishedFile.Find(
                    trans, it->GetTableKey(), errMsg );

                if ( itRetCode == Data_UpdStatRequestFile::Iterator::Okay ) {

                    finishedFile.SetRequestProcessed( trans, true, errMsg );
                    finishedFile.SetEndTimeStamp( trans, errMsg );

                } else {
                    // FIXME: error handling
                    ok = false;
                }

                tgg00_Surrogate surr;
                surr.rawAssign( it->GetAuxFileNo().GetCharArray() );
                bd998DestroyTempFile( trans, surr );
                trans.trError_gg00 = e_ok;
                SrvTasks_JobCollectTableStatistics* pJob = &(it->GetJob());
                m_runQueue.Delete( it );
                destroy (
                    pJob,
                    *reinterpret_cast<SAPDBMem_IRawAllocator*>(
                        m_trans.trAllocator_gg00 )
                    );
                return ok;
            }
        }
        return false;
    }

    /****************************************************************/

    SAPDB_Bool MarkJobFailed(
        tgg00_TransContext trans,
        SAPDB_Int          jobId,
        Msg_List&          errMsg)
    {
        // FIXME: make sure to clear all pending jobs for same table
        // once we go parallel
        RunQueue::Iterator it = m_runQueue.Begin();
        for ( RunQueue::Iterator itEnd = m_runQueue.End();
              it != itEnd; ++it ) {
            if ( jobId == it->GetJobId() ) {
                SAPDB_Bool ok = true;

                // file is marked not processed in request file,
                // nothing to do there...

                tgg00_Surrogate surr;
                surr.rawAssign( it->GetAuxFileNo().GetCharArray() );
                bd998DestroyTempFile( trans, surr );
                trans.trError_gg00 = e_ok;

                SrvTasks_JobCollectTableStatistics* pJob = &(it->GetJob());
                m_runQueue.Delete( it );
                destroy (
                    pJob,
                    *reinterpret_cast<SAPDBMem_IRawAllocator*>(
                        m_trans.trAllocator_gg00 )
                    );
                return ok;
            }
        }
        return false;
    }

    /****************************************************************/

    void BuildTempFileId(
        tgg00_TransContext& trans,
        tgg00_FileId&       fileId,
        SAPDB_Int4          curFileIndex)
    {
        g04build_temp_tree_id( fileId, trans );
        fileId.fileTfn_gg00().becomes( tfnTemp_egg00 );
        fileId.fileTfnTemp_gg00().becomes( ttfnTempTable_egg00 );
        fileId.fileTempCnt_gg00() = curFileIndex;
        fileId.fileType_gg00().addElement( ftsConcurrent_egg00 );
        fileId.fileType_gg00().addElement( ftsShared_egg00 );
    }
};

/**************************************************************************/

Data_PrimTreeStatisticsCoordinator::Data_PrimTreeStatisticsCoordinator(
    tgg00_TransContext&     transContext,
    Data_UpdStatRequestFile& updStatRequestFile)
    :
    m_trans( transContext ),
    m_updStatRequestFile( updStatRequestFile )
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_PrimTreeStatisticsCoordinator::Data_PrimTreeStatisticsCoordinator",
        DataTree_Trace, 5 );

	if ( e_ok != transContext.trError_gg00 ) {
        return;
    }
}

/**************************************************************************/

Data_PrimTreeStatisticsCoordinator::~Data_PrimTreeStatisticsCoordinator()
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_PrimTreeStatisticsCoordinator::~Data_PrimTreeStatisticsCoordinator",
        DataTree_Trace, 5 );

   // Call Destructor for FileList
}

/**************************************************************************/

Data_PrimTreeStatisticsCoordinator::ReturnCode
Data_PrimTreeStatisticsCoordinator::UpdateStatistics(
    Msg_List& errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_PrimTreeStatisticsCoordinator::UpdateStatistics",
        DataTree_Trace, 5 );

    RTETask_ITask& task = *RTETask_ITask::ByTaskID( m_trans.trTaskId_gg00 );

    SrvTasks_WaitingTask waitingTask ( task );

    Srv_StatJobList
        serverJobList ( *reinterpret_cast<SAPDBMem_IRawAllocator*>(
                           m_trans.trAllocator_gg00 ) );

    SAPDB_Int sUsableServerCount;
    if ( SQLMan_UpdateStatistics::GetNoOfParallelServers() == 0 ) {
        sUsableServerCount = bd01UsableServerForParallelExecution();
    } else {
        sUsableServerCount = SQLMan_UpdateStatistics::GetNoOfParallelServers();
    }

    SAPDB_UInt usableServerCount =
        sUsableServerCount > 1 ? sUsableServerCount : 1;

    if ( ! serverJobList.Reserve( usableServerCount ) ) {
        errMsg =
            Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_OUT_OF_MEMORY,
                Msg_Arg( "ALLOC_SIZE",
                         usableServerCount
                         * sizeof( Srv_StatJobList::ValueType) )
                );
        return SystemError;
    }

    const SAPDB_Int c_maxRetries = 10;
    bool  ok                     = true;
    bool  retryStart             = false;
    SAPDB_UInt retryCounter      = 0;

    ReturnCode rc      = Okay;
    JobQueue jobQueue(
        m_trans, m_updStatRequestFile, usableServerCount, errMsg );

    while ( (! jobQueue.AllJobsDone()) && ok ) {

        if ( retryStart
             || jobQueue.PrepareNextJob( m_trans, waitingTask, errMsg ) ) {

            retryStart = false;
            JobInfo& newJobInfo = jobQueue.GetPreparedJob();

            SrvTasks_Job::ExecuteByServerResult retCode =
                newJobInfo.GetJob().ExecuteByServer( task, errMsg );

            if ( retCode == SrvTasks_Job::executeOK ) {

                newJobInfo.SetIsRunning( true );
                serverJobList.InsertEnd( &newJobInfo.GetJob() );
                retryCounter = 0;

            } else if ( retCode != SrvTasks_Job::executeNoServerAvailable ) {

                if ( retryCounter <= c_maxRetries ) {

                    retryStart = true;
                    ++retryCounter;

                } else {
                    // FIXME: add message to knldiag
                    // process job yourself
                    if ( this->SelfProcessJob( jobQueue, newJobInfo, errMsg )
                         != Okay ) {
                        ok = false;
                    }
                }

            } else {
                ok = false;
            }
        }

        if ( ok
             && ( serverJobList.IsFull()
                  || ( serverJobList.GetSize() >= usableServerCount )
                  || (! jobQueue.JobsAvailable())
                  || retryStart )
             && ( ! serverJobList.IsEmpty() ) ) {

            Srv_StatJobList::Iterator
                finishedJob;
            if ( serverJobList.WaitForAnyJobFinished(
                     waitingTask, finishedJob )
                ) {

                rc = ProcessResult( jobQueue, *(*finishedJob), errMsg );
                if ( rc != Okay ) {
                    ok = false;
                }

                serverJobList.Delete( finishedJob );

            } else {

                errMsg =
                    Msg_List(
                        Msg_List::Error,
                        SDBMSG_DATA_UPDSTAT_JOBLIST_ERROR,
                        Msg_Arg("SIZE", serverJobList.GetSize() )
                        );
                SAPDBERR_ASSERT_STATE( false );
                rc = SystemError;
                ok = false;
            }
        }

        if ( m_trans.trRteCommPtr_gg00->to_cancel ) {
            SAPDBTRACE_WRITELN(
                DataTree_Trace, 1, "trans was cancelled " );
            ok = false;
            rc = Cancelled;
        }
    }

    if ( ! ok ) {
        ok = true;
        serverJobList.AbortAll( waitingTask );

        if ( rc != Cancelled ) {
            // add message that clearly states that the command is aborting
            errMsg.AddDetail( SDBMSG_DATA_UPDSTAT_ABORTING );
        }

        // make sure we have an error code set
        if ( rc == Okay ) {
            rc = SystemError;
        }
    }

    return rc;
}

/**************************************************************************/

Data_PrimTreeStatisticsCoordinator::ReturnCode
Data_PrimTreeStatisticsCoordinator::SelfProcessJob(
    JobQueue& jobQueue,
    JobInfo&  jobInfo,
    Msg_List& errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_PrimTreeStatisticsCoordinator::SelfProcessJob",
        DataTree_Trace, 5 );

    SAPDBERR_ASSERT_STATE( jobInfo.IsValid() );

    SrvTasks_JobCollectTableStatistics& job = jobInfo.GetJob();

    job.Execute( m_trans );

    return ProcessResult( jobQueue, job, errMsg );
}

/**************************************************************************/

Data_PrimTreeStatisticsCoordinator::ReturnCode
Data_PrimTreeStatisticsCoordinator::ProcessResult(
    JobQueue&                           jobQueue,
    SrvTasks_JobCollectTableStatistics& job,
    Msg_List&                           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_PrimTreeStatisticsCoordinator::SelfProcessJob",
        DataTree_Trace, 5 );

    ReturnCode rc = Okay;

    SrvTasks_JobCollectTableStatistics::ReturnCode childRc =
        job.GetReturnCode( errMsg );

    if ( childRc == SrvTasks_JobCollectTableStatistics::ResultOkay ) {

        jobQueue.MarkJobDone(
            m_trans, job.GetId(), errMsg );

    } else {

        switch ( childRc ) {
            case SrvTasks_JobCollectTableStatistics::ResultTableFileNotFound:
            case SrvTasks_JobCollectTableStatistics::ResultCancelled:
                jobQueue.MarkJobFailed(
                    m_trans, job.GetId(), errMsg );
                break;

            case SrvTasks_JobCollectTableStatistics::ResultOldFileVersion:
                errMsg =
                    Msg_List(
                        Msg_List::Info,
                        SDBMSG_DATA_UPDSTAT_OLD_FILEVERSION,
                        Msg_Arg(
                            "FILENO",
                            FileDir_FileNoToString( job.GetTableFileNo() ) )
                        );
                RTE_Message( errMsg );
                errMsg.ClearMessageList();
                jobQueue.MarkJobFailed(
                    m_trans, job.GetId(), errMsg );
                break;

            case SrvTasks_JobCollectTableStatistics::ResultSystemError:
                rc = SystemError;
                break;

            default:
                errMsg.AddDetail(
                    SDBMSG_DATA_UPDSTAT_UNEXPECTED_JOB_RESULT,
                    Msg_Arg("RETURNCODE", SAPDB_ToString( rc ) ) );

                SAPDBERR_ASSERT_STATE( false );
                rc = SystemError;
        }
    }

    return rc;
}

/**************************************************************************/

// void Data_PrimTreeStatisticsCoordinator::Start(
//     Msg_List& errMsg)
// {
//     tgg00_FileId QueueId;
//     this->BuildTempFileId( QueueId, 0 ); // FIXME
//     Data_PageCount  auxLeaves;
//     Data_PageCount  auxRecords;
//     tgg00_ColumnDistinctValues  distinctValues;

//     Data_UpdStatRequestFile::DistinctValueVector
//         distinctVals(
//             *reinterpret_cast<SAPDBMem_IRawAllocator*>(
//                 m_trans.trAllocator_gg00 ) );

//     FileDir_Table tableFile;
//     FileDir_IDirectory& fDir =
//         FileDir_IDirectory::Instance( m_trans.trIndex_gg00 );

//     Data_UpdStatRequestFile::Iterator it =
//         m_updStatRequestFile->GetIterator( m_trans );
//     SAPDB_Bool ok = true;
//     while ( it.Next( m_trans, errMsg ) && ok ) {
//         Data_UpdStatRequestFile::StatisticsRequestInfo& requestInfo =
//             it.GetStatisticsRequest( errMsg );

//         fDir.GetTableFile(
//             m_trans.trTaskId_gg00,
//             requestInfo.tableFileNo,
//             tableFile,
//             errMsg );

//         if ( ! tableFile.IsAssigned() ) {
//             continue;
//         }

//         tgg00_FileId tableFileId;
//         tableFile.BuildFileId( tableFileId );

//         tbd_current_tree current;
//         tbd_current_tree tempCurrent;

//         bd30GetTree( m_trans, tableFileId, current,
//                      ! LOCK_TREE_EXCL_BD00, m_select );

//         b01treset_file( m_trans, QueueId );
//         bd30GetTree( m_trans, QueueId, tempCurrent,
//                      ! LOCK_TREE_EXCL_BD00, m_nil );

//         Data_PrimTreeStatistic  primTreeStatistic( current );

//         SAPDB_Int2 columnCount = requestInfo.columnCount;
//         tgg00_StEntryAddr pColumnDescription =
//             reinterpret_cast<tgg00_StEntryAddr>( &requestInfo.stackEntries[0] );

//         Data_SerializableScanTreeAccess scanTreeAccessor( current );
//         primTreeStatistic.CalculateColumnStatistic(
//             tempCurrent,
//             scanTreeAccessor,
//             columnCount, 1, columnCount,
//             pColumnDescription, auxLeaves, auxRecords,
//             distinctValues );

//         bd30ReleaseTree (tempCurrent);
//         bd30ReleaseTree (current);

//         distinctVals.Clear();
//         if ( ! distinctVals.Reserve( columnCount ) ) {
//             // FIXME: error handling
//             ok = false;
//             continue;
//         }

//         for ( int i = 0; i < columnCount; ++i ) {
//             distinctVals.InsertEnd( distinctValues[ i ] );
//         }

//         Data_UpdStatRequestFile::ReturnCode retCode;
//         retCode = it.SetStatistics(
//             m_trans,
//             auxLeaves, auxRecords,
//             distinctVals, errMsg );

//         if ( retCode != Data_UpdStatRequestFile::Okay ) {
//             // FIXME: error handling
//             ok = false;
//             continue;
//         }
//     }
// }

/*---------------------------------------------------------------------------*/

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
