/*****************************************************************************/
/*!
  @file         Data_UpdStatRequestFile.cpp
  @author       MartinKi
  @ingroup

  @brief        Implementation of class Data_UpdStatRequestFile.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2003 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"

#include "DataAccess/Data_UpdStatRequestFile.hpp"
#include "DataAccess/Data_Types.hpp"
#include "SAPDBCommon/Messages/SDBMsg_Data.h"

#include "FileDirectory/FileDir_IDirectory.hpp"
#include "FileDirectory/FileDir_Table.hpp"
#include "FileDirectory/FileDir_FileNo.hpp"

#include "Messages/Msg_List.hpp"
#include "RunTime/System/RTESys_Time.h"
#include "RunTime/RTE_Types.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"

#include "hbd02.h"
#include "hbd01.h"
#include "hbd998.h"
#include "hgg04.h"
#include "ggg00.h"
#include "heo52.h" // vtimestamp
#include "hsp41.h" // s41p4int

const char* c_fileComment = "UpdStatRequestFile";

struct RequestRecord {
    SAPDB_Int2  recLen;
    SAPDB_Int2  keyLen;
    SAPDB_Int2  varColOffset;
    SAPDB_Int2  varColCount;
    //
    SAPDB_Byte  tableFileSize[ 6 ];
    SAPDB_Byte  tableFileNo[ 8 ];
    //
    SAPDB_Byte      infoLen[ 2 ];
    SAPDB_Byte      infoDefByte;
    SAPDB_Byte      columnCount[ 2 ];
    SAPDB_Byte      pageCntToSample[ 4 ];
    RTESys_SQLTimestamp executeStart;
    RTESys_SQLTimestamp executeEnd;
    SAPDB_Byte      isDone;
    SAPDB_Byte      varPart;
    //
    // the following entries are part of the record but their position
    // depends on the number of columns:
    //
    // SAPDB_Int2  stackDescLen;
    // SAPDB_Byte  stackDescDefByte;
    //SAPDB_Byte  stackDesc;
};

/**************************************************************************/

Data_UpdStatRequestFile::Data_UpdStatRequestFile()
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Data_UpdStatRequestFile", DataTree_Trace, 5 );
}

/**************************************************************************/

FileDir_FileNo Data_UpdStatRequestFile::Create(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)

{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Create", DataTree_Trace, 5 );

    tgg00_FileId fileId;
    g04build_temp_tree_id( fileId, trans );
    fileId.fileType_gg00().addElement( ftsConcurrent_egg00 );
    fileId.fileType_gg00().addElement( ftsShared_egg00 );

    bd998CreateTempFile( trans, fileId );
    if ( trans.trError_gg00 == e_ok ) {
        FileDir_FileNo fileNo( fileId.fileTabId_gg00() );
        FileDir_IDirectory& fDir =
            FileDir_IDirectory::Instance( trans.trIndex_gg00 );
        fDir.GetTableFile( trans.trTaskId_gg00, fileNo, m_file, errMsg );
        m_file.SetComment( c_fileComment );

        m_file.BuildFileId( m_fileId );
    } else {
        m_file.Detach();
    }

    if ( m_file.IsAssigned() ) {
        return m_file.GetFileNo();
    } else {
        return FileDir_FileNo();
    }
}

/**************************************************************************/

Data_UpdStatRequestFile::ReturnCode Data_UpdStatRequestFile::Open(
    tgg00_TransContext&   trans,
    const FileDir_FileNo& fileNo,
    Msg_List&             errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Open", DataTree_Trace, 5 );

    if ( m_file.IsAssigned() ) {
        m_file.Detach();
    }

    FileDir_IDirectory& fDir =
        FileDir_IDirectory::Instance( trans.trIndex_gg00 );
    fDir.GetTableFile( trans.trTaskId_gg00, fileNo, m_file, errMsg );

    if ( ! m_file.IsAssigned() ) {
        return FileNotFound;
    }

    m_file.BuildFileId( m_fileId );

    SAPDBERR_ASSERT_STATE( m_file.GetComment() == c_fileComment );
    return Okay;
}

/**************************************************************************/

Data_UpdStatRequestFile::ReturnCode Data_UpdStatRequestFile::Destroy(
    tgg00_TransContext&   trans,
    Msg_List&             errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Destroy", DataTree_Trace, 5 );

    if ( ! m_file.IsAssigned() ) {
        return FileNotFound;
    }

    tgg00_Surrogate fileSurrogate;
    fileSurrogate.rawAssign( m_file.GetFileNo().GetCharArray() );
    bd998DestroyTempFile ( trans, fileSurrogate );

    trans.trError_gg00 = e_ok;
    return Okay;
}

/**************************************************************************/

FileDir_FileNo Data_UpdStatRequestFile::GetFileNo() const
{
    if ( ! m_file.IsAssigned() ) {
        // return invalid fileno
        return FileDir_FileNo();
    }

    return m_file.GetFileNo();
}

/**************************************************************************/

Data_UpdStatRequestFile::ReturnCode Data_UpdStatRequestFile::AddTable(
    tgg00_TransContext&     transContext,
    const FileDir_FileNo&   tableFileNo,
    const SAPDB_Int4        noOfPagesToSample,
    const SAPDB_Int4        columnCount,
    const SAPDB_Int2        columnNumbers[],
    const tgg00_StEntryAddr pColumnDescriptions,
    Msg_List&               errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::AddTable", DataTree_Trace, 5 );

    tgg00_Rec recBuf;
    RequestRecord* rec = reinterpret_cast<RequestRecord*>(
        &recBuf.recBuf_gg00()[ 0 ] );
    char* pos =  reinterpret_cast<char*>( rec );

    SAPDB_Int  colDescLen = columnCount * sizeof( tgg00_StackEntry );
    SAPDB_Int2 infoLen =
        3                      // def + len bytes
        + sizeof( rec->columnCount )
        + sizeof( rec->pageCntToSample )
        + columnCount * ( sizeof(SAPDB_Int2) + sizeof(SAPDB_Int4) ) // statistics info
        + 2 * sizeof( RTESys_SQLTimestamp ) // timestamps
        + sizeof ( rec->isDone )
        + colDescLen;                   // column descriptions

    // get table size
    FileDir_Table tableFile;
    FileDir_IDirectory& fDir =
        FileDir_IDirectory::Instance( transContext.trIndex_gg00 );
    fDir.GetTableFile( transContext.trTaskId_gg00,
                       tableFileNo, tableFile, errMsg );

    if ( ! tableFile.IsAssigned() ) {
        return FileNotFound;
    }

    const pasbool c_countRecords = true;
    SAPDB_Int4 leafPageCount;
    SAPDB_Int4 dummyCount;
    tgg00_FileId tableFileId;
    tableFile.BuildFileId( tableFileId );
    bd01GetLeafPageAndRecordCount(
        transContext, tableFileId, ! c_countRecords, leafPageCount, dummyCount);

    if ( transContext.trError_gg00 != e_ok ) {
        transContext.trError_gg00 = e_ok;

        if ( tableFile.GetFileState() == Data_FileDeleted ) {
            return FileNotFound;
        }

        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_GETTABLESIZE_FAILED,
                Msg_Arg( "FILENO", FileDir_FileNoToString( tableFileNo ) )
            );
        return SystemError;
    }
    // end of get table size

    rec->keyLen = sizeof( rec->tableFileSize ) + sizeof( tableFileNo );

    rec->varColOffset = 0;
    rec->varColCount  = 0;

    rec->recLen = cgg_rec_key_offset + rec->keyLen + infoLen;

    // insert table size (key column)
    tsp00_NumError numErr;
    s41p4int(
        rec,
        RTE_OFFSET_OF( RequestRecord, tableFileSize ) + 1,
        leafPageCount, numErr );
    SAPDBERR_ASSERT_STATE( numErr == num_ok );

    // insert table file no (key column)
    SAPDB_MemCopyNoCheck(
        rec->tableFileNo, tableFileNo.GetCharArray(), sizeof( tableFileNo ) );

    // insert column count
    SAPDB_MemCopyNoCheck( &rec->infoLen, &infoLen, sizeof( rec->infoLen ) );
    rec->infoDefByte = csp_ascii_blank;
    SAPDB_Int2 colCount = static_cast<SAPDB_Int2>( columnCount );
    SAPDB_MemCopyNoCheck(
        &rec->columnCount, &colCount, sizeof( rec->columnCount ) );

    // insert sample size
    SAPDB_MemCopyNoCheck(
        &rec->pageCntToSample, &noOfPagesToSample,
        sizeof( rec->pageCntToSample ) );

    // set empty timestamp
    //
    // build non-zero terminated empty timestamp
    RTESys_SQLTimestamp c_emptyTimestamp = "                   ";
    c_emptyTimestamp[ sizeof( c_emptyTimestamp ) - 1 ] = ' ';

    SAPDB_MemCopyNoCheck(
        rec->executeStart, c_emptyTimestamp, sizeof( c_emptyTimestamp ) );
    SAPDB_MemCopyNoCheck(
        rec->executeEnd, c_emptyTimestamp, sizeof( c_emptyTimestamp ) );

    // insert status (not done)
    rec->isDone = 0;

    // insert column descriptions
    pos = reinterpret_cast<char*>( &rec->varPart );
    for ( int i = 0; i < columnCount; ++i ) {
        const SAPDB_Int4 zeroCount = 0;
        SAPDB_MemCopyNoCheck(
            pos, &columnNumbers[ i ], sizeof( SAPDB_Int2 ) );
        SAPDB_MemCopyNoCheck(
            pos + sizeof(SAPDB_Int2), &zeroCount, sizeof( SAPDB_Int4 ) );
        pos += sizeof(SAPDB_Int2) + sizeof(SAPDB_Int4);
    }

    // insert stack descriptions
    SAPDB_MemCopy(
        pos, pColumnDescriptions, colDescLen,
        "Data_UpdStatRequestFile", __FILE__, __LINE__, errMsg );

    // now insert record
    b02add_record( transContext, m_fileId, recBuf );
    if ( transContext.trError_gg00 != e_ok ) {
        SAPDBERR_ASSERT_STATE( false );
        transContext.trError_gg00 = e_ok;
    }

    return Okay;
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator Data_UpdStatRequestFile::GetIterator(
    tgg00_TransContext& trans)
{
    return Iterator( trans, *this );
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::Iterator(
    tgg00_TransContext&      trans,
    Data_UpdStatRequestFile& file)
    : m_file( file ),
      m_statisticsRequestInfo(
          *reinterpret_cast<SAPDBMem_IRawAllocator*>(
              trans.trAllocator_gg00 ) ),
      m_tableStatistics(
          *reinterpret_cast<SAPDBMem_IRawAllocator*>(
              trans.trAllocator_gg00 ) )
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator", DataTree_Trace, 5 );

//    m_curKey.keyLen_gg00() = 0;

    m_curKey.keyLen_gg00() = MAX_KEYLEN_GG00;
    m_curKey.keyVal_gg00() = b01fullkey;

}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::Iterator(
    const Iterator& it)
    : m_file( it.m_file ),
      m_statisticsRequestInfo(
          it.m_statisticsRequestInfo.stackEntries.GetAllocator() ),
      m_tableStatistics(
          it.m_statisticsRequestInfo.stackEntries.GetAllocator() ),
      m_curKey( it.m_curKey )
{
    if ( it.IsValid() ) {
        Msg_List errMsg;
        SAPDB_MemCopyNoCheck( &m_recBuf, &it.m_recBuf, sizeof( m_recBuf ) );
        this->ParseRecord( errMsg );
    }
}

/**************************************************************************/

void Data_UpdStatRequestFile::Iterator::SetStartKey()
{
    m_curKey.keyLen_gg00() = MAX_KEYLEN_GG00;
    m_curKey.keyVal_gg00() = b01fullkey;
}

/**************************************************************************/

void Data_UpdStatRequestFile::Iterator::SetEndKey()
{
    m_curKey.keyLen_gg00() = 0;
}

/**************************************************************************/

SAPDB_Bool Data_UpdStatRequestFile::Iterator::IsValid() const
{
    return
        (m_curKey.keyLen_gg00() != MAX_KEYLEN_GG00)
        && (m_curKey.keyLen_gg00() != 0);
}

/**************************************************************************/

void Data_UpdStatRequestFile::Iterator::SetFirst()
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::SetFirst", DataTree_Trace, 5 );

    this->SetStartKey();
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::Find(
    tgg00_TransContext&                      trans,
    const Data_UpdStatRequestFile::TableKey& tableKey,
    Msg_List&                                errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Find", DataTree_Trace, 5 );

    // build key for tableKey
    m_curKey.keyRecLenSpace_gg00() =
        cgg_rec_key_offset
        + sizeof( tableKey.tableFileSize )
        + sizeof( tableKey.tableFileNo );
    m_curKey.keyLen_gg00()         =
        sizeof( tableKey.tableFileSize )
        + sizeof( tableKey.tableFileNo );
    m_curKey.keyVarOffSpace_gg00() =
        cgg_rec_key_offset
        + sizeof( tableKey.tableFileSize )
        + sizeof( tableKey.tableFileNo );
    m_curKey.keyVarCntSpace_gg00() = 0;

    SAPDB_MemCopy(
        reinterpret_cast<char*>( &m_curKey.keyVal_gg00() ),
        tableKey.tableFileSize,
        sizeof( tableKey.tableFileSize ),
        "Data_UpdStatRequestFile", __FILE__, __LINE__, errMsg );

    SAPDB_MemCopy(
        reinterpret_cast<char*>( &m_curKey.keyVal_gg00() )
        + sizeof( tableKey.tableFileSize ),
        tableKey.tableFileNo.GetCharArray(),
        sizeof( tableKey.tableFileNo ),
        "Data_UpdStatRequestFile", __FILE__, __LINE__, errMsg );

    //
    // retrieve record
    const pasbool c_inclusive = true;
    b02next_record(
        trans, m_file.m_fileId, m_curKey, c_inclusive, m_recBuf );

    if ( (trans.trError_gg00 == e_no_next_record)
         || (trans.trError_gg00 == e_key_not_found) ) {
        trans.trError_gg00 = e_ok;
        return EndOfFile;
    }

    if ( trans.trError_gg00 != e_ok ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_POSITION_ITERATOR_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_file.GetFileNo() ) ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
            );

        trans.trError_gg00 = e_ok;
        return SystemError;
    }

    trans.trError_gg00 = e_ok;

    if ( ! this->ParseRecord( errMsg ) ) {
        return SystemError;
    }

    return Okay;
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::Next(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Next", DataTree_Trace, 5 );

    // invalidate current entry
    m_tableKey.tableFileNo.Invalidate();

    const pasbool c_inclusive = true;

    b02prev_record(
        trans, m_file.m_fileId, m_curKey, ! c_inclusive, m_recBuf );

    if ( trans.trError_gg00 == e_no_prev_record ) {
        trans.trError_gg00 = e_ok;
        this->SetStartKey();
        return EndOfFile;
    }

    if ( (trans.trError_gg00 != e_ok)
         && (trans.trError_gg00 != e_key_not_found) ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_POSITION_ITERATOR_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_file.GetFileNo() ) ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
            );
        trans.trError_gg00 = e_ok;
        this->SetEndKey();
        return SystemError;
    }

    trans.trError_gg00 = e_ok;

    // update key
    SAPDB_MemCopy(
        &m_curKey, m_recBuf.recBuf_gg00(),
        m_recBuf.recKeyLen_gg00() + cgg_rec_key_offset,
        "Data_UpdStatRequestFile", __FILE__, __LINE__, errMsg );

    if ( ! this->ParseRecord( errMsg ) ) {
        return SystemError;
    }

    return Okay;
}

/**************************************************************************/

const Data_UpdStatRequestFile::TableKey&
Data_UpdStatRequestFile::Iterator::GetTableKey() const
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::GetStatisticsRequest",
        DataTree_Trace, 5 );

    return m_tableKey;
}

/**************************************************************************/

Data_UpdStatRequestFile::StatisticsRequestInfo&
Data_UpdStatRequestFile::Iterator::GetStatisticsRequest(
    Msg_List& errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::GetStatisticsRequest",
        DataTree_Trace, 5 );

    return m_statisticsRequestInfo;
}

/**************************************************************************/

const Data_UpdStatRequestFile::TableStatistics&
Data_UpdStatRequestFile::Iterator::GetStatistics(Msg_List& errMsg) const
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::GetStatistics", DataTree_Trace, 5 );

    return m_tableStatistics;
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::SetStatistics(
    tgg00_TransContext&  trans,
    Data_PageCount       leafPageCount,
    Data_PageCount       recordCount,
    DistinctValueVector& distinctValues,
    Msg_List&            errMsg)
{
    // set statistics in record
    RequestRecord* rec =
        reinterpret_cast<RequestRecord*>(&m_recBuf.recBuf_gg00()[ 0 ] );
    char* pos = reinterpret_cast<char*>( &rec->varPart );

    DistinctValueVector::Iterator it    = distinctValues.Begin();
    DistinctValueVector::Iterator itEnd = distinctValues.End();
    while ( it != itEnd ) {
        pos += sizeof( SAPDB_Int2 );

        SAPDB_MemCopy(
            pos, it, sizeof( SAPDB_Int4 ),
            "Data_UpdStatRequestFile", __FILE__, __LINE__, errMsg );
        pos += sizeof( SAPDB_Int4 );

        ++it;
    }

    // update file
    b02repl_record( trans, m_file.m_fileId, m_recBuf );

    if ( trans.trError_gg00 != e_ok ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_UPDATERECORD_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_file.GetFileNo() ) ),
                Msg_Arg( "KEYASHEX",
                         m_curKey.keyVal_gg00(),
                         m_curKey.keyLen_gg00(),
                         m_curKey.keyLen_gg00(),
                         SAPDB_ToStringClass::hex ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
            );
        trans.trError_gg00 = e_ok;
        return SystemError;
    }

    if ( ! this->ParseRecord( errMsg ) ) {
        return SystemError;
    }

    return Okay;
}

/**************************************************************************/

SAPDB_Int4
Data_UpdStatRequestFile::Iterator::GetNoOfPagesToSample() const
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::GetNoOfPagesToSample",
        DataTree_Trace, 5 );

    return m_pageCntToSample;
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::SetRequestProcessed(
    tgg00_TransContext& trans,
    const SAPDB_Bool    isProcessed,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::SetBeginTimeStamp",
        DataTree_Trace, 5 );

    RequestRecord* rec =
        reinterpret_cast<RequestRecord*>(&m_recBuf.recBuf_gg00()[ 0 ] );

    if ( isProcessed ) {
        rec->isDone = 1;
    } else {
        rec->isDone = 0;
    }

    return this->UpdateRecord( trans, errMsg );
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::SetBeginTimeStamp(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::SetBeginTimeStamp",
        DataTree_Trace, 5 );

    vtimestamp( *reinterpret_cast<tsp00_Timestamp*>(&m_beginTimeStamp) );

    RequestRecord* rec =
        reinterpret_cast<RequestRecord*>(&m_recBuf.recBuf_gg00()[ 0 ] );
    SAPDB_MemCopyNoCheck(
        rec->executeStart, m_beginTimeStamp, sizeof( m_beginTimeStamp ) );

    return this->UpdateRecord( trans, errMsg );
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::SetEndTimeStamp(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::SetEndTimeStamp",
        DataTree_Trace, 5 );

    vtimestamp( *reinterpret_cast<tsp00_Timestamp*>(&m_endTimeStamp) );

    RequestRecord* rec =
        reinterpret_cast<RequestRecord*>(&m_recBuf.recBuf_gg00()[ 0 ] );
    SAPDB_MemCopyNoCheck(
        rec->executeEnd, m_endTimeStamp, sizeof( m_endTimeStamp ) );

    return this->UpdateRecord( trans, errMsg );
}

/**************************************************************************/

const RTESys_SQLTimestamp&
Data_UpdStatRequestFile::Iterator::GetBeginTimeStamp(
    Msg_List& errMsg) const
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::GetBeginTimeStamp",
        DataTree_Trace, 5 );

    return m_beginTimeStamp;
}

/**************************************************************************/

const RTESys_SQLTimestamp&
Data_UpdStatRequestFile::Iterator::GetEndTimeStamp(
    Msg_List& errMsg) const
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::GetEndTimeStamp",
        DataTree_Trace, 5 );

    return m_endTimeStamp;
}

/**************************************************************************/

SAPDB_Bool Data_UpdStatRequestFile::Iterator::ParseRecord(
    Msg_List& errMsg)
{
    RequestRecord* rec =
        reinterpret_cast<RequestRecord*>(&m_recBuf.recBuf_gg00()[ 0 ] );
    char* pos =  reinterpret_cast<char*>( rec );

    SAPDB_MemCopyNoCheck(
        m_tableKey.tableFileSize, rec->tableFileSize,
        sizeof( rec->tableFileSize ) );

    tgg00_Surrogate tableSurrogate;
    SAPDB_MemCopyNoCheck(
        &tableSurrogate, rec->tableFileNo, sizeof( rec->tableFileNo ) );
    m_tableKey.tableFileNo = tableSurrogate;

    // column count
    SAPDB_MemCopyNoCheck(
        &m_columnCount, &rec->columnCount, sizeof( m_columnCount ) );

    // page count
    SAPDB_MemCopyNoCheck(
        &m_pageCntToSample, &rec->pageCntToSample,
        sizeof( m_pageCntToSample ) );

    m_statisticsRequestInfo.stackEntries.Clear();
    m_tableStatistics.columnStatistics.Clear();
    if ( (! m_statisticsRequestInfo.stackEntries.Reserve( m_columnCount ) )
         || (! m_tableStatistics.columnStatistics.Reserve( m_columnCount )) ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_OUT_OF_MEMORY,
                Msg_Arg( "ALLOC_SIZE",
                         SAPDB_ToString( m_columnCount ) )
            );
        return false;
    }


    // timestamps
    SAPDB_MemCopyNoCheck(
        m_beginTimeStamp, rec->executeStart, sizeof( m_beginTimeStamp ) );
    SAPDB_MemCopyNoCheck(
        m_endTimeStamp, rec->executeEnd, sizeof( m_endTimeStamp ) );

    // isDone
    m_isDone = (rec->isDone == 1);

    // column statistics
    int i;
    m_tableStatistics.tableFileNo = m_tableKey.tableFileNo;
    pos = reinterpret_cast<char*>( &rec->varPart );
    for ( i = 0; i < m_columnCount; ++i ) {
        SAPDB_Int2 columnNumber;
        SAPDB_Int4 columnValue;
        SAPDB_MemCopyNoCheck( &columnNumber, pos, sizeof( columnNumber ) );

        SAPDB_MemCopyNoCheck(
            &columnValue, pos + sizeof( columnNumber ), sizeof( columnValue ) );

        m_tableStatistics.columnStatistics.InsertEnd(
            ColumnStatistic( columnNumber, columnValue ) );

        pos += sizeof( columnNumber ) + sizeof( columnValue );
    }

    // get column description
    m_statisticsRequestInfo.tableFileNo = m_tableKey.tableFileNo;
    m_statisticsRequestInfo.columnCount = m_columnCount;
    m_statisticsRequestInfo.pageSampleSize  = m_pageCntToSample;
    for ( i = 0; i < m_columnCount; ++i ) {
        tgg00_StackEntry stackEntry;

        SAPDB_MemCopyNoCheck( &stackEntry, pos, sizeof( stackEntry ) );
        m_statisticsRequestInfo.stackEntries.InsertEnd( stackEntry );

        pos += sizeof( stackEntry );
    }

    return true;
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::Delete(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::Delete",
        DataTree_Trace, 5 );

    if ( ! this->IsValid() ) {
        return KeyNotFound;
    }

    // delete current record, returns true, if the iterator is still valid
    b02del_record( trans, m_file.m_fileId, m_curKey );

    SAPDBTRACE_WRITELN(
        DataTree_Trace, 5,
        "del returned:"
        << SQLManErr_Interface::GetInstance().GetBaseErrorText(
            trans.trError_gg00) );

    if ( (trans.trError_gg00 != e_ok)
         && (trans.trError_gg00 != e_key_not_found) ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_DELETE_ENTRY_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_file.GetFileNo() ) ),
                Msg_Arg( "KEYASHEX",
                         m_curKey.keyVal_gg00(),
                         m_curKey.keyLen_gg00(),
                         m_curKey.keyLen_gg00(),
                         SAPDB_ToStringClass::hex ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
            );
        SAPDBERR_ASSERT_STATE( false );
        trans.trError_gg00 = e_ok;
        return SystemError;
    }
    trans.trError_gg00 = e_ok;

    return this->Next( trans, errMsg );
}

/**************************************************************************/

Data_UpdStatRequestFile::Iterator::ReturnCode
Data_UpdStatRequestFile::Iterator::UpdateRecord(
    tgg00_TransContext& trans,
    Msg_List&           errMsg)
{
    SAPDBTRACE_METHOD_DEBUG(
        "Data_UpdStatRequestFile::Iterator::UpdateRecord",
        DataTree_Trace, 5 );

    if ( ! this->IsValid() ) {
        return KeyNotFound;
    }

    // update file
    b02repl_record( trans, m_file.m_fileId, m_recBuf );
    if ( trans.trError_gg00 != e_ok ) {
        errMsg = Msg_List(
                Msg_List::Error,
                SDBMSG_DATA_UPDSTAT_UPDATERECORD_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_file.GetFileNo() ) ),
                Msg_Arg( "KEYASHEX",
                         m_curKey.keyVal_gg00(),
                         m_curKey.keyLen_gg00(),
                         m_curKey.keyLen_gg00(),
                         SAPDB_ToStringClass::hex ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
            );
        trans.trError_gg00 = e_ok;
        return SystemError;
    }
    return Okay;
}
