/*!**************************************************************************

  module      : FBM_ClusteredDataVolume.cpp

  -------------------------------------------------------------------------

  responsible : Henrik

  auhtor      : Henrik

  special area: FreeBlockManagement (FBM)
  description : 


  last changed: 2000-03-10  10:00
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG



    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "FreeBlockManagement/FBM_ClusteredDataVolume.hpp"
#include "FreeBlockManagement/FBM_Dump.hpp"
#include "FreeBlockManagement/FBM_Exception.hpp"
#include "FreeBlockManagement/FBM_Messages.hpp"
#include "KernelCommon/Kernel_Dump.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"


/*===========================================================================*
 *  DEFINES                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASS METHOD DEFINITION                                                  *
 *===========================================================================*/

/*---------------------------------------------------------------------------*
 *  CLASS FBM_ClusteredDataVolume                                              *
 *---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::InitVolume ()
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::InitVolume", FBM_Trace, 5);

    SAPDBERR_ASSERT_STATE( 0 < m_LastBlock );

    /* set members to their start values */
    m_NumBlocksUsed                   = 0;
    m_NumBlocksFreeAfterSVP           = 0;
    m_NumBlocksMarkedForBackup        = 0;
    m_NumBlocksToAddUntilOptIsReached = 0;
    m_VolumeIsInCompressingMode       = false;

    /* m_LastFragmentedBlock remains undefined,and gets only a value when */
    /* it is used by the data writer to compress and balance the devices  */

    /* initialize the list containing the block states */
    /* and set all block states in the list to free    */
    if(! m_BlockStateList.Create( m_LastBlock + 1 ))
    {
        FBM_Exception errMsg( __CONTEXT__, FBM_ERR_NO_MORE_MEMORY );
        RTE_Crash( errMsg );
    }

    /* initialize free cluster list */
    m_ClusterFreeList.RemoveAll();
    m_ClusterFreeList.AddFreeCluster(0, m_LastBlock + 1);

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::RestoreAllBlockStatesMarkedForBackup (IOMan_BlockCount &NumBlocksRestoredToFreeAfterSVP)
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::RestoreBlockStatesMarkedforBackup", FBM_Trace, 5);

    SAPDBERR_ASSERT_STATE (!m_VolumeIsInCompressingMode);

    IOMan_BlockNo     iBlock;
    SAPDB_Bool     bRestoredStateIsFreeAfterSVP;

    NumBlocksRestoredToFreeAfterSVP = 0;

    for (iBlock=0; iBlock<=m_LastBlock;  ++iBlock)
    {
        if (GetBlockState (iBlock) == BlockState_BackUp)
        {
            /* restore the state to the one it had before the backup */
            RestoreBlockStateMarkedForBackup (iBlock, bRestoredStateIsFreeAfterSVP);
            if (bRestoredStateIsFreeAfterSVP)
                ++NumBlocksRestoredToFreeAfterSVP;

            if (m_NumBlocksMarkedForBackup == 0) break;
        }
    }
    SAPDBERR_ASSERT_STATE (m_NumBlocksMarkedForBackup == 0);

    /* reset marker of last block read for back up */
    m_ActBlockNoForBackup.Invalidate();

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/

IOMan_BlockNo
FBM_ClusteredDataVolume::GetMultFreeBlocks  (const IOMan_BlockCount  NumBlocksWanted,
                                    IOMan_BlockCount       &NumBlocksSupplied)
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::GetMultFreeBlocks", FBM_Trace, 5);

    /* this routine is called to get  a cluster of free blocks.  */
    /* before the routine is called it is already obvious that   */
    /* there must be at least one free block on this device      */

    IOMan_BlockNo FreeBlockNo;

    /* find free BlockNo */
    m_ClusterFreeList.SearchCluster(NumBlocksWanted, FreeBlockNo, NumBlocksSupplied);

    if(NumBlocksSupplied > NumBlocksWanted)
    {
        NumBlocksSupplied = NumBlocksWanted;
    }

    /* change state of free block to occupied */
    for (IOMan_BlockNo block = FreeBlockNo; block < FreeBlockNo + NumBlocksSupplied; block++)
    {
        m_BlockStateList.SetBlockState (block, BlockState_Occupied);
        m_ClusterFreeList.RemoveFreeBlock(block);
    }

    /* write found block no into vtrace */
    SAPDBTRACE_WRITELN( FBM_Trace, 5, "FreeBlockNo: " << FreeBlockNo
                        << " #WantedFreeB: "  << NumBlocksWanted
                        << " #FoundFreeB: "  << NumBlocksSupplied);

    /* update counter of used blocks */
    m_NumBlocksUsed += NumBlocksSupplied;

    /* update the counter which controlls the distribution of blocks onto devices */
    m_NumBlocksToAddUntilOptIsReached -= NumBlocksSupplied;

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif

    return FreeBlockNo; // and NumBlocksSupplied
}

/*---------------------------------------------------------------------------*/

IOMan_BlockNo 
FBM_ClusteredDataVolume::GetFreeBlock  ()
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::GetFreeBlock", FBM_Trace, 5);
    SAPDBERR_ASSERT_STATE ((m_LastBlock + 1 - m_NumBlocksUsed) > 0);

    /* this routine is called to get the adress of a free block. */
    /* before the routine is called it is already obvious that   */
    /* there must be at least one free block on this device      */

    IOMan_BlockNo FreeBlockNo;

    m_ClusterFreeList.SearchBlock(FreeBlockNo);

    /* check that a free block was found */
    SAPDBERR_ASSERT_STATE (m_BlockStateList.GetBlockState(FreeBlockNo) == BlockState_Free);

    /* change state of free block to occupy */
    m_BlockStateList.SetBlockState (FreeBlockNo,BlockState_Occupied);

    m_ClusterFreeList.RemoveFreeBlock(FreeBlockNo);

    /* write found block no into vtrace */
    SAPDBTRACE_WRITELN( FBM_Trace, 5, "FreeSBlockNo: " << FreeBlockNo);

    /* update counter of used blocks */
    ++m_NumBlocksUsed;

    /* update the counter which controlls the distribution of blocks onto devices */
    --m_NumBlocksToAddUntilOptIsReached;

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif

    return FreeBlockNo;
}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::SetAllBlocksMarkedAsFreeAfterSVPToFree ()
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::SetAllBlocksMarkedAsFreeAfterSVPToFree", FBM_Trace, 5);

    IOMan_BlockNo     iBlock;

    for (iBlock=0; iBlock <= m_LastBlock; iBlock++)
    {
        if (m_BlockStateList.GetBlockState (iBlock) == BlockState_FreeAfterSVP)
        {
            m_BlockStateList.SetBlockState (iBlock,BlockState_Free);

            m_ClusterFreeList.AddFreeBlock(iBlock);

            --m_NumBlocksUsed;
            --m_NumBlocksFreeAfterSVP;
        }
    }
    /* check that no blocks remain in the state free after save point */
    SAPDBERR_ASSERT_STATE (m_NumBlocksFreeAfterSVP == 0);

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/


void
FBM_ClusteredDataVolume::GetNextBlocksForBackUp (const IOMan_BlockCount   MaxNumBlocksWanted,
                                        IOMan_BlockCount        &SuppliedNumBlocks,
                                        IOMan_BlockNo        &BlockNo)

{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::GetNextBlocksForBackUp", FBM_Trace, 5);

    /* check that the device ist not in the compression mode */
    SAPDBERR_ASSERT_STATE (!m_VolumeIsInCompressingMode);

    /* initialize output parameter */
    SuppliedNumBlocks = 0;

    if (m_ActBlockNoForBackup.IsInvalid())
    {
        BlockNo.Invalidate();
        return;
    }

    /* search next block marked for back up starting from the current backup block no */
    SAPDB_Bool bBlockFound    = false;

    /* go through current section and find the first block which is marked as back up */
    while (m_ActBlockNoForBackup <= m_LastBlock)
    {
        if (GetBlockState (m_ActBlockNoForBackup) == BlockState_BackUp)
        {
            /* check if the found block is really marked for backup and not for compression */
            SAPDBERR_ASSERT_STATE ((GetBlockStateAfterBackup (m_ActBlockNoForBackup) == BlockState_Occupied) ||
                                   (GetBlockStateAfterBackup (m_ActBlockNoForBackup) == BlockState_FreeAfterSVP));
            bBlockFound = true;
            break;
        }
        ++m_ActBlockNoForBackup;
    }

    /* goto next section which contains blocks marked for backup */
    if (!bBlockFound)
    {
        m_ActBlockNoForBackup.Invalidate();
    }

    /* set return value where the next set of neighbouring blocks marked for back up starts */
    BlockNo = m_ActBlockNoForBackup;

    /* check if there is really a next block which is          */
    /* marked for back up and not yet fetched by an datawriter */
    if (bBlockFound)
    {
        /* check out if the following blocks are also in the state BlockState_BackUp */
        SuppliedNumBlocks = 1;
        ++m_ActBlockNoForBackup;

        while ((SuppliedNumBlocks      <  MaxNumBlocksWanted) &&
                (m_ActBlockNoForBackup  <= m_LastBlock)    &&
                (GetBlockState (m_ActBlockNoForBackup) == BlockState_BackUp))
        {

            /* check if the found block is really marked for backup and not for compression */
            SAPDBERR_ASSERT_STATE ((GetBlockStateAfterBackup (m_ActBlockNoForBackup) == BlockState_Occupied) ||
                                   (GetBlockStateAfterBackup (m_ActBlockNoForBackup) == BlockState_FreeAfterSVP));
            ++ SuppliedNumBlocks;
            ++ m_ActBlockNoForBackup;
        }

        /* set marker to display that there are no more blocks for backup */
        if (m_ActBlockNoForBackup > m_LastBlock )
            m_ActBlockNoForBackup.Invalidate();
    }

    /* check consistency of the device */
#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::GetDeviceStatistics (IOMan_BlockCount &NumBlocks,
                                     IOMan_BlockCount &NumBlocksFree,
                                     IOMan_BlockCount &NumBlockFreeAfterSVP) const
{
    NumBlocks            = m_LastBlock + 1;
    NumBlocksFree        = m_LastBlock + 1 - m_NumBlocksUsed;
    NumBlockFreeAfterSVP = m_NumBlocksFreeAfterSVP;
}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::Verify ()
{
    SAPDBTRACE_ROUTINE_DEBUG  ("FBM_ClusteredDataVolume::Verify", FBM_Trace, 5);

    SAPDBERR_ASSERT_STATE ((m_LastBlock + 1) == m_BlockStateList.Capacity());

    IOMan_BlockCount NumBlocksFree        = 0;
    IOMan_BlockCount NumBlockFreeAfterSVP = 0;

    IOMan_BlockNo FirstFreeBlock                = m_LastBlock + 1;
    IOMan_BlockNo LastUsedBlock                 = 0;
    IOMan_BlockCount NumBlocksMarkedForBackUp      = 0;
    IOMan_BlockCount NumBlocksMarkedForCompression = 0;

    /* go through all entries in the blocklist and count the occurence of each state */
    for (IOMan_BlockNo iBlockNo=0; iBlockNo <= m_LastBlock; ++iBlockNo)
    {
        FBM_BlockState BlockStateAfterBackUp = GetBlockStateAfterBackup(iBlockNo);

        switch (GetBlockState(iBlockNo))
        {
        case BlockState_Free:
            ++NumBlocksFree;
            if (FirstFreeBlock > iBlockNo)
                FirstFreeBlock = iBlockNo;
            SAPDBERR_ASSERT_STATE (BlockState_Free == BlockStateAfterBackUp);
            break;
        case BlockState_Occupied:
            if (LastUsedBlock < iBlockNo)
                LastUsedBlock = iBlockNo;
            SAPDBERR_ASSERT_STATE (BlockState_Free == BlockStateAfterBackUp);
            break;
        case BlockState_FreeAfterSVP:
            ++NumBlockFreeAfterSVP;
            if (LastUsedBlock < iBlockNo)
                LastUsedBlock = iBlockNo;
            SAPDBERR_ASSERT_STATE (BlockState_Free == BlockStateAfterBackUp);
            break;
        case BlockState_BackUp:
            ++ NumBlocksMarkedForBackUp;
            if (m_VolumeIsInCompressingMode)
            {
                //blocks are marked for compression
                SAPDBERR_ASSERT_STATE ((BlockState_BackUp  == BlockStateAfterBackUp) ||
                                       (BlockState_Free    == BlockStateAfterBackUp));
            }
            else
            {
                //blocks are marked for back up
                SAPDBERR_ASSERT_STATE ((BlockState_Occupied     == BlockStateAfterBackUp) ||
                                       (BlockState_FreeAfterSVP == BlockStateAfterBackUp));
            }

            if (LastUsedBlock < iBlockNo)
                LastUsedBlock = iBlockNo;
            if ((BlockState_BackUp == BlockStateAfterBackUp) ||
                    (BlockState_Free   == BlockStateAfterBackUp  ))
                ++NumBlocksMarkedForCompression;
            break;
        default:
            /* write error message into knldiag and vtrace */
            FBM_IllegalState IllegalState(__CONTEXT__, iBlockNo, GetBlockState (iBlockNo), BlockStateAfterBackUp);
            RTE_Crash(IllegalState);
        }
    }

    /* check counters and pointers */
    SAPDBERR_ASSERT_STATE (NumBlocksMarkedForBackUp == m_NumBlocksMarkedForBackup);
    SAPDBERR_ASSERT_STATE (m_NumBlocksUsed == (m_LastBlock + 1 - NumBlocksFree));
    SAPDBERR_ASSERT_STATE (m_NumBlocksFreeAfterSVP == NumBlockFreeAfterSVP);

    /* check that either all blocks are marked for backup or for compression */
    SAPDBERR_ASSERT_STATE ((NumBlocksMarkedForCompression == NumBlocksMarkedForBackUp) ||
                           (NumBlocksMarkedForCompression == 0));

    /* check that the device is not used for a backup and gets compressed at the same time */
    SAPDBERR_ASSERT_STATE (m_VolumeIsInCompressingMode ==  (NumBlocksMarkedForCompression>0));
    SAPDBERR_ASSERT_STATE (m_VolumeIsInCompressingMode  || (m_LastFragmentedBlock.IsInvalid()));
    SAPDBERR_ASSERT_STATE (!m_VolumeIsInCompressingMode || (m_ActBlockNoForBackup.IsInvalid()));

}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::RestoreBlockStateMarkedForBackup (IOMan_BlockNo  BlockNo,
        SAPDB_Bool &bRestoredStateIsFreeAfterSVP)
{
    /* check if the state to be restored is correct */
    SAPDBERR_ASSERT_STATE ((GetBlockState             (BlockNo) == BlockState_BackUp)       &&
                           ((GetBlockStateAfterBackup (BlockNo) == BlockState_Occupied) ||
                            (GetBlockStateAfterBackup  (BlockNo) == BlockState_FreeAfterSVP)));

    /* update counter of blocks marked for back up */
    --m_NumBlocksMarkedForBackup;

    /* restore the state to the one it had before the backup. If  */
    /* the restored state (=return value of the restore function) */
    /* is FreeAfterSVP the appropriate counter is updated         */
    if (m_BlockStateList.RestoreBlockStateMarkedForBackup (BlockNo) != BlockState_FreeAfterSVP)
        bRestoredStateIsFreeAfterSVP = false;
    else
    {
        ++m_NumBlocksFreeAfterSVP;
        bRestoredStateIsFreeAfterSVP = true;
    }

#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/

void
FBM_ClusteredDataVolume::SetBlockStateMarkedForCompressionToFreeAfterSVP (IOMan_BlockNo BlockNo)
{

    /* check if the FBM is in the compression mode where this state can only exist */
    SAPDBERR_ASSERT_STATE (m_VolumeIsInCompressingMode);

    /* check if the transition is correct */
    SAPDBERR_ASSERT_STATE ((GetBlockStateAfterBackup (BlockNo) == BlockState_BackUp) &&
                           (GetBlockState            (BlockNo) == BlockState_BackUp));

    /* set new block state */
    m_BlockStateList.SetBlockStateAfterBackup (BlockNo, BlockState_Free);
    m_BlockStateList.SetBlockState            (BlockNo, BlockState_FreeAfterSVP);

    /* update counter of blocks marked for back up */
    --m_NumBlocksMarkedForBackup;

    /* remove compression mode if there are no more blocks to compress */
    m_VolumeIsInCompressingMode = (m_NumBlocksMarkedForBackup != 0);

    /* update counter of blocks marked as free after savepoint */
    ++m_NumBlocksFreeAfterSVP;

#   ifdef SAPDB_SLOW 
    if( FBM_Check.ChecksLevel( 5 ))
        Verify ();
#   endif
}

/*---------------------------------------------------------------------------*/

// fixme: dumpvar anpassen / neue Struktur???
void
FBM_ClusteredDataVolume::Dump(
    Kernel_Dump  &dump,
    const IOMan_DeviceNo devNo) const
{
    struct FBMDataDevice   fbmDev;

    fbmDev.dmpDevNo                           = devNo;
    fbmDev.dmpLastFragmentedBlock             = m_LastFragmentedBlock;
    fbmDev.dmpNumBlocks                       = m_LastBlock + 1;
    fbmDev.dmpNumBlocksUsed                   = m_NumBlocksUsed;
    fbmDev.dmpNumBlocksMarkedForBackup        = m_NumBlocksMarkedForBackup;
    fbmDev.dmpNumBlocksFreeAfterSVP           = m_NumBlocksFreeAfterSVP;
    fbmDev.dmpNumBlocksToAddUntilOptIsReached = m_NumBlocksToAddUntilOptIsReached;
    fbmDev.dmpActBlockNoForBackup             = m_ActBlockNoForBackup;
    fbmDev.dmpDeviceIsInCompressingMode       = m_VolumeIsInCompressingMode;
    fbmDev.dmpVolumeToDrop                    = m_IsToDrop;
    fbmDev.dmpFiller2                         = 0;

    dump.InsertEntry( Kernel_Dump::DmpFBMDataDevice,
                      Kernel_DumpPage::Entry( &fbmDev, sizeof( fbmDev )));
}
