/*!**************************************************************************

  module      : FBM_Manager.cpp

  -------------------------------------------------------------------------

  responsible : TorstenS

  auhtor      : AlexanderK

  special area: FreeBlockManagement (DDBM)
  description : 


  last changed: 2000-03-10  11:00
  see also    : 

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG



    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

*****************************************************************************/

/*===========================================================================*
*  INCLUDES                                                                 *
*===========================================================================*/

#include <new.h>                                  // SYSTEM: required for placement new
#include "hgg08.h"                                // region identifier
#include "hgg01.h"                                // region identifier
#include "Container/Container_Vector.hpp"
#include "Logging/Log_Types.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "FreeBlockManagement/FBM_Manager.hpp"
#include "FreeBlockManagement/FBM_Dump.hpp"
#include "FreeBlockManagement/FBM_Types.hpp"
#include "FreeBlockManagement/FBM_Exception.hpp"
#include "FreeBlockManagement/FBM_DataVolume.hpp"
#include "FreeBlockManagement/FBM_ClusteredDataVolume.hpp"
#include "FreeBlockManagement/FBM_SequentialDataVolume.hpp"
#include "Pager/Pager_Controller.hpp"
#include "KernelCommon/Kernel_Dump.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/Tasking/RTETask_ITask.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"

/*===========================================================================*
 *  DEFINES                                                                   *
 *===========================================================================*/

#define FBM_MEM_INITIAL_SIZE            (8192 * 8)
#define FBM_MEM_SUPPLEMENT_SIZE         (8192 * 8)

/*===========================================================================*
 *  PUBLIC METHODS                                                           *
 *===========================================================================*/

FBM_Manager* FBM_Manager::m_Instance = NULL;

/*---------------------------------------------------------------------------*/

FBM_Manager&
FBM_Manager::CreateInstance()
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::CreateInstance", FBM_Trace, 5);

    if (m_Instance == NULL)
    {
        m_Instance = new (RTEMem_Allocator::Instance())
                     FBM_Manager(RTEMem_Allocator::Instance());

        if( NULL == m_Instance )
        {
            FBM_Exception errMsg( __CONTEXT__, FBM_ERR_NO_MORE_MEMORY_TO_BUILD_FBM );
            RTE_Crash( errMsg );
        }
    }
    return *m_Instance;
}

/*---------------------------------------------------------------------------*/

FBM_Manager::FBM_Manager(SAPDBMem_IRawAllocator &Allocator)
        :
        m_Allocator( UTF8( "FBM_Manager" ), Allocator ),
        m_DataVolume( m_Allocator ),
        m_ArchiveVolumes( m_Allocator ),
        m_ArchiveVolumeGroups( m_Allocator ),
        m_DBFullWaitQueue( FBM_SUSPEND_HANDLE_DB_FULL )
{
    ResetMembersToNullAndZero();
}

/*---------------------------------------------------------------------------*/

FBM_IManager&
FBM_IManager::Instance ()
{
    return FBM_Manager::Instance();
}

/*---------------------------------------------------------------------------*/

FBM_ReturnCode
FBM_Manager::Restart(
    const tsp00_TaskId                TaskId,
    const IOMan_DeviceNo              MaxNumDev,
    const IOMan_ReservedBlockAddress  &ReservedBlocks,
    IOMan_BlockAddress          *RestartPageAddress )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::Restart", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM (TaskId);

    if( m_Active )
    {
        if(( m_MaxNumDev != MaxNumDev ))
        {
            FBM_Exception errMsg( __CONTEXT__, FBM_ERR_VOLUME_COUNT_INCOMPATIBLE );
            RTE_Message( errMsg );

            return( FBM_IncompatibleVolumeCount );
        }
        return( FBM_AlreadyStarted );
    }

    m_MaxNumDev      = MaxNumDev;
    m_ReservedBlocks = ReservedBlocks;
    m_RestartPageAddress = RestartPageAddress;

    if (g01is_archive())
    {
        m_ArchiveVolumes.Reserve(m_MaxNumDev);

        RTEConf_Parameter::Integer value;
        SAPDBErr_MessageList errorList;
        RTEConf_ParameterAccess::Instance()->GetInteger((RTEConf_Parameter::Name)"DATA_VOLUME_GROUPS", value, errorList);
        m_NumberVolumeGroups = (SAPDB_Int) value;

        m_ArchiveVolumeGroups.Initialize(m_NumberVolumeGroups, 0);
    }

    /* construct array containing the number of free pages for each section */
    SAPDBTRACE_WRITELN( FBM_Trace, 5, "MaxNumDev: " << m_MaxNumDev);

    m_DataVolume.Initialize (m_MaxNumDev, FirstVolNo());
    m_Active = SAPDB_TRUE;

    return( FBM_Okay );
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::Shutdown( const tsp00_TaskId taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::Shutdown", FBM_Trace, 5);

    /* this method releases all memory resources */

    {
        FBM_SynchObject SynchronizeFBM (taskId);

        /* release memory allocated by the FBM_Manager and its member */
        m_DataVolume.Shutdown();

        /* reset all member variables, usage of the FBM_Manager leads to errors now */
        ResetMembersToNullAndZero();
    }
    m_DBFullWaitQueue.ResumeAllWaitingTasks( taskId );
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::RegisterVolume(
    const tsp00_TaskId         TaskId,
    const IOMan_DeviceNo       VolNo,
    const IOMan_BlockCount     DevSize,
    const RTE_VolumeAccessMode VolMode)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RegisterDevice", FBM_Trace, 5);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "VolNo: " << VolNo << " DevSize: "  << DevSize);

    /* check validity of the device number */
    if ((VolNo < FirstVolNo()) || (VolNo > MaxVolNo()))
    {
        /* write error message into knldiag and vtrace */
        FBM_WrongDevNo WrongDevNo(__CONTEXT__, VolNo, FirstVolNo(), MaxVolNo());
        RTE_Message (WrongDevNo);
        return false;
    }
    {
        FBM_SynchObject SynchronizeFBM (TaskId);

        if (RTE_VolumeAccessModeSequential == VolMode)
        {
            m_DataVolume.Set(VolNo, new (m_Allocator) FBM_SequentialDataVolume(m_Allocator, DevSize, VolMode));
        }
        else
        {
            if (!Pager_Controller::GetInstance().IsClusteredIOConfigured())
            {
                m_DataVolume.Set(VolNo, new (m_Allocator) FBM_DataVolume(m_Allocator, DevSize, VolMode));
            }
            else
            {
                m_DataVolume.Set(VolNo, new (m_Allocator) FBM_ClusteredDataVolume(m_Allocator, DevSize, VolMode));
            }
        }

        /* initialize device description and fill it with start values */
        m_DataVolume [VolNo].InitVolume();

        // update archiveVolumes
        if (RTE_VolumeAccessModeSequential == VolMode)
        {
            SAPDBERR_ASSERT_STATE (g01is_archive());

            m_ArchiveVolumes.InsertEnd(VolNo);
            // recalc VolumeGroup startOffsets
            for (SAPDB_UInt loop=0; loop < m_NumberVolumeGroups; loop++)
            {
                m_ArchiveVolumeGroups[loop] = (m_ArchiveVolumes.GetSize() / m_NumberVolumeGroups) * loop;
            }
        }

        /* update global block counters */
        m_NumDev             ++;
        if (VolNo > m_LastVolume)
        {
            m_LastVolume = VolNo;
        }
        m_TotalNumBlocks     += DevSize;
        m_TotalNumBlocksFree += DevSize;

        /* reserve blocks on the device for example for the restart record */
        IOMan_ReservedBlockAddressIterator Iter( m_ReservedBlocks, VolNo );

        for( Iter.Begin(); !Iter.End(); ++Iter )
        {
            m_TotalNumBlocksFree --;
            m_DataVolume[VolNo].SetBlockStateToOccupied( Iter.GetBlockAddress().GetBlockNo());
        }
        if (m_RestartPageAddress->GetDeviceNo() == VolNo)
        {
            m_DataVolume[VolNo].SetBlockStateToOccupied( m_RestartPageAddress->GetBlockNo());
        }

        // After marking the reserved blocks the volume is usable
        m_NextDevToUseForNewBlock = VolNo;
    }
    m_DBFullWaitQueue.ResumeAllWaitingTasks( TaskId );

    return true;
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::DeRegisterVolume(
        const RTETask_ITask     &task,
        Msg_List                &msgList,
        const IOMan_DeviceNo    volNo)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::DeRegisterVolume", FBM_Trace, 5);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "VolNo: " << volNo);

    /* check validity of the device number */
    if ((volNo < FirstVolNo()) || (volNo > MaxVolNo()))
    {
        msgList.AppendNewMessage( FBM_WrongDevNo( __CONTEXT__, 
            volNo, FirstVolNo(), MaxVolNo()));
        return false;
    }

    FBM_IVolume * volPtr = &m_DataVolume[volNo];
    destroy(volPtr, m_Allocator);

    m_DataVolume.Set(volNo,NULL);

    return true;
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::PrepareVolumeForDrop(
        const RTETask_ITask     &task,
        Msg_List                &msgList,
        const IOMan_DeviceNo    volNo)
{
        FBM_SynchObject SynchronizeFBM (task.ID());

        /* check device number and get a reference to the current device */
        if (NULL == &m_DataVolume[volNo])
        { 
            msgList.AppendNewMessage( FBM_WrongDevNo( __CONTEXT__, 
                volNo, FirstVolNo(), MaxVolNo()));
            return false;
        }

        FBM_IVolume& IVolume = m_DataVolume [volNo];

        if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        {
            msgList.AppendNewMessage( FBM_Exception( __CONTEXT__, 
                FBM_ERR_NO_CLASSIC_VOLUME, SAPDB_ToString( volNo, _T_d )));
            return false;
        }

        FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

        Volume.PrepareForDrop();
        IOMan_BlockCount numBlocksUsed = Volume.GetNumBlocksUsed();
        IOMan_BlockCount numBlocks = Volume.GetNumBlocks();

        /* update global block counters */
        m_NumDev             --;
        m_TotalNumBlocks     -= numBlocks;
        m_TotalNumBlocksFree -= numBlocks - numBlocksUsed;

        return true;
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::RestoreVolumeForDrop(
        const RTETask_ITask     &task,
        Msg_List                &msgList,
        const IOMan_DeviceNo    volNo)
{
        FBM_SynchObject SynchronizeFBM (task.ID());

        /* check device number and get a reference to the current device */
        if (NULL == &m_DataVolume[volNo])
        {
            msgList.AppendNewMessage( FBM_WrongDevNo( __CONTEXT__, 
                volNo, FirstVolNo(), MaxVolNo()));
            return false;
        }

        FBM_IVolume & IVolume = m_DataVolume [volNo];

        if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        {
            msgList.AppendNewMessage( FBM_Exception( __CONTEXT__, 
                FBM_ERR_NO_CLASSIC_VOLUME, SAPDB_ToString( volNo, _T_d )));
            return false;
        }

        FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

        Volume.RestoreForDrop();

        IOMan_BlockCount numBlocksUsed = Volume.GetNumBlocksUsed();
        IOMan_BlockCount numBlocks = Volume.GetNumBlocks();

        /* update global block counters */
        m_NumDev             ++;
        m_TotalNumBlocks     += numBlocks;
        m_TotalNumBlocksFree += numBlocks - numBlocksUsed;

        return true;
}

/*---------------------------------------------------------------------------*/

bool
FBM_Manager::IsVolumeToDrop(
        const IOMan_DeviceNo    volNo) 
{

        /* check device number and get a reference to the current device */
        if (NULL == &m_DataVolume[volNo])
        {
            return false;
        }

        FBM_IVolume & IVolume = m_DataVolume [volNo];

        if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        {
            return false;
        }

        FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

        return Volume.IsToDrop();
}


void
FBM_Manager::SetClusterStateToFree(
    const tsp00_TaskId         TaskId,
    const IOMan_ClusterAddress &ClusterAddress )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetClusterStateToFree", FBM_Trace, 5);

    IOMan_BlockCount pageCnt = ClusterAddress.GetBlockCount();

    for (SAPDB_Int4 loop=0; loop < pageCnt; loop++){
        SetBlockStateToFree(TaskId,ClusterAddress.GetBlockAddress(loop));
    }
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetBlockStateToFree(
    const tsp00_TaskId       TaskId,
    const IOMan_BlockAddress &BlockAddress )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToFree", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetFree: DevNo: " << DevNo << " BlockNo: " << BlockNo);

    /* check that the block containing the restart record is never set to free */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    {
        FBM_SynchObject SynchronizeFBM (TaskId);

        /* check device number and get a reference to the current device */
        FBM_IVolume & IVolume = m_DataVolume [DevNo];

        if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
            return;

        FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

        /* check that the current state is  occupied */
        switch (Volume.GetBlockState (BlockNo))
        {
        case BlockState_Occupied:
            /* set block state to free and update all auxilliary structures */
            if (!Volume.IsToDrop())     
            {
                m_TotalNumBlocksFree ++;
            }
            Volume.SetBlockStateToFree (BlockNo);
            break;
        case BlockState_BackUp:
            if (Volume.GetBlockStateAfterBackup (BlockNo) == BlockState_BackUp)
            {
                /* set  back up block state to free */
                Volume.SetBlockStateAfterBackup (BlockNo, BlockState_Free);
                break;
            }
            // fatal error use default action
        default:
            {
                /* if none of the cases above was fulfilled an error occured */
                /* write error message into knldiag and vtrace               */
                FBM_IllegalTransition ErrorMessage (__CONTEXT__, DevNo, BlockNo,
                                                    Volume.GetBlockState (BlockNo),
                                                    Volume.GetBlockStateAfterBackup (BlockNo),
                                                    BlockState_Free,
                                                    BlockState_Free);
                RTE_Crash(ErrorMessage);

            }
        }
    }
    if((0 < GetNumberOfFreeBlocks()) && ( ! m_DBFullWaitQueue.IsEmpty())){
        m_DBFullWaitQueue.ResumeAllWaitingTasks( TaskId );
    }
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupied (
    const tsp00_TaskId        TaskId,
    IOMan_IBlockAddressIterator &BlockIterator,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToOccupied", FBM_Trace, 5);

    FBM_SynchObject SynchronizeFBM (TaskId);

    while (BlockIterator.hasMoreElements()){
        IOMan_BlockAddress BlockAddress = BlockIterator.getNextElement();

        if (!SetBlockStateToOccupiedUnsynched ( BlockAddress, bAbortIfErrorOccured))
        {
            return SAPDB_FALSE;
        }
    }
    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupied(
    const tsp00_TaskId        TaskId,
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    FBM_SynchObject SynchronizeFBM (TaskId);
    return SetBlockStateToOccupiedUnsynched ( BlockAddress, bAbortIfErrorOccured);
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToOccupiedUnsynched (
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToOccupied", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetOcc: DevNo: " << DevNo << " BlockNo: " << BlockNo );

    /* check device number and get a reference to the current device */
    FBM_IVolume & IVolume = m_DataVolume [DevNo];

    /* check that the current state is free */
    if (IVolume.GetBlockState (BlockNo) == BlockState_Free)
    {
        if(RTE_VolumeAccessModeNormal == IVolume.VolMode()) {
            SAPDBERR_ASSERT_STATE ((((FBM_IDataVolume&)IVolume).GetBlockStateAfterBackup (BlockNo) == BlockState_Free))
        }

        m_TotalNumBlocksFree --;

        /* set block state to occupied and update all auxilliary structures */
        IVolume.SetBlockStateToOccupied (BlockNo);
        return( SAPDB_TRUE );
    }

    if(RTE_VolumeAccessModeSequential == IVolume.VolMode()) {
        return( SAPDB_TRUE );
    }
    else
    {
        /* if none of the cases above was fulfilled an error occured */
        /* write error message into knldiag and vtrace               */

        FBM_BlockState backupState;
        if(RTE_VolumeAccessModeNormal == IVolume.VolMode())
        {
            backupState = ((FBM_IDataVolume&)IVolume).GetBlockStateAfterBackup (BlockNo);
        }
        else
        {
            backupState = BlockState_Free;
        }

        FBM_IllegalTransition errMsg (__CONTEXT__, DevNo, BlockNo,
                                      IVolume.GetBlockState (BlockNo), backupState,
                                      BlockState_Occupied, BlockState_Free);

        if( FBM_ABORT_IF_ERROR == bAbortIfErrorOccured )
            RTE_Crash( errMsg );
        else
            RTE_Message( errMsg );

        return( SAPDB_FALSE );
    }
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
FBM_Manager::SetBlockStateToBackup (
    const tsp00_TaskId        TaskId,
    const IOMan_BlockAddress &BlockAddress,
    const SAPDB_Bool          bAbortIfErrorOccured )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToBackup", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetBackU DevNo " << DevNo << " BlockNo: " << BlockNo );

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check that the block containing the restart record is never set to backup */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    /* check device number and get a reference to the current device */
    FBM_IVolume & IVolume = m_DataVolume [DevNo];

    if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        return( SAPDB_TRUE );

    FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

    /* check that no more blocks are marked after reading of the marked blocks has started     */
    /* which is displayed by the iterator m_ActBlockNoForBackup indicating the last read block */
    SAPDBERR_ASSERT_STATE (Volume.GetActBlockNoForBackup().IsInvalid());

    /* check that the current state is occupied */
    if (BlockState_Occupied == Volume.GetBlockState (BlockNo))
    {
        Volume.SetBlockStateToBackUp(BlockNo);
        return( SAPDB_TRUE );
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace */
    FBM_IllegalTransition errMsg (__CONTEXT__, DevNo, BlockNo,
                                  Volume.GetBlockState (BlockNo),
                                  Volume.GetBlockStateAfterBackup (BlockNo),
                                  BlockState_BackUp,
                                  BlockState_Occupied);

    if( FBM_ABORT_IF_ERROR == bAbortIfErrorOccured )
        RTE_Crash( errMsg );
    else
        RTE_Message( errMsg );

    return( SAPDB_FALSE );
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetBlockStateToFreeAfterSVP(
    const tsp00_TaskId       TaskId,
    const IOMan_BlockAddress &BlockAddress )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetBlockStateToFreeAfterSVP", FBM_Trace, 5);

    const SAPDB_UInt  DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt  BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "SetFSVP: DevNo: " << DevNo << " BlockNo: " << BlockNo );

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check that the block containing the restart record is never set to fsvp */
    SAPDBERR_ASSERT_STATE ( ! m_ReservedBlocks.IsMember( BlockAddress ));

    /* check volume number and get a reference to the current volume */
    FBM_IVolume & IVolume = m_DataVolume [DevNo];

    if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        return;

    FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

    /* check that the current state is  occupied */
    switch (Volume.GetBlockState (BlockNo))
    {
    case BlockState_Occupied:
        Volume.SetBlockStateToFreeAfterSVP (BlockNo);
        ++m_TotalNumBlocksFreeAfterSVP;
        return;
    case BlockState_BackUp:
        {
            switch (Volume.GetBlockStateAfterBackup (BlockNo))
            {
            case BlockState_Occupied:
                Volume.SetBlockStateAfterBackup (BlockNo, BlockState_FreeAfterSVP);
                return;
            case BlockState_BackUp:
                ++m_TotalNumBlocksFreeAfterSVP;
                Volume.SetBlockStateMarkedForCompressionToFreeAfterSVP (BlockNo);
                return;
            }
        }
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace */
    FBM_IllegalTransitionToFreeAfterSVP ErrorMessage (__CONTEXT__, DevNo, BlockNo,
            Volume.GetBlockState (BlockNo),
            Volume.GetBlockStateAfterBackup (BlockNo));
    RTE_Crash(ErrorMessage);
}

/*---------------------------------------------------------------------------*/

IOMan_ClusterAddress
FBM_Manager::GetMultFreeBlocks(
    const tsp00_TaskId        taskId,
    const IOMan_BlockCount    NumFreeBlocksWanted,
    const SAPDB_Bool          bReqSequential )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetMultFreeBlocks", FBM_Trace, 5);

    IOMan_BlockAddress      block;
    IOMan_ClusterAddress    cluster;
    SAPDB_Bool              bTriggerSavepoint = false;
    IOMan_BlockCount        NumFreeBlocksSupplied = 0;

    /* This following block is needed to leave the FBM region before entering the     */
    /* log region in the block below. This is done to avoid a dead lock. The FBM      */
    /* region is left when the FBM_SynchObject is destructed at the end of this block */

    while( m_Active )
    {
        {
            FBM_SynchObject SynchronizeFBM( taskId );

            /* get number of a device containing free blocks */
            const IOMan_DeviceNo DevNo = GetDeviceWithFreeBlocks (NumFreeBlocksWanted, bReqSequential);

            if (DevNo.IsInvalid())
            {
                FBM_Exception NoMoreSpace( __CONTEXT__, FBM_NO_MORE_FREE_BLOCKS );
                RTE_Message( NoMoreSpace );
            }
            else
            {
                block.SetAddress( DevNo, m_DataVolume [DevNo].GetMultFreeBlocks( NumFreeBlocksWanted,
                                  NumFreeBlocksSupplied ));
                cluster.SetCluster( block, NumFreeBlocksSupplied );

                SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free DevNo  " << cluster.GetDeviceNo()
                                    << " Free BlockNo: " << cluster.GetBlockNo( 0 ));

                SAPDBTRACE_WRITELN( FBM_Trace, 5, "Wanted Num: " << NumFreeBlocksWanted
                                    << " Suppl.Num: " << NumFreeBlocksSupplied );

                /* decrease the counter of free blocks and check if there are too */
                /* few free blocks which means that a savepoint must be triggered */
                m_TotalNumBlocksFree -= NumFreeBlocksSupplied;
                if( SavepointNeeded() )
                {
                    bTriggerSavepoint      = true;
                    m_SavepointIsRequested = true;
                }
                break; // usable cluster found => okay
            }
        }
        m_DBFullWaitQueue.InsertTaskAndWait( taskId );
    }
    if( m_Active && bTriggerSavepoint )
    {
        TriggerSavepoint( taskId );
    }
    return cluster;
}

/*---------------------------------------------------------------------------*/

IOMan_BlockAddress
FBM_Manager::GetFreeBlock(
    const tsp00_TaskId  taskId,
    const SAPDB_Bool    bReqSequential)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetFreeBlock", FBM_Trace, 5);

    SAPDB_Bool         bTriggerSavepoint = false;
    IOMan_BlockAddress BlockAddress;

    /* This following block is needed to leave the FBM region before entering the     */
    /* log region in the block below. This is done to avoid a dead lock. The FBM      */
    /* region is left when the FBM_SynchObject is destructed at the end of this block */

    while( m_Active )
    {
        {
            FBM_SynchObject SynchronizeFBM( taskId );

            /* get number of a device containing free blocks */
            const IOMan_BlockCount cNumFreeBlocksWanted = 1;
            const IOMan_DeviceNo DevNo = GetDeviceWithFreeBlocks (cNumFreeBlocksWanted, bReqSequential);

            if (DevNo.IsInvalid())
            {
                FBM_Exception NoMoreSpace( __CONTEXT__, FBM_NO_MORE_FREE_BLOCKS );
                RTE_Message( NoMoreSpace );
            }
            else
            {
                BlockAddress.SetAddress (DevNo, m_DataVolume[DevNo].GetFreeBlock());

                SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free DevNo: " << BlockAddress.GetDeviceNo()
                                    << " Free BlockNo: " << BlockAddress.GetBlockNo() );

                /* decrease the counter of free blocks and check if there are too */
                /* few free blocks which means that a savepoint must be triggered */
                -- m_TotalNumBlocksFree;

                if( SavepointNeeded() )
                {
                    bTriggerSavepoint      = true;
                    m_SavepointIsRequested = true;
                }
                break; // usable cluster found => okay
            }
        }
        m_DBFullWaitQueue.InsertTaskAndWait( taskId );
    }
    if( m_Active && bTriggerSavepoint )
    {
        TriggerSavepoint( taskId );
    }
    return BlockAddress;
}

/*---------------------------------------------------------------------------*/

IOMan_DeviceNo
FBM_Manager::GetFreeVolumeInGroup(
    const SAPDB_UInt       CurrentVolumeGroup,
    const IOMan_BlockCount NumRequestedBlocks )
{
    SAPDB_UInt     VolumesPerGroup;
    IOMan_DeviceNo VolWithFreeBlocks;
    SAPDB_UInt     VolIndex;

    SAPDBERR_ASSERT_STATE (g01is_archive());

    if (0 == m_ArchiveVolumes.GetSize() % m_NumberVolumeGroups)
    {
        VolumesPerGroup = m_ArchiveVolumes.GetSize() / m_NumberVolumeGroups;
    }
    else
    {
        VolumesPerGroup = m_ArchiveVolumes.GetSize() / m_NumberVolumeGroups + 1;
    }

    if (CurrentVolumeGroup * VolumesPerGroup >= m_ArchiveVolumes.GetSize())
    {
        VolWithFreeBlocks.Invalidate();
        return VolWithFreeBlocks;
    }

    VolIndex = m_ArchiveVolumeGroups[CurrentVolumeGroup];

    do
    {
        SAPDBERR_ASSERT_STATE (RTE_VolumeAccessModeSequential == m_DataVolume[m_ArchiveVolumes[VolIndex]].VolMode())

        if (m_DataVolume[m_ArchiveVolumes[VolIndex]].GetNumBlocksFree() < NumRequestedBlocks)
        {
            SAPDBTRACE_WRITELN( FBM_Trace, 5, "Archive Volume Full: " << m_ArchiveVolumes[VolIndex] );
            VolIndex++;
            if ((VolIndex >= ((CurrentVolumeGroup+1)*VolumesPerGroup)) || (VolIndex >=m_ArchiveVolumes.GetSize()))
            {
                VolIndex = CurrentVolumeGroup*VolumesPerGroup;
            }
        }
        else
        {
            VolWithFreeBlocks = m_ArchiveVolumes[VolIndex];
        }
    } while ((VolWithFreeBlocks.IsInvalid()) && (VolIndex != m_ArchiveVolumeGroups[CurrentVolumeGroup]));

    m_ArchiveVolumeGroups[CurrentVolumeGroup] = VolIndex;

    return VolWithFreeBlocks;
}

/*---------------------------------------------------------------------------*/

IOMan_DeviceNo
FBM_Manager::GetDeviceWithFreeBlocks(
    const IOMan_BlockCount NumRequestedBlocks,
    const SAPDB_Bool bReqSequential )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetDeviceWithFreeBlocks", FBM_Trace, 5);

    IOMan_DeviceNo  iVolNo;
    IOMan_DeviceNo  VolWithFreeBlocks;
    SAPDB_Bool      VolumeFound        = SAPDB_FALSE;

    if (bReqSequential)// archive Volume
    {
        SAPDB_UInt StartVolumeGroup = m_CurArchiveVolumeGroup;

        do
        {
            m_CurArchiveVolumeGroup++;
            if (m_CurArchiveVolumeGroup >= m_NumberVolumeGroups)
            {
                m_CurArchiveVolumeGroup = 0;
            }

            VolWithFreeBlocks = GetFreeVolumeInGroup(m_CurArchiveVolumeGroup, NumRequestedBlocks);

            if (VolWithFreeBlocks.IsValid())
            {
                VolumeFound = SAPDB_TRUE;
            }
        }while(!VolumeFound && (m_CurArchiveVolumeGroup != StartVolumeGroup));
    }
    else
    {
        /* search a data device which has still NumRequestedBlocks free blocks */
        IOMan_DeviceNo  SearchBegDevNo     = m_NextDevToUseForNewBlock;

        iVolNo                         = SearchBegDevNo;

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "FirstVolNo: " << FirstVolNo()
                            << " LastVolNo: " << m_LastVolume );

        SAPDBTRACE_WRITELN( FBM_Trace, 5, "SearchBegDev: " << SearchBegDevNo
                            << " BlocksWanted: " << NumRequestedBlocks );

        /* loop over all devices until a device with NumRequestedBlocks free blocks is found */
        do
        {
            if (NULL != &m_DataVolume[iVolNo])
            {
                if (RTE_VolumeAccessModeNormal == m_DataVolume[iVolNo].VolMode())
                {
                    FBM_IDataVolume & Volume = (FBM_IDataVolume&) m_DataVolume[iVolNo];
                    if (Volume.GetNumBlocksFree() < NumRequestedBlocks)
                    {
                        SAPDBTRACE_WRITELN( FBM_Trace, 5, "Device Full: " << iVolNo);
                    }
                    else
                    {
                        VolumeFound       = SAPDB_TRUE;
                        VolWithFreeBlocks = iVolNo;

                        /* if this device contains too few used blocks this device is choosen imediately.   */
                        /* Otherwise this device is skipped this time and the search is continued. However, */
                        /* m_NumBlocksToAddUntilOptIsReached is changed so that this device is choosen      */
                        /* next time it is looked at                                                        */
                        if (Volume.GetNumBlocksToAddUntilOptIsReached() > 0)
                            break;
                        else
                            Volume.SetNumBlocksToAddUntilOptIsReached(1);
                    }

                }
            }
            /* goto next volume */
            if ((++iVolNo) > MaxVolNo())
            {
                iVolNo = FirstVolNo();
            }

        }
        while (iVolNo != SearchBegDevNo);

        /* return device or throw exception if there is no device with enough free blocks */
        if (VolumeFound)
        {
            /* set device where to start the next search */
            if (VolWithFreeBlocks == MaxVolNo())
            {
                m_NextDevToUseForNewBlock = FirstVolNo();
            }
            else
            {
                m_NextDevToUseForNewBlock = VolWithFreeBlocks+1;
            }
        }
    }
    if (!VolumeFound)
    {
        VolWithFreeBlocks.Invalidate();
    }

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "Free_Device: " << VolWithFreeBlocks);

    return (VolWithFreeBlocks) ;
}

/*---------------------------------------------------------------------------*/

IOMan_BlockCount
FBM_Manager::NumBlocksMarkedForBackup( const tsp00_TaskId taskId ) const
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::NumBlocksMarkedForBackup", FBM_Trace, 5);

    IOMan_BlockCount NumAllBlocksMarkedForBackUp = 0;

    FBM_SynchObject SynchronizeFBM( taskId );

    for (IOMan_DeviceNo iVolNo=FirstVolNo(); iVolNo<=MaxVolNo(); ++iVolNo)
    {
        if (NULL != &m_DataVolume[iVolNo])
        {
            const FBM_IVolume & IVolume = m_DataVolume[iVolNo];

            if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
                continue;

            const FBM_IDataVolume & Volume = (FBM_IDataVolume &) IVolume;

#           ifdef SAPDB_SLOW 
            if (Volume.GetNumBlocksMarkedForBackup() > 0)
                SAPDBTRACE_WRITELN( FBM_Trace, 5, "bup blocks  " <<  Volume.GetNumBlocksMarkedForBackup());
#           endif

            NumAllBlocksMarkedForBackUp += Volume.GetNumBlocksMarkedForBackup();
        }
    }
    return NumAllBlocksMarkedForBackUp;
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreAllBlockStatesMarkedForBackup( const tsp00_TaskId TaskId )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreAllBlockStatesMarkedForBackup", FBM_Trace, 5);

    IOMan_DeviceNo      iDevNo;
    IOMan_BlockCount    NumBlocksRestoredToFreeAfterSVP;

    FBM_SynchObject SynchronizeFBM (TaskId);

    for (iDevNo=FirstVolNo(); iDevNo<=MaxVolNo(); iDevNo++)
    {
        if (NULL != &m_DataVolume[iDevNo])
        {
            FBM_IVolume &IVolume = m_DataVolume[iDevNo];

            if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
                continue;

            FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

            Volume.RestoreAllBlockStatesMarkedForBackup(NumBlocksRestoredToFreeAfterSVP);
            m_TotalNumBlocksFreeAfterSVP += NumBlocksRestoredToFreeAfterSVP;
        }
    }
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreBlockStateMarkedForBackup(
    const tsp00_TaskId       TaskId,
    const IOMan_BlockAddress &BlockAddress)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreBlockStateMarkedForBackup", FBM_Trace, 5);

    const SAPDB_UInt DevNo   = BlockAddress.GetDeviceNo();
    const SAPDB_UInt BlockNo = BlockAddress.GetBlockNo();

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "RestoreB Dev: " << DevNo
                        << " BlockNo: " <<BlockNo );

    SAPDB_Bool bRestoredStateIsFreeAfterSVP;

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_IVolume & IVolume = m_DataVolume [DevNo];

    if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        return;

    FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

    if (BlockState_BackUp == Volume.GetBlockState(BlockNo))
    {
        Volume.RestoreBlockStateMarkedForBackup (BlockNo, bRestoredStateIsFreeAfterSVP);

        if (bRestoredStateIsFreeAfterSVP)
            ++m_TotalNumBlocksFreeAfterSVP;
        return;
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalRestoreAfterBackup ErrorMesage (__CONTEXT__, DevNo, BlockNo,
            Volume.GetBlockState(BlockNo),
            Volume.GetBlockStateAfterBackup(BlockNo));
    RTE_Crash(ErrorMesage);
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::RestoreBlockStateMarkedForCompression(
    const tsp00_TaskId      TaskId,
    const IOMan_DeviceNo    DevNo,
    const IOMan_BlockNo     BlockNo,
    SAPDB_Bool             &bBlockInUse )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::RestoreBlockStateMarkedForCompression", FBM_Trace, 5);

    /* this method checks if a given block is in the state  [backup,backup] */
    /* which means that it is marked for compression and still in use then  */
    /* the original state is restored, the block can also be in the state   */
    /* [freeaftersavepouint, free] which means that there is nothing to do  */

    /* initialize return value */
    bBlockInUse = false;

    FBM_SynchObject SynchronizeFBM (TaskId);

    /* check device number and get a reference to the current device */
    FBM_IVolume & IVolume = m_DataVolume [DevNo];

    if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
        return;

    FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

    const FBM_BlockState BlockState            = Volume.GetBlockState           (BlockNo);
    const FBM_BlockState BlockStateAfterBackup = Volume.GetBlockStateAfterBackup(BlockNo);

    switch (BlockState)
    {
    case  BlockState_BackUp:
        {
            switch (BlockStateAfterBackup)
            {
            case BlockState_BackUp:
                /* restore the state to occupied*/
                Volume.SetBlockStateMarkedForCompressionToOccupied (BlockNo);
                bBlockInUse = true;
                return;
            case BlockState_Free:
                /* restore state to free */
                Volume.SetBlockStateMarkedForCompressionToFree (BlockNo);
                return;
            }
        }
        break;
    case BlockState_FreeAfterSVP:
        SAPDBERR_ASSERT_STATE (BlockStateAfterBackup == BlockState_Free);
        return;
    }

    /* if none of the cases above was fulfilled an error occured */
    /* write error message into knldiag and vtrace               */
    FBM_IllegalRestoreAfterCompression ErrorMesage (__CONTEXT__, DevNo, BlockNo, BlockState, BlockStateAfterBackup);
    RTE_Crash(ErrorMesage);
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::SetAllBlocksMarkedAsFreeAfterSVPToFree( const tsp00_TaskId TaskId )
{
    /* go through all registered devices an set the states of all */
    /* blocks which are marked as free after savepoint to free    */

    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::SetAllBlocksMarkedAsFreeAfterSVPToFree", FBM_Trace, 5);

    {
        FBM_SynchObject SynchronizeFBM (TaskId);

        IOMan_DeviceNo      iDevNo;

        for (iDevNo=FirstVolNo(); iDevNo<=MaxVolNo(); iDevNo++)
        {
            if (NULL != &m_DataVolume[iDevNo])
            {
                FBM_IVolume &IVolume = m_DataVolume[iDevNo];

                if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
                    continue;

                FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

                if (Volume.GetNumBlocksFreeAfterSVP() > 0)
                {
                    /* update global counters of all blocks  free and free after save point */
                    m_TotalNumBlocksFreeAfterSVP -= Volume.GetNumBlocksFreeAfterSVP();
                    if (!Volume.IsToDrop())
                    {
                        m_TotalNumBlocksFree         += Volume.GetNumBlocksFreeAfterSVP();
                    }

                    Volume.SetAllBlocksMarkedAsFreeAfterSVPToFree();
                }
            }
        }

        /* update counters which control which devices are prefered when free blocks are   */
        /* requested  (getfreeblock) depending on the number of used blocks on each device */
        ChangeDeviceUsagePreferences ();

        /* reset the trigger flag,i.e. from now the routine GetFreeBlock */
        /* can trigger again a savepoint if there are too few blocks     */
        m_SavepointIsRequested = false;

        /* check that the global counter of all blocks which are free are svp is zero now */
        SAPDBERR_ASSERT_STATE (m_TotalNumBlocksFreeAfterSVP == 0)
    }
    if( 0 < GetNumberOfFreeBlocks() && ! m_DBFullWaitQueue.IsEmpty()){
        m_DBFullWaitQueue.ResumeAllWaitingTasks( TaskId );
    }
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::GetNextBlocksForBackUp(
    const tsp00_TaskId     taskId,
    const IOMan_DeviceNo   DevNo,
    const IOMan_BlockCount MaxNumBlocksWanted,
    IOMan_BlockCount      &SuppliedNumBlocks,
    IOMan_BlockNo         &BlockNo,
    tgg00_BasisError      &trError)
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::GetNextBlocksForBackUp", FBM_Trace, 5);

    if( ! m_Active )
    {
        trError = e_shutdown;
        return;
    }
    /* this routine supplies a set of SuppliedNumBlocks  */
    /* neighbouring blocks  marked for backup            */
    /* see method BeginReadingBlocksMarkedForBackUp      */

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << DevNo);

    FBM_SynchObject SynchronizeFBM( taskId );

    /* check device number and get a reference to the current device */
    FBM_IVolume & iDevice = m_DataVolume [DevNo];

    if (RTE_VolumeAccessModeNormal != iDevice.VolMode())
        return;

    FBM_IDataVolume &Device = (FBM_IDataVolume&) iDevice;

    /* get blocks marked for backup */
    Device.GetNextBlocksForBackUp (MaxNumBlocksWanted, SuppliedNumBlocks, BlockNo);

    SAPDBTRACE_WRITELN( FBM_Trace, 5, "num_found: " << SuppliedNumBlocks
                        << " BlockNo: " << BlockNo );
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::BeginReadingBlocksMarkedForBackUp( const tsp00_TaskId TaskId )
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::BeginReadingBlocksMarkedForBackUp", FBM_Trace, 5);

    /* this routine sets the  position (iterator)                                     */
    /* m_ActBlockNoForBackup from where to start to look for blocks marked for backup */
    /* first position on the device. All blocks marked for backup can than be fetched */
    /* by calling GetNextBlocksForBackUp until this function returns no more blocks   */

    FBM_SynchObject SynchronizeFBM (TaskId);

    for (int iDevNo=FirstVolNo(); iDevNo<=MaxVolNo(); ++iDevNo)
    {
        if (NULL != &m_DataVolume[iDevNo])
        {
            FBM_IVolume & IVolume = m_DataVolume[iDevNo];

            if (RTE_VolumeAccessModeNormal != IVolume.VolMode())
                continue;

            FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;

            SAPDBERR_ASSERT_STATE (Volume.GetActBlockNoForBackup().IsInvalid());

            if (Volume.GetNumBlocksMarkedForBackup() > 0)
            {
                Volume.SetActBlockNoForBackup(0);
            }
        }
    }
}
/*---------------------------------------------------------------------------*/

void
FBM_Manager::Dump(
    const tsp00_TaskId    taskId,
    Kernel_Dump           &dump ) const
{
    struct FBMManager   fbmMan;

    FBM_SynchObject SynchronizeFBM( taskId );

    fbmMan.dmpActive                     = m_Active;
    fbmMan.dmpSvpIsTriggered             = m_SavepointIsRequested;
    fbmMan.dmpFiller1                    = 0;
    fbmMan.dmpMaxNumDev                  = m_MaxNumDev;
    fbmMan.dmpNumDev                     = m_NumDev;
    fbmMan.dmpLastVolume                 = m_LastVolume;
    fbmMan.dmpTotalNumBlocksFreeAfterSVP = m_TotalNumBlocksFreeAfterSVP;
    fbmMan.dmpTotalNumBlocksFree         = m_TotalNumBlocksFree;
    fbmMan.dmpTotalNumBlocks             = m_TotalNumBlocks;
    fbmMan.dmpTotalUsedBlocks            = m_TotalNumBlocks - m_TotalNumBlocksFree
                                           - m_TotalNumBlocksFreeAfterSVP;
    fbmMan.dmpNextDevToUseForNewBlock    = m_NextDevToUseForNewBlock;
    fbmMan.dmpNextDevToUseForFragBlock   = m_NextDevToUseForFragBlock;

    dump.InsertEntry( Kernel_Dump::DmpFBMManager,
                      Kernel_DumpPage::Entry( &fbmMan, sizeof( fbmMan )));

    if( m_Active )
    {
        for( IOMan_DeviceNo devNo = FirstVolNo(); devNo <= MaxVolNo(); ++devNo )
        {
            if (NULL != &m_DataVolume[devNo])
            {
                m_DataVolume[ devNo ].Dump( dump, devNo );
            }
        }
    }
}

/*===========================================================================*
 *  PRIVATE METHODS                                                          *
 *===========================================================================*/

void
FBM_Manager::ChangeDeviceUsagePreferences ()
{
    SAPDBTRACE_ROUTINE_DEBUG ("FBM_Manager::ChangeDeviceUsagePreferences", FBM_Trace, 5);

    /* update counters which control which devices are prefered when free blocks are   */
    /* requested  (getfreeblock) depending on the number of used blocks on each device */
    /* the prefrences are calculated to get optimum parallel read accesses that means  */
    /* all devices should accomodate the same number of blocks                         */
    IOMan_BlockCount MaxNumBlocksUsed = 0;
    IOMan_DeviceNo   iDevNo;

    /* search maximum number of blocks on one device */
    //    MaxNumBlocksUsed = m_DataVolume [FirstVolNo()].GetNumBlocksUsed();
    for (iDevNo=FirstVolNo(); iDevNo<=MaxVolNo(); ++iDevNo)
    {
        if (NULL != &m_DataVolume[iDevNo])
        {
            FBM_IVolume &IVolume = m_DataVolume [iDevNo];

            if (RTE_VolumeAccessModeNormal == IVolume.VolMode())
            {
                FBM_IDataVolume & Volume = (FBM_IDataVolume&) IVolume;
                if (MaxNumBlocksUsed < Volume.GetNumBlocksUsed()){
                    MaxNumBlocksUsed = Volume.GetNumBlocksUsed();
                }
            }
        }
    }

    /* calculate the number of used blocks to add to each */
    /* device that all devices have the same occupation   */

    for (iDevNo=FirstVolNo(); iDevNo<=MaxVolNo(); iDevNo++)
    {
        if (NULL != &m_DataVolume[iDevNo])
        {
            FBM_IVolume & IVolume = m_DataVolume [iDevNo];

            if (RTE_VolumeAccessModeNormal == IVolume.VolMode())
            {
                FBM_IDataVolume &Volume = (FBM_IDataVolume&) IVolume;
                Volume.SetNumBlocksToAddUntilOptIsReached((MaxNumBlocksUsed - Volume.GetNumBlocksUsed()));

                SAPDBTRACE_WRITELN( FBM_Trace, 5, "DevNo: " << iDevNo
                                    << " OptAddNumBlo: " << Volume.GetNumBlocksToAddUntilOptIsReached() );
            }
        }
    }

    m_NextDevToUseForFragBlock = FirstVolNo();
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::ResetMembersToNullAndZero()
{
    m_Active                     = false;
    m_MaxNumDev.Invalidate();
    m_TotalNumBlocksFreeAfterSVP = 0;
    m_TotalNumBlocksFree         = 0;
    m_TotalNumBlocks             = 0;
    m_NumDev                     = 0;
    m_LastVolume                 = 0;
    m_SavepointIsRequested       = false;
    m_NextDevToUseForNewBlock.Invalidate();
    m_NextDevToUseForFragBlock.Invalidate();
    m_LastArchiveVolumeUsed.Invalidate();
    m_CurArchiveVolumeGroup      = 0;
    m_NumberVolumeGroups         = 0;

    m_ArchiveVolumes.Clear();
    m_ArchiveVolumeGroups.Clear();
}

/*---------------------------------------------------------------------------*/

void
FBM_Manager::TriggerSavepoint( const tsp00_TaskId taskId )
{
    Msg_List errlist;
    if ( Log_Savepoint::ok != Log_Savepoint::Instance().StartSavepoint
            ( *RTETask_ITask::ByTaskID(taskId), Log_SVPReasonFBM, errlist ) )
    {
        RTE_Message (errlist);
    }
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
