/*!
    @file       IOMan_DataArea.cpp
    @author     TorstenS
    @ingroup    IOManagement
    @brief      This module is used to manage the data area.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (c) 2001-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/


/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "IOManager/IOMan_DataArea.hpp"
#include "IOManager/IOMan_Messages.hpp"
#include "IOManager/IOMan_MessageList.hpp"

#include "Converter/Converter_IManager.hpp"

#include "DataAccess/Data_PageFrame.hpp"

#include "FreeBlockManagement/FBM_IManager.hpp"

#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/InputOutput/RTEIO_VolumeInfo.hpp"
#include "RunTime/Tasking/RTETask_ITask.hpp"

#include "SAPDBCommon/SAPDB_MemCopyMove.hpp"
#include "SAPDBCommon/SAPDB_ToString.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTION PROTOTYPES                                        *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

SAPDB_Bool
IOMan_DataArea::AddVolume(
    const tsp00_TaskId      taskId,
    const SAPDB_Bool        bAddVolumeOnly,
    tsp00_VFilename         &devName,
    const IOMan_BlockCount  configuredSize,
    const IOMan_DeviceNo    devNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::AddVolume", IOMan_Trace, 6 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 6, "Volume: " << devName <<
                        " VolSize: " << configuredSize << " DevNo: " << devNo );

    if( ! IsDataVolumeNoValid( devNo ) )
    {
        IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_DATA_DEVNO_OUT_OF_RANGE,
                                  SAPDB_ToString( devNo, _T_d ),
                                  SAPDB_ToString( MaxVolumes(), _T_d ));
        RTE_Message( errMsg );
        return SAPDB_FALSE;
    }

    IOMan_DataVolume  &volume      = Get( devNo );
    IOMan_DataVolume  *pPrevVolume = NULL;
    IOMan_DataVolume  *pNextVolume = NULL;

    // Add data volume to the control structures of the RTE, because at the
    // time of kernel restart this volume was not stored within data base
    // configuration file. The volume was made persistent by the dbmsrv after
    // starting the db kernel, therefore vcold couldn't find the volume!

    if(! volume.Add( devName, configuredSize ))
        return SAPDB_FALSE;

    const SAPDB_Byte *pDBIdentifier = DetermineLinkage( volume, &pPrevVolume, &pNextVolume );

    if( NULL == pDBIdentifier )
        return SAPDB_FALSE;

    if( IOMan_Okay == volume.Create( taskId, pDBIdentifier ))
    {
        if( IOMan_Okay == volume.Open( taskId ))
        {
            if( bAddVolumeOnly || FBM_IManager::Instance().RegisterVolume( taskId,
                    volume.GetLogicalDevNo(), volume.GetInternalSize(), volume.GetAccessMode()))
            {
                if( NULL != pPrevVolume )
                    CreateLinkage( taskId, *pPrevVolume, volume );

                if( NULL != pNextVolume )
                    CreateLinkage( taskId, volume, *pNextVolume );

                m_TotalDataPages += volume.GetUsableSize();
                m_UsedVolumes++;
                m_ConfiguredVolumes++;

                // At this moment the blocks of the new volume are usable. 
                // If the converter expansion failed (crash) the new volume 
                // is still valid!
                {
                    IOMan_MessageList msgList( __CONTEXT__, IOMAN_INFO_ADD_VOLUME,
                                               volume.GetTypeName(),
                                               SAPDB_ToString( volume.GetLogicalDevNo(), _T_d ),
                                               SAPDB_ToString( volume.GetConfiguredSize(), _T_d ));
                    RTE_Message( msgList );
                }
                // try to expand converter if data base is online to increase
                // the number of usable pageNo's. Note that it is not fatal if
                // the expand fails, because the converter will grow during the
                // next restart.

                if( ! bAddVolumeOnly )
                {
                    const RTETask_ITask &task = *RTETask_ITask::ByTaskID( taskId );

                    Converter_IManager::Instance().Expand( task, volume.GetUsableSize());
                }
                return SAPDB_TRUE; // normal termination
            }
        }
    }
    volume.Close( taskId );
    volume.Del();
    return SAPDB_FALSE;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_DataArea::DropVolume(
    const tsp00_TaskId      taskId,
    const IOMan_DeviceNo    volNo )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::DropVolume", IOMan_Trace, 6 );

    SAPDBTRACE_WRITELN( IOMan_Trace, 6, "Volume: " <<  volNo );

    if( ! IsDataVolumeNoValid( volNo ) )
    {
        IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_DATA_DEVNO_OUT_OF_RANGE,
                                  SAPDB_ToString( volNo, _T_d ),
                                  SAPDB_ToString( MaxVolumes(), _T_d ));
        RTE_Message( errMsg );
        return SAPDB_FALSE;
    }

    IOMan_DataVolume  &volume = Get( volNo );


    DropLinkage(taskId, volume );

    m_TotalDataPages -= volume.GetUsableSize();
    m_UsedVolumes--;
    m_ConfiguredVolumes--;

    volume.Close( taskId );
    volume.Del();

    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

void
IOMan_DataArea::CloseAll( const tsp00_TaskId taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::CloseAll", IOMan_Trace, 6 );

    ConstIterator   endIter = End();

    for( Iterator iter = Begin(); iter != endIter; ++iter )
        ( *iter ).Close( taskId );

    m_TotalDataPages = 0;
    m_UsedVolumes    = 0;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_DataArea::CreateAll(
    const tsp00_TaskId taskId,
    const SAPDB_Byte   *pDBIdentifier )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::CreateAll", IOMan_Trace, 6 );

    IOMan_MessageList msgList;

    if( 0 != m_TotalDataPages )
    {
        IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_DATA_AREA_ALREADY_ONLINE );
        RTE_Crash( errMsg );
    }

    if( ! FormatAll( taskId, msgList ))
    {
        RTE_Message( msgList );
        return SAPDB_FALSE;
    }

    Iterator        prevIter = End();
    ConstIterator   endIter  = End();

    for( Iterator iter = Begin(); iter != endIter; ++iter )
    {
        IOMan_ReturnCode retCode = ( *iter ).CreateInfoPage( taskId, pDBIdentifier );

        if( IOMan_Okay == retCode )
        {
            if( IOMan_Okay != (( *iter ).Open( taskId )))
                return SAPDB_FALSE;
            else
            {
                if( prevIter != endIter )
                    CreateLinkage( taskId, *prevIter, *iter );

                prevIter          = iter;
                m_TotalDataPages += ( *iter ).GetUsableSize();
                m_UsedVolumes++;
            }
        }
        else if( IOMan_NotConfigured == retCode )
            continue;
        else
            return SAPDB_FALSE;
    }
    SAPDBERR_ASSERT_STATE( m_ConfiguredVolumes == m_UsedVolumes );
    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_DataArea::GetVolumeNo(
    tsp00_VFilename  &devName,
    IOMan_DeviceNo   &devNo ) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::GetVolumeNo", IOMan_Trace, 6 );

    tsp00_VFilename     auxName;
    ConstIterator       endIter = End();

    for( ConstIterator iter = Begin(); iter != endIter; ++iter )
    {
        if(( *iter ).GetName( auxName ))
        {
            if( 0 == memcmp( devName, auxName, sizeof( devName )))
            {
                devNo = (*iter).GetLogicalDevNo();
                return SAPDB_TRUE;
            }
        }
    }
    return SAPDB_FALSE;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_DataArea::OpenAll( const tsp00_TaskId taskId )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::OpenAll", IOMan_Trace, 6 );

    IOMan_BlockCount    totalDataPages = 0;
    SAPDB_UInt          usedVolumes    = 0;
    Iterator            prevIter       = End();
    ConstIterator       endIter        = End();

    for( Iterator iter = Begin(); iter != endIter; ++iter )
    {
        IOMan_ReturnCode retCode = ( *iter ).Open( taskId );

        if( IOMan_Okay == retCode )
        {
            if( prevIter == endIter )
            {
                // iter points to the first volume within configuration file
                if( ( *iter ).IsPrevLinkageValid())
                {
                    IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_MISSING_PREDECESSOR,
                                              ( *iter ).GetTypeName(),
                                              SAPDB_ToString( ( *iter ).GetLogicalDevNo(), _T_d ),
                                              SAPDB_ToString( ( *iter ).GetPrevLogicalDevNo(), _T_d ));
                    RTE_Message( errMsg );
                    return SAPDB_FALSE;
                }
            }
            else
            {
                if( !( *prevIter ).IsLinkageValid( taskId, *iter ))
                    return SAPDB_FALSE;

                if( !( *prevIter ).IsDBIdentifierValid( taskId, *iter ))
                    return SAPDB_FALSE;
            }
            prevIter        = iter;
            totalDataPages += ( *iter ).GetUsableSize();
            usedVolumes++;
        }
        else if( IOMan_NotConfigured == retCode )
            continue;
        else
            return SAPDB_FALSE;
    }
    if( ( *prevIter ).IsNextLinkageValid())
    {
        IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_MISSING_SUCCESSOR,
                                  ( *prevIter ).GetTypeName(),
                                  SAPDB_ToString( ( *prevIter ).GetLogicalDevNo(), _T_d ),
                                  SAPDB_ToString( ( *prevIter ).GetNextLogicalDevNo(), _T_d ));
        RTE_Message( errMsg );
        return SAPDB_FALSE;
    }

    SAPDBERR_ASSERT_STATE( m_ConfiguredVolumes == usedVolumes );

    // Set the data area capacity after opening all volumes
    m_TotalDataPages = totalDataPages;
    m_UsedVolumes    = usedVolumes;

    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

IOMan_ReturnCode
IOMan_DataArea::MigrateCapacity(
    const tsp00_TaskId  taskId,
    IOMan_MessageList   &msgList )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::MigrateCapacity", IOMan_Trace, 6 );

    ConstIterator   endIter = End();

    for( Iterator iter = Begin(); iter != endIter; ++iter )
    {
        const IOMan_ReturnCode retCode = ( *iter ).MigrateCapacity( taskId, msgList );

        if(( IOMan_Okay == retCode ) || ( IOMan_NotConfigured == retCode ))
            continue;
        else
            return retCode;
    }
    return IOMan_Okay;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool
IOMan_DataArea::Initialize(
    const tsp00_TaskId    taskId,
    const SAPDB_Int       maxDataVolumes )
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::Initialize", IOMan_Trace, 6 );

    if(! m_Volumes.Resize( maxDataVolumes ))
        return SAPDB_FALSE;

    IOMan_DeviceNo  devNo   = IOMan_DataArea::GetFirstDataVolumeNo();
    Iterator        iter    = Begin();
    ConstIterator   endIter = End();

    while( iter != endIter )
    {
        Data_PageFrame frame = m_PageAllocator.New( taskId );

        if( ! frame.IsAssigned() )
            return SAPDB_FALSE;

        ( *iter ).Initialize( devNo, frame );

        // Determine the number of configured data volumes at this point
        // because same compontents are interested in this information
        // before the corresponding data volumes are started.

        if( ( *iter ).IsConfigured() )
            ++m_ConfiguredVolumes;
        ++iter;
        ++devNo;
    }
    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

IOMan_DeviceNo
IOMan_DataArea::GetConfiguredVolumeNo(SAPDB_Int pos) const
{
    SAPDBTRACE_ROUTINE_DEBUG( "IOMan_DataArea::GetConfiguredVolume", IOMan_Trace, 6 );

    IOMan_DeviceNo  devNo   = IOMan_DataArea::GetFirstDataVolumeNo();
    ConstIterator   iter    = Begin();
    ConstIterator   endIter = End();
    SAPDB_Int counter = 0;

    while( iter != endIter )
    {
        if( ( *iter ).IsConfigured() )
        {
            counter ++;
            if (counter == pos)
            {
                return devNo;
            }
        }
        ++iter;
        ++devNo;
    }
    return IOMan_DeviceNo(); // invalid
}

/*===========================================================================*
 *  PRIVATE METHODS                                                          *
 *===========================================================================*/

SAPDB_Bool
IOMan_DataArea::FormatAll(
    const tsp00_TaskId  taskId,
    IOMan_MessageList   &msgList )
{
    RTEIO_VolumeInfoVector  infoVector( m_Allocator );

    if( ! infoVector.Reserve( m_ConfiguredVolumes )){
        return SAPDB_FALSE;
    }

    ConstIterator   iter;
    ConstIterator   endIter = End();

    for( iter = Begin(); iter != endIter; ++iter )
    {
        if(( *iter ).IsConfigured())
        {
            infoVector.InsertEnd( new (m_Allocator) RTEIO_DataVolumeInfo(
                                      (*iter).GetLogicalDevNo(),
                                      (*iter).GetInternalSize()));
        }
    }

    // format all volumes
    if ( ! RTEIO_FormatVolumeVector( taskId, infoVector, msgList )){
        return SAPDB_FALSE;
    }

    return SAPDB_TRUE;
}

/*---------------------------------------------------------------------------*/

void
IOMan_DataArea::CreateLinkage(
    const tsp00_TaskId  taskId,
    IOMan_DataVolume    &prevVolume,
    IOMan_DataVolume    &volume )
{
    if( prevVolume.SetNextLinkage( taskId, volume ))
        if( volume.SetPrevLinkage( taskId, prevVolume ))
            return;

    IOMan_MessageList errMsg( __CONTEXT__, IOMAN_ERR_CREATE_LINKAGE,
                              volume.GetTypeName(),
                              SAPDB_ToString( prevVolume.GetLogicalDevNo(), _T_d ),
                              SAPDB_ToString( volume.GetLogicalDevNo(), _T_d ));
    RTE_Crash( errMsg );
}

/*---------------------------------------------------------------------------*/

void
IOMan_DataArea::DropLinkage(
    const tsp00_TaskId  taskId,
    IOMan_DataVolume    &volume )
{
    IOMan_DataVolume  *pPrevVolume = NULL;
    IOMan_DataVolume  *pNextVolume = NULL;

    DetermineLinkage( volume, &pPrevVolume, &pNextVolume );

    if( NULL != pPrevVolume )
    {
        if( NULL == pNextVolume )
            pPrevVolume->SetNextLinkageToInvalid( taskId );
        else
            pPrevVolume->SetNextLinkage( taskId, *pNextVolume );
    }
    if( NULL != pNextVolume )
    {
        if( NULL == pPrevVolume )
            pNextVolume->SetPrevLinkageToInvalid( taskId );
        else
            pNextVolume->SetPrevLinkage( taskId, *pPrevVolume );
    }
}

/*---------------------------------------------------------------------------*/

const SAPDB_Byte*
IOMan_DataArea::DetermineLinkage(
    const IOMan_DataVolume    &volume,
    IOMan_DataVolume          **pPrevVolume,
    IOMan_DataVolume          **pNextVolume )
{
    ConstIterator   endIter = End();

    for( Iterator iter = Begin(); iter != endIter; ++iter )
    {
        if(( *iter ).IsOnline())
        {
            if(( *iter ).GetLogicalDevNo() < volume.GetLogicalDevNo())
                *pPrevVolume = iter;
            else
                if(( *iter ).GetLogicalDevNo() > volume.GetLogicalDevNo())
                {
                    *pNextVolume = iter;
                    break;
                }
        }
    }
    if( NULL != *pPrevVolume )
        return( **pPrevVolume ).GetDBIdentifier();
    else if( NULL != *pNextVolume )
        return( **pNextVolume ).GetDBIdentifier();
    else
        return NULL;
}

/*===========================================================================*
*  END OF CODE                                                              *
*===========================================================================*/
