/*!
  @file    Msg_OutputFile.cpp
  @author  JoergM
  @ingroup Common
  @brief   Message output file class
  
  All message output classes are derived from class Msg_IOutput.

  This class is implementing the file output channel.
  
  The output file channel handles three files in two versions.

  The three files are
    the 'starting phase messages' extension .start
    the 'running phase messages' extension .run
    the 'termination phase messages' extension .stop

  The two versions are destinct by an additional extension .old

  The common used file name for kernel messages is 'knldiag'

  The following files were found in the rundirectory before

  'knldiag' and if ever started before 'knldiag.old'

  The following file are found now in the rundirectory

  'knldiag.start', 'knldiag.run', 'knldiag.stop'

  and if ever started before 'knldiag.start.old',  'knldiag.run.old', 'knldiag.stop.old'

  To convert from 'knldiag.start + knldiag.run + knldiag.stop' into knldiag use the tool

  'TransformIntoOldKnldiag DatabaseName'. This tool will convert the new <xml> alike output into
  a chronological correct knldiag in 'table form'.

  -------------------------------------------------------------------------

  \if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
  \endif
 */

#include "Messages/Msg_OutputFile.hpp"
#include "Messages/Msg_List.hpp"
#include "Messages/Msg_FileProcessor.hpp"

#include "SAPDBCommon/SAPDB_sprintf.h"
#include "RunTime/System/RTESys_IO.h"

#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"

#include "SAPDBCommon/SAPDB_Names.h"

#define FILE_EXTENSION_SIZE         4096
// Runtime internal used phase changing routines



SAPDB_Bool Msg_OutputFile::ExtendFile(SAPDB_UInt8 size,Msg_List &errList)
{
    tsp00_VfReturn_Param returnStatus;
    RTE_FileOffset newPosition;

    SAPDB_Char *clear = (SAPDB_Char *)alloca((size_t)FILE_EXTENSION_SIZE);
    if( NULL == clear )
    {
        return false;
    }
    memset(clear,' ',(size_t)FILE_EXTENSION_SIZE);

    RTESys_IOSeek ( m_FileHandle,
                    0,
                    RTESys_IOSeekEnd,
                    newPosition,
                    returnStatus );
    if( vf_notok == returnStatus )
    {
        errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"seek failed",SAPDB_ToString(RTESys_GetLastError()));
        return false;
    }
//  append the file with a block of spaces of the desired size
    RTE_FileOffset writtenSizeInBytes=0;
    RTE_FileOffset writtenSizeInBytesThisTime;
    for(SAPDB_UInt8 i=0;i<size/FILE_EXTENSION_SIZE;i++)
    {
        RTESys_IOWrite ( m_FileHandle,
                        clear,
                        FILE_EXTENSION_SIZE,
                        writtenSizeInBytesThisTime,
                        returnStatus );
        if( vf_notok == returnStatus )
        {
            errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"write failed",SAPDB_ToString(FILE_EXTENSION_SIZE),SAPDB_ToString((SAPDB_UInt8)writtenSizeInBytes),SAPDB_ToString(RTESys_GetLastError()));
            break;
        }
        else
        {
            writtenSizeInBytes += writtenSizeInBytesThisTime;
        }
    }
//  now reposition to the previous file position
    RTESys_IOSeek ( m_FileHandle,
                    m_FilePosition,
                    RTESys_IOSeekSet,
                    newPosition,
                    returnStatus );
    if( vf_notok == returnStatus )
    {
        errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"seek failed",SAPDB_ToString(RTESys_GetLastError()));
    }
    m_currentFileSize += writtenSizeInBytes;
    return ( size == writtenSizeInBytes );
}

/*!
    @brief spinlock protected open
    @param errList [out] filled in case of problem
    @return true if open succeded, false if still closed (errList filled)
    */
SAPDB_Bool Msg_OutputFile::TryOpen(SAPDB_Char *fileName,Msg_List &errList)
{
    tsp00_VfReturn_Param returnStatus;
    SAPDB_Bool fileDidExist = true;
    RTESys_IOOpen ( m_FileHandle,
                    fileName,
                    RTESys_IOWriteOnly/*Exclusive*/,
                    false/*DoCreate*/,
                    DEFAULT_SECURITY_RTESYS_IO,
                    returnStatus );
    if(vf_ok != returnStatus)
    {
        RTESys_IOOpen ( m_FileHandle,
                        fileName,
                        RTESys_IOWriteOnly/*Exclusive*/,
                        true/*DoCreate*/,
                        DEFAULT_SECURITY_RTESYS_IO,
                        returnStatus );
        fileDidExist = false;
    }
    if(vf_ok == returnStatus)
    {
        m_currentFileSize = 0;
        m_FilePosition = 0;
        // pre-fill the file when it is
        //  1.) written cyclically (== has a fixed size)
        //  AND
        //  2.) re-created on application start (== existing output is not continued)
        //  AND
        //  3.) it is the "body" file (== we are between phases starting and running, current phase is still "starting")
        if(0 != m_FixedSizeInBytes && (ContinueExistingOutput != m_HandleExistingOutput || !fileDidExist) && (m_CurrentPhase == Starting))
        {
            if(!ExtendFile(m_FixedSizeInBytes,errList))
            {
                errList = errList + Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"file initialization failed");
            }
        }
        else
        {
            if(ContinueExistingOutput == m_HandleExistingOutput)
            {
                RTE_FileOffset endOfFilePosition;
                RTESys_IOSeek ( m_FileHandle,
                                0,
                                RTESys_IOSeekEnd,
                                endOfFilePosition,
                                returnStatus );
                if( vf_notok == returnStatus )
                {
                    errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"seek failed",SAPDB_ToString(RTESys_GetLastError()));
                    return false;
                }
                else
                {
                    m_currentFileSize = endOfFilePosition;
                }
                // TODO: m_currentFileSize = RTESys_IOGetFileSize()
            }
        }
        return true;
    }
    else
    {
        errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"file open failed",SAPDB_ToString(RTESys_GetLastError()));
        return false;
    }
}
    
/*!
    @brief spinlock protected close
    @param errList [out] filled in case of problem
    @return true if close succeded, false if still open (errList filled)
    */
SAPDB_Bool Msg_OutputFile::TryClose(Msg_List &errList)
{
    if( 0 == m_FileHandle )
    {
        return true;
    }
    else
    {
        tsp00_VfReturn_Param returnStatus;
        RTESys_IOClose ( m_FileHandle,
                        returnStatus );
        
        if(vf_ok == returnStatus)
        {
            return true;
        }
        else
        {
            errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"closing failed",SAPDB_ToString(RTESys_GetLastError()));
            return false;
        }
    }
}
    
/*!
    @brief spinlock protected output
    @brief the output routine for informational messages

    @param outputMsgList [in] the message to be written
    @param errList [out] filled if error occured

    @return the sequence number of the output message, or 0 if output failed
    */
#define GUESSED_MESSAGE_LENGTH 256

// practice what you preach: supply RTETask_ITask object reference to find out taskId and uktId cheaply?
SAPDB_Bool Msg_OutputFile::TryOutput(Msg_List /*const*/ &outputMsgList, Msg_List &errList,SAPDB_Bool emergency)
{
    SAPDB_Bool bufferOnStack = true;
    // TODO: lock it?
    SAPDB_Char buffer1[GUESSED_MESSAGE_LENGTH];
    for(int i=0;i<GUESSED_MESSAGE_LENGTH;i++)
        buffer1[i]=' ';

    SAPDB_Char *pbuffer;
    SAPDB_UInt4 neededSize;
    RTE_FileOffset writtenSizeInBytes;    
    tsp00_VfReturn_Param returnStatus;

    // generate a xml line from outputMsgList
    outputMsgList.XML_Output(  GUESSED_MESSAGE_LENGTH,
                                buffer1,
                                neededSize,
                                errList);

    // if the first buffer was not big enough, allocate a new one with the right size 
    if(neededSize > GUESSED_MESSAGE_LENGTH)
    {
        pbuffer = (SAPDB_Char *)RTEMem_RteAllocator::Instance().Allocate(neededSize);
        // could not allocate so much memory.... must use the truncated one on the stack
        if( NULL == pbuffer )
        {
            errList = SAPDBErr_MessageList(RTE_CONTEXT,RTEERR_GENERAL,"could not allocate enough memory for message. message is truncated");
            pbuffer = buffer1;
            neededSize = strlen(buffer1);   // XML_Output() generated string ends with zero byte...
        }
        else
        {
            bufferOnStack = false;
            outputMsgList.XML_Output(  neededSize,
                                        pbuffer,
                                        neededSize,
                                        errList);
        }
    }
    else
    {
        pbuffer = buffer1;
    }
    if( neededSize > 0 )
    {
        pbuffer[neededSize-1] = '\n';
    }
// if cyclic, check if there is enough space. If not, fill the file with spaces (?), and seek to start of file
    if((HasFixedSize()/*GetOutputStyle()==Msg_IOutput::Triple*/) && Running == m_CurrentPhase)
    {
        if(m_FilePosition + neededSize > m_FixedSizeInBytes)
        {
            SAPDB_UInt8 leftover = m_FixedSizeInBytes - m_FilePosition;
            SAPDB_Char *clear = (SAPDB_Char *)alloca((size_t)leftover);
            memset(clear,' ',(size_t)leftover);
            RTESys_IOWrite ( m_FileHandle,
                            clear,
                            leftover,
                            writtenSizeInBytes,
                            returnStatus );
            if(vf_ok != returnStatus || leftover != writtenSizeInBytes)
            {
                errList = Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"clear until end of file failed",SAPDB_ToString(RTESys_GetLastError()));
            }
            RTE_FileOffset newPosition;
            RTESys_IOSeek ( m_FileHandle,
                            0,
                            RTESys_IOSeekSet,
                            newPosition,
                            returnStatus );
            if(vf_ok != returnStatus || 0 != newPosition)
            {
                errList = errList + Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"seek to start of file failed",SAPDB_ToString(RTESys_GetLastError()));
            }
            else
            {
                m_FilePosition = 0;
            }
        }
    }
    else    // in the other cases (no cyclic writes!), allocate a security buffer at the end of the file
            // (to be able to write a "file system full" message)
    {
        if(!emergency)
        {
            if(m_currentFileSize - m_FilePosition <= m_sizeofEmergencyMessage)
            {
                if(!ExtendFile(m_sizeofEmergencyMessage,errList))
                {
                    errList = errList + Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"file extension failed");
                    Msg_List dummyList;
                    TryOutput(errList,dummyList,true);   // write the "emergency" message
                }
            }
        }
    }
// Now write the message to the file
    RTESys_IOWrite ( m_FileHandle,
                     pbuffer,
                     neededSize,
                     writtenSizeInBytes,
                     returnStatus );
    if(!bufferOnStack)
    {
        RTEMem_RteAllocator::Instance().Deallocate(pbuffer);
    }
    m_FilePosition += writtenSizeInBytes;
    m_SequenceNumber++;
    if(vf_ok != returnStatus || neededSize != writtenSizeInBytes)
    {
        errList = errList + Msg_List(RTE_CONTEXT,RTEERR_GENERAL,"write failed",SAPDB_ToString(neededSize),SAPDB_ToString((SAPDB_UInt8)writtenSizeInBytes),SAPDB_ToString(RTESys_GetLastError()));
        return false;
    }
    else
    {
        return true;
    }
}

/*!
    @brief preparing for starting phase
    @param errList [out] filled in case of problems
    @return true if preparation succeeded, false if failed
    */
SAPDB_Bool Msg_OutputFile::PrepareStartingPhase(Msg_List &errList)
{
    RTE_Path fileName;
    // build the file name: base file name + ".start" for cyclically written files (fixedSize != 0), 
    // simply the base file name for not cyclically written files (fixedSize == 0)
    if(HasFixedSize()/*GetOutputStyle()==Msg_IOutput::Triple*/ && ContinueExistingOutput != m_HandleExistingOutput)
    {
        SAPDB_sprintf(fileName,sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_STARTING_PHASE_SUFFIX)+1,"%s%s",m_Filename,SAPDB_OUTPUTFILE_STARTING_PHASE_SUFFIX);
    }
    else
    {
        SAPDB_sprintf(fileName,sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_COMPLETE_SUFFIX)+1,"%s%s",m_Filename,SAPDB_OUTPUTFILE_COMPLETE_SUFFIX);
    }
    // treatment of existing files: if not to be continued, delete them, making a backup copy if desired
    if((RenameExistingOutput == m_HandleExistingOutput)||(RemoveExistingOutput == m_HandleExistingOutput))
    {
        tsp00_VfReturn_Param returnStatus;
        RTE_Path *temporaryOldFileNames;
        RTE_Path *temporaryNewFileNames;
        SAPDB_Int4 numberOfFiles;
        SAPDB_Int4 counter;
        if(HasFixedSize()/*GetOutputStyle()==Msg_IOutput::Triple*/ && ContinueExistingOutput != m_HandleExistingOutput)
        {
            numberOfFiles = 3;
            temporaryOldFileNames = (RTE_Path *)alloca(sizeof(RTE_Path) * numberOfFiles);
            SAPDB_sprintf(temporaryOldFileNames[0],sizeof(RTE_Path),"%s"SAPDB_OUTPUTFILE_STARTING_PHASE_SUFFIX,m_Filename);
            SAPDB_sprintf(temporaryOldFileNames[1],sizeof(RTE_Path),"%s"SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX,m_Filename);
            SAPDB_sprintf(temporaryOldFileNames[2],sizeof(RTE_Path),"%s"SAPDB_OUTPUTFILE_TERMINATING_PHASE_SUFFIX,m_Filename);

            temporaryNewFileNames = (RTE_Path *)alloca(sizeof(RTE_Path) * numberOfFiles);
            SAPDB_sprintf(temporaryNewFileNames[0],sizeof(RTE_Path),"%s%s"SAPDB_OUTPUTFILE_STARTING_PHASE_SUFFIX,m_Filename,SAPDB_OUTPUTFILE_OLD_FILE_SUFFIX);
            SAPDB_sprintf(temporaryNewFileNames[1],sizeof(RTE_Path),"%s%s"SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX,m_Filename,SAPDB_OUTPUTFILE_OLD_FILE_SUFFIX);
            SAPDB_sprintf(temporaryNewFileNames[2],sizeof(RTE_Path),"%s%s"SAPDB_OUTPUTFILE_TERMINATING_PHASE_SUFFIX,m_Filename,SAPDB_OUTPUTFILE_OLD_FILE_SUFFIX);
        }
        else
        {
            numberOfFiles = 1;
            temporaryOldFileNames = (RTE_Path *)alloca(sizeof(RTE_Path));
            SAPDB_sprintf(temporaryOldFileNames[0],sizeof(RTE_Path),"%s",m_Filename);

            temporaryNewFileNames = (RTE_Path *)alloca(sizeof(RTE_Path));
            SAPDB_sprintf(temporaryNewFileNames[0],sizeof(RTE_Path),"%s"SAPDB_OUTPUTFILE_OLD_FILE_SUFFIX,m_Filename);
        }
        for(counter=0;counter<numberOfFiles;counter++)
        {
            if(RenameExistingOutput == m_HandleExistingOutput)
            {
                RTESys_IORename(temporaryOldFileNames[counter],
                        temporaryNewFileNames[counter],
                        returnStatus);
            }
            RTESys_IOUnlink ( temporaryOldFileNames[counter],
                            returnStatus );
            if(vf_ok != returnStatus)
            {
                // TODO!!
            }

        }
    }

    if(/*GetOutputStyle()!=Msg_IOutput::Triple && */ContinueExistingOutput == m_HandleExistingOutput)
    {
        // seek to the write position if existing files are to be continued
        Msg_FileProcessor fileProcessor(m_Filename,Msg_IOutput::Single);
        RTE_FileOffset writePosition;
        SAPDB_UInt8 biggestSequenceNumber;
//DebugBreak();
        fileProcessor.SearchWritePosition(writePosition,biggestSequenceNumber,errList);

        m_SequenceNumber = biggestSequenceNumber+1;
        if(!TryOpen(fileName,errList))
            return false;
        RTE_FileOffset newPosition;
        tsp00_VfReturn_Param returnStatus;
        RTESys_IOSeek ( m_FileHandle,
                        writePosition,
                        RTESys_IOSeekSet,
                        newPosition,
                        returnStatus );
        m_FilePosition = newPosition;
    }
    else
    {
        if(!TryOpen(fileName,errList))
            return false;
    }


    m_CurrentPhase = Starting;
    return true;
}

/*!
    @brief preparing for running phase
    @param errList [out] filled in case of problems
    @return true if preparation succeeded, false if failed
    */
SAPDB_Bool Msg_OutputFile::PrepareRunningPhase(Msg_List &errList)
{
    if(HasFixedSize()/*GetOutputStyle()==Msg_IOutput::Triple*/)
    {
        if(!TryClose(errList))
            return false;
        SAPDB_Char *fileName;
        
        if(ContinueExistingOutput == m_HandleExistingOutput)
        {
            fileName = (SAPDB_Char *)alloca(sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_COMPLETE_SUFFIX)+1);
            SAPDB_sprintf(fileName,sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_COMPLETE_SUFFIX)+1,"%s%s",m_Filename,SAPDB_OUTPUTFILE_COMPLETE_SUFFIX);
            // seek to the write position if existing files are to be continued
            Msg_FileProcessor fileProcessor(m_Filename,Msg_IOutput::Single);
            RTE_FileOffset writePosition;
            SAPDB_UInt8 biggestSequenceNumber;
    //DebugBreak();
            fileProcessor.SearchWritePosition(writePosition,biggestSequenceNumber,errList);

            m_SequenceNumber = biggestSequenceNumber+1;
            if(!TryOpen(fileName,errList))
                return false;
            RTE_FileOffset newPosition;
            tsp00_VfReturn_Param returnStatus;
            RTESys_IOSeek ( m_FileHandle,
                            writePosition,
                            RTESys_IOSeekSet,
                            newPosition,
                            returnStatus );
            RTE_FileOffset writtenSizeInBytes;
            RTESys_IOWrite ( m_FileHandle,
                             "\n",
                             1,
                             writtenSizeInBytes,
                             returnStatus );
            m_FilePosition = newPosition + 1;
        }
        else
        {
            fileName = (SAPDB_Char *)alloca(sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX)+1);
            SAPDB_sprintf(fileName,sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX)+1,"%s%s",m_Filename,SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX);

            if(!TryOpen(fileName,errList))
                return false;
        }
    }
    m_CurrentPhase = Running;
    return true;
}

/*!
    @brief preparing for terminating phase
    @param errList [out] filled in case of problems
    @return true if preparation succeeded, false if failed
    */
SAPDB_Bool Msg_OutputFile::PrepareTerminatingPhase(Msg_List &errList)
{
    if ( m_CurrentPhase != Stopping )
    {
        if(HasFixedSize()/*GetOutputStyle()==Msg_IOutput::Triple*/)
        {
            if(!TryClose(errList))
                return false;
            SAPDB_Char *fileName = (SAPDB_Char *)alloca(sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_TERMINATING_PHASE_SUFFIX)+1);
            SAPDB_sprintf(fileName,sizeof(m_Filename)+sizeof(SAPDB_OUTPUTFILE_RUNNING_PHASE_SUFFIX)+1,"%s%s",m_Filename,SAPDB_OUTPUTFILE_TERMINATING_PHASE_SUFFIX);
            if(!TryOpen(fileName,errList))
                return false;
        }
    }
    return true;
}
