/*!
  @file    Msg_OutputKnlDiag.cpp
  @author  Robin
  @ingroup Common
  @brief   KnlDiag Message output class
  
  implementation of diagnostic output used in the database kernel


  -------------------------------------------------------------------------

  \if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
  \endif
 */

#include "Messages/Msg_OutputFile.hpp"
#include "heo07.h"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"
#include "KernelCommon/ParameterNames/KernelParam_RunTime.h"
#include "geo007_2.h"
#include "RunTime/System/RTESys_Time.h"
#include "SAPDBCommon/SAPDB_sprintf.h"
#include "hsp100.h"
#include "RunTime/RTE_Speed.h"
#include "geo007_1.h"

#define MAX_MSG_LINE_LEN    512

class BooleanTLS
{
public:
    BooleanTLS ()
    {
        teo07_ThreadErr terr;
        tsp00_ErrTextc ttext;
        sqlcreatetls(&m_Key, ttext, &terr);
        sqlsettls(m_Key,0,ttext, &terr);
    }

    SAPDB_Bool Set(SAPDB_Bool value)
    {
        teo07_ThreadErr terr;
        tsp00_ErrTextc ttext;
        sqlsettls(m_Key,(void *)value,ttext, &terr);
        return ( THR_OK_EO07 == terr );
    }

    SAPDB_Bool Get(void)
    {
        return 0 != sqlgettls(m_Key);
    }
private:
    tsp00_Int4 m_Key;
};


class OutputFileKnlDiagErr : public Msg_OutputFile
{
public:
    OutputFileKnlDiagErr ()
     : Msg_OutputFile   ( "knldiag.err"/*MSG_OUTPUTFILE_COMPLETE_SUFFIX*/   /*SAPDB_Char const      *filename*/
                        , (SAPDB_UTF8 *)"knldiag_err_lock"              /*SAPDB_UTF8 const      *lockname*/
                        , ContinueExistingOutput                        /*HandleExistingOutput const handleExistingOutput*/
                        , 0                                             /*SAPDB_UInt8 const      fixedSizeInBytes*/
                        )
    {

    }

    virtual SAPDB_Bool Output(Msg_List  &outputMsgList, Msg_List &errList )
    {
        if(m_InUse.Get())
        {
            // TODO: pass to emergency output
            return false;
        }
        m_InUse.Set(true);
        SAPDB_Bool retval = Msg_OutputFile::Output(outputMsgList, errList );
        m_InUse.Set(false);
        return retval;
    }

    virtual SAPDB_Bool PrepareStartingPhase(Msg_List &errList)
    {
        if(!Msg_OutputFile::PrepareStartingPhase(errList))
        {
            return false;
        }

        SAPDB_Char szOpenLine[MAX_MSG_LINE_LEN + 1];
        tsp00_Versionc   currentVersion;
        SAPDB_Char	speed[RTE_MAXSPEEDLEN + 1];	
        RTESys_Timestamp timeBuffer;
        RTESys_Timestamp timeBufferGMT;
        RTESys_FillProtocolTimeStamp(timeBuffer);
        RTESys_FillProtocolTimeStamp(timeBufferGMT);
        memset (speed, 0, RTE_MAXSPEEDLEN + 1);
		strcpy (speed, RTE_GetSpeed ());
        if (!strcmp (speed, RTE_FAST) )
        {
            strcpy (speed, " ");
        }
        else
        {
            strcat (speed, " ");
        }

		sp100_GetVersionString( speed, 
	    						s100buildnumber,
		    					currentVersion);
		SAPDB_sprintf( szOpenLine,MAX_MSG_LINE_LEN + 1,
                  ERR_FILE_STARTING_TEXT,
                  &timeBufferGMT[0], &currentVersion[0] );
        Msg_List theList = Msg_List( Msg_List::Info, 27, MSG_MSGOUTPUT_COMPONENT, __FILE__, __LINE__, szOpenLine);
        return Output( theList, errList );
    }
private:
    BooleanTLS m_InUse;
};

class OutputFileKnlDiag : public Msg_OutputFile
{
public:
    OutputFileKnlDiag (  const SAPDB_UInt8 fixedSizeInBytes, Msg_IOutput *pErrorArchive )
        : Msg_OutputFile    ( "knldiag"                     /*SAPDB_Char const      *filename*/
                            , (SAPDB_UTF8 *)"knldiag_lock"  /*SAPDB_UTF8 const      *lockname*/
                            , RenameExistingOutput          /*HandleExistingOutput const handleExistingOutput*/
                            , fixedSizeInBytes              /*SAPDB_UInt8 const      fixedSizeInBytes*/
                            , pErrorArchive
                            )
    {
    }

    virtual SAPDB_Bool Output(Msg_List  &outputMsgList, Msg_List &errList, PassToErrorArchive toArchive )
    {
        if(m_InUse.Get())
        {
            // TODO: pass to emergency output
            return false;
        }
        m_InUse.Set(true);
        outputMsgList.SetOutputIdentification( 0        /*m_SequenceNumber - will be really set in the Output() metmod*/
                                      ,sqlgetthreadid() /* SAPDB_UInt4 const outputThreadId=0*/
                                                        /*, SAPDB_UInt4 const outputTaskId=0*/
                                                        /*, SAPDB_UInt4 const outputSchedulerId=0 */);
        SAPDB_Bool retval = Msg_OutputFile::Output(outputMsgList, errList, toArchive );
        m_InUse.Set(false);
        return retval;
    }

    virtual SAPDB_Bool BeginOfTerminationPhase(Msg_List &errList)
    {
        if(m_InUse.Get())
        {
            m_MessageLock.Unlock();
        }
        {
            RTESync_LockedScope lock(m_MessageLock);
            return Msg_OutputFile::BeginOfTerminationPhase(errList);
        }
    }

private:
    BooleanTLS m_InUse;
};

static OutputFileKnlDiag *pInstance=NULL;
/*!
    @brief return reference to default diagnostic output channel

    This reference can be used to 
*/

Msg_IOutput &Msg_IOutput::GetDiagnosticOutput()
{
    if(!pInstance)
    {
        // first create the error archive (aka knldiag.err): existing file is continued, file size is not limited (no cyclic write)
        SAPDBMEM_STATIC_RAW_ALLOCATE(KnlDiagErr_InstanceData, sizeof(OutputFileKnlDiagErr));
        Msg_IOutput *pErrorArchive = NULL;
        pErrorArchive = new(KnlDiagErr_InstanceData) OutputFileKnlDiagErr ();
        // now create the diagnostic file itself (aka knldiag): existing files are renamed, file size is limited (cyclic write)
        SAPDBMEM_STATIC_RAW_ALLOCATE(KnlDiag_InstanceData, sizeof(OutputFileKnlDiag)-4);
        Msg_List errList;
        RTEConf_Parameter::Integer sizeFromParams;
        if(!(RTEConf_ParameterAccess::Instance())->GetInteger(KERNELPARAM_KERNELDIAGSIZE,sizeFromParams,errList))
            sizeFromParams = 16;
        pInstance = new(KnlDiag_InstanceData) OutputFileKnlDiag   ( sizeFromParams*1024,pErrorArchive );
    }
    return *pInstance;
}


/*!
    @brief return reference to emergency diagnostic output channel

    This reference can be used in case the normal diagnostic messages output call returns an error message.
    */
Msg_IOutput &Msg_IOutput::GetEmergencyOutput()
{
    return *pInstance;// TODO: replace by something sensible
}

extern "C" void sql60_strt_msg_wrt_cycle_old ( void );


extern "C" void Msg_OutputToKnlDiag(SAPDB_Long messageID,SAPDB_ULong eventType,const SAPDB_Char *pMessage)
{
    Msg_IOutput &diagnosticMessages = Msg_IOutput::GetDiagnosticOutput();
    Msg_List errList;
    Msg_List::MessageType messageType;
    switch(eventType)
    {
    case IERR_TYPE:
    case ERR_TYPE:
        messageType = Msg_List::Error;
        break;
    case WRN_TYPE:
        messageType = Msg_List::Warning;
        break;
    case INFO_TYPE:
        messageType = Msg_List::Info;
        break;
    }
#ifndef NO_OLD_DIAG_FILES
    Msg_List theList = Msg_List( messageType, messageID, MSG_MSGOUTPUT_COMPONENT, __FILE__, __LINE__, pMessage);
#else
    Msg_List theList = Msg_List( messageType, messageID, "RTEKERN", __FILE__, __LINE__, pMessage);
#endif
    diagnosticMessages.Output( theList, errList );
}

extern "C" void Msg_StartupPhaseCompleted(void)
{
    Msg_IOutput &diagnosticMessages = Msg_IOutput::GetDiagnosticOutput();
    Msg_List errList;
    diagnosticMessages.StartupPhaseCompleted(errList);
}


extern "C" void Msg_BeginOfTerminationPhase(void)
{
    Msg_List tmpMsgList;
    if ( !Msg_IOutput::GetDiagnosticOutput().BeginOfTerminationPhase(tmpMsgList) )
    {
        Msg_List dummyTmpMsgList;
//        Msg_IOutput::GetEmergencyOutput().Output(tmpMsgList, dummyTmpMsgList);
    }
}

extern "C" void Msg_OutputToKnlDiagVArgs(SAPDB_Long messageID,SAPDB_ULong eventType,const SAPDB_Char *pFormatString,va_list args)
{
    SAPDB_Char message[256];
#ifdef WIN32
    _vsnprintf      ( message, sizeof(message) - 1, pFormatString, args );
#else
    SAPDB_sprintf   ( message, sizeof(message), pFormatString, args ) ;
#endif
    Msg_OutputToKnlDiag(messageID,eventType,&message[0]);
}
