/*!
  @file    Msg_ProtConv.cpp
  @author  Robin
  @ingroup Messages
  @brief   Message file processing class
  
  processing of message list files (pseudo-xml to well-formed xml)

  -------------------------------------------------------------------------

  \if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
  \endif
 */

#include "Messages/Msg_FileProcessor.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessSimple.h"
#include "heo01.h"
#include "SAPDBCommon/SAPDB_StandardArgumentParser.hpp"
#include "Messages/Msg_ProtFileInfo.hpp"
#include "KernelCommon/ParameterNames/KernelParam_RunTime.h"

#include <stdio.h>


/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

class  ProtConvArgParser : public SAPDB_StandardArgumentParser
{
public:
    ProtConvArgParser(  const SAPDB_Int4            argc,
                        const RTE_ArgChar* const*   argv );

    enum { optID_Help, optID_DBName, optID_SortMode, optID_OutputFormat, optID_OutputFile, optID_FileKey, optID_Path };

    RTE_DBName &                GetDBName()             { return m_DBName; }

    Msg_FileProcessor::SortMode GetSortMode()           { return m_SortMode; }

    Msg_FileProcessor::Format   GetFormat()             { return m_Format; }

    RTE_Path &                  GetDestinationFileName(){ return m_destinationFileName; }

    RTE_Path &                  GetPath()               { return m_Path; }

    SAPDB_Bool                  IncludeClassicHeader()  { return m_ClassicHeader;}

    Msg_ProtFileInfo           *GetFileTag()            { return &m_fileTag;}

    Msg_ProtFileInfo           &GetFileInfo()           { return m_fileTag;}
protected:
    virtual SAPDB_Bool HandleNextOption  ( const SAPDB_Int4       optionID,
                                           const RTE_ArgChar*     optionArgument = 0 );
private:
    RTE_DBName                  m_DBName;
    Msg_FileProcessor::SortMode m_SortMode;
    Msg_FileProcessor::Format   m_Format;
    RTE_Path                    m_destinationFileName;
    RTE_Path                    m_Path;
    SAPDB_Bool                  m_ClassicHeader;
    Msg_ProtFileInfo            m_fileTag;
};

/*---------------------------------------------------------------------------*/

static ProtConvArgParser::OptionDesc optionDesc[] = 
       {{ ProtConvArgParser::optID_Help,            RTE_ArgChar('h'),      (RTE_ArgChar*)("help"),           OptionArgumentIsOptional | OptionIsOptional,  (RTE_ArgChar*)("[<Option> | LONG]"),   (RTE_ArgChar*)("help") }
       ,{ ProtConvArgParser::optID_Help,            RTE_ArgChar('?'),      (RTE_ArgChar*)("help"),           OptionArgumentIsOptional | OptionIsOptional,  (RTE_ArgChar*)("[<Option> | LONG]"),   (RTE_ArgChar*)("help") }
       ,{ ProtConvArgParser::optID_DBName,          RTE_ArgChar('d'),      (RTE_ArgChar*)("DBName"),         OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<DBName>"),            (RTE_ArgChar*)("name of the database") }
       ,{ ProtConvArgParser::optID_SortMode,        RTE_ArgChar('s'),      (RTE_ArgChar*)("SortMode"),       OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<SortMode>"),          (RTE_ArgChar*)("sort mode (number|time(default))") }
       ,{ ProtConvArgParser::optID_OutputFile,      RTE_ArgChar('o'),      (RTE_ArgChar*)("OutputFile"),     OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<OutputFile>"),        (RTE_ArgChar*)("output file") }
       ,{ ProtConvArgParser::optID_OutputFormat,    RTE_ArgChar('f'),      (RTE_ArgChar*)("OutputFormat"),   OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<OutputFormat>"),      (RTE_ArgChar*)("output format (short(default)|long|classic)") }
       ,{ ProtConvArgParser::optID_FileKey,         RTE_ArgChar('t'),      (RTE_ArgChar*)("FileKey"),        OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<FileKey>"),           (RTE_ArgChar*)("file key") }
       ,{ ProtConvArgParser::optID_Path,            RTE_ArgChar('p'),      (RTE_ArgChar*)("Path"),           OptionHasAnArgument | OptionIsOptional,       (RTE_ArgChar*)("<InputFilePath>"),     (RTE_ArgChar*)("path to read input files from") }
       };

/*---------------------------------------------------------------------------*/

ProtConvArgParser::ProtConvArgParser (  const SAPDB_Int4            argc,
                                        const RTE_ArgChar* const*   argv )
    :SAPDB_StandardArgumentParser ( argc, argv, ProtConvArgParser::optID_Help, optionDesc, NUM_OF_OPTIONS_DESC(optionDesc) )
    ,m_Format(Msg_FileProcessor::WithoutLongText)
    ,m_SortMode(Msg_FileProcessor::MessagesByTimeStamp)
    ,m_ClassicHeader(false)
{
    m_DBName[0] = '\0';
    m_destinationFileName[0] = '\0';
    m_Path[0]   = '\0';
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool ProtConvArgParser::HandleNextOption (    const SAPDB_Int4       optionID,
                                                    const RTE_ArgChar*     optionArgument )
{
    SAPDB_Bool   ok = true;
    Msg_List errList;

    switch (optionID)
    {
        case optID_DBName:
            strncpy ( m_DBName, optionArgument, sizeof(m_DBName) - 1 );
            m_DBName[sizeof(m_DBName) - 1] = '\0';
            break;

        case optID_SortMode:
            if( 0 == strcmp(optionArgument,"time"))
            {
                m_SortMode = Msg_FileProcessor::MessagesByTimeStamp;
            }
            else
            {
                m_SortMode = Msg_FileProcessor::MessageListsByNo;
            }
            break;

        case optID_OutputFormat:
            if( 0 == strcmp(optionArgument,"long"))
            {
                m_Format = Msg_FileProcessor::WithLongText;
                m_ClassicHeader = false;
            }
            else if( 0 == strcmp(optionArgument,"classic"))
            {
                m_Format = Msg_FileProcessor::WithoutLongText;
                m_ClassicHeader = true;
            }
            else
            {
                m_Format = Msg_FileProcessor::WithoutLongText;
                m_ClassicHeader = false;
            }
            break;

        case optID_OutputFile:
            SAPDB_strncpy_zero_terminated(m_destinationFileName,optionArgument,sizeof(RTE_Path));
            break;

        case optID_FileKey:
            m_fileTag = Msg_ProtFileInfo::ByTagName((SAPDB_Char *)optionArgument);
            break;

        case optID_Path:
            SAPDB_strncpy_zero_terminated(m_Path,optionArgument,sizeof(m_Path));
            break;

        default:
            printf ( ("Unknown option identfier '%d'\n"), optionID  );
            ok = false;
            break;
    }
    return ok;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool BuildFilePath(RTE_DBName &DBName,const SAPDB_Char *fileName,RTE_Path filePath)
{
const SAPDB_Char *pDBName = (const SAPDB_Char *)&DBName;
    tsp00_ErrText   errText;
    RTEConf_ParamaterAccessGetParam  ( *((tsp00_DbNamec *)pDBName)      /*const tsp00_DbNamec &DBNameC */,
                                       *((tsp00_C256c *)KERNELPARAM_RUNDIRECTORY) /*const tsp00_C256c   &XpKeyC*/,
                                       *((tsp00_C256c *)filePath)       /*tsp00_C256c   &XpValueC*/,
                                       errText                          /*tsp00_ErrText &ErrText */);
    SAPDB_Char delimiter[2]={PATH_DELIMITER_EO01,0};
    if(strlen(filePath)+strlen(delimiter)+strlen(fileName) > sizeof(RTE_Path))
    {
        return false;
    }
    else
    {
        strcat(filePath,delimiter);
        strcat(filePath,fileName);
        return true;
    }
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool BuildFilePath(const RTE_Path &path,const SAPDB_Char *fileName,RTE_Path filePath)
{
    SAPDB_strncpy(filePath,path,sizeof(RTE_Path));
    if(filePath[strlen(filePath)-1] != PATH_DELIMITER_EO01)
    {
        SAPDB_Char delimiter[2]={PATH_DELIMITER_EO01,0};
        if(strlen(filePath)+strlen(delimiter) > sizeof(RTE_Path))
        {
            return false;
        }
        strcat(filePath,delimiter);
    }
    if(strlen(filePath)+strlen(fileName) > sizeof(RTE_Path))
    {
        return false;
    }
    else
    {
        strcat(filePath,fileName);
        return true;
    }
}

/*---------------------------------------------------------------------------*/

int main(int argc,char **argv)
{
    ProtConvArgParser ArgParser( argc,argv );

    if(ArgParser.StartParsing())
    {
        RTE_Path filePath;
        RTE_Filename fileName = {0};
        Msg_ProtFileInfo &fileInfo = ArgParser.GetFileInfo();

        SAPDB_Bool filePathBuilt = false; 

        fileInfo.GetCompleteFileName(fileName,/**(RTE_DBName *)*/ArgParser.GetDBName());
        if( 0 == fileName[0])
        {
            SAPDB_strncpy_zero_terminated(fileName , "knldiag",sizeof(fileName));
        }

        if(ArgParser.GetDBName()[0])
        {
            filePathBuilt = BuildFilePath(ArgParser.GetDBName(),fileName,filePath);
        }

        if(ArgParser.GetPath()[0])
        {
            filePathBuilt = BuildFilePath(ArgParser.GetPath(),fileName,filePath);
        }

        if(filePathBuilt)
        {
            Msg_FileProcessor fileProcessor(filePath,fileInfo.GetStyle(),fileInfo.LongColumns());
            RTE_Path &destinationFileName = ArgParser.GetDestinationFileName();
            Msg_List errList;
            if(destinationFileName[0]) 
                fileProcessor.CreateDestinationFile(destinationFileName,errList);
            fileProcessor.BuildWellFormedXMLFile(errList,ArgParser.GetSortMode(),ArgParser.GetFormat(),ArgParser.IncludeClassicHeader());
        }
    }
}

