/*!
 * \file    OMS_Globals.cpp
 * \author  MarkusSi, PeterG, Roterring
 * \brief   Global methods and variables.
 */
/*

    ========== licence begin  GPL
    Copyright (c) 2003-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
*/

#include "Oms/OMS_Defines.h"
#include "Oms/OMS_Globals.hpp"
#include "Oms/OMS_DummyKernelCallback.hpp"
#include "Oms/OMS_KernelDefaultInterface.hpp"
#include "Oms/OMS_DumpInterface.hpp"
#include "Oms/OMS_VersionDictionary.hpp"
#include "Oms/OMS_MonitorDirectory.hpp"
#include "Oms/OMS_Session.hpp"
#include "Oms/OMS_ObjectContainer.hpp"
#include "Oms/OMS_AFX.h"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "SAPDBCommon/SAPDB_MemCopyMove.hpp"


/*----------------------------------------------------------------------*/

static OMS_DummyKernelCallback DummySink;

IliveCacheSink *OMS_GetProxySinkAddr() 
{
  return &DummySink; 
}

/*----------------------------------------------------------------------*/

extern "C" void co10_initSingletons()
{
  // initialize all singletons to sensible values - on load in xregcomp
  OMS_Globals::InitSingletons();
}

/*----------------------------------------------------------------------*/

extern "C" void* co10_GetInterface() 
{
  if (OMS_Globals::m_libOmsInterfaceInstance == NULL)
    OMS_Globals::InitSingletons();
  return reinterpret_cast<void*>(OMS_Globals::m_libOmsInterfaceInstance);
}

/*----------------------------------------------------------------------*/

extern void omsGetLiveCacheVersion(char* pBuf, int bufSize)
{
  if (NULL == OMS_Globals::KernelInterfaceInstance->GetKnlVersion(pBuf, bufSize))
  {
    OMS_Globals::Throw(e_buffer_too_small, "omsGetLiveCacheVersion", __MY_FILE__, __LINE__);
  }
}

/*----------------------------------------------------------------------*/

extern unsigned int omsGetLCVersion()
{
  return <MAJOR_VERSION><MINOR_VERSION><CORRECTION_LEVEL><FORMATED3_BUILD_PREFIX>;
}

/*----------------------------------------------------------------------*/

extern void omsSimConsoleMessage(const char* msg)
{
  if (OMS_Globals::KernelInterfaceInstance) OMS_Globals::KernelInterfaceInstance->ConsoleMessage(msg);
}

/*----------------------------------------------------------------------*/

extern void AFX_EXT_CLASS omsResetForSimulator()
{
  OMS_Globals::m_globalsInstance->ResetInfo();
}

/*----------------------------------------------------------------------*/

OMS_Globals			            *OMS_Globals::m_globalsInstance = NULL;
OMS_GetSinkPtrFunc		      OMS_Globals::GetSinkAddrFunc = OMS_GetProxySinkAddr;
LVC_KernelInterface*		    OMS_Globals::KernelInterfaceInstance = NULL;
OMS_LibOmsInterfaceInstance	*OMS_Globals::m_libOmsInterfaceInstance = NULL;

static RTESync_Spinlock omsSyncDumpError;

static double			omsGlobalsInstanceSpace[sizeof(OMS_Globals) / sizeof(double) + 1];
static double			omsVersionDictInstanceSpace[sizeof(OMS_VersionDictionary) / sizeof(double) + 1];
static double			omsMonitorDirInstanceSpace[sizeof(OMS_MonitorDirectory) / sizeof(double) + 1];
static void       *omsLibOmsInterfaceInstanceSpace = NULL;

/*----------------------------------------------------------------------*/
/* PTS 1115231 */
externCpp
bool omsIsUnicodeInstance()
{
    IliveCacheSink* pSink = OMS_Globals::GetCurrentLcSink();
    bool  isUnicodeInstance = false;
    if (pSink)
    {
        short error;
        pSink->IsUnicodeInstance(isUnicodeInstance, error);
        if (0 != error)
        {
            OMS_Globals::Throw(error, "omsIsUnicodeInstance", __MY_FILE__, __LINE__);
        }
    }
    return isUnicodeInstance;
}

/*----------------------------------------------------------------------*/
OMS_Globals::OMS_Globals()
  : m_versionDictionary(*reinterpret_cast<OMS_VersionDictionary*>(omsVersionDictInstanceSpace))
  , m_monitorDirectory(*reinterpret_cast<OMS_MonitorDirectory*>(omsMonitorDirInstanceSpace))
  , m_versionIter() 
  , m_host(HostUnknown)
  #if defined(OMSTST)
  , m_isOmsTestLib(true)
  #else
  , m_isOmsTestLib(false)
  #endif 
  , m_cntDumpErrors(0)
  , m_heapThreshold(100)
  , m_omsVersionThreshold(MAX_INT4_SP00)
  , m_tracer(0)
  , m_resetSink() 
  , m_nilOid()
  , m_versionBuffer()  // PTS 1128597
{
  new(&m_versionDictionary) OMS_VersionDictionary();
  new(&m_monitorDirectory) OMS_MonitorDirectory();
  m_nilSeq.gg91SetNilRef();
}

void OMS_Globals::InitSingletons()
{
  if (m_globalsInstance == NULL) 
    m_globalsInstance = new(&omsGlobalsInstanceSpace) OMS_Globals();
  if (KernelInterfaceInstance == NULL) 
    KernelInterfaceInstance = &OMS_KernelDefaultInterface::Instance();
  if (m_libOmsInterfaceInstance == NULL)
    m_libOmsInterfaceInstance = new(&omsLibOmsInterfaceInstanceSpace) OMS_LibOmsInterfaceInstance();
}

void OMS_Globals::MakeKnlIdentifier (const char* Source, tsp00_KnlIdentifier& Dest) {
  size_t copyLen = strlen(Source);
  if (copyLen > sizeof(Dest)) {
    OMS_Globals::Throw(e_OMS_identifier_too_long, "OMS_Globals::MakeKnlIdentifier", __MY_FILE__, __LINE__);
    //copyLen = sizeof(Dest);
  }
  SAPDB_MemCopyNoCheck(&Dest[0], Source, copyLen);
  memset(&Dest[copyLen], ' ', sizeof(Dest) - copyLen);
}

/*----------------------------------------------------------------------*/

void OMS_Globals::ResetInfo()
{
  // TODO: this is not perfect, but it will do the job...
  m_versionDictionary.Shutdown();
}

/*----------------------------------------------------------------------*/

void OMS_Globals::Dump(OMS_DumpInterface& dumpObj)
{
  this->DumpVersionDirectory(dumpObj);
  OMS_Session* pSession = REINTERPRET_CAST(OMS_Session*, dumpObj.GetOmsSession());
  if (NULL != pSession)
  {
    pSession->Dump(dumpObj);
  }
}

/*----------------------------------------------------------------------*/

bool OMS_Globals::DumpRequested(tsp00_Int4 errorNo) 
{
  RTESync_LockedScope sync(omsSyncDumpError);
  for (int ix = 0; ix < m_cntDumpErrors; ++ix)
  {
    if (m_dumpOnError[ix] == errorNo)
    {
      /* remove error to be sure that error is only dumped once */
      while (ix < m_cntDumpErrors - 1)
      {
        m_dumpOnError[ix] = m_dumpOnError[ix+1];
      }
      --m_cntDumpErrors;
      return true;
    }
  }
  return false;
}

/*----------------------------------------------------------------------*/

IliveCacheSink* OMS_Globals::GetCurrentLcSink()
{
  return KernelInterfaceInstance->GetSinkPtr();
}

/*----------------------------------------------------------------------*/

void OMS_Globals::DumpVersionDirectory(OMS_DumpInterface& dumpObj)
{
  dumpObj.SetDumpLabel(LABEL_OMS_VERSION);
   
  OMS_VersionDictionary::Iter iter(dumpObj.Synchronize() ? OMS_LOCK_SHARED : OMS_NO_LOCK);
  while (true){
    OMS_Context *pContext = iter.GetFirstInSlot();
    if (pContext == NULL){
      break;
    }

    while (pContext){
    pContext->Dump(dumpObj);
      pContext = iter.GetNextInSlot();
    }
  }
}

/*----------------------------------------------------------------------*/

bool OMS_Globals::InSimulator() const
{
    return (m_host == HostSimulator);
}

/*----------------------------------------------------------------------*/

bool OMS_Globals::InProcServer() const
{
  if (m_host == HostSimulator) {
    IliveCacheSink *sink = GetCurrentLcSink();
    return sink && (sink->GetSinkType() == 2);
  } else {
    return (m_host == HostProcServer);
  }
}

/*----------------------------------------------------------------------*/

void OMS_Globals::SetDumpError(tsp00_Int4 e)
{
  RTESync_LockedScope sync(omsSyncDumpError);
  if (m_cntDumpErrors < MAX_DUMP_ERRORS)
  {
    m_dumpOnError[m_cntDumpErrors] = e;
    ++m_cntDumpErrors;
  }
}

/*----------------------------------------------------------------------*/

void OMS_Globals::SetHost(bool isKernel, bool isProcServer)
{
  m_host = isKernel ? HostKernel : (isProcServer ? HostProcServer : HostSimulator);
}

/*----------------------------------------------------------------------*/

void OMS_Globals::SetOmsVersionThreshold(tsp00_Uint4 threshold) /* PTS 1110149 */
{
  // 100 has been wrong default, set to wanted default
  m_omsVersionThreshold = (100 == threshold) ? 2097152 : threshold;
}

/*----------------------------------------------------------------------*/

OMS_TraceInterface* OMS_Globals::GetTracer()
{
  return m_tracer;
}

/*----------------------------------------------------------------------*/

void OMS_Globals::SetTracer(OMS_TraceInterface* pTraceObj)
{
  m_tracer = pTraceObj;
}

/*----------------------------------------------------------------------*/

void OMS_Globals::Throw(short errorNo, const char* msg, const char* pFile, unsigned int line, OMS_Session* pSession)
{
  Throw(errorNo, msg, OmsObjectId(), pFile, line, pSession);
}

/*----------------------------------------------------------------------*/

void OMS_Globals::Throw(short errorNo, const char* msg, const OmsObjectId& oid, const char* pFile, unsigned int line, OMS_Session* pSession)
{
  IliveCacheSink* pLcSink = NULL;
  if (pSession == NULL){
    // Get session and sink pointer if not already given
    pLcSink = GetCurrentLcSink();
    tsp00_TaskId     taskId;
    pasbool*         pToCancel;
    bool             optimizedStreamCommunication; // PTS 1130281
    pLcSink->GetDefaultContext(REINTERPRET_CAST(void**, &pSession), &pToCancel, optimizedStreamCommunication, taskId);
  }
  else {
    pLcSink = pSession->m_lcSink;
  }

  if (pLcSink && m_globalsInstance->DumpRequested(errorNo)){
    tsp00_Int4      BufLen      = sizeof(errorNo);
    short           DummyError  = 0;
    OMS_HResult hr  = pLcSink->MultiPurpose (m_outcopy, mm_nil, &BufLen, 
                                            (unsigned char*)&errorNo, 
                                            &DummyError);
  }

  // Group errors into error classes
  switch (errorNo) {
  case e_new_failed: // PTS 1109338
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw DbpError with ErrorNo=" << errorNo 
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    if (pSession != 0)
      pSession->IncOutOfMemory();
    throw DbpError(DbpError::DB_ERROR, errorNo, msg, oid, pFile, line);
    break;

  case e_object_dirty:
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw OmsOutOfDate with ErrorNo=" << errorNo
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    if (pSession)
      pSession->IncOutOfDate();
    throw OmsOutOfDate(oid, msg, pFile, line);
    break;

  case e_duplicate_hash_key:
  case e_key_still_visible:
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw OmsDuplicateKey with ErrorNo=" << errorNo
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    throw OmsDuplicateKey(errorNo, oid, msg, pFile, line);
    break;

  case e_hash_key_in_use:
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw OmsKeyInUse with ErrorNo=" << errorNo 
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    throw OmsKeyInUse(e_hash_key_in_use, oid, msg, pFile, line);
    break;

  case e_OMS_lock_collision :
  case e_OMS_request_timeout : /* PTS 1102829 */
    // NOTE: do not forget to test the same codes in OMS_Context::TryLockResult to prevent
    // omsTryLock from throwing OmsLockTimeout exceptions.
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw OmsLockTimeout with ErrorNo=" << errorNo 
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    if (pSession)
      pSession->IncTimeout();
    throw OmsLockTimeout(oid, msg, pFile, line);
    break;

  case e_wrong_class_id:
  case e_wrong_object_version:
  case e_wrong_object_state:
  case e_too_old_oid:
  case e_no_next_object:
  case e_hash_not_found:
  case e_hash_key_not_found:
  case e_unknown_consist_vers:
  case e_object_not_found:
  case e_obj_history_not_found:
  case e_illegal_object_length:
  case e_illegal_object_pos:
  case e_container_dropped:
  case e_consview_cancelled:
    // Combine all error-codes, which corresponds to a 'not found', in a single error-class // PTS 1124935 
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw OmsObjectNotFound with ErrorNo=" << errorNo
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    throw OmsObjectNotFound(errorNo, oid, msg, pFile, line); 
    break;

  default :
    if (pLcSink)
      OMS_TRACE(omsTrError, pLcSink, "throw DbpError with ErrorNo=" << errorNo 
                                     << "  Msg=\""  << msg << "\""
                                     << "  " << oid
                                     << "  (" << pFile << ":" << line << ")");
    bool cancelledByProgram   = false;
    bool provokedByTestkernel = false;
    if (pSession){
      pSession->IncExceptions();
      if (errorNo==e_OMS_cancelled && pSession->IsCancelledByProgram()){
        cancelledByProgram   = true;
        pSession->SetCancelledByProgram(false);
      }
      if ((errorNo==e_requested_dbperror || errorNo==e_OMS_cancelled)
         && pSession->IsProvokedByTestkernel()){
        provokedByTestkernel = true;
        pSession->SetProvokedByTestkernel(false);
      }
    }
    throw DbpError(DbpError::DB_ERROR, errorNo, msg, oid, pFile, line,
                   cancelledByProgram, provokedByTestkernel);
  }
}

/*----------------------------------------------------------------------*/

//PTS 1117690
// strlen on strings of type OmsTypeWyde
int OMS_Globals::WideStrlen(const OmsTypeWyde* s)
{
  int len=-1;
  while (s[++len]);
  return len;
}

/*----------------------------------------------------------------------*/
// PTS 1124170
int OMS_Globals::GetLockIdForVersionDirectory (const OmsVersionId &versionId) const { 
  return m_versionDictionary.GetSingleLockId(versionId);
}

/*----------------------------------------------------------------------*/
// PTS 1124170
bool OMS_Globals::GetLockMethodForVersionDictionary() const{  
  return m_versionDictionary.UseRWLocks();
}

/*----------------------------------------------------------------------*/

void OmsObjectContainer::error(char* msg, OmsObjectContainer*p)
{
  char buf[64];
  IliveCacheSink* pLcSink = OMS_Globals::GetCurrentLcSink();
  DbpBase base(pLcSink);
  base.dbpOpError(msg);
  sp77sprintf(buf, sizeof(buf), "HashNext: %p", p->m_hashnext);
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "Oid: %d.%d(%d)", p->m_oid.getPno(), p->m_oid.getPagePos(), p->m_oid.getGeneration());
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "ObjSeq: %d.%d", p->m_objseq.gg91RefPno(), p->m_objseq.gg91RefPos());
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "State: %d", p->m_state);
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "VerState: %d", p->m_verstate);
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "BeforeImages: %d", p->m_beforeImages);
  base.dbpOpError(buf);
#if defined(BIT64)
  base.dbpOpError(buf);
  sp77sprintf(buf, sizeof(buf), "Filler: %d", p->filler);
#endif
  sp77sprintf(buf, sizeof(buf), "ContainerInfo: %p", p->m_pContainerInfo);
  base.dbpOpError(buf);
}


