/****************************************************************************

  module      : Pager_Controller.cpp

  -------------------------------------------------------------------------

  responsible : Henrik

  special area:
  description : This description is currently not used, as
    documentation is only generated from header files

  last changed: 2002-02-26
  see also    :

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2001-2004 SAP AG


    ========== licence begin  GPL
    Copyright (c) 2001-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "Pager/Pager_Controller.hpp"
#include "Pager/Pager_Exception.hpp"
#include "Pager/Pager_Messages.hpp"
#include "Pager/Pager_Lock.hpp"

#include "DataAccess/Data_BasePage.hpp"
#include "IOManager/IOMan_IDataIO.hpp"
#include "Converter/Converter_ISavePoint.hpp"

#include "KernelCommon/ParameterNames/KernelParam_Pager.hpp"
#include "KernelCommon/Kernel_Dump.hpp"
#include "KernelCommon/Kernel_Trace.hpp"

#include "RunTime/Configuration/RTEConf_ParameterAccess.hpp"
#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"
#include "RunTime/MemoryManagement/RTEMem_Allocator.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/Tasking/RTETask_ITask.hpp"
#include "RunTime/Tasking/RTETask_SuspendReason.h"
#include "FileDirectory/FileDir_ISharedDirectory.hpp"

#include "ggg00.h"
#include "ggg01.h" // OMS errors
#include "gbd00.h"          // tbd00_NodePtrList
#include "hgg01.h"          // g01maxdatawriter
#include "heo56.h"          // vsuspend/vsleep
#include "hgg04.h"
#include "hbd20_5.h"
#include "gbd200.h"
#include "hbd01_1.h" // b01niltree_id
#include "hbd13_3.h" // bd13LoadNode


/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define PAGER_WAIT_FOR_FLUSHCACHE 253

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/

// singleton
Pager_Controller* Pager_Controller::m_Instance = NULL;

/*---------------------------------------------------------------------------*/

Pager_Controller& Pager_Controller::GetInstance(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::GetInstance", Pager_Trace, 5 );

    if ( NULL == m_Instance ) {
        Pager_Controller::CreateInstance();
    }

    return *m_Instance;
}

/*---------------------------------------------------------------------------*/

Pager_IRestart& Pager_IRestart::GetInstance(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_IRestart::GetInstance", Pager_Trace, 5 );

    return Pager_Controller::GetInstance();
}

Pager_ISavepoint& Pager_ISavepoint::GetInstance(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_ISavepoint::GetInstance", Pager_Trace, 5 );

    return Pager_Controller::GetInstance();
}

Pager_IDump& Pager_IDump::GetInstance(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_IDump::GetInstance", Pager_Trace, 5 );

    return Pager_Controller::GetInstance();
}


/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/


SAPDB_Int Pager_Controller::RegisterNextPagerWriter(const tsp00_TaskId taskId,
        Pager_Writer *writer){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::RegisterNextPagerWriter", Pager_Trace, 5 );

    Pager_Lock lock(taskId);

    SAPDBERR_ASSERT_ARGUMENT(m_registeredPagerWriter < m_numPagerWriter);

    m_pagerWriter[m_registeredPagerWriter] = writer;
    m_activePagerWriter[m_registeredPagerWriter] = true;

    return m_registeredPagerWriter++;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::BeginSavepoint (const tsp00_TaskId taskId) {
    if (IsClusteredIOConfigured())
    {
        BeginClusteredSavepoint(taskId);
    }
    else
    {
        BeginUnclusteredSavepoint(taskId);
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::EndSavepoint ( const tsp00_TaskId    taskId,
                                      const SAPDB_Bool      disablePagerWriter){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::EndSavepoint", Pager_Trace, 5 );


    Pager_Lock lock(taskId);

    SAPDBERR_ASSERT_STATE(m_svpActive);

    if (!IsClusteredIOConfigured()) {
        CleanupActiveCaches();
    }
    m_pagerWriterDisabled = disablePagerWriter;
    m_preparePagerWriter  = false;
    m_svpActive          = false;
    m_pagerOrder         = poNil;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushDataCacheBeforePrepare (const tsp00_TaskId taskId,
        SAPDB_UInt &numPagesWritten,
        SAPDB_UInt &numIOActions){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushDataCacheBeforePrepare", Pager_Trace, 5 );

    SAPDBERR_ASSERT_STATE(m_svpActive);
    SAPDBERR_ASSERT_STATE(poNil == m_pagerOrder);

    if (IsClusteredIOConfigured()){
        m_pagerOrder   = poFlushClusteredDataCacheBeforePrepare;
    } else{
        m_pagerOrder   = poFlushDataCacheBeforePrepare;
    }

    m_dataIOTotal.SetValue(0);
    m_dataPagesTotal.SetValue(0);

    Pager_Exception StartWriteData(__CONTEXT__, PAGER_SVP_START_WRITE);
    RTE_Message (StartWriteData);

    if (IsClusteredIOConfigured()){
        bd200_IOSorter::GetInstance().CollectDataPages(taskId);
    }

    ResumeAllPagerWriter (taskId, !SET_CHECK_FBM, false);
    WaitForPagerWriterReply (taskId);

    if (IsClusteredIOConfigured()){
        bd200_IOSorter::GetInstance().EndSave(taskId);
    }

    Pager_Exception StopDataIO(__CONTEXT__, PAGER_SVP_STOP_DATA_IO,
                               SAPDB_ToString(m_dataPagesTotal.GetValue(), _T_d),
                               SAPDB_ToString(m_dataIOTotal.GetValue(), _T_d));
    RTE_Message (StopDataIO);

    numPagesWritten = m_dataPagesTotal.GetValue();
    numIOActions =    m_dataIOTotal.GetValue();
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::WaitForLastSvpSyncOperation (
    const tsp00_TaskId taskId,
    SAPDB_UInt         &numPagesWritten,
    SAPDB_UInt         &numIOActions){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::WaitForLastSvpSyncOperation", Pager_Trace, 5 );

    SAPDB_Bool1     dummy;
    SAPDB_Bool      bSplitLocksExist;
    SAPDB_Int2      pageCnt;
    SAPDB_Int       currCache;
    SAPDB_Int       loopCnt;
    SAPDB_UInt      totalPages;
    SAPDB_UInt      totalIO;

    tbd00_NodePtrList       NodePtrList;
    tbd2_data_cb_flush_list cbptrList;
    tgg00_BasisError        trError = e_ok;

    bSplitLocksExist = true;

    loopCnt          = 0;
    totalIO          = 0;
    totalPages       = 0;

    Pager_Exception WaitForLastSvpSyncOperation(__CONTEXT__, PAGER_SVP_WAIT_LAST_SPLIT,
            SAPDB_ToString(taskId, _T_d));

    RTE_Message (WaitForLastSvpSyncOperation);

    while ((e_ok == trError) && bSplitLocksExist){

        if  (bd20ExclusiveLocks(taskId)) {
            bd20FullestDataCacheSegment(taskId, currCache);

            pageCnt = 1;

            bd20GetPageForIO (taskId, currCache, trError, pageCnt,
                              NodePtrList, cbptrList, dummy);

            if (e_ok == trError) {
                if (pageCnt > 0) {
                    totalIO++;
                    totalPages += pageCnt;
                    WriteDataPage (taskId, NodePtrList[0]);
                    if (e_ok == trError) {
                        bd20ResetIOState (taskId, currCache, pageCnt, cbptrList);
                    }
                }
                else {
                    if (loopCnt >= 5) {
                        vsleep (taskId, 1);
                        loopCnt = 0;
                    }
                    else {
                        loopCnt++;
                        vsleep (taskId, 0);
                    }
                }
            }
        }
        else {
            bSplitLocksExist = false;
        }
    }   // END while
    Pager_Exception StopWaitLastSplit(__CONTEXT__, PAGER_SVP_STOP_WAIT_LS,
                                      SAPDB_ToString(totalPages, _T_d),
                                      SAPDB_ToString(totalIO, _T_d));
    RTE_Message (StopWaitLastSplit);

    numPagesWritten = totalPages;
    numIOActions =    totalIO;

    if (e_ok != trError){
        Pager_Exception Split_Error(__CONTEXT__, PAGER_ERROR_IN_SPLIT,
                                    SAPDB_ToString(trError, _T_d));
        RTE_Crash (Split_Error);
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushDataCacheAfterPrepare (const tsp00_TaskId taskId,
        SAPDB_UInt &numPagesWritten,
        SAPDB_UInt &numIOActions){


    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushDataCacheAfterPrepare", Pager_Trace, 5 );

    SAPDBERR_ASSERT_STATE(m_svpActive);

    if (IsClusteredIOConfigured()){
        m_pagerOrder   = poFlushClusteredDataCacheAfterPrepare;
    } else{
        m_pagerOrder   = poFlushDataCacheAfterPrepare;
    }

    m_dataIOTotal.SetValue(0);
    m_dataPagesTotal.SetValue(0);

    Pager_Exception StartWriteData(__CONTEXT__, PAGER_SVP_START_AFTER_PREPARE);
    RTE_Message (StartWriteData);

    if (IsClusteredIOConfigured()){
        bd200_IOSorter::GetInstance().CollectDataPages(taskId);
    }

    ResumeAllPagerWriter (taskId, !SET_CHECK_FBM, false);
    WaitForPagerWriterReply (taskId);

    if (IsClusteredIOConfigured()){
        bd200_IOSorter::GetInstance().EndSave(taskId);
    }

    Pager_Exception StopAfterPrepare(__CONTEXT__, PAGER_SVP_STOP_AFTER_PREPARE,
                                     SAPDB_ToString(m_dataPagesTotal.GetValue(), _T_d),
                                     SAPDB_ToString(m_dataIOTotal.GetValue(), _T_d));
    RTE_Message (StopAfterPrepare);

    numPagesWritten = m_dataPagesTotal.GetValue();
    numIOActions =    m_dataIOTotal.GetValue();
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushConverterParallel ( const tsp00_TaskId TaskId,
        SAPDB_UInt &numPagesWritten,
        SAPDB_UInt &numIOActions)
{
    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushConverterParallel", Pager_Trace, 5 );

    m_pagerOrder = poFlushConverterCache;
    m_convIOTotal.SetValue(0);
    m_convPagesTotal.SetValue(0);

    Pager_Exception StartFlushConverter(__CONTEXT__, PAGER_SVP_START_CONV_AFTER_PREPARE);
    RTE_Message (StartFlushConverter);

    ResumeAllPagerWriter(TaskId, !SET_CHECK_FBM, true);
    WaitForPagerWriterReply (TaskId);

    Pager_Exception StopAfterFlushConverter(__CONTEXT__, PAGER_SVP_STOP_CONV_AFTER_PREPARE,
                                            SAPDB_ToString(m_convPagesTotal.GetValue(), _T_d),
                                            SAPDB_ToString(m_convIOTotal.GetValue(), _T_d));

    RTE_Message (StopAfterFlushConverter);

    numPagesWritten = m_convPagesTotal.GetValue();
    numIOActions =    m_convIOTotal.GetValue();

    // done
    m_pagerOrder = poNil;
}

/*---------------------------------------------------------------------------*/

bool
Pager_Controller::ReadPagesForDrop(
    const RTETask_ITask &task,
    Msg_List            &msgList,
    IOMan_DeviceNo      volNo )
{
    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::ReadPagesForDrop", Pager_Trace, 5 );

    m_pageIter    = Converter_ISavePoint::Instance().BeginBlockIterator(volNo);
    m_pageEndIter = Converter_ISavePoint::Instance().EndBlockIterator(volNo);

    {
        Pager_Exception StartDropVolume(__CONTEXT__, PAGER_START_DROP_VOLUME,
                                        SAPDB_ToString(volNo, _T_d));
        RTE_Message (StartDropVolume);
    }
    const bool retCode = DropVolume( task, msgList );
    {
        Pager_Exception StopDropVolume(__CONTEXT__, PAGER_STOP_DROP_VOLUME,
                                       SAPDB_ToString(volNo, _T_d));
        RTE_Message (StopDropVolume);
    }
    return retCode;
}

/*---------------------------------------------------------------------------*/

bool
Pager_Controller::DropVolume(
    const RTETask_ITask &task,
    Msg_List            &msgList )
{

    tgg00_FileId tempFileId = b01niltree_id;
    tgg00_FileId permFileId = b01niltree_id;
    tgg00_FileId permStaticFileId = b01niltree_id;

    tempFileId.fileTfn_gg00().becomes( tfnTemp_egg00 );
    tempFileId.fileType_gg00().clear();
    tempFileId.fileType_gg00().addElement (ftsTemp_egg00);

    permFileId.fileTfn_gg00().becomes( tfnTable_egg00 );
    permFileId.fileType_gg00().clear();
    permFileId.fileType_gg00().addElement (ftsPerm_egg00);

    permStaticFileId.fileTfn_gg00().becomes( tfnObj_egg00 );
    permStaticFileId.fileType_gg00().clear();
    permStaticFileId.fileType_gg00().addElement (ftsObject_egg00);

    Data_PageNo             pageNo;
    Data_PageRecoveryMode   recMode;
    tgg00_BasisError        baseError = e_ok; // because of bd13LoadNode

    while (GetNextPageToDrop(pageNo, recMode) && (e_ok == baseError))
    {
        tgg00_RecoveryMode recoveryMode;

        if (recMode.UseStaticPageAddressing())
        {
            recoveryMode = tgg00_RecoveryMode::fromConst(rmStatic_egg00);
        }
        else if (recMode.PageIsRecoverable())
        {
            recoveryMode = tgg00_RecoveryMode::fromConst(rmPerm_egg00);
        }
        else
        {
            recoveryMode = tgg00_RecoveryMode::fromConst(rmTemp_egg00);
        }

        const SAPDB_Int maxRetries = 4;
        SAPDB_Int       retryCount = 0;
        bool            bDoRetry    = false;

        tgg00_PageType pageType = tgg00_PageType::fromConst( ptNil_egg00 );

        do {
            switch (recoveryMode)
            {
            case rmTemp_egg00:
                bd13LoadNode (task.ID(), baseError, tempFileId, pageType, pageNo, rmTemp_egg00, true);
                break;
            case rmPerm_egg00:
                bd13LoadNode (task.ID(), baseError, permFileId, pageType, pageNo, rmPerm_egg00, true);
                break;
            case rmStatic_egg00:
                bd13LoadNode (task.ID(), baseError, permStaticFileId, pageType, pageNo, rmStatic_egg00, true);
                break;
            }

            switch (baseError)
            {
            case e_no_converter_entry: // continue
                baseError = e_ok;
                break;

            case e_illegal_page_no:
            case e_shutdown:
            case e_inconsistent_nodetype:
            case e_bad_datapage:
            case e_corrupted_datapage_from_io:
            case e_no_more_memory:
            case e_invalid_oid:
                break; // abnormal termination

            case e_sysbuffer_overflow:
                retryCount++;
                vsleep( task.ID(), 3 );
                bDoRetry = true;
                break;
            }
        } while ((retryCount < maxRetries) && bDoRetry);

        if( e_ok == baseError && pageType == ptFileDir_egg00 ){
            FileDir_ISharedDirectory::Instance().MarkPageForWrite( task.ID(), pageNo);
        }
    }
    if( e_ok == baseError ){
        return true;
    }

    Pager_Exception abortDropVolume(__CONTEXT__, PAGER_ABORT_DROP_VOLUME,
        Kernel_BasisErrorToResultString( baseError ));
    
    msgList.AppendNewMessage( abortDropVolume );

    return false;
}

/*---------------------------------------------------------------------------*/

SAPDB_Bool Pager_Controller::IsClusteredIOConfigured()
{

    if (cioUndef == m_isClusteredIO) {
        RTEConf_Parameter::String paramValue;
        SAPDBErr_MessageList errMsg;

        if( ! RTEConf_ParameterAccess::Instance()->GetString( UTF8( KERNELPARAM_CLUSTERED_IO ), paramValue, errMsg ))
        {
            RTE_Crash( errMsg );
        }

        if( 0 == strcmp( "NO", ( SAPDB_Char* ) paramValue ))
        {
            m_isClusteredIO = cioNo; // CLUSTERED_IO == NO
        }
        else
        {
            m_isClusteredIO = cioYes; // CLUSTERED_IO == NO
        }
    }
    return (cioYes == m_isClusteredIO);
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::BeginUnclusteredSavepoint (const tsp00_TaskId taskId) {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::BeginUnclusteredSavepoint", Pager_Trace, 5 );

    { // locked scope
        Pager_Lock lock(taskId);

        if (m_preparePagerWriter) { // don't call me twice!
            Pager_Exception SVP_Error(__CONTEXT__, PAGER_ERROR_SVP_ACTIVE,
                                      SAPDB_ToString(taskId, _T_d));
            RTE_Crash (SVP_Error);
        }

        m_preparePagerWriter = true;

        m_svpCoordinator = cgg_nil_pid;
        m_pagerOrder   = poNil;
    }

    WaitForPagerWriterReply(taskId);

    m_svpActive = true;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::BeginClusteredSavepoint (const tsp00_TaskId taskId) {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::BeginClusteredSavepoint", Pager_Trace, 5 );

    SAPDB_Bool doSuspend = false;
    { // locked scope
        Pager_Lock lock(taskId);

        if (m_preparePagerWriter) { // don't call me twice!
            Pager_Exception SVP_Error(__CONTEXT__, PAGER_ERROR_SVP_ACTIVE,
                                      SAPDB_ToString(taskId, _T_d));
            RTE_Crash (SVP_Error);
        }
        m_preparePagerWriter = true;

        // wait until flushDataCache is finished
        if (m_flushActive) {
            m_waitingForFlush = taskId;
            doSuspend = true;
        }
    }

    if(doSuspend) {
        vsuspend (taskId, PAGER_WAIT_FOR_FLUSHCACHE);
    }

    m_svpActive = true;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::Dump( const tsp00_TaskId taskId,
                             Kernel_Dump        &dump )
{
    struct PagerDump pcd;

    pcd.dmpNumPagerWriter        = m_numPagerWriter;
    pcd.dmpRegisteredPagerWriter = m_registeredPagerWriter;
    pcd.dmpSuspendPagerWriter    = m_suspendCnt;
    pcd.dmpPagerOrder            = m_pagerOrder;
    pcd.dmpSVPIsActive           = m_svpActive;
    pcd.dmpFlushIsActive         = m_flushActive;
    pcd.dmpPreparePagerWriter    = m_preparePagerWriter;
    pcd.dmpPagerWriterDisabled   = m_pagerWriterDisabled;

    dump.InsertEntry( Kernel_Dump::DmpPagerController,
                      Kernel_DumpPage::Entry( &pcd, sizeof( pcd )));
};

/*---------------------------------------------------------------------------*/

void Pager_Controller::ReadConverter(
    const tsp00_TaskId    taskId,
    const SAPDB_Bool      bReadSnapShotConverter )
{

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::ReadConverter", Pager_Trace, 5 );

    if (m_pagerOrder != poNil) { // idle ?
        Pager_Exception CVT_Error(__CONTEXT__, PAGER_ERROR_CVT_ACTIVE,
                                  SAPDB_ToString(taskId, _T_d));
        RTE_Crash (CVT_Error);
    }

    m_convIOTotal.SetValue(0);
    m_convPagesTotal.SetValue(0);

    if( bReadSnapShotConverter ){
        m_pagerOrder = poReadSnapShotConverter;
        Pager_Exception StartReadConverter(__CONTEXT__, PAGER_START_READ_SNAPSHOT_CONVERTER);
        RTE_Message (StartReadConverter);
    }
    else{
        m_pagerOrder = poReadConverterCache;
        Pager_Exception StartReadConverter(__CONTEXT__, PAGER_START_READ_CONVERTER);
        RTE_Message (StartReadConverter);
    }

    ResumeAllPagerWriter(taskId, !SET_CHECK_FBM, true);
    WaitForPagerWriterReply (taskId);


    Pager_Exception StopAfterReadConverter(__CONTEXT__, PAGER_STOP_READ_CONVERTER,
                                           SAPDB_ToString(m_convPagesTotal.GetValue(), _T_d),
                                           SAPDB_ToString(m_convIOTotal.GetValue(), _T_d));

    RTE_Message (StopAfterReadConverter);

    // done
    m_pagerOrder = poNil;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::SuspendPagerWriter(SAPDB_UInt    pagerIndex,
        tsp00_TaskId  taskId){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::SuspendPagerWriter", Pager_Trace, 5 );

    tsp00_TaskId resumeTaskId = cgg_nil_pid;

    { // locked scope
        Pager_Lock lock(taskId);

        SAPDBERR_ASSERT_STATE(m_activePagerWriter[pagerIndex]);
        SAPDBERR_ASSERT_STATE(!visresume(taskId));

        m_activePagerWriter[pagerIndex] = false;
        m_suspendCnt.Increment();

        if((m_suspendCnt.GetValue() == m_numPagerWriter) && (m_svpCoordinator != cgg_nil_pid)){
            // -> sl_prepare_data_writer is true
            resumeTaskId      = m_svpCoordinator;
            m_svpCoordinator = cgg_nil_pid;
        }
    }

    if(resumeTaskId != cgg_nil_pid) {
        vresume (resumeTaskId);
    }

    vsuspend (taskId, TSK_VSUSPEND_REASON_NO_WORK);
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::EnablePagerWriter (){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::EnablePagerWriter", Pager_Trace, 5 );

    RTESync_LockedScope Lock(m_SpinLock);

    m_pagerWriterDisabled = false;
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushDataCacheRegion ( const tsp00_TaskId taskId,
        const SAPDB_Int   regOffset) {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushDataCacheRegion", Pager_Trace, 5 );

    const SAPDB_UInt pagerIndex = regOffset % m_registeredPagerWriter;

    SAPDB_Bool doResume = false;

    if(!m_pagerWriterDisabled && !m_preparePagerWriter) { // PTS 1111179 TS 2001-07-30
        m_pagerWriter[pagerIndex]->ReqFlushForDataCache(regOffset);
        if (!m_activePagerWriter[pagerIndex]) {
            doResume = false;
            {
                Pager_Lock lock(taskId);
                if (!(m_preparePagerWriter || m_pagerWriterDisabled)) {
                    if (!m_activePagerWriter[pagerIndex]) {
                        m_activePagerWriter[pagerIndex] = true;
                        m_suspendCnt.Decrement();
                        doResume   = true;
                    }
                }
            }
            if  (doResume) {
                vresume (m_pagerWriter[pagerIndex]->GetTaskId());
            }
        }
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushTempPagesFromLRURechain(const tsp00_TaskId taskId){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushTempPagesFromLRURechain", Pager_Trace, 5 );

    tsp00_TaskId resumeTaskId = cgg_nil_pid;

    { // locked scope
        Pager_Lock lock(taskId);

        if ((m_pagerOrder != poNil) || m_svpActive || m_preparePagerWriter) { // idle ?
            return;
        }

        SAPDBERR_ASSERT_STATE(!m_flushActive);
        m_flushActive = true;

        m_pagerOrder   = poFlushTempPagesFromLRURechain;
        m_dataIOTotal.SetValue(0);
        m_dataPagesTotal.SetValue(0);

        //      Pager_Exception StartWriteData(__CONTEXT__, PAGER_FLUSHLRU_START_WRITE);
        //      RTE_Message (StartWriteData);
    }

    ResumeAllPagerWriter (taskId, !SET_CHECK_FBM, false);
    WaitForPagerWriterReply (taskId);

    { // locked scope
        Pager_Lock lock(taskId);

        Pager_Exception StopDataIO(__CONTEXT__, PAGER_FLUSHLRU_STOP_DATA_IO,
                                   SAPDB_ToString(m_dataPagesTotal.GetValue(), _T_d),
                                   SAPDB_ToString(m_dataIOTotal.GetValue(), _T_d));
        RTE_Message (StopDataIO);

        m_pagerOrder   = poNil;
        m_flushActive = false;

        // wake up BeginSavepoint
        if (m_waitingForFlush != cgg_nil_pid) {
            resumeTaskId      = m_waitingForFlush;
            m_waitingForFlush = cgg_nil_pid;
        }
    }

    if(resumeTaskId != cgg_nil_pid) {
        vresume (resumeTaskId);
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::FlushDataCache( const tsp00_TaskId taskId) {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::FlushDataCache", Pager_Trace, 5 );

    tsp00_TaskId resumeTaskId = cgg_nil_pid;

    { // locked scope
        Pager_Lock lock(taskId);

        if ((m_pagerOrder != poNil) || m_svpActive || m_preparePagerWriter) { // idle ?
            return;
        }

        SAPDBERR_ASSERT_STATE(!m_flushActive);
        m_flushActive = true;

        bd200_IOSorter::GetInstance().CollectDataPages(taskId);

        m_pagerOrder   = poFlushDataCacheBeforePrepare;
        m_dataIOTotal.SetValue(0);
        m_dataPagesTotal.SetValue(0);

        //      Pager_Exception StartWriteData(__CONTEXT__, PAGER_PRED_START_WRITE);
        //      RTE_Message (StartWriteData);
    }

    ResumeAllPagerWriter (taskId, !SET_CHECK_FBM, false);
    WaitForPagerWriterReply (taskId);

    { // locked scope
        Pager_Lock lock(taskId);

        bd200_IOSorter::GetInstance().EndSave(taskId);

        Pager_Exception StopDataIO(__CONTEXT__, PAGER_PRED_STOP_DATA_IO,
                                   SAPDB_ToString(m_dataPagesTotal.GetValue(), _T_d),
                                   SAPDB_ToString(m_dataIOTotal.GetValue(), _T_d));
        RTE_Message (StopDataIO);

        m_pagerOrder   = poNil;
        m_flushActive = false;

        // wake up BeginSavepoint
        if (m_waitingForFlush != cgg_nil_pid) {
            resumeTaskId      = m_waitingForFlush;
            m_waitingForFlush = cgg_nil_pid;
        }
    }

    if(resumeTaskId != cgg_nil_pid) {
        vresume (resumeTaskId);
    }
}

/*---------------------------------------------------------------------------*/

bool
Pager_Controller::GetNextPageToDrop(
    Data_PageNo             &pageNo,
    Data_PageRecoveryMode   &recMode)
{
    if (m_pageIter == m_pageEndIter)
    {
        return false;
    }
    else
    {
        m_pageIter.GetPageNo(pageNo, recMode);
        ++m_pageIter;
    }
    return true;
}

/*---------------------------------------------------------------------------*/

// private:

Pager_Controller::Pager_Controller(SAPDBMem_IRawAllocator &allocator):
        m_SpinLock((SAPDB_UTF8*)"Pager_Controller::m_SpinLock"),
        m_pagerWriter(allocator),
        m_activePagerWriter(allocator),
        m_numPagerWriter(g01maxdatawriter()),
        m_registeredPagerWriter(0),
        m_svpCoordinator(cgg_nil_pid),
        m_waitingForFlush(cgg_nil_pid),
        m_svpActive(false),
        m_preparePagerWriter(false),
        m_flushActive(false),
        m_pagerWriterDisabled(SAPDB_TRUE),
        m_isClusteredIO(cioUndef),
        m_pagerOrder(poNil)
{
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::CreateInstance(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::CreateInstance", Pager_Trace, 5 );

    if ( NULL == m_Instance ) {
        m_Instance = new( RTEMem_Allocator::Instance() ) Pager_Controller(RTEMem_Allocator::Instance());
        if (NULL == m_Instance ){
            RTE_Crash(SAPDBErr_Exception(__FILE__, __LINE__, SAPDBERR_ASSERT_STATE_FAILED, "Pager_Controller::CreateInstance() alloc failed"));
        }
        m_Instance->Initialize();
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::Initialize() {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::Initialize", Pager_Trace, 5 );

    m_pagerWriter.Initialize(m_numPagerWriter);
    m_activePagerWriter.Initialize(m_numPagerWriter,false);
    Pager_Exception InitMode(__CONTEXT__, PAGER_START_MODE, (IsClusteredIOConfigured() ? "with clustered IO" : "without clustered IO"));
    RTE_Message (InitMode);
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::ResumeAllPagerWriter (
    const tsp00_TaskId  taskId,
    const SAPDB_Bool1   set_check_fbm,
    const SAPDB_Bool    resumeForConverter){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::ResumeAllPagerWriter", Pager_Trace, 5 );

    SAPDB_Bool doResume = false;

    for (SAPDB_UInt pagerIndex = 0; pagerIndex < m_numPagerWriter; pagerIndex++) {
        doResume = false;

        { // locked scope
            Pager_Lock lock(taskId);

            if(resumeForConverter || !m_pagerWriterDisabled){
                if (!m_activePagerWriter[pagerIndex]) {
                    m_activePagerWriter[pagerIndex] = true;
                    m_suspendCnt.Decrement();
                    doResume   = true;
                }
            }
        }

        if  (doResume) {
            vresume (m_pagerWriter[pagerIndex]->GetTaskId());
        }
    }
}

/*---------------------------------------------------------------------------*/

/*
void Pager_Controller::ResumePagerWriterForConverter (const tsp00_TaskId taskId,
        const SAPDB_UInt pagerIndex) {

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::ResumePagerWriterForConverter", Pager_Trace, 5 );

    SAPDB_Bool doResume = false;

    if(pagerIndex < m_numPagerWriter) {
        doResume = false;

        { // locked scope
            Pager_Lock lock(taskId);

            if (!m_activePagerWriter[pagerIndex]) {
                m_activePagerWriter[pagerIndex] = true;
                m_suspendCnt.Decrement();
                doResume = true;
            }
        }

        if (doResume) {
            vresume (m_pagerWriter[pagerIndex]->GetTaskId());
        }
    }
}
*/

/*---------------------------------------------------------------------------*/

void Pager_Controller::CleanupActiveCaches(){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::CleanupActiveCaches", Pager_Trace, 5 );

    SAPDB_UInt pagerIndex;

    for(pagerIndex = 0; pagerIndex < m_numPagerWriter; pagerIndex++){
        m_pagerWriter[pagerIndex]->CleanupActiveCache();
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::WaitForPagerWriterReply (const tsp00_TaskId taskId){

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::WaitForPagerWriterReply", Pager_Trace, 5 );

    SAPDB_Bool doSuspend = false;

    { // locked scope
        Pager_Lock lock(taskId);

        if(m_suspendCnt.GetValue() != m_numPagerWriter) {
            SAPDBERR_ASSERT_STATE(m_svpCoordinator == cgg_nil_pid);
            SAPDBERR_ASSERT_STATE(!visresume(taskId));

            m_svpCoordinator = taskId;
            doSuspend    = true;
        }
    }

    if(doSuspend) {
        vsuspend (taskId, 54);
    }
}

/*---------------------------------------------------------------------------*/

void Pager_Controller::WriteDataPage (const tsp00_TaskId    taskId,
                                      tbd_nodeptr           &pNode ) const{

    SAPDBTRACE_ROUTINE_DEBUG("Pager_Controller::WriteDataPage", Pager_Trace, 5 );

    Data_BasePage   Page( pNode,  !gg04IsStaticPage( pNode->nd_pt2() ),
                          pNode->nd_pmode().includes( pmTemp_egg00 ) );

    IOMan_IDataIO::GetInstance().WriteDataPage( taskId, Page );
};

/*===========================================================================*
*  END OF CODE                                                              *
*===========================================================================*/

