/*!
    @ingroup        Restart
    @file           Rst_RedoReadTask.cpp
    @author         UweH
    @brief          defines a singleton to handle a task which reads the log
                    and creates redo files
*/
/*

    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


*/
#include "ggg92.h"   // used by hkb51.h
#include "heo52.h"   // vclock
#include "hgg01_3.h" // g01restart_time
#include "hkb51.h"   // kb51GetRedoTransIndex
#include "hkb50.h"   // k50GetLogTransaction, k50RegisterTransaction
#include "hkb57.h"   // kb57RedoTransNoIncrement
#include "hkb57_1.h" // k57restartrec

#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_IRawAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "Container/Container_Vector.hpp"
#include "KernelCommon/Kernel_IAdminInfo.hpp"
#include "KernelCommon/Kernel_IAdminConfig.hpp"
#include "KernelCommon/Kernel_Migration.hpp"
#include "KernelCommon/Kernel_Version.hpp"
#include "KernelCommon/Kernel_Trace.hpp"
#include "Restart/Rst_RedoReadTask.hpp"
#include "Restart/Rst_IRedoTrafficControl.hpp"
#include "Restart/Rst_Exceptions.hpp"
#include "Restart/Rst_LogEntryIterator.hpp"
#include "Logging/Log_VolumeIterator.hpp"
#include "Logging/Log_SeqTapeReader.hpp"
#include "Logging/Log_Volume.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "Logging/Log_AfterImage.hpp"
#include "Logging/Log_ActionUndoRedo.hpp"
#include "Logging/Log_ActionPartialRollback.hpp"
#include "Logging/Log_ActionSavepoint.hpp"
#include "Logging/Log_OpenTransFile.hpp"
#include "Logging/Log_Savepoint.hpp"
#include "Transaction/Trans_Context.hpp"

/// EndOfTrans will check the IOSequence of the inserted job
#define CHECK_ASCENDING_EOT_SEQUENCE            true   // PTS 1113641 UH 2002-01-25 

/// every NUMBER_OF_ENTRIES_FOR_NEXT_WRITE entries, a message will be written in knldiag
/// also it is the checked, whether the restart_time has been expired, so that a savepoint
/// has to be written
#define NUMBER_OF_ENTRIES_FOR_NEXT_WRITE        50000

/// Defines the state of the iterator
class Rst_CopyState
{
public:
    /// Default CTOR. This must not used for the first initialization, because
    /// m_isFirstEntry has to be set to TRUE for the first entry. 
    Rst_CopyState()
        :
        m_copyPending(false),
        m_isMultipleEntry(false),
        m_isFirstEntry(false),
        m_transaction(NULL),
        m_alreadyCopied(false)
    {}
    
    /// initializes the state for starting the redo. I.e. the state isFirstEntry has to be
    /// set to true
    void InitializeForRedo()
    {
        *this = Rst_CopyState();
        this->m_isFirstEntry = true;
    }

    /// Writes Information of the copying-state into the Kernel_VTrace
    void WriteToTrace(const char * text)
    {
        Kernel_VTrace() << " === Rst_CopyState :: " << text << NewLine ;
        Kernel_VTrace() <<  " copy pending = " << m_copyPending <<
                            " multipleEntry = " << m_isMultipleEntry <<
                            " already copied = " << m_alreadyCopied <<
                            " is first entry = " << m_isFirstEntry << 
                            " transno = " << m_transaction->GetRedoTransNo().gg90GetInt4() << NewLine ;
    }        

    /// Releases a not completely copied Log_Entry, which cannot be completed.
    /// Releases all used ressources and locks.
    SAPDB_Bool ReleasePendingEntry( SAPDBErr_MessageList   &errlist )
    {
        if ( LogTrans_Trace.TracesLevel(9) )
            WriteToTrace("ReleasePendingEntry");
        if (m_copyPending && !m_alreadyCopied )
        {
            SAPDBERR_ASSERT_STATE( m_transaction != NULL );
            if ( ! m_isMultipleEntry )
            {
                if ( ! m_transaction->ReleaseIncompleteRedoEntry(errlist)) 
                {
                    m_writer.WriteToTrace("writer");
                    return false;
                }
            }
            else
            { 
                m_transaction->ResetLocalRedoEntries();
            }

        }
        *this = Rst_CopyState();
        return true;
    }

    /// Finishes the copying of a RedoEntry by releasing all used pages and
    /// locks. Needed values are given as parameters, because members are
    /// not yet given (i.e. called from StartCopyEntry)
    SAPDB_Bool MaterializeEntryInRedoFile(const Log_IOSequenceNo        ioSequence,
                                                Log_Transaction        &transaction,
                                                Data_SplitSpaceWriter  &writer,
                                                SAPDB_Bool             alreadyCopied,
                                                SAPDB_Bool             multipleEntry,
                                                SAPDBErr_MessageList   &errlist)
    {
        if ( LogTrans_Trace.TracesLevel(9) )
        {
            m_transaction = &transaction;
            m_alreadyCopied = alreadyCopied;
            m_isMultipleEntry = multipleEntry;
            
            WriteToTrace("MaterializeEntryInRedoFile");
        }
            
        if ( ! alreadyCopied)
        {
            if ( ! multipleEntry )
            {
                if ( ! transaction.FinishCopiedRedoEntry(errlist)) 
                {
                    writer.WriteToTrace("writer");
                    return false;
                }
            }
            else
            {
                if (! transaction.RestoreLocalRedoEntries (ioSequence,errlist))
                {
                    errlist = errlist + Log_Exception(__CONTEXT__,SAPDBERR_ASSERT_STATE_FAILED,"transaction.RestoreLocalRedoEntries");
                    writer.WriteToTrace("writer");
                    return false;
                }
            }
        }
        *this = Rst_CopyState();
        return true;
    }



    /// Finishes the copying of a RedoEntry by releasing all used pages and
    /// locks. Needed values are taken from the member-variables. (called from
    /// ContinueCopyEntry)
    SAPDB_Bool MaterializeEntryInRedoFile(const Log_IOSequenceNo        ioSequence,
                                                SAPDBErr_MessageList   &errlist )
    {
        return  MaterializeEntryInRedoFile(ioSequence, *m_transaction, m_writer, 
                                           m_alreadyCopied, m_isMultipleEntry, errlist);
    }
    
    /// Sets the state of the entry to CopyPending. This means not the
    /// whole entry could be yet copied and more calls to ContinueCopyEntry
    /// are needed in order to finish copying of this entry.
    void SetCopyPending(          Log_Transaction        *transaction,
                        const     Data_SplitSpaceWriter   splitSpaceWriter,
                        const     SAPDB_Bool              alreadyCopied,
                        const     SAPDB_Bool              isMultipleEntry)
    {
        m_copyPending = true;
        m_transaction = transaction;
        m_writer = splitSpaceWriter;
        m_alreadyCopied = alreadyCopied;
        m_isMultipleEntry  = isMultipleEntry;

        if ( LogTrans_Trace.TracesLevel(9) )
            WriteToTrace("SetCopyPending");
    }

    /// Returns true, if the associated entry is not complete copied
    SAPDB_Bool IsCopyPending () const
    {
        return m_copyPending;
    }

    /// Returns the transaction of the currently copied entry
    Log_Transaction & GetTransaction() const
    {
        SAPDBERR_ASSERT_STATE (m_transaction != NULL);
        return *m_transaction;
    }

    /// returns the writer, which is used to copy this entry into the Redofile
    Data_SplitSpaceWriter & GetWriter() 
    {
        return m_writer;
    }

    /// returns true, if this entry has already been copied and thus can
    /// be skipped by the Log_Reader
    SAPDB_Bool AlreadyCopied() const
    {
        return m_alreadyCopied;
    }

    /// returns true, if this entry is the first one read by this logqueue
    SAPDB_Bool IsFirstEntryCopied() const
    {
        return m_isFirstEntry;
    }

    /// Resets the flag, which defines, whether this is the first entry read 
    void SetFirstEntryFinished()
    {
        m_isFirstEntry = false;
    }
        
private:
    SAPDB_Bool            m_copyPending;     ///< true if the last entry has been copied completely
    SAPDB_Bool            m_isMultipleEntry; ///< true if the last entry has been a multiple entry
    SAPDB_Bool            m_alreadyCopied;   ///< true if the entry has been copied already (redo of redo)
    SAPDB_Bool            m_isFirstEntry;    ///< true if this is the first entry read from this logqueue (PTS 1133580 mb 2005-01-25)
    Log_Transaction      *m_transaction;     ///< pointer to the current transaction
    Data_SplitSpaceWriter m_writer;          ///< the writer, which is used to append parts of logentries
};



/*!
    @class Rst_AfterImageHeadSpace
    @brief This is used to implement the interface Data_ISplitSpace for reading purpose.
 */
class Rst_AfterImageHeadSpace : public Data_ISplitSpace
{
public:
	/// returns the number of current part
    virtual PartNo CurrentPart () const
    {
        return 0;
    }
    /// get one part
    virtual void GetPart (PartNo       Part,
                          SAPDB_UInt  &Length,
                          SAPDB_Byte* &Value) const
    {
        Length = m_Length;
        Value  = m_Value;
    }
	/// returns lenght of complete splitspace
    virtual SAPDB_UInt Length() const
    {
        return m_TotalLength;
    }
	/// checks, whether an value is assigned or not
    virtual bool IsAssigned() const
    {
        return m_Value != 0;
    }
    /// Appends the given information to the errorlist
    virtual void AppendContextInfo (SAPDB_Char  const * const  filename,
                                    SAPDB_UInt4 const          linenumber,
                                    SAPDBErr_MessageList      &errlist,
                                    SAPDB_Char  const * const  title = 0) const
	{
        errlist = errlist + Restart_Exception(filename,linenumber,RST_AFTERIMAGE_HEAD_INFO,
                                              title==0?"":title,
                                              m_Value==0?"":(SAPDB_Char const *)SAPDB_ToString(m_Value, m_Length, SAPDB_ToStringClass::hex),
                                              SAPDB_ToString(m_Length),
                                              SAPDB_ToString(m_TotalLength));
	}
    /// default constructor
    Rst_AfterImageHeadSpace ()
    : m_Value       (NULL),
      m_Length      (0),
      m_TotalLength (0)
    {}
    /*!
        @brief  assigns a new space
        @param  iter the logentry, which will be assigned to the Afterimage
     */
    void Assign (const Rst_LogEntryIterator& iter)
    {
        m_Value       = iter.GetValuePtr();
        m_Length      = iter.Length();
        m_TotalLength = iter.RestLength();
    }
private:
    /// pointer to the value of the current entry
    SAPDB_Byte *m_Value;
    /// length of the current entry
    SAPDB_UInt m_Length;
    /// total length of the current entry
    SAPDB_UInt m_TotalLength;
};
/* --------------------------------------------------------------------------- */
Rst_RedoReadTask::Rst_RedoReadTask ( Rst_LogEntryIterator   &logIter,
                                     volatile bool          &redoIsAborted )
: m_logIter (logIter),
  m_redoIsAborted(redoIsAborted)
{}

/* --------------------------------------------------------------------------- */
Rst_RedoReadTask::~Rst_RedoReadTask()
{
    m_logIter.DeleteIter();
}

/* --------------------------------------------------------------------------- */
Rst_IRedoReadTask::ReadLogResult
Rst_RedoReadTask::ReadLogAndCreateRedoFiles (tsp00_TaskId          taskid,
                                             tgg00_TransContext   &oldcontext,
                                             SAPDBErr_MessageList &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::ReadLogAndCreateRedoFiles", LogTrans_Trace, 5);

    SAPDBMem_IRawAllocator  &allocator      = *(reinterpret_cast <SAPDBMem_IRawAllocator*>
                                                (oldcontext.trAllocator_gg00));
    Rst_IRedoTrafficControl &coordinator    = Rst_IRedoTrafficControl::Instance();
    Log_Volume              &log            = Log_Volume::Instance();

    
    if ( ! ReCreateOpenTransactions (taskid, oldcontext, allocator, errlist) )
	{
	    coordinator.SetEndOfLogRead (taskid);
        return RedoReadReCreateOpenTransFailed;
	}

    m_logIter.Initialize(taskid, allocator, &(oldcontext.trRteCommPtr_gg00->to_cancel) ) ;

    const ReadLogInternResult result =
        InternReadLogAndCreateRedoFile( taskid, allocator, oldcontext, coordinator, m_logIter, errlist );
    switch (result)
    {
        case ReadLogOk:
            break;
        case ReadLogAborted:
            m_redoIsAborted = true;
            break;
        case ReadLogSavepointNotFound:
            return RedoReadSavepointNotFound;
        case ReadLogDifferentEotFound:
        case ReadLogReadActionError:
        case ReadLogError:
        case ReadLogMaterializeError:
        case ReadLogPartialRollbackError:
        case ReadLogEndOfTransHandlingFailed:
        case ReadLogRegisterTransactionFailed:
        case ReadLogCloseOpenTransFailed:
            return RedoReadTransHandlingError;
        case ReadLogSoftwareTooOld:
            return RedoReadSoftwareTooOld;
        case ReadLogReleaseRedoEntryFailed:
            return RedoReadCloseOpenTransFailed;
        default:
            errlist = errlist + Restart_Exception(__CONTEXT__, RST_UNEXPECTED_RESULT,
                                                  "InternReadLogAndCreateRedoFile",
                                                  SAPDB_ToString(result));
    }

    if ( m_redoIsAborted )
    {
        coordinator.SetEndOfLogRead (taskid);
        errlist = Restart_Exception(__CONTEXT__, RST_REDO_ABORTED);
        return RedoReadAborted;
    }

    if ( ! CloseOpenTransactions (taskid, allocator, m_logIter.GetLastProcessedIOSeq(), errlist) )
        return RedoReadCloseOpenTransFailed;
        
    coordinator.SetEndOfLogRead (taskid);
    return RedoReadOk;
}
/* --------------------------------------------------------------------------- */
Rst_RedoReadTask::ReadLogInternResult
Rst_RedoReadTask::InternReadLogAndCreateRedoFile(tsp00_TaskId             taskid,
                                                 SAPDBMem_IRawAllocator  &allocator,
                                                 tgg00_TransContext      &oldcontext,
                                                 Rst_IRedoTrafficControl &coordinator,
                                                 Rst_LogEntryIterator    &iter,
                                                 SAPDBErr_MessageList    &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::InternReadLogAndCreateRedoFile", LogTrans_Trace, 5);
    
    Rst_AfterImageHeadSpace  entryHeadSpace;
    Log_AfterImage           afterImage;
    bool                     isOK            = true;
    bool                     checkForSavepointEntry
                             = Log_Savepoint::GetLastSavepointEntryType() != Log_NoOp; // for migration
    bool                     endOfTransAlreadyProcessed = false; // PTS 1123461 UH 2003-08-28
    tgg00_TransIndex         currTransIndex;
    Log_Transaction         *pTransaction    = NULL;
    SAPDB_UInt               readEntryCount  = 0;
    SAPDB_Int                timeInSec;
    SAPDB_Int                dummyTimeInMSec;
    SAPDB_Int                lastSavepointTriggerTime;
    Log_IOSequenceNo         lastIOSequence;

    vclock (&lastSavepointTriggerTime, &dummyTimeInMSec);

    Container_Vector<Rst_CopyState> iterState (allocator);

    // allocate initial array for the queue-states (PTS 1133580 mb 2005-01-25)
    
    Log_QueueID configuredQueueCount = Log_Volume::Instance().GetQueueCountFromParameter();
    iterState.Resize(Log_Volume::Instance().GetQueueCountFromParameter());
    
    Log_QueueID queueTmpId;
    for (queueTmpId = 0; queueTmpId < configuredQueueCount; queueTmpId++)
    {
        iterState[queueTmpId].InitializeForRedo();
    }


    ReadLogInternResult result = ReadLogOk; // PTS 1133003 mb 2004-12-15 
    while (    (iter.GetState() == Log_VolumeIterator::Iter_ok )
            && (ReadLogOk == result ))
    {
        if ( m_redoIsAborted )
        {
            Rst_IRedoTrafficControl::Instance().SetEndOfLogRead (taskid);
            errlist = Restart_Exception(__CONTEXT__, RST_REDO_ABORTED);
            result = ReadLogAborted;
            break;
        }

        iter.CheckSuspendForStandby(taskid);

        // PTS 1130336 mb 2004-07-15 multi-log-queue
        if ((Log_QueueID)iterState.GetSize() < (iter.GetQueueId()+1))
        // react flexible, if the number of queues increases
        {
            iterState.Resize(iter.GetQueueId()+1);
            for (queueTmpId = 0; queueTmpId < iterState.GetSize(); queueTmpId++)
            {
                iterState[queueTmpId].InitializeForRedo();
            }
        }

        Rst_CopyState &currentCopyState(iterState[iter.GetQueueId()]);

        // check for an interrupted entry
        if (    iter.QueueSeqNo() == MIN_IOSEQUENCE 
             && currentCopyState.IsCopyPending() )
        {
            // an interrupted entry must be released
            if (!currentCopyState.ReleasePendingEntry(errlist))
            {
                errlist = errlist + Restart_Exception(__CONTEXT__,
                                                      RST_ERROR_DURING_RELEASE_SPACE );
                result = ReadLogReleaseRedoEntryFailed;
                break;
            }
        }

        lastIOSequence = iter.IOSeqNo();

        if ( ! currentCopyState.IsCopyPending() )
        // a new entry is starting
        {
            if (currentCopyState.IsFirstEntryCopied())
            // PTS 1133580 mb 2005-01-25
            // due to interleaving of multiple queues, a logentry may be interrupted by the savepointentry.
            // In this case the incomplete entry has to be ignored.
            {
                if ( checkForSavepointEntry || ! iter.IsTailOfAnEntry())
                // the first read entry has to be the savepoint-entry or, in case of a redo-savepoint,
                // the start of another valid entry. Thus it must be finished!
                // Also for every queue the state has to be reset if the first complete entry is found
                {
                    currentCopyState.SetFirstEntryFinished();
                }
                else
                {
                    Log_QueueID currentQueue = iter.GetQueueId();
                    
                    // skip all incomplete entries of this queue thy may be spawn among multiple pages and
                    // can also be interleaved with pages from other queues
                    while (iter.IsTailOfAnEntry() && (iter.GetQueueId() == currentQueue))
                    {
                        ++iter;
                    }
                    
                    // if the whole entry has been skipped this has never be done again - for this queue
                    if (iter.GetQueueId() == currentQueue)
                    {
                        currentCopyState.SetFirstEntryFinished();
                    }
                }
            }

            entryHeadSpace.Assign           (iter);
            Data_SplitSpaceReader reader    (entryHeadSpace);
            if(!afterImage.ReadPersistentFormat (reader, errlist))
            {
                errlist = errlist + Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                                      "Error during reading of after image");
                iter.AppendContextInfo(__CONTEXT__,SAPDB_MAX_UINT4, 0, errlist);
                result = ReadLogError;
                break;
            }
        
            if ( checkForSavepointEntry )
            // check the type of the entry, if it is the savepoint of this restart
            {
                checkForSavepointEntry = false;
                
                if ( afterImage.GetActionType() != Log_Savepoint::GetLastSavepointEntryType() )
                {
                    errlist = Restart_Exception(__CONTEXT__, RST_SAVEPOINT_NOT_FOUND,
                                                Log_ActionTypeStrings[Log_Savepoint::GetLastSavepointEntryType()]);
                    iter.AppendContextInfo(__CONTEXT__, afterImage.GetSequence(),
                                            afterImage.GetTransNo().gg90GetInt4(),
                                            errlist );
                    errlist = errlist + Restart_Exception(__CONTEXT__, RST_REDO_BEGIN_NOT_FOUND);
                    errlist = errlist + Restart_Exception(__CONTEXT__, RST_HINT_RESTORE_LOG);
                    result = ReadLogSavepointNotFound;
                    break;
                }
            }

            if ( ! afterImage.GetTransNo().gg90IsNil() )
            {
                if ( ! GetOrCreateRedoTransaction ( taskid,
                                                    allocator,
                                                    oldcontext,
                                                    afterImage.GetTransNo(),
                                                    afterImage.GetSequence(),
                                                    currTransIndex,
                                                    pTransaction,
                                                    errlist ) )
                {
                    result = ReadLogRegisterTransactionFailed;
                    break;
                }
            }
            else
            {
                currTransIndex = cgg_nil_transindex;
            }
            
            // The read task builds then redo files with a correct transid.
            oldcontext.trWriteTransId_gg00 = afterImage.GetTransNo();
            oldcontext.trIndex_gg00        = currTransIndex;

            // save current read entry position for redo-savepoint
            iter.SetCurrentEntry (afterImage.GetActionType());
            ++readEntryCount;
            
            // write progress messages
            if ( LogTrans_Trace.TracesLevel(9)
                 ||
                 (readEntryCount % NUMBER_OF_ENTRIES_FOR_NEXT_WRITE == 0) )
                 iter.Message ( "redo-read", readEntryCount,
                                         afterImage.GetTransNo().gg90GetInt4(),
                                         afterImage.GetSequence() );

            if ( readEntryCount % NUMBER_OF_ENTRIES_FOR_NEXT_WRITE == 0 )
            {
                vclock (&timeInSec, &dummyTimeInMSec);
                if ( lastSavepointTriggerTime + g01restart_time <= timeInSec )
                {
                    lastSavepointTriggerTime = timeInSec;
                    Trans_Context &context = Trans_Context::GetContext(oldcontext);
                    if ( Log_Savepoint::ok != Log_Savepoint::Instance().StartSavepoint (context.Task(), Log_SVPReasonDistance, context.ErrorList()) )
                    {
                        RTE_Message(context.ErrorList());
                        context.ErrorList().ClearMessageList();
                        // ignore this error
                    }
                }
            }
            
            endOfTransAlreadyProcessed = false; // PTS 1123461 UH 2003-08-28
            
            if ( pTransaction != 0
                 &&
                 pTransaction->GetTransType() != Log_Transaction::redoRead )
            {
                // PTS 1123461 UH 2003-08-28 new if-clause
                // The context is already given to the redo tasks.
                // Do not insert it again to the redo traffic control.
                // But check the correct EndOfTrans handling.
                if ( (pTransaction->GetTransState() == Log_Transaction::committed
                      &&
                      afterImage.GetActionType() != Log_Commit)
                      /* PTS 1132673 UH 2004-11-25
                         if trans is marked as rollbacked the next entry may be of any kind
                         but it must be the last entry in the log !!!
                      ||
                      (pTransaction->GetTransState() == Log_Transaction::rollbacked
                      &&
                      afterImage.GetActionType() != Log_Rollback)
                      */
                      ||
                      (pTransaction->GetTransState() == Log_Transaction::opened) )
                {
                    const char* foundStr =  afterImage.GetActionType() == Log_Commit
                                           ? "COMMIT"
                                           : "ROLLBACK";
                    
                    errlist = Restart_Exception(__CONTEXT__, RST_DIFFERENT_EOT_FOUND,
                                                foundStr,
                                                SAPDB_ToString(pTransaction->GetTransState()) );
                    iter.AppendContextInfo          ( __CONTEXT__,
                                                      afterImage.GetSequence(),
                                                      afterImage.GetTransNo().gg90GetInt4(),
                                                      errlist );
                    pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
                    result = ReadLogDifferentEotFound;
                    break;
                }
                endOfTransAlreadyProcessed = true;
            }
            
            switch ( afterImage.GetActionType() )
            {
            case Log_Commit:
            {
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                Log_ActionEndOfTransaction commitAction (Log_Commit);

                if ( iter.RestLength()
                     ==
                     commitAction.GetPersistentLength() + Log_AfterImage::GetPersistentLength() )
                {
                    // PTS 1124045 mb 2003-09-08 commit-logentry with timestamp for point-in-time recovery
                    if(!commitAction.ReadPersistentFormat (reader, allocator, errlist))
                    {
                        errlist = errlist + Restart_Exception(__CONTEXT__, RST_ERROR_DURING_READ_ACTION);
                        iter.AppendContextInfo          ( __CONTEXT__, afterImage.GetSequence(),
                                                          afterImage.GetTransNo().gg90GetInt4(),
                                                          errlist );
                        pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
                        result = ReadLogReadActionError;
                        break;
                    }
                }
                
                if ( ! iter.UntilDateIsReached(commitAction.GetCommitDate(),commitAction.GetCommitTime()))
                {
                    if ( ! endOfTransAlreadyProcessed ) // PTS 1123461 UH 2003-08-28
                    {
                        if ( ! EndOfTransaction ( taskid,
                                           allocator,
                                           pTransaction,
                                           afterImage.GetActionType() == Log_Commit
                                           ? Log_Transaction::committed
                                           : Log_Transaction::rollbacked,
                                           iter.IOSeqNo(),
                                           currTransIndex,
                                           CHECK_ASCENDING_EOT_SEQUENCE, // PTS 1113641 UH 2002-01-25
                                           errlist ) )
                        {
                            result = ReadLogEndOfTransHandlingFailed;
                            break;
                        }
                    }
                }
                break;
            }
            case Log_Rollback:
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );

                if ( ! endOfTransAlreadyProcessed ) // PTS 1123461 UH 2003-08-28
                {
                    if ( ! EndOfTransaction ( taskid,
                                       allocator,
                                       pTransaction,
                                       afterImage.GetActionType() == Log_Commit
                                       ? Log_Transaction::committed
                                       : Log_Transaction::rollbacked,
                                       iter.IOSeqNo(),
                                       currTransIndex,
                                       CHECK_ASCENDING_EOT_SEQUENCE, // PTS 1113641 UH 2002-01-25
                                       errlist ) )
                    {
                        result = ReadLogEndOfTransHandlingFailed;
                        break;
                    }
                }
                break;
            case Log_FinishOpenTrans:
            {
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );

                // PTS 1121011 mb 2003-03-11 extract kernel-version from savepoint
                Log_ActionSavepoint SVPAction (Log_SVPReasonRestart);
                if(!SVPAction.ReadPersistentFormat (reader, allocator, errlist))
                {
                    errlist = errlist + Restart_Exception(__CONTEXT__, RST_ERROR_DURING_READ_ACTION);
                    iter.AppendContextInfo          ( __CONTEXT__,
                                                      afterImage.GetSequence(),
                                                      afterImage.GetTransNo().gg90GetInt4(),
                                                      errlist );
                    result = ReadLogReadActionError;
                    break;
                }
                if (SVPAction.HasKernelVersionString())
                {
                    if ( ! Kernel_IAdminConfig::Instance().MigrationHandler().
                                CheckCompatibility( Kernel_Migration::log,
                                                    Kernel_Version(SVPAction.GetKernelVersionString()),
                                                    errlist ) )
                    {
                        errlist = errlist + Restart_Exception(__CONTEXT__, RST_SOFTWARE_VERSION_TOO_OLD)
                                  +
                                  Restart_Exception(__CONTEXT__, RST_SAVEPOINT_VERSION,
                                                    SVPAction.GetKernelVersionString());
                        iter.AppendContextInfo     ( __CONTEXT__,
                                                     afterImage.GetSequence(),
                                                     afterImage.GetTransNo().gg90GetInt4(),
                                                     errlist );
                        result = ReadLogSoftwareTooOld;
                        break;
                    }
                }

                // PTS 1130336 mb 2004-07-15 multi-log-queue
                if ( ! ReleaseAllPendingEntries (iterState, errlist) )
                {
                    result = ReadLogReleaseRedoEntryFailed;
                    break;
                }

                if ( ! CloseOpenTransactions (taskid, allocator, lastIOSequence, errlist) )
                {
                    errlist = Restart_Exception(__CONTEXT__, RST_CLOSE_OPEN_TRANS_FAILED);
                    iter.AppendContextInfo     ( __CONTEXT__, 
                                                 afterImage.GetSequence(),
                                                 afterImage.GetTransNo().gg90GetInt4(),
                                                 errlist );
                    result = ReadLogCloseOpenTransFailed;
                    break;
                }
                break;
            }
            case Log_PartialRollback:
            {
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                Log_ActionPartialRollback Action;
                if(!Action.ReadPersistentFormat (reader, allocator, errlist))
                {
                    errlist = Restart_Exception(__CONTEXT__, RST_ERROR_DURING_READ_ACTION);
                    iter.AppendContextInfo          ( __CONTEXT__, 
                                                      afterImage.GetSequence(),
                                                      afterImage.GetTransNo().gg90GetInt4(),
                                                      errlist );
                    pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
                    result = ReadLogReadActionError;
                    break;
                }
                if ( Action.GetRedoStopSequence() // PTS 1111 UH 2000-08-10
                     ==
                     afterImage.GetSequence() )
                {
                    // this should be prevented by Log_Transaction,
                    // but for more safety it is checked here again
                    SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "truncate: stopseq: "
                                                           << Action.GetRedoStopSequence()
                                                           << ", nothing to do.");
                    break;
                }
                SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "truncate: inclusive stopseqence: "
                                                       << Action.GetRedoStopSequence());
                // PTS 1113230 UH 2001-12-20 begin
                if ( ! pTransaction->RedoPartialRollback ( Action.GetUndoStopSequence(),
                                                           Action.GetRedoStopSequence(),
                                                           errlist ) )
                {
                    errlist = errlist + Restart_Exception(__CONTEXT__, RST_ERROR_DURING_PARTIAL_ROLLBACK);
                    iter.AppendContextInfo          ( __CONTEXT__, 
                                                      afterImage.GetSequence(),
                                                      afterImage.GetTransNo().gg90GetInt4(),
                                                      errlist );
                    result = ReadLogPartialRollbackError;
                    break;
                }
                // PTS 1113230 UH 2001-12-20 end
                break;
            }
            case Log_SavepointEntry:
                SAPDBERR_ASSERT_STATE( iter.RestLength() == iter.Length() );
                 // nothing to do
                break;
            default:
                if (! StartCopyEntry (*pTransaction, iter, currentCopyState, afterImage, errlist))
                {
                    errlist = errlist + Restart_Exception(__CONTEXT__, RST_ERROR_DURING_COPY_ENTRY);
                    result = ReadLogReadActionError;
                    break;
                }
                break;
            }//switch
        }
        // continuation of a logentry
        else
        {
            if (! ContinueCopyEntry (iter, currentCopyState, afterImage,errlist))
            {
                errlist = errlist + Restart_Exception(__CONTEXT__, RST_ERROR_DURING_COPYCONT_ENTRY);
                result = ReadLogReadActionError;
                break;
            }
        }

        if (   (Log_VolumeIterator::Iter_ok == iter.GetState() )
            && (ReadLogOk == result ))
        {
            ++iter;
        }
    } //endwhile
    iter.Message ( "last-redo-read", readEntryCount,
                             afterImage.GetTransNo().gg90GetInt4(),
                             afterImage.GetSequence() );
    // PTS 1130336 mb 2004-07-15 multi-log-queue
    // PTS 1133003 always release pending entries
    if ( ! ReleaseAllPendingEntries (iterState, errlist) )
    {
        return ReadLogReleaseRedoEntryFailed;
    }
	return result;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::CloseOpenTransactions (tsp00_TaskId            taskid,
                                              SAPDBMem_IRawAllocator &allocator,
                                              Log_IOSequenceNo        lastIOSequence,
                                              SAPDBErr_MessageList   &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::CloseOpenTransactions", LogTrans_Trace, 5);
    
    tgg00_TransIndex  openTransIndex = cgg_nil_transindex;
    Log_Transaction  *pTransaction   = 0;

    if ( lastIOSequence.IsInvalid() )
    {
        // PTS 1124684 UH 2003-10-16
        lastIOSequence = Rst_IRedoTrafficControl::Instance().GetLastIOSequence(taskid);
        if ( lastIOSequence.IsInvalid() )
            lastIOSequence  = 0;
    }
    
    kb51GetFirstOpenRedoTrans (taskid, openTransIndex);

    while ( openTransIndex != cgg_nil_transindex )
    {
        pTransaction = reinterpret_cast<Log_Transaction*>(k50GetLogTransaction (openTransIndex));
        if ( 0 == pTransaction )
        {
            errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                         "k50GetLogTransaction() returned nothing");
            return false;
        }

        if ( pTransaction->GetRedoEotSequence().IsValid() )
        {
            errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                         "eot sequence is valid although forced to be rollbacked");
            return false;
        }

        if ( ! EndOfTransaction ( taskid,
                                  allocator,
                                  pTransaction,
                                  Log_Transaction::rollbacked,
                                  lastIOSequence,
                                  openTransIndex,
                                  ! CHECK_ASCENDING_EOT_SEQUENCE, // PTS 1113641 UH 2002-01-25
                                  errlist ) )
        {
            return false;
        }

        // force search at the beginning of the transactions list
        // until there exists no more open transaction
        openTransIndex = cgg_nil_transindex;
        kb51GetFirstOpenRedoTrans (taskid, openTransIndex);
    }//while
    // PTS 1129746 UH 2004-06-18 added wait for all transactions which end at this time
    Rst_IRedoTrafficControl::Instance().WaitForAllRedoTasks(taskid);
    return true;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::ReCreateOpenTransactions (tsp00_TaskId            taskid,
                                                 tgg00_TransContext     &oldcontext,
                                                 SAPDBMem_IRawAllocator &allocator,
                                                 SAPDBErr_MessageList   &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::ReCreateOpenTransactions", LogTrans_Trace, 5);

    tkb00_SaveptParam &saveptParam = k57restartrec->rstLastSavept_kb00();

    Log_OpenTransFile file ( oldcontext, saveptParam.svpOpenTransRoot_kb00 );

    if ( ! file.IsCreated() )
        return true; // nothing to do
    
    errlist.ClearMessageList();
    m_OldestKnownEOTSequence = file.GetOldestKnownIOSequence(errlist); // PTS 1124684 UH 2003-10-16
    if(!errlist.IsEmpty())
        return false;
    if ( LogTrans_Trace.TracesLevel(6) )
        Kernel_VTrace() << "ReCreateOpenTransactions: opentransfile.GetOldestKnownIOSequence() "
                        << m_OldestKnownEOTSequence << NewLine ;
    if (m_OldestKnownEOTSequence.IsInvalid()) 
    // for migration or if no transactions had been open during the savepoint
    {
        if ( saveptParam.svpOpenTransCnt_kb00 > 0 ) // PTS ? 2004-08-04 (opentrans file contains only rollbacked trans)
            m_OldestKnownEOTSequence = 0;
        else
            m_OldestKnownEOTSequence = saveptParam.svpIOsequence_kb00; // PTS 1127009 UH 2004-01-09, PTS 1127704
        if ( LogTrans_Trace.TracesLevel(6) )
            Kernel_VTrace() << "ReCreateOpenTransactions: m_OldestKnownEOTSequence was nil now set to "
                            << m_OldestKnownEOTSequence << NewLine;
    }
    // PTS 1114727 mb 2003-11-04 corrected settting of OldestKnownEOTSequence 
    // PTS 1127009 UH 2004-01-09 update of oldestknown for Log_Volume removed

    errlist.ClearMessageList();
    Log_OpenTransFile::Iterator &iter = file.GetFirstOpenTrans(saveptParam.svpId_kb00,errlist);
    if(!errlist.IsEmpty())
        return false;

    tgg00_TransIndex currentTransIndex = cgg_nil_transindex;

    while ( iter.IsValid() )
    {
        Log_Transaction* newtrans = new (allocator) Log_Transaction ();

        if ( newtrans == NULL )
        {
            errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                         "allocate of Log_Transaction");
            return false;
        }

        newtrans->ReadPersistentFormat( *iter, file.GetSpaceSize() );

        if ( ! newtrans->IsRedoRelevant() )
        {
            // The transaction entry can be deleted with all its empty root pages.
            if ( ! newtrans->InitializeForRedo (oldcontext) )
            {
                errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                             "InitializeForRedo() called for delete");
                return false;
            }
            if(!newtrans->DeleteAfterRedo(errlist))
                return false;
            destroy (newtrans, allocator);
            *iter;
            continue;
        }
        
        if ( ! newtrans->InitializeForRedoCopy (oldcontext,
                                                newtrans->GetRedoTransNo(),
                                                allocator,
                                                errlist) )
        {
            errlist = errlist + Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                                 "InitializeForRedoCopy()");
            return false; 
        }

        if ( ! newtrans->UpdateUndoRedoInfoForRestart(m_OldestKnownEOTSequence, errlist) )
        {
            errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                         "UpdateUndoRedoInfoForRestart()");
            return false;
        }

        tgg00_BasisError error = e_ok;
        
        kb51GetRedoTransIndex (taskid,
                               newtrans->GetRedoTransNo(),
                               currentTransIndex,
                               error);

        if ( error != e_ok )
        {
            Kernel_AppendBasisError(error,errlist,"kb51GetRedoTransIndex");
            if ( e_too_many_lock_requests == error )
            {
                errlist = errlist
                          +
                          Restart_Exception(__CONTEXT__, RST_NO_TRANS_ENTRY_AVAILABLE,
                                            SAPDB_ToString(newtrans->GetRedoTransNo().gg90GetInt4()))
                          +
                          Restart_Exception(__CONTEXT__, RST_NO_TRANS_ENTRY_AVAILABLE);
            }
            return false;
        }
        k50RegisterTransaction (currentTransIndex, newtrans);

        if ( LogTrans_Trace.TracesLevel(6) )
            newtrans->WriteToTrace("OpenTrans");

        if ( newtrans->GetRedoEotSequence().IsValid()
             && // PTS 1114946 UH 2002-03-20
             newtrans->GetTransState() != Log_Transaction::opened )
        {
            // The transaction was collected during savepoint before the commit/rollback was written.
            // Between setting TransState and writing eot logentry may the trans been collected by savepoint.
            // Then the trans has a TransState which says that the transaction has ended, but the eot
            // logentry must be found after the savepoint entry.
            // This assumes a fixed sequence of operations in the Log_Savepoint::Coordinator().
            // First Savepoint Entry writing, then transactions collecting.

            if ( ! EndOfTransaction ( taskid,
                               allocator,
                               newtrans,
                               newtrans->GetTransState(),
                               newtrans->GetRedoEotSequence(),
                               currentTransIndex,
                               ! CHECK_ASCENDING_EOT_SEQUENCE, // PTS 1113641 UH 2002-01-25
                               errlist ) )
            {
                return false;
            }
        }

        if(!iter.PlusPlus(errlist))
        {
            return false;
        }
    }
	
    // PTS 1124684 UH 2003-10-16 added the following two lines
    if ( LogTrans_Trace.TracesLevel(6) )
        Rst_IRedoTrafficControl::Instance().
            WriteToTrace("End of Rst_RedoReadTask::ReCreateOpenTransactions()");
    Rst_IRedoTrafficControl::Instance().EnableRedoProcessing(taskid);
    return true;
}
/* --------------------------------------------------------------------------- */
bool
Rst_RedoReadTask::GetOrCreateRedoTransaction ( tsp00_TaskId             taskid,
                                               SAPDBMem_IRawAllocator  &allocator,
                                               tgg00_TransContext      &oldcontext,
                                               const tgg91_TransNo     &redoTransNo,
                                               const Log_EntrySequence &entrySequence,
                                               tgg00_TransIndex        &currTransIndex,
                                               Log_Transaction*        &redoTrans,
                                               SAPDBErr_MessageList    &errlist )
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::GetOrCreateRedoTransaction", LogTrans_Trace, 5);

    redoTrans      = 0;
    currTransIndex = cgg_nil_transindex;
    
    Rst_IRedoTrafficControl::Instance().CheckJobCount(taskid); // PTS ? 2005-01-03 added to reserve some transentries

//    // PTS ??? UH 2004-12-10
//    const bool transactionMustExistOrEntryMustBeSkipped = entrySequence > LOG_FIRST_ENTRY_SEQUENCE;

    kb51GetRedoTransIndex ( taskid,
                            redoTransNo,
//                            transactionMustExistOrEntryMustBeSkipped ? 1 : 0, // PTS ??? UH 2004-12-10
                            currTransIndex,
                            oldcontext.trError_gg00 );

//    if ( transactionMustExistOrEntryMustBeSkipped
//         &&
//         currTransIndex == cgg_nil_transindex )
//	{
//        // PTS ??? UH 2004-12-10
//        // This is a redo of a redo and this entry is read again, but the 
//        return skipEntry;
//	}

    if ( oldcontext.trError_gg00 != e_ok
         ||
         currTransIndex == cgg_nil_transindex )
    {
        errlist = Restart_Exception(__CONTEXT__, RST_REDO_READ_GET_TRANSENTRY_FAILED,
                                    SAPDB_ToString(Kernel_BasisErrorToResultCode(oldcontext.trError_gg00)),
                                    Kernel_BasisErrorToResultString(oldcontext.trError_gg00),
                                    SAPDB_ToString(currTransIndex),
                                    SAPDB_ToString(redoTransNo.gg90GetInt4()));
        return false;
    }

    redoTrans = reinterpret_cast<Log_Transaction*>(k50GetLogTransaction(currTransIndex));

    if ( 0 == redoTrans )
    {
        redoTrans = new (allocator) Log_Transaction ();
        if ( 0 == redoTrans )
        {
            errlist = Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                         "Could not allocate memory for new redo transaction.");
            return false;
        }

        k50RegisterTransaction (currTransIndex, redoTrans);

        if ( ! redoTrans->InitializeForRedoCopy(oldcontext,redoTransNo,allocator,errlist) )
        {
            errlist = errlist + Restart_Exception(__CONTEXT__, SAPDBERR_ASSERT_STATE_FAILED,
                                                "Could not allocate memory for redo transaction. (InitializeForRedoCopy)");
            return false;
        }

        kb57RedoTransNoIncrement (taskid, redoTransNo);
    }

    return true;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::StartCopyEntry (Log_Transaction        &transaction, // PTS 1115065 UH 2002-04-10 return bool
                                       Rst_LogEntryIterator   &iter,
                                       Rst_CopyState          &copyState,
                                       Log_AfterImage         &afterimage,
                                       SAPDBErr_MessageList   &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::StartCopyEntry", LogTrans_Trace, 5);
    SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "StartCopyEntry: length:" << iter.RestLength() <<
                                           ", minimal: " << Log_ActionGetMinCopyLength(afterimage.GetActionType()));

    // If entryAlreadyCopied is true then nothing is to do - this is a redo of a redo.
    // This entry was already copied into redo file.
    const bool entryAlreadyCopied = transaction.GetLastCopiedRedoEntrySequence().IsValid()
                                    &&
                                    afterimage.GetSequence() <= transaction.GetLastCopiedRedoEntrySequence();
        
    // PTS 1131107 UH 2004-08-24 added handling for Log_MultipleActions
    const bool isMultipleAction = Log_MultipleActions == afterimage.GetActionType();

	errlist.ClearMessageList();

    Data_SplitSpaceWriter writer = entryAlreadyCopied
                                ? Data_SplitSpaceWriter()   // PTS 1127732 UH 2004-02-11 use dummy writer
                                : isMultipleAction
                                  ? transaction.GetWriterForLocalRedoEntries
                                          (iter.GetValuePtr() + Log_AfterImage::GetPersistentLength(),
                                           iter.RestLength() - Log_AfterImage::GetPersistentLength(),
                                           errlist)
                                  : transaction.ReserveSpaceForRedoEntryCopy
                                            ( iter.RestLength(),
                                              Log_ActionGetMinCopyLength (afterimage.GetActionType(), afterimage.GetPersistentVersion())
                                              +
                                              Log_AfterImage::GetPersistentLength(),
                                              afterimage.GetSequence(),
                                              iter.IOSeqNo(),
                                              errlist );
    if ( ! errlist.IsEmpty() )
	{
        iter.AppendContextInfo (__CONTEXT__,
                                transaction.GetRedoTransNo().gg90GetInt4(),
                                afterimage.GetSequence(),
                                errlist);
        return false;
    }
    
    Data_SplitSpaceWriter::Result writeResult      = Data_SplitSpaceWriter::ok;

    if ( ! entryAlreadyCopied ) // PTS 1127732 UH 2004-02-11
    {
        SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "StartCopyEntry: write:" << iter.Length());

        if ( ! writer.IsAssigned() )
        {
            const SAPDB_UInt reserveLength = Log_AfterImage::GetPersistentLength()
                                             +
                                             Log_ActionGetMinCopyLength(afterimage.GetActionType());
            errlist = Restart_Exception(__CONTEXT__, RST_COPY_ENTRY_RESERVE_FAILED,
                                        SAPDB_ToString(reserveLength));
            Kernel_VTrace() << "ReserveSpaceForRedoEntryCopy failed, Writer is not assigned.";
            iter.WriteToTrace("iterator");
            writer.WriteToTrace("writer");
            iter.AppendContextInfo (__CONTEXT__,
                                    transaction.GetRedoTransNo().gg90GetInt4(),
                                    afterimage.GetSequence(),
                                    errlist);
            transaction.AppendContextInfo ( __CONTEXT__, errlist );
            return false;
        }


        if (isMultipleAction)
        {
            const SAPDB_UInt firstTimeSkipLength = Log_AfterImage::GetHeaderSize()
                                                   +
                                                   Log_ActionMultiple::GetHeaderSize();
            if ( firstTimeSkipLength < iter.Length() )
            {
                // copy the entry without the header of the multiple entry
                SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "StartCopyEntry: multiple write: length = " << iter.Length() << " - " << firstTimeSkipLength);
                writeResult = writer.Write ( iter.GetValuePtr() + firstTimeSkipLength,
                                             iter.Length()      - firstTimeSkipLength );
            }
            else
            {
                // only the header of the multiple entry is in this part and the
                // header needs not to be copied
                SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "CopyEntry: NO WRITE firstTimeSkipLength:" \
                                                        << firstTimeSkipLength << " < iter.Length():" \
                                                        << iter.Length());
                writeResult = Data_SplitSpaceWriter::moreSpaceAvailable;
            }
        }
        else
        {
            SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "StartCopyEntry: simple write: length = " << iter.Length());
            writeResult = writer.Write (iter.GetValuePtr(), iter.Length());
        }
        
        if ( writeResult != Data_SplitSpaceWriter::moreSpaceAvailable
             &&
             writeResult != Data_SplitSpaceWriter::ok )
        {
            Kernel_VTrace() << "writeResult: " << writeResult;
            iter.WriteToTrace("iterator");
            writer.WriteToTrace("writer");
            errlist = Restart_Exception(__CONTEXT__, RST_WRITE_FAILED,
                                        SAPDB_ToString(writeResult));
            iter.AppendContextInfo (__CONTEXT__,
                                    transaction.GetRedoTransNo().gg90GetInt4(),
                                    afterimage.GetSequence(),
                                    errlist);
            writer.AppendContextInfo(__CONTEXT__,errlist);
            transaction.AppendContextInfo(__CONTEXT__,errlist);
            return false;
        }
    }
    
    const bool entryIsNotCompleteCopied = ((iter.RestLength() > iter.Length())
                                            &&
                                           (writeResult == Data_SplitSpaceWriter::moreSpaceAvailable));
    if (entryIsNotCompleteCopied)
    {
        copyState.SetCopyPending(&transaction,
                                  writer,
                                  entryAlreadyCopied,
                                  isMultipleAction);
    }
    else
    {
        return copyState.MaterializeEntryInRedoFile(iter.IOSeqNo(), transaction, writer, 
                                                    entryAlreadyCopied, isMultipleAction, errlist);
    }
    return true;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::ContinueCopyEntry (Rst_LogEntryIterator   &iter,
                                          Rst_CopyState          &copyState,
                                          Log_AfterImage         &afterimage,
                                          SAPDBErr_MessageList   &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::ContinueCopyEntry", LogTrans_Trace, 5);

    // If copyState.AlreadyCopied() is true then nothing is to do - this is a redo of a redo.
    // This entry was already copied into redo file.

    SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "ContinueCopyEntry: length:" << iter.RestLength() <<
                                           ", minimal: " << Log_ActionGetMinCopyLength(afterimage.GetActionType()));

    errlist.ClearMessageList();

    Data_SplitSpaceWriter         &writer              = copyState.GetWriter();
    Data_SplitSpaceWriter::Result  writeResult         = Data_SplitSpaceWriter::ok;
    bool                           morePartsExpected   = false;
    
    if ( copyState.AlreadyCopied() ) 
    {
        morePartsExpected = iter.RestLength() > iter.Length();
    }
    else
    {
        SAPDBTRACE_WRITELN (LogTrans_Trace, 7, "ContinueCopyEntry: write:" << iter.Length());

        writeResult = writer.Write (iter.GetValuePtr(), iter.Length());
        
        if ( writeResult != Data_SplitSpaceWriter::moreSpaceAvailable
             &&
             writeResult != Data_SplitSpaceWriter::ok )
        {
            Kernel_VTrace() << "writeResult: " << writeResult;
            iter.WriteToTrace("iterator");
            writer.WriteToTrace("writer");
            errlist = Restart_Exception(__CONTEXT__, RST_WRITE_FAILED,
                                        SAPDB_ToString(writeResult));
            iter.AppendContextInfo (__CONTEXT__,
                                    copyState.GetTransaction().GetRedoTransNo().gg90GetInt4(),
                                    afterimage.GetSequence(),
                                    errlist);
            writer.AppendContextInfo(__CONTEXT__,errlist);
            copyState.GetTransaction().AppendContextInfo(__CONTEXT__,errlist);
            return false;
        }
        morePartsExpected = writeResult == Data_SplitSpaceWriter::moreSpaceAvailable
                            &&
                            iter.RestLength() > iter.Length();
    }

    if ( ! morePartsExpected )
    {
        return copyState.MaterializeEntryInRedoFile(iter.IOSeqNo(), errlist);
    }
    
    return true;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::EndOfTransaction
        ( tsp00_TaskId                 redoReaderTaskid,
          SAPDBMem_IRawAllocator      &allocator,
          Log_Transaction*            &pTransaction,
          Log_Transaction::TransState  state,
          Log_IOSequenceNo             eotSequence,
          tgg00_TransIndex             transindex,
          bool                         checkAscendingSequence, // PTS 1113641 UH 2002-01-25
          SAPDBErr_MessageList        &errlist )
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::EndOfTransaction", LogTrans_Trace, 5);

    pTransaction->SetEOTInRedo ( state, eotSequence );

    if ( LogTrans_Trace.TracesLevel(7) )
        pTransaction->WriteToTrace ("EndOfRedoTrans");

    // PTS 1114946 UH 2002-03-21 begin
    bool             transIsAlreadyInRedoList = false;
    Log_Transaction* pPublicTrans = reinterpret_cast<Log_Transaction*>(k50GetLogTransaction(transindex));

    if ( 0 == pPublicTrans )
    {
        errlist = Restart_Exception(__CONTEXT__, RST_EOT_HANDLING_NO_TRANSENTRY);
        pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
        return false;
    }

    if ( pPublicTrans != pTransaction )
    {
        // there is another log transcontext registered in the public transaction entry in locklist
        // Check if this correct.
        if ( pTransaction->GetTransState()      != pPublicTrans->GetTransState()
             ||
             pTransaction->GetRedoEotSequence() != pPublicTrans->GetRedoEotSequence() )
        {
            errlist = Restart_Exception(__CONTEXT__, RST_EOT_HANDLING_WRONG_CONTEXT);
            pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
            pPublicTrans->AppendContextInfo ( __CONTEXT__, errlist );
            return false;
        }
        transIsAlreadyInRedoList = true;
    }
    else
    {
        // 2004-02-18 UH added
        // This can happen if trans was open in last redo savepoint and last entry read was the commit.
        // So ReCreate has already inserted the transaction into the redo list and after that
        // the redo reader finds the commit.
        // In this case the transaction need not to be inserted.
        transIsAlreadyInRedoList =
            pPublicTrans->GetTransState() == Log_Transaction::illegalTransType;
    }
    // PTS 1114946 UH 2002-03-21 end
    
    bool           isOK    = true;
    Log_ActionType eottype = state ==  Log_Transaction::committed
                                       ? Log_Commit
                                       : Log_Rollback; // <- default

    if ( (eottype == Log_Commit
          &&
          ! pTransaction->EntriesToRedoExist())
         ||
         (eottype == Log_Rollback
          &&
          ! pTransaction->WasOpenInLastSavepoint()) )
    {
        tgg00_TransState dummyState;
        dummyState.clear();
        k51remove_locks ( redoReaderTaskid, transindex, dummyState );
        k51del_transentry_from_locklist ( redoReaderTaskid, transindex,  pTransaction->GetOldTransContext());
        if(!pTransaction->Delete(errlist))
            return false;
    }
    else
    {
        if ( ! pTransaction->DeleteAfterRedoCopy(errlist))
        {
            return false;
        }

        if ( ! transIsAlreadyInRedoList ) // PTS 1114946 UH 2002-03-21
        {
            if ( ! Rst_IRedoTrafficControl::Instance().
                        InsertJob ( redoReaderTaskid,
                                    transindex, eotSequence, eottype,
                                    m_OldestKnownEOTSequence, // PTS 1124684 UH 2003-10-16
                                    checkAscendingSequence,   // PTS 1113641 UH 2002-01-25
                                    *pTransaction,
                                    errlist ) )
            {
                errlist = errlist + Restart_Exception(__CONTEXT__, RST_EOT_HANDLING_INSERT_JOB_FAILED);
                pTransaction->AppendContextInfo ( __CONTEXT__, errlist );
                return false;
            }
        }
    }

    allocator.Deallocate(pTransaction);
    pTransaction = NULL;
    return true;
}
/* --------------------------------------------------------------------------- */
bool Rst_RedoReadTask::ReleaseAllPendingEntries ( Container_Vector<Rst_CopyState>  &iterState, 
                                                  SAPDBErr_MessageList             &errlist)
{
    SAPDBTRACE_METHOD_DEBUG ("Rst_RedoReadTask::ReleaseAllPendingEntries ", LogTrans_Trace, 5);

    for ( unsigned int i = 0; i < iterState.GetSize() ; i++)
    {
        if ( iterState[i].IsCopyPending() )
        {
            if (!iterState[i].ReleasePendingEntry(errlist))
            {
                errlist = errlist + Restart_Exception(__CONTEXT__,
                                                      RST_ERROR_DURING_RELEASE_SPACE );
                return false;
            }
        }
    }
	return true;
}

