/*!*********************************************************************

    @file         RTESec_Authentication.cpp

    @author       DanielD

    @brief

\if EMIT_LICENCE
    ========== licence begin  GPL
    Copyright (c) 2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
\endif
***********************************************************************/

#include "RunTime/Security/RTESec_Authentication.hpp"

#include <string.h>

#include "RunTime/RTE_Random.hpp"
#include "RunTime/Security/RTESec_SCRAMMD5.h"
#include "RunTime/Security/RTESec_DumpHex.hpp"

/*----------------------------------------*/

static const char *
stripUnicodeBlanks (
    const char * password,
    const char * passwordEnd,
    char c0,
    char c1)
{
    while ((passwordEnd > password)
           && (passwordEnd [-1] == c0) && (passwordEnd [-2] == c1))
    {
        passwordEnd -= 2;
    }
    return passwordEnd;
}

/*----------------------------------------*/

static const char *
stripAsciiBlanks (
    const char * password,
    const char * passwordEnd)
{
    while ((passwordEnd > password) && (passwordEnd [-1] == ' '))
    {
        --passwordEnd;
    }
    return passwordEnd;
}

/*----------------------------------------*/

static int
stripTrailingBlanks (
    const void * password_in,
    int pwdlen)
{
    const char * password = (const char *) password_in;
    const char * passwordEnd = password + pwdlen;
    const char * strippedEnd;

    strippedEnd = stripUnicodeBlanks (password, passwordEnd, '\0', ' ');
    if (strippedEnd != passwordEnd) {
        return strippedEnd - password;
    }
    strippedEnd = stripUnicodeBlanks (password, passwordEnd, ' ', '\0');
    if (strippedEnd != passwordEnd) {
        return strippedEnd - password;
    }
    strippedEnd = stripAsciiBlanks (password, passwordEnd);
    return strippedEnd - password;
}

/*----------------------------------------*/

RTESec_Authentication::
RTESec_Authentication (
    const char * name_init)
: name (name_init)
{
}

/*----------------------------------------*/

RTESec_Authentication::
~RTESec_Authentication ()
{
}

/*----------------------------------------*/

const char * RTESec_Authentication::
Name () const
{
    return this->name;
}

/*----------------------------------------*/

/*!

  @class RTESec_ScramMD5

 */
class RTESec_ScramMD5
    : public RTESec_Authentication
{
public:
    RTESec_ScramMD5 (const char * name_init);

    virtual bool DataForChallengeRequest (RTESec_ClientAuthenticationBuffer & clientBuffer,
                                const char *  username, int namelen,
                                const void *& requestData, int & outlen) const;

    virtual bool CreateChallenge (
        RTESec_ServerAuthenticationBuffer & serverBuffer,
        const void *& challenge, int & challengeLen,
        const void * seed, int seedLen) const;

    virtual bool CreateResponse (
        RTESec_ClientAuthenticationBuffer & clientBuffer,
        const void *& response, int & responseLen,
        const void * serverChallenge, int serverChLen,
        const char * password, int passwordLen) const;

    virtual bool         VerifyResponse (
        RTESec_ServerAuthenticationBuffer & serverBuffer,
        const void * clientResponse, int clientResponseLen,
        const void * password, int pwdlen) const;
};

/*----------------------------------------*/

typedef struct ScramMD5_ClientBufferT {
    char            requestData [64];
    SCRAMMD5_CLIENT responseData;
    unsigned char   serverproof [256];
} ScramMD5_ClientBufferT;


/*----------------------------------------*/

typedef struct ScramMD5_ServerBufferT {
    char            requestData [64];
    char            challenge [40];
    SCRAMMD5_CLIENT responseData;
    unsigned char   serverproof [256];
} ScramMD5_ServerBufferT;


/*----------------------------------------*/

RTESec_ScramMD5::
RTESec_ScramMD5 (
    const char * name_init)
: RTESec_Authentication (name_init)
{
}

/*----------------------------------------*/

bool RTESec_ScramMD5::
DataForChallengeRequest (
    RTESec_ClientAuthenticationBuffer & clientBuffer,
    const char *  username,
    int namelen,
    const void *& requestData,
    int & outlen) const
{
    ScramMD5_ClientBufferT * clientBuf = reinterpret_cast <ScramMD5_ClientBufferT*> (&clientBuffer);
    if (namelen > sizeof (clientBuf->requestData)) {
        namelen = sizeof (clientBuf->requestData);
    }
    memcpy (clientBuf->requestData, username, namelen);
    RTE_Random::Generator.fillBuf (clientBuf->requestData + namelen,
        sizeof (clientBuf->requestData) - namelen);
    requestData = clientBuf->requestData;
    outlen = sizeof (clientBuf->requestData);
    return true;
}

/*----------------------------------------*/

bool RTESec_ScramMD5::
CreateChallenge (
    RTESec_ServerAuthenticationBuffer & serverBuffer,
    const void *& challenge, int & challengeLen,
    const void * seed, int seedLen) const
{
    ScramMD5_ServerBufferT * serverBuf = reinterpret_cast<ScramMD5_ServerBufferT*> (& serverBuffer);

    /*
     * memorize client data
     */
    if (seedLen > sizeof (serverBuf->requestData)) {
        seedLen = sizeof (serverBuf->requestData);
    }
    memcpy (serverBuf->requestData, seed, seedLen);
    /*
     * set salt
     */
    char * target = serverBuf->challenge;
    RTE_Random::Generator.fillBuf (target, SCRAMMD5_SALTSIZE);
    /*
     * set key
     */
    target += SCRAMMD5_SALTSIZE;
    RTE_Random::Generator.fillBuf (target, sizeof (serverBuf->challenge) - SCRAMMD5_SALTSIZE);
    /*
     * return
     */
    challenge = serverBuf->challenge;
    challengeLen = sizeof (serverBuf->challenge);
    return true;
}

/*----------------------------------------*/

bool RTESec_ScramMD5::
CreateResponse (
    RTESec_ClientAuthenticationBuffer & clientBuffer,
    const void *& response,
    int & responseLen,
    const void * serverChallenge,
    int serverChLen,
    const char * password,
    int passwordLen) const

{
    ScramMD5_ClientBufferT * clientBuf = reinterpret_cast<ScramMD5_ClientBufferT*> (&clientBuffer);
    passwordLen = stripTrailingBlanks (password, passwordLen);
    int rc = SCRAMMD5GenClient (&clientBuf->responseData,
        clientBuf->requestData, sizeof (clientBuf->requestData),
        (const char *) serverChallenge, serverChLen,
        password, passwordLen,
        clientBuf->serverproof);
    if (rc != 0) {
        return false;
    }
    response = &clientBuf->responseData;
    responseLen = SCRAMMD5_DATASIZE;
    return true;
}

/*----------------------------------------*/

bool RTESec_ScramMD5::
VerifyResponse (
    RTESec_ServerAuthenticationBuffer & serverBuffer,
    const void * clientResponse,
    int clientResponseLen,
    const void * password,
    int pwdlen) const
{
    ScramMD5_ServerBufferT * serverBuf = reinterpret_cast<ScramMD5_ServerBufferT*> (&serverBuffer);
    SCRAMMD5_VRFY verifier;

    if (clientResponseLen != sizeof (SCRAMMD5_CLIENT)) {
        return false;
    }
    pwdlen = stripTrailingBlanks (password, pwdlen);
    const unsigned char * salt = (const unsigned char *) serverBuf->challenge;
    SCRAMMD5GenVerifier (&verifier, salt, (const char*) password, pwdlen, 0);
    int rc = SCRAMMD5Gen ((SCRAMMD5_CLIENT*) clientResponse,
        serverBuf->requestData, sizeof (serverBuf->requestData),
        serverBuf->challenge, sizeof (serverBuf->challenge),
        (const char * )&verifier, sizeof (SCRAMMD5_VRFY), SCRAM_VERIFY,
        serverBuf->serverproof);
    return rc == 0;
}

/*----------------------------------------*/

static const RTESec_ScramMD5 scramMD5 ("SCRAMMD5");

/*----------------------------------------*/
#if 0
/*!

  @class RTESec_PlainText

 */
class RTESec_PlainText
    : public RTESec_Authentication
{
public:
    RTESec_PlainText (const char * name_init);
    virtual bool DataForChallengeRequest (RTESec_ClientAuthenticationBuffer & clientBuffer,
                                const void *& requestData, int & outlen) const;
    virtual bool CreateChallenge (void * challenge,
                                int challengeSize, int & challengeLen,
                                const void * seed, int seedLen) const;
    virtual bool CreateResponse (void * response, int responseSize, int & responseLen,
                                const void * clientChallenge, int clientChLen,
                                const void * serverChallenge, int serverChLen,
                                const char * password, int passwordLen,
                                void * serverProof) const;
};

/*----------------------------------------*/

RTESec_PlainText::
RTESec_PlainText (
    const char * name_init)
: RTESec_Authentication (name_init)
{
}

/*----------------------------------------*/

bool RTESec_PlainText::
DataForChallengeRequest (
    RTESec_ClientAuthenticationBuffer & clientBuffer,
    const void *& requestData,
    int & outlen) const
{
    static char * dummyData = "";
    requestData = dummyData;
    outlen = 0;
    return true;
}

/*----------------------------------------*/

bool RTESec_PlainText::
CreateChallenge (
    void * challenge,
    int    challengeSize,
    int  & challengeLen,
    const void * seed,
    int    seedLen) const
{
    challengeLen = 0;
    return true;
}

/*----------------------------------------*/

bool RTESec_PlainText::
CreateResponse (
    void * response,
    int responseSize,
    int & responseLen,
    const void * clientChallenge,
    int clientChLen,
    const void * serverChallenge,
    int serverChLen,
    const char * password,
    int passwordLen,
    void * serverProof) const
{
    return false;
}

/*----------------------------------------*/

static const RTESec_PlainText plainText ("PLAINTEXT");

#endif

/*----------------------------------------*/

static const RTESec_Authentication * allAuthentications [] = {
    &scramMD5,
    // &plainText,
    0
};

/*----------------------------------------*/

const RTESec_Authentication * RTESec_Authentication::
FindByName (
    const char * name,
    int len)
{
    if (len == -1) {
        len = strlen (name);
    }
    RTESec_Authentication::Iterator iter;
    const RTESec_Authentication * current;
    while (iter.next (current)) {
        const char * currentName = current->Name ();
        if ((len == strlen (currentName)) && (memcmp (name, currentName, len) == 0) ) {
            return current;
        }
    }
    return 0;
}

/*----------------------------------------*/

RTESec_Authentication::Iterator::
Iterator ()
{
    this->index = 0;
}

/*----------------------------------------*/

bool RTESec_Authentication::Iterator::
next (
    const RTESec_Authentication * & current)
{
    if (allAuthentications [this->index] == 0) {
        return false;
    }
    current = allAuthentications [this->index];
    ++this->index;
    return true;
}

/*----------------------------------------*/

extern "C" const char *
RTESec_Authentication_Name (
    const RTESec_Authentication * authentication)
{
    return authentication->Name ();
}

/*----------------------------------------*/

extern "C" bool
RTESec_Authentication_DataForChallengeRequest (
    const RTESec_Authentication * authentication,
    RTESec_ClientAuthenticationBuffer * clientBuffer,
    const char  * username,
    int           namelen,
    const void ** buf,
    int  * outlen)
{
    return authentication->DataForChallengeRequest (*clientBuffer,
        username, namelen,
        *buf, *outlen);
}

/*----------------------------------------*/

extern "C" bool
RTESec_Authentication_CreateResponse (
    const RTESec_Authentication * authentication,
    RTESec_ClientAuthenticationBuffer * clientBuffer,
    const void ** response,
    int * responseLen,
    const void * serverChallenge,
    int serverChLen,
    const char * password,
    int passwordLen)
{
    return authentication->CreateResponse (
        *clientBuffer, *response, *responseLen,
        serverChallenge, serverChLen,
        password, passwordLen);
}


/*----------------------------------------*/

extern "C" bool
RTESec_Authentication_VerifyResponse (
    const RTESec_Authentication * authentication,
    RTESec_ServerAuthenticationBuffer * serverBuffer,
    const void * clientResponse,
    int clientResponseLen,
    const void * password,
    int pwdlen)
{
    return authentication->VerifyResponse (
        *serverBuffer,
        clientResponse, clientResponseLen,
        password, pwdlen);
}

/*----------------------------------------*/

extern "C" const RTESec_Authentication *
RTESec_Authentication_FindByName (const char * name, int len)
{
    return RTESec_Authentication::FindByName (name, len);
}

/*----------------------------------------*/

extern "C" int
RTESec_Authentication_Next (
    const RTESec_Authentication ** current,
    int * index)
{
    if (allAuthentications [*index] == NULL) {
        return 0;
    }
    *current = allAuthentications [*index];
    *index += 1;
    return 1;
}

/*----------------------------------------*/

#if defined (RTESec_DumpHex)

/* has been defined as an empty macro */

#else

/*
 This is an array of lines to be overwritten in cycles.
 That way, a debugger can show a bit of history.
 */
static char dumpLines [20] [300];
static int currentLine = 0;
static char * dumpLine;

void
RTESec_DumpHex (const char * comment,
    const void * buf_in,
    int          buflen)
{
    static const char * hexdigits = "0123456789abcdef";
    const unsigned char * buf = (const unsigned char *) buf_in;
    dumpLine = dumpLines [currentLine];
    sprintf (dumpLine, "%s (%d):\n", comment, buflen);
    unsigned char * target = (unsigned char *) dumpLine + strlen (dumpLine);
    for (int i = 0; i < buflen; ++i) {
        if (i % 16 == 0) {
            target [0] = ' ';
            ++target;
        }
        unsigned char current = buf [i];
        target [0] = hexdigits [current / 16];
        target [1] = hexdigits [current % 16];
        target += 2;
    }
    target [0] = '\0';
    fprintf (stdout, "%s\n", dumpLine);
    fflush (stdout);
    ++currentLine;
    if (currentLine >= 20) {
        currentLine = 0;
    }
}

#endif

/*----------------------------------------*/


