/*!
  @file           RTETask_WorkerJob.cpp
  @author         Raymond
  @ingroup        Runtime
  @brief          Scheduling

  System Independend Worker Thread For Time Consuming Or Blocking Task Work
  Needed i.e. during initialization of HotStandby 

\if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 2001-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


\endif
*/
#include "RunTime/Tasking/RTETask_IWorkerJob.hpp"
#include "RunTime/Tasking/RTETask_WorkerJob.hpp"
#include "RunTime/System/RTESys_AtomicOperation.hpp"
#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_MessageList.hpp"
#define KERNEL_LZU
#include "geo50_0.h"
#include "heo56.h"

#define WORKER_COMMAND_THRESHOLD       2

class RTETask_WorkerPool
{
public:
    RTETask_WorkerPool() : m_NumOfWorker(0), m_pWorkerPool(0) {}

    static RTETask_WorkerPool&  Instance();

    void AddJob( RTETask_IWorkerJob &job );

    SAPDB_UInt4 CreateAllWorker ( SAPDB_UInt4            numOfWorker,
                                  SAPDB_UInt4            workerStackSize,
                                  SAPDBErr_MessageList   &messageList );

private:
    RTETask_Worker**           m_pWorkerPool;
    SAPDB_UInt4                m_NumOfWorker;
    static RTETask_WorkerPool* m_pInstance;
};

//--------------------------------------------------------

RTETask_WorkerPool & RTETask_WorkerPool::Instance()
{
  SAPDBMEM_STATIC_RAW_ALLOCATE(Space, sizeof(RTETask_WorkerPool));

  if ( m_pInstance == NULL )
    m_pInstance = new (Space) RTETask_WorkerPool();

  return *m_pInstance;
}

RTETask_WorkerPool * RTETask_WorkerPool::m_pInstance = NULL;


//-----------------------------------------------------------------

SAPDB_UInt4 RTETask_WorkerPool::CreateAllWorker ( SAPDB_UInt4            numOfWorker,
                                                  SAPDB_UInt4            workerStackSize,
                                                  SAPDBErr_MessageList   &messageList )
{
    m_NumOfWorker = numOfWorker;

    m_pWorkerPool = (RTETask_Worker**)RTEMem_RteAllocator::Instance().Allocate(numOfWorker * sizeof(RTETask_Worker*));

    if ( 0 == m_pWorkerPool )
    {
        messageList = SAPDBErr_MessageList( RTE_CONTEXT, RTEERR_MALLOC_FAILED_FOR, "TaskWorkerPool" );
        return false;
    }

    SAPDB_UInt4 worker;

    for ( worker = 0; worker < numOfWorker; worker++ )
    {
        m_pWorkerPool[worker] = RTETask_Worker::BeginThread( workerStackSize, messageList );

        if ( 0 == m_pWorkerPool[worker] )
        {
            m_NumOfWorker = worker;
            break;
        }
    }

    return worker;
}

//-----------------------------------------------------------------

void RTETask_WorkerPool::AddJob( RTETask_IWorkerJob &job )
{
    SAPDB_UInt4 worker = 0;

    if ( m_pWorkerPool[worker]->CommandsPending() >= WORKER_COMMAND_THRESHOLD )
    {
        // Find shortest command queue
        for ( SAPDB_UInt4 workerIdx = 1; workerIdx < m_NumOfWorker; workerIdx++ )
        {
            if ( m_pWorkerPool[workerIdx]->CommandsPending() < 
                 m_pWorkerPool[workerIdx-1]->CommandsPending() )
            {
                worker = workerIdx;

                if ( m_pWorkerPool[worker]->CommandsPending() < WORKER_COMMAND_THRESHOLD )
                   break;
            }
        }
    }

    m_pWorkerPool[worker]->AddJob(job);
}


//--------------------------------------------------------

struct WorkerQueueEntry
{
    struct WorkerQueueEntry *pPreviousEntry;
    RTETask_IWorkerJob      &theJob;

    WorkerQueueEntry(RTETask_IWorkerJob &job) : pPreviousEntry(0), theJob(job) {}
};

//--------------------------------------------------------

void RTETask_Worker::AddJob(RTETask_IWorkerJob &job)
{
    WorkerQueueEntry newEntry(job);

    m_NumOfPendingCommands.Increment();

    // enqueue new job and trigger worker thread if not already busy
    //---- loop until we have successfully replaced tail of queue entry
    do
    {
        newEntry.pPreviousEntry = m_pTailOfQueue;

    } while ( !RTESys_CompareAndExchange( m_pTailOfQueue, 
#ifdef OSF1
                                          newEntry.pPreviousEntry,
                                          &newEntry,
#else
                                          const_cast<WorkerQueueEntry * const>(newEntry.pPreviousEntry),
                                          const_cast<WorkerQueueEntry * const>(&newEntry),
#endif
                                          newEntry.pPreviousEntry ) );

    //----- signal new entry
    sqlsignalsem(m_QueueSemaphore);

    // stop until worker thread has done its job
    vstop(job.GetTaskId());
}


//-----------------------------------------------------------------
    
SAPDB_Int4  RTETask_Worker::KernelThreadMain()
{
    for ( ; m_LoopForever ; )
    {
        while ( !m_pTailOfQueue )
        {
            sqlwaitsem(m_QueueSemaphore);
        }

        // first top entry of queue and pointer to pointer of top entry
        WorkerQueueEntry *pEntry        = m_pTailOfQueue;
        WorkerQueueEntry **ppFirstEntry = &m_pTailOfQueue;

        while ( pEntry->pPreviousEntry )
        {
            ppFirstEntry = &(pEntry->pPreviousEntry);
            pEntry       = pEntry->pPreviousEntry;
        }

        // clear pointer to top entry of queue (if only one element, tail of queue pointer is cleared!
        *ppFirstEntry = 0;

        pEntry->theJob.DoJob();

        m_NumOfPendingCommands.Decrement();

        vcontinue(pEntry->theJob.GetTaskId());

    }
    return 0;
}

//-----------------------------------------------------------------

RTETask_Worker* RTETask_Worker::BeginThread( SAPDB_UInt4            stackSize,
                                             SAPDBErr_MessageList   &messageList )
{
    tsp00_ErrTextc           errText;
    teo07_ThreadErr          errCode;
    RTETask_Worker*          pInstance;

    pInstance  = new (RTEMem_RteAllocator::Instance()) RTETask_Worker();

    if ( 0 == pInstance )
    {
        messageList = SAPDBErr_MessageList( RTE_CONTEXT, RTEERR_MALLOC_FAILED_FOR, "RTETask_Worker" );
        return 0;
    }

    sqlcreatesem( &pInstance->m_QueueSemaphore, 0, errText, &errCode);

    if ( THR_OK_EO07 != errCode )
    {
        pInstance->DestroyObject();
        messageList = SAPDBErr_MessageList(RTE_CONTEXT, RTEERR_ALLOCATION_FAILED, "worker thread semaphore");
        return 0;
    }

    if ( pInstance->Create( RTETask_Worker::Normal, stackSize,
                            messageList ) != RTETask_Worker::NoError )
    {
        pInstance->DestroyObject();
        return 0;
    }

    return pInstance;
}

//-----------------------------------------------------------------

//--------- to be removed if ven81.c and vos81k.c are rewritten in C++
extern "C" void RTETask_CWrapperForCreateWorkerThread( SAPDB_Int4 workerStackSize )
{
    SAPDBErr_MessageList   messageList;
    SAPDB_Int4             numOfWorker = 4;
    SAPDB_Int4             numOfWorkerStarted;

    numOfWorkerStarted = RTETask_WorkerPool::Instance().CreateAllWorker ( numOfWorker, workerStackSize, messageList );

    if ( 0 == numOfWorkerStarted )
        RTE_Crash( messageList );
    else if ( numOfWorkerStarted < numOfWorker )
        RTE_Message ( messageList, MSG_DIAG_CONSOLE );
}

//--------------------------------------------------------

extern "C" void RTETask_CallWorker( RTETask_IWorkerJob &job )
{
    RTETask_WorkerPool::Instance().AddJob(job);
}
