/*!========================================================================

  @file         RTETask_WorkerJob.hpp
  @ingroup      
  @author       RaymondR

  @brief        System Independend Task Worker Thread.

                System Independend Worker Thread For Time Consuming Or Blocking Task Work
                Needed i.e. during initialization of HotStandby 

  @since        2005-01-10  18:33
  @sa           

  ==========================================================================

  \if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (C) 2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
  \endif
 ============================================================================*/


#ifndef RTETASK_WORKERJOB_HPP
#define RTETASK_WORKERJOB_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "heo07.h"
#include "RunTime/Threading/RTEThread_KernelThread.hpp"
#include "RunTime/MemoryManagement/RTEMem_RteAllocator.hpp"
#include "RunTime/Synchronisation/RTESync_InterlockedCounter.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/


/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

#if defined(__cplusplus)

class  RTETask_IWorkerJob;
struct WorkerQueueEntry;

/*!---------------------------------------------------------------------------
  @class   RTETask_Worker
  @brief   Task worker thread.

  The worker thread is listening on its worker queue and does blocking calls
  instead of tasks.
 ----------------------------------------------------------------------------*/

class RTETask_Worker : public RTEThread_KernelThread
{
public:
    /*!
    @description    Startup a new task worker

                    This call creates the worker thread, that is listening on its
                    worker queue and does blocking calls instead of tasks. This 
                    routine is called once during kernel initialization, where no 
                    task is running.

    @param          stackSize   [in] - Stack size in bytes.
    @return value   if no error an instance pointer else null
    */
    static RTETask_Worker*  BeginThread( SAPDB_UInt4            stackSize,
                                         SAPDBErr_MessageList   &messageList );

    /*!
    @description    Enqueue worker job and wait for completion

                    This call enqueues the task in a waiting queue and calls 
                    the dispatcher to allow the next task in the task group 
                    to execute. Only if no more tasks are runnable, the task
                    will suspend itself. It will do so by internally calling 
                    a 'vstop()'. The worker thread will issue the corresponding
                    'vcontinue()'. If the worker thread is running, will find 
                    the job in the queue, otherwise it will be awoken while
                    sleeping on its internal semaphore.

                    @param job [in] the job to enqueue
    */
   void AddJob( RTETask_IWorkerJob &job );

    /*!
      @brief          Get number of pending commands
      @return         Number of pending commands
      */
    SAPDB_UInt4    CommandsPending() const  { return m_NumOfPendingCommands.GetValue(); }


private:
    /*!
       @brief          Constructor
     */
    RTETask_Worker() 
         : RTEThread_KernelThread("TaskWorker"), 
           m_pTailOfQueue(0),
           m_LoopForever(true)
    {
        m_NumOfPendingCommands.SetValue(0);
    }

    /*!
    @description    The created thread runs the code of the methode 'ThreadMain'
    */
    virtual SAPDB_Int4  KernelThreadMain();

  /*!
    @description    Destroy object
    */
   virtual void  DestroyObject() { this->~RTETask_Worker();
                                   RTEMem_RteAllocator::Instance().Deallocate(this); }

private:
    teo07_ThreadSemaphore                   m_QueueSemaphore;
    struct WorkerQueueEntry*                m_pTailOfQueue;
    RTESync_InterlockedCounter<SAPDB_UInt4> m_NumOfPendingCommands;
    SAPDB_Bool                              m_LoopForever;
};

#endif

extern "C" void RTETask_CWrapperForCreateWorkerThread( SAPDB_Int4 workerStackSize );

#endif  /* RTETASK_WORKERJOB_HPP */
