/*!*********************************************************************

    @file         SAPDBFields_VarData.cpp

    @author       DanielD

    @brief

\if EMIT_LICENCE
    ========== licence begin  GPL
    Copyright (c) 2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
\endif
***********************************************************************/

#include "SAPDBCommon/Fields/SAPDBFields_VarData.hpp"

#include "vsp001.h"



/*----------------------------------------*/

inline void
writeInt2 (
    unsigned char * target,
    int             value)
{
    target [0] = value / 256;
    target [1] = value % 256;
}

/*----------------------------------------*/

inline int
readInt2 (
    const unsigned char * source)
{
    return source [0] * 256 + source [1];
}

/*----------------------------------------*/

inline bool
lenRequires2Bytes (
    unsigned int len)
{
    return len <= csp1_fi_max_1byte_length;
}

/*----------------------------------------*/

SAPDBFields_VarData::Reader::
Reader (
    const void * buf_init)
{
    unsigned char * buf = (unsigned char *) buf_init;
    this->fieldsRemaining = readInt2 (buf);
    this->readPos = buf + 2;
}

/*----------------------------------------*/

bool SAPDBFields_VarData::Reader::
next (
    const void * & fieldptr,
    int          & fieldlen,
    int          & indicator)
{
    if (this->fieldsRemaining <= 0) {
        return false;
    }
    switch (this->readPos [0]) {
        case csp1_fi_ignored:
        case csp1_fi_special_null:
        case csp1_fi_blob_description:
        case csp1_fi_default_value:
        case csp1_fi_null_value:
            fieldptr = NULL;
            fieldlen = 0;
            indicator = this->readPos [0];
            this->readPos += 1;
            break;
        case csp1_fi_2byte_length:
            fieldptr = readPos + 3;
            fieldlen = readInt2 (this->readPos + 1);
            indicator = 0;
            this->readPos += 3 + fieldlen;
            break;
        default:
            fieldptr = readPos + 1;
            fieldlen = this->readPos [0];
            indicator = 0;
            this->readPos += 1 + fieldlen;
            break;
    }
    --this->fieldsRemaining;
    return true;
}

/*----------------------------------------*/

class Verifier
    : public SAPDBFields_VarData::Reader
{
public:
    Verifier (const void * buf, int buflen);
    int verify ();
private:
    unsigned char * endPos;
};

/*----------------------------------------*/

Verifier::
Verifier (
    const void * buf,
    int buflen)
: SAPDBFields_VarData::Reader (buf)
{
    this->endPos = ((unsigned char*) buf) + buflen;
}

/*----------------------------------------*/

int Verifier::
verify ()
{
    const void * fieldptr;
    int          fieldlen;
    int          indicator;

    while (this->next (fieldptr, fieldlen, indicator)) {
        if (this->readPos > this->endPos) {
            return 1;
        }
    }
    if (this->readPos == this->endPos) {
        return 0;
    }
    else {
        return -1;
    }
}

/*----------------------------------------*/

bool SAPDBFields_VarData::
Verify (
    const void * buf,
    int buflen,
    bool allowUnderflow)
{
    Verifier ver (buf, buflen);
    int result = ver.verify ();
    if (allowUnderflow) {
        return result <= 0;
    }
    else {
        return result == 0;
    }
}

/*----------------------------------------*/

SAPDBFields_VarData::Writer::
Writer (
    void * buf,
    int bufsize)
{
    this->fieldCount = 0;
    this->startPos = (unsigned char *) buf;
    this->writePos = startPos + 2;
    this->endPos = startPos + bufsize;
}

/*----------------------------------------*/

bool SAPDBFields_VarData::Writer::
addField (
    const void * field,
    unsigned int fieldlen)
{
    int indicatorLen;
    if (lenRequires2Bytes (fieldlen)) {
        indicatorLen = 1;
    }
    else {
        indicatorLen = 3;
    }
    if ((this->writePos + indicatorLen + fieldlen) > this->endPos) {
        return false;
    }
    if (lenRequires2Bytes (fieldlen)) {
        this->writePos [0] = (unsigned int) fieldlen;
    }
    else {
        this->writePos [0] = csp1_fi_2byte_length;
        writeInt2 (this->writePos + 1, fieldlen);
    }
    memcpy (this->writePos + indicatorLen, field, fieldlen);
    this->writePos += indicatorLen + fieldlen;
    ++this->fieldCount;
    return true;
}

/*----------------------------------------*/

bool SAPDBFields_VarData::Writer::
addSpecial (
    int indicator)
{
    if ((this->endPos - this->writePos) < 1) {
        return false;
    }
    this->writePos [0]   = (unsigned int) indicator;
    ++this->fieldCount;
    ++this->writePos;
    return true;
}

/*----------------------------------------*/

void SAPDBFields_VarData::Writer::
close (
    int & fieldCount_out,
    int & buflen)
{
    writeInt2 (this->startPos, this->fieldCount);
    fieldCount_out = this->fieldCount;
    buflen = this->writePos - this->startPos;
}

/*----------------------------------------*/


