/*!***************************************************************************

  module      : QueryRewrite_Rules.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / QueryRewrite
  description : 

  last changed: 2004-11-17  13:45
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Optimizer/QueryRewrite/QueryRewrite_Rules.hpp"
#include "SAPDB/SQLManager/Query/Query_Graph.hpp"
#include "SAPDB/SQLManager/Query/Query_Quantifier.hpp"
#include "SAPDB/SQLManager/Query/Query_Node.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeSelect.hpp"

// ------------------------------------------------------------------------------------------------

inline void RuleTrace( const char* const RuleName, Query_Node* NodePtr )
{
    if (QueryRewrite_Trace.TracesLevel(3))
    {
        Kernel_VTrace() 
            << "    > Apply rule " 
            << RuleName 
            << " for node " 
            << NodePtr->NodeID() << FlushLine;
    }
}

inline void RuleTrace( const char* const RuleName, Query_Node* NodePtr, Query_Quantifier* Quantifier )
{
    if (QueryRewrite_Trace.TracesLevel(3))
    {
        Kernel_VTrace() 
            << "    > Apply rule " 
            << RuleName 
            << " for node " 
            << NodePtr->NodeID()
            << " childnode "
            << Quantifier->NodeRef()->NodeID() << FlushLine;
    }
}

// ------------------------------------------------------------------------------------------------

bool Rule_MergeSelects::IsApplicable( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    if (( Node->GetNodeType() == NT_Select ) &&
        ( Quantifier->QuantifierType() == QT_F ) &&
        ( Quantifier->NodeType() == NT_Select ) &&
        ( ( Node->HeadDistinct() == true ) ||
          ( Node->BodyDistinctType() == DT_Permit ) ||
          ( Quantifier->NodeRef()->BodyDistinctType() != DT_Enforce ) ))
    {
        if ( (!((Query_SelectNode*)Node)->HasForUpdate()) &&
             !Node->OuterJoin() &&
             !Quantifier->OuterJoin() &&
             !Quantifier->Aggregates() )
        {

            if ( !((Query_SelectNode*)Quantifier->NodeRef())->HasGroupBy() )
            {
                return true;
            }
            else
            {
                if ( Node->GetQuantifier().GetSize() == 1 )
                    return true;
            }
        }
    }
    return false;
}

bool Rule_MergeSelects::Apply( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    RuleTrace(GetRuleName(), Node, Quantifier);

    if ( Merge( (Query_SelectNode*)Node, QuantifierNo ) )
    {
        if (( Node->BodyDistinctType() != DT_Permit ) &&
            ( Quantifier->NodeRef()->BodyDistinctType() == DT_Enforce ))
        {
            if ( Node->BodyDistinctType() != DT_Enforce )
            {
                if ( ((Query_SelectNode*)Node)->AddDistinctNode() )
                {
                    Node->BodyDistinctType() = DT_Enforce;
                    ((Query_SelectNode*)Node)->All() = false;
                    return true;
                }
                return false;
            }
        }
        return true;
    }
    return false;
}

bool Rule_MergeSelects::Merge( Query_SelectNode* Node, SAPDB_UInt2 QuantifierNo ) const
{
    
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];
    Query_SelectNode* LowerNode = (Query_SelectNode*) Quantifier->NodeRef();

    bool rc = true;
    rc &= CopyBodyQuery(Node, LowerNode);
    rc &= CopyFromPartSyntaxTree(Node, LowerNode);
    rc &= AdjustSelectListSyntaxTree(Node, LowerNode);

    rc &= Node->PullUpSearchCond(LowerNode);

    rc &= DeleteLowerNodeSyntaxTree(Node, LowerNode);
    rc &= DeleteLowerNodeQuery(Node, QuantifierNo, LowerNode);

    return true;
}

bool Rule_MergeSelects::CopyBodyQuery( Query_SelectNode* UpperNode, Query_SelectNode* LowerNode ) const
{
    unsigned qc = LowerNode->GetQuantifier().GetSize();
    UpperNode->GetQuantifier().Reserve(UpperNode->GetQuantifier().GetSize() + qc );
    for ( unsigned i = 0; i < qc; i++ )
    {
        UpperNode->AddQuantifier(LowerNode->GetQuantifier()[i]);
    }
    LowerNode->GetQuantifier().Delete();
    return true;
}

bool Rule_MergeSelects::CopyFromPartSyntaxTree( Query_SelectNode* UpperNode, Query_SelectNode* LowerNode ) const
{
    Query_SyntaxTree& SyntaxTree = mGraph.SyntaxTree();

    SyntaxTreeNodeNo LowerFromNode = SyntaxTree[LowerNode->FromPartStart()].n_lo_level;
    SyntaxTree[LowerNode->FromPartStart()].n_lo_level = 0;

    SyntaxTreeNodeNo ToBeSubstituted = SyntaxTree[LowerNode->SyntaxNode()].n_refback;
    SyntaxTree.SubstituteNodeBySubtree(ToBeSubstituted, LowerFromNode);

    SyntaxTree.RemoveSubSelectNode( ToBeSubstituted ); // DDT : to be replaced 
//    UpperNode->RemoveSubSelectNode(ToBeSubstituted, true);
    SyntaxTree.Changes() = true;

    return true;
}

bool  Rule_MergeSelects::AdjustSelectListSyntaxTree( Query_SelectNode* UpperNode, Query_SelectNode* LowerNode ) const
{
    return false;
}

bool  Rule_MergeSelects::DeleteLowerNodeSyntaxTree( Query_SelectNode* UpperNode, Query_SelectNode* LowerNode ) const
{
    Query_SyntaxTree& SyntaxTree = mGraph.SyntaxTree();
//    SyntaxTree.UnlinkNode( LowerNode->SyntaxNode() );  // DDT ??? warum auskommentiert ?
    SyntaxTree.SetUnusedNode( LowerNode->SyntaxNode(), /*Recursive*/ true );
    SyntaxTree.Changes() = true;
    return true;
}

bool Rule_MergeSelects::DeleteLowerNodeQuery( Query_SelectNode* UpperNode, SAPDB_UInt2 QuantifierNo, Query_SelectNode* LowerNode ) const
{
    Container_Vector<Query_Quantifier*>::Iterator Iter = UpperNode->GetQuantifier().Position(QuantifierNo);
    UpperNode->GetQuantifier().Delete(Iter);
    LowerNode->~Query_SelectNode();    
    return true;
}

// ------------------------------------------------------------------------------------------------

bool Rule_MergeUnions::IsApplicable( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    if (( Node->GetNodeType() == NT_Select ) &&
        ( Quantifier->QuantifierType() == QT_F ) &&
          Quantifier->NodeTypeSet() )
    {
        if ( !((Query_SelectNode*)Node)->HasForUpdate() &&
             !Node->OuterJoin() &&
             !Node->IsCorrelated() && // DDT
             !Quantifier->OuterJoin() &&     // DDT : keiner der Quantifier darf ein OuterJoin sein!? Jedenfalls nicht die "benachbarten"
             !Quantifier->Aggregates() )
        {
            SAPDB_Int2 n = Node->GetQuantifier().GetSize();
            for ( SAPDB_Int2 i=0; i<n; i++ )
            {
                if ( Node->GetQuantifier()[i]->OuterJoin() )  // Check for OuterJoins
                    return false;
            }


            if ( ((Query_SelectNode*)Quantifier->NodeRef())->HasGroupBy() )
            {
                return false;
            }
            else
            {
                if ( Node->GetQuantifier().GetSize() == 1 )
                    return true;
            }
        }
    }
    return false;
}

bool Rule_MergeUnions::Apply( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_SyntaxTree& SyntaxTree = mGraph.SyntaxTree();
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];
    Query_Node* LowerNode        = Quantifier->NodeRef();
    RuleTrace(GetRuleName(), Node, Quantifier);

    SAPDB_Int2 n = Node->GetQuantifier().GetSize();
    for ( SAPDB_Int2 i=0; i<n; i++)
    {
        if ( i != QuantifierNo )
        {
            Query_Quantifier* Q = Node->GetQuantifier()[i];
            Container_Vector<Query_Quantifier*>::Iterator Iter = Node->GetQuantifier().Position(i);
            if ( Q->QuantifierType() == QT_F )
            {
                SyntaxTreeNodeNo APNode = Q->NodeRef()->SyntaxNode();
                while ( ! ((SyntaxTree[APNode].n_proc == a66) && 
                        ((SyntaxTree[APNode].n_subproc() == cak_x_table_spec) || 
                         (SyntaxTree[APNode].n_subproc() == cak_x_select_in_from_part)) ) )
                {
                    APNode = SyntaxTree[APNode].n_refback;
                }
                SyntaxTree.UnlinkNode(APNode, /*SetUnused*/ false);
                if ( i<QuantifierNo )
                {
                    Node->GetQuantifier().Delete(Iter);
                    if ( !LowerNode->PullDownFQuantifierBegin(Q, APNode) )
                    {
                        mGraph.SetError("Rule_MergeUnions::PullDownFQuantifierBegin");
                        return false;
                    }
                }
                if ( i>QuantifierNo )
                {
                    Node->GetQuantifier().Delete(Iter);
                    if ( !LowerNode->PullDownFQuantifierEnd(Q, APNode) )
                    {
                        mGraph.SetError("Rule_MergeUnions::PullDownFQuantifierEnd");
                        return false;
                    }
                }
                Q->SetDisabled();
            }
        }
    }
    return true;
}

// ------------------------------------------------------------------------------------------------

bool Rule_EtoF_SelMerge::IsApplicable( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    return false;
}

bool Rule_EtoF_SelMerge::Apply( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    RuleTrace(GetRuleName(), Node, Quantifier);

    return false;
}

// ------------------------------------------------------------------------------------------------

bool Rule_PushDownConditions::IsApplicable( Query_Node* Node ) const
{
    if ( Node->GetNodeType() == NT_Select )
    {
        if (  Node->HasSearchCondition() && 
             (Node->FQuantifierCnt()==1) && // DDT >=1
             (Node->WQuantifierCnt()==0) && // DDT: No selects within the where part for now
             !Node->IsCorrelated() )     // DDT
        {
			return true;
            //SAPDB_Int2 n = Node->GetQuantifier().GetSize();
            //for ( SAPDB_Int2 i=0; i<n; i++ )
            //{
            //    Query_Quantifier* Q = Node->GetQuantifier()[i];
            //    if (Q->NodeTypeSelectOrSet())
            //        return true;
            //}
        }
    }
    return false;
}

bool Rule_PushDownConditions::Apply( Query_Node* Node ) const
{
    RuleTrace(GetRuleName(), Node);
    bool rc = ((Query_SelectNode*)Node)->PushDownSearchCond();
    if ( rc )
        mGraph.SyntaxTree().Changes() = true;
    return rc;
}

// ------------------------------------------------------------------------------------------------
// ok

bool Rule_PullUpDistinct::IsApplicable( Query_Node* Node ) const
{
    if ( Node->GetNodeType() == NT_Select && Node->BodyDistinctType() != DT_Preserve )
    {
        bool Cond = true;
        for ( unsigned i=0; i < Node->GetQuantifier().GetSize() && Cond; i++)
        {
            Query_Quantifier* Q = Node->GetQuantifier()[i];
            if ( Q->QuantifierType() == QT_F )
                Cond &= ( Q->OneTupleCondition() || Q->NoDuplicateCondition() );
        }
        return Cond;
    }
    return false;
}

bool Rule_PullUpDistinct::Apply( Query_Node* Node ) const
{
    RuleTrace(GetRuleName(), Node);

    if ( Node->BodyDistinctType() == DT_Enforce )
        if ( !((Query_SelectNode*)Node)->RemoveDistinctNode() )
            return false;

    Node->HeadDistinct()         = true;
    Node->BodyDistinctType()     = DT_Preserve;
    ((Query_SelectNode*)Node)->All()= true;
    return true;  
}

// ------------------------------------------------------------------------------------------------

bool Rule_EXEC2NEXIST::IsApplicable( Query_Node* Node ) const
{
    return false;
}

bool Rule_EXEC2NEXIST::Apply( Query_Node* Node ) const
{
    RuleTrace(GetRuleName(), Node);

    return false;
}

// ------------------------------------------------------------------------------------------------

//bool Rule_AddKeys::IsApplicable( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
//{
//    Query_Quantifier& Quantifier = Node->GetQuantifier()[QuantifierNo];
//
//    return false;
//}
//
//bool Rule_AddKeys::Apply( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
//{
//    Query_Quantifier& Quantifier = Node->GetQuantifier()[QuantifierNo];
//
//    RuleTrace(GetRuleName(), Node, Quantifier);
//
//    return true;
//}

// ------------------------------------------------------------------------------------------------

bool Rule_INT2EXIST::IsApplicable( Query_Node* Node ) const
{
    return false;
}

bool Rule_INT2EXIST::Apply( Query_Node* Node ) const
{
    RuleTrace(GetRuleName(), Node);

    return false;
}

// ------------------------------------------------------------------------------------------------
// ok

bool Rule_DISTPDTO::IsApplicable( Query_Node* Node ) const
{
    Query_NodeType BT = Node->GetNodeType();

    if  (( BT == NT_Select ) ||
         ( BT == NT_Union )  ||
         ( BT == NT_Intersect) ||
         ( BT == NT_Difference))
    {
        if ( Node->BodyDistinctType() != DT_Permit )
            if ( Node->FatherQuantifier()->DistinctType() == DT_Permit )
                return true;
    }
    return false;
}

bool Rule_DISTPDTO::Apply( Query_Node* Node ) const
{
    RuleTrace(GetRuleName(), Node);

    if ( Node->BodyDistinctType() == DT_Enforce )
        if ( !((Query_SelectNode*)Node)->RemoveDistinctNode() )
            return false;

    Node->BodyDistinctType()     = DT_Permit;
    ((Query_SelectNode*)Node)->All()= true;
    return true;
}

// ------------------------------------------------------------------------------------------------
// ok

bool Rule_EorAPDFR::IsApplicable( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];

    if ( Node->GetNodeType() == NT_Select && Quantifier->DistinctType() != DT_Permit )
    {
        if (( Quantifier->QuantifierType() == QT_E ) ||
            ( Quantifier->QuantifierType() == QT_A ))  
            return true;
    }
    return false;
}

bool Rule_EorAPDFR::Apply( Query_Node* Node, SAPDB_UInt2 QuantifierNo ) const
{
    Query_Quantifier* Quantifier = Node->GetQuantifier()[QuantifierNo];
    RuleTrace(GetRuleName(), Node, Quantifier);

    Quantifier->DistinctType() = DT_Permit;
    return true;
}

// ------------------------------------------------------------------------------------------------
// ok

bool Rule_DISTPDFR::IsApplicable( Query_Node* Node ) const
{
    Query_NodeType BT = Node->GetNodeType();
    Query_DistinctType DT = Node->BodyDistinctType();

    if (   (( BT == NT_Select ) ||
            ( BT == NT_Union )  ||
            ( BT == NT_Intersect) ||
            ( BT == NT_Difference)) &&
           ((DT == DT_Permit) || 
            (DT == DT_Enforce)))
    {
        for ( unsigned i=0; i < Node->GetQuantifier().GetSize(); i++)
        {
            Query_Quantifier* Q = Node->GetQuantifier()[i];
            if (( Q->QuantifierType() == QT_F ) && ( Q->DistinctType() != DT_Permit ))
            {
                if ( Q->DistinctType() != DT_Permit )
                    return true;
            }
        }
    }
    return false;
}

bool Rule_DISTPDFR::Apply( Query_Node* Node ) const
{
    bool applied = false;
    for ( unsigned i=0; i < Node->GetQuantifier().GetSize(); i++)
    {
        Query_Quantifier* Q = Node->GetQuantifier()[i];
        if (( Q->QuantifierType() == QT_F ) && ( Q->DistinctType() != DT_Permit ))
        {
            RuleTrace(GetRuleName(), Node, Q);
            Q->DistinctType() = DT_Permit;
            applied |= true;
        }
    }
    return applied;
}

// ------------------------------------------------------------------------------------------------

//bool Rule_CopyNodes::IsApplicable( Query_Node* Node ) const
//{
//    return false;
//}
//
//bool Rule_CopyNodes::Apply( Query_Node* Node ) const
//{
//    RuleTrace(GetRuleName(), Node);
//    return true;
//}

// ------------------------------------------------------------------------------------------------
