/*!***************************************************************************

  module      : Query_Graph.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Query/Query_Graph.hpp"
#include "SAPDB/SQLManager/Query/Query_Node.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeRoot.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeTable.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeSelect.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeCursor.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeUnion.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeIntersect.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeDifference.hpp"

#include "SAPDB/SQLManager/Catalog/Catalog_ViewDescriptor.hpp"

#include "SAPDB/RunTime/RTE_CurrentStack.hpp"

#include "hak05.h"
#include "hak06.h"
#include "hak35.h"
#include "hak542.h"
#include "hak61.h"

// ------------------------------------------------------------------------------------------------

	SQLMan_Identifier Query_QueryGraph::GetIdentifier( SyntaxTreeNodeNo NodeNo )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_SyntaxTree::GetIdentifier", QueryRewrite_Trace, 6);
    SAPDBERR_ASSERT_ARGUMENT( NodeNo > 0 );

    if ( NodeNo > 0 )
    {
        SQLMan_Identifier Identifier;
		a05_identifier_get_from_buffer(mContext.SQLManContext(), 
			mContext.GetSQLStatementPtr(), mContext.GetSQLStatementSize(),
			mSyntaxTree[NodeNo].n_pos(), mSyntaxTree[NodeNo].n_length(), mSyntaxTree[NodeNo].n_special.includes(ni_complex_string_id),
			sizeof(SQLMan_Identifier), Identifier);
        return Identifier;
    }
    return SQLMan_Identifier();
}

SQLMan_Identifier Query_QueryGraph::GetSchemaName( SyntaxTreeNodeNo& NodeNo )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_SyntaxTree::GetSchemaName", QueryRewrite_Trace, 6);
    SAPDBERR_ASSERT_ARGUMENT( NodeNo > 0 );

    if ( NodeNo > 0 )
    {
        SQLMan_Identifier Identifier;
        int N = NodeNo;
		a06get_username_from_buffer(mContext.SQLManContext(), N,
			mContext.GetSQLStatementPtr(), mContext.GetSQLStatementSize(),
			Identifier);
        NodeNo = N;
        return Identifier;
    }
    return SQLMan_Identifier();
}

// ------------------------------------------------------------------------------------------------

bool Query_QueryGraph::Initialize( bool& NeedOptimization )
{
    SAPDBTRACE_METHOD_DEBUG("Query_QueryGraph::Initialize", QueryRewrite_Trace, 5);

    char* StatementPtr = (char*) mContext.SQLManContext().GetSQLStatementPtr(); // just for debugging

    if ( mRootNode )
        destroy(mRootNode, mAllocator);

    mSyntaxTree.SetRefBack(0);  // this is the first thing to do in order to avoid recursion for some of the later function calls !!!

    if ( mRootNode = new(mAllocator) Query_RootNode(mContext, *this) )
    {
        CreateQueryNodes( mSyntaxTree[0].n_lo_level, mRootNode, P_Start, /*ComplexViewReplacement=*/false, /*SubQueryType*/cak_x_subquery );
        if ( IsOk() )
        {
            SetProgress(PG_Nodes);
    
            if ( mRootNode->Semantics(/*UniqueOutputColumns=*/false, /*ComplexViewReplacement=*/false) )
            {
                NeedOptimization = mNeedOptimization;
                SetProgress(PG_Semantics);

/* DDT */
                if ( !mRootNode->ReplaceComplexViews() )
                    SetError("ComplexViewReplacements failed");
/* DDT */
                //if ( NormalizeSearchConditions(0) )
                //    SetProgress(PG_Normalized);
            }
            else
            {
                SetError("Semantics failed");
            }
        }
    }
    else
    {
        SetError("Initialize");
    }
    return (IsOk());
}

// ------------------------------------------------------------------------------------------------

void Query_QueryGraph::PrintToTrace( void )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_QueryGraph::PrintToTrace", QueryRewrite_Trace, 5);
    SAPDBERR_ASSERT_STATE(  ((mProgress == PG_Error) &&  mError ) || 
                            ((mProgress != PG_Error) && !mError ) );

    if ( QueryRewrite_Trace.TracesLevel(1) )
    {
        bool Check = false;
        switch ( mProgress )
        {   
        case PG_Init:
            mVTrace << "-- Query: Initial State -------------------------" << FlushLine;
            break;
        case PG_Nodes:
            mVTrace << "-- Query: QueryNodes created --------------------" << FlushLine;
            Check = true;
            break;
        case PG_SubstituteViews:
            mVTrace << "-- Query: Views substituted ---------------------" << FlushLine;
            Check = true;
            break;
        case PG_Semantics:
            mVTrace << "-- Query: SemanticInfo added --------------------" << FlushLine;
            Check = true;
            break;
        case PG_Normalized:
            mVTrace << "-- Query: SearchCond normalized -----------------" << FlushLine;
            Check = true;
            break;
        case PG_Optimized:
            mVTrace << "-- Query: Query optimized -----------------------" << FlushLine;
            Check = true;
            break;
        case PG_Error:
            mVTrace << "-- Query: ERRORS !!! ----------------------------" << FlushLine;
            break;
        default:
            mVTrace << "-- Query: Unknown state !!! ---------------------" << FlushLine;
            break;
        }

	    if ( QueryRewrite_Trace.TracesLevel(2) )
		{
			SyntaxTree().PrintToTrace();
			PrintQueryToTrace();
		}

#if defined(SAPDB_SLOW)
        if ( Check )
        {
            SyntaxTreeNodeNo ErrNode = SyntaxTree().CheckNodes();
            if ( ErrNode != cak_qrewrite_ref_init )
            {

                if ( !QueryRewrite_Trace.TracesLevel(2) )
                    SyntaxTree().PrintToTrace();
                if ( QueryRewrite_Trace.TracesLevel(1) )
                    mVTrace << "Error in Node " << ErrNode << FlushLine; 

// DDT                SAPDBERR_ASSERT_STATE( 0 ); // Never fullfilled ! slowknl will alway break
            }
        }
#endif
    }
}

void Query_QueryGraph::PrintQueryToTrace( void )
{
    if ( mRootNode )
    {
        if ( QueryRewrite_Trace.TracesLevel(1) )
        {
            mRootNode->PrintToTrace(mVTrace);
            if ( QueryRewrite_Trace.TracesLevel(2) )
            mVTrace << "+-------------------------------" << FlushLine;
        }
    }
}

// ------------------------------------------------------------------------------------------------

void Query_QueryGraph::CreateQueryNodes( SyntaxTreeNodeNo NodeNo, Query_Node* ActNode, Query_Position Position, bool ComplexViewReplacement, SAPDB_Int2 SubQueryType )
{	// DDT : RECURSIVE
    SAPDBTRACE_METHOD_DEBUG ("Query_QueryGraph::CreateQueryNodes", QueryRewrite_Trace, 8);

    {
        SAPDB_Int4 StackFree = RTE_CurrentStackFree();

        if ( StackFree < 4096 )
            SetError("Stack exhausted", /*PrintToTrace=*/false);

		//if ( QueryRewrite_Trace.TracesLevel(8) )
	    //	mVTrace << "  Stack Free      : " << StackFree << FlushLine;
    }

	if ( NodeNo <= 0 || !IsOk() )
	    return;

    SyntaxTreeNodeNo low  = mSyntaxTree[NodeNo].n_lo_level;
    SyntaxTreeNodeNo same = mSyntaxTree[NodeNo].n_sa_level;
/**/	Query_Node*  LoNode = ActNode;
/**/	Query_Node*  SaNode = ActNode;
/**/	Query_Position   LoPosition = Position;
/**/	Query_Position   SaPosition = Position;
/**/	SAPDB_Int2       LoSubQueryType = SubQueryType;
/**/	SAPDB_Int2       SaSubQueryType = SubQueryType;

    switch ( mSyntaxTree[NodeNo].n_proc )
    {

    case no_proc:
        switch ( mSyntaxTree[NodeNo].n_symb )
        {
		case s_parameter_name:
			break;

		case s_authid:
            break;

        case s_reference_name:
            // no break here !!
        case s_columnname:
            break;

        case s_identifier:
            // no break here !!
        case s_tablename:
            break;

        case s_asterisk:
            break;

        case s_all:
			SaPosition = (Position==P_Where?P_WhereAll:(Position==P_Having?P_HavingAll:P_Unknown));
            break;
        case s_any:
            break;

        case s_all_count:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Count() = true;
                else
                    SetError("s_all_count");
                break;
        case s_count:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Count() = true;
                else
                    SetError("s_count");
                break;
        case s_sum:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Sum() = true;
                else
                    SetError("s_sum");
                break;
        case s_avg:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Avg() = true;
                else
                    SetError("s_avg");
                break;
        case s_min:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Min() = true;
                else
                    SetError("s_min");
                break;
        case s_max:
                if ( ActNode->GetNodeType() == NT_Select )
                    ((Query_SelectNode*)ActNode)->Max() = true;
                else
                    SetError("s_max");
                break;
                
        default:
            break;
        }
        break;

    case a16:   // CREATE
        SetNotYetImplemented("a16");
        break;

    case a43:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
		case cak_x_explain:				// 4
    			break;
		default:
		        SetNotYetImplemented("a43");
				break;
        }
        break;

	case a51:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
      //cak_x_connect             =  1;
      //cak_x_current_user        =  2;
      //cak_x_set_lang            =  3;
      //cak_x_set_format          =  6;
      //cak_x_set_degree          =  7;
      //cak_x_get_special_register=  8;
      //cak_x_set_isolevel        =  9;
      //cak_x_nls_date_format     =  11;
      //cak_x_set_session_trace   =  14;
      //cak_x_set_role            =  15;
      //cak_x_authentication_init =  16;
      //cak_x_set_session_unicode     =  20;
      //cak_x_set_session_replication = 21;

		case cak_i_isolation:	// 221
			break;
		default:
		    SetNotYetImplemented("a51");
			break;
        }
        break;

	case a52:
      //cak_x_commit              =   1;
      //cak_x_rollback            =   2;
      //cak_x_release             =   3;
      //cak_x_lock                =   4;
      //cak_x_nowait              =   5;
      //cak_x_table               =   6;
      //cak_x_rows                =   7;
      //cak_x_row                 =   8;
      //cak_x_exclusive           =   9;
      //cak_x_share               =  10;
      //cak_x_unlock              =  12;
      //cak_x_begin_subtrans      =  13;
      //cak_x_end_subtrans        =  14;
      //cak_x_rollback_subtrans   =  15;
      //cak_x_optimistic          =  16;
      //cak_x_force_savepoint     =  17;
		SetNotYetImplemented("a52");
		break;

	case a55:
		switch ( mSyntaxTree[NodeNo].n_subproc() )
		{
		case cak_x_search_clause:       //  1
			break;
		case cak_x_keyspec_list:        //  2
			break;
		case cak_x_keyspec:             //  3
			break;
		case cak_x_current_of:          //  4
			break;
		case cak_x_pos_of:              //  5
			break;
		default:
			SetNotYetImplemented("a55");
			break;
		}
		break;

    case a56:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
            //case cak_x_normal_insert       =  1;
            //case cak_x_enum_values         =  2;
            //case cak_x_column_list         =  3;
        case cak_x_value_list:  // 4
            break;

			//case cak_x_ignore_dupl         =  5;
            //case cak_x_update_dupl         =  6;
            //case cak_x_reject_dupl         =  7;
            //case cak_x_named_values        =  8;
            //case cak_x_set_clause          =  9;
            //case cak_x_insert_select       = 10;
            //case cak_x_minsert             = 11;
            //case cak_x_pseudo_ins_select   = 12;
        default:
            SetNotYetImplemented("a56");
            break;
        }
        break;

    case a57:
    //    switch ( mSyntaxTree[NodeNo].n_subproc() )
    //    {
    //        case cak_x_update              =  1;
    //        case cak_x_update_set          =  2;
    //        case cak_x_one_update_col      =  3;
    //        case cak_x_mupdate             =  4;
    //        case cak_x_append              =  7;
    //    }
        SetNotYetImplemented("a57");
        break;

    case a58:
    //    switch ( mSyntaxTree[NodeNo].n_subproc() )
    //    {
    //        case cak_x_delete              =  1;
    //        case cak_x_mdelete             =  2;            }
    //            break;
    //    }
        SetNotYetImplemented("a58");
        break;

    case a60:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
        case cak_x_select_list:         //   1
            ActNode->SetSelectListStart(NodeNo);
			LoPosition = P_SelectList;
            break;
        case cak_x_select_column:       //   2
            ActNode->AppendOutputColumn(NodeNo);
			LoSubQueryType = cak_x_one_val_subquery;
            break;
        case cak_x_single_select:       //   3  == "into :a"
            if ( ActNode->GetNodeType() != NT_Select )
                SetError("cak_x_single_select");
            break;

		case cak_x_lock_nowait:         //   4
            break;
		case cak_x_lock_exclusive:      //   5
			break;
		case cak_x_lock_share:          //   6
			break;
		case cak_x_lock_optimistic:     //   7
			break;
		case cak_x_lock_ignore:         //   8
			break;
		
		default:
            SetNotYetImplemented("a60");
            break;
        }
        break;

    case a62:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
        case cak_x_select_direct:       //   1
			SetError("cak_x_select_direct");
		    break;
		default:
			SetNotYetImplemented("a62");
			break;
        }
        break;

    case a63:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
        case cak_x_mass_select:         //   1
            LoNode = AddSelectNode(ActNode, Position, NodeNo, ComplexViewReplacement, LoSubQueryType);
			LoPosition = P_Select;
            break;
        case cak_x_distinct:            //   2  : Indicates that the select is distinct ! 
            if ( ActNode->GetNodeType() == NT_Select )
            {
                ((Query_SelectNode*)ActNode)->All()                 = false;
                ((Query_SelectNode*)ActNode)->HeadDistinct()        = true;
                ((Query_SelectNode*)ActNode)->BodyDistinctType()    = DT_Enforce;
            }
            else
            {
                SetError("cak_x_distinct");
            }
            break;
        case cak_x_from_part:           //   3
            ActNode->SetFromPartStart(NodeNo);
			LoPosition = P_FromPart;
            break;

        case cak_x_decl_cursor:         //   4
            LoNode = new(mAllocator) Query_CursorNode(mContext, *this, Position, NodeNo, true); 
            if ( LoNode )
            {
                if ( !ActNode->AddQuantifier(LoNode, DT_Preserve, Position, DetermineQuantifierType(Position)) )
                {
                    SetError("cak_x_decl_cursor 1");
                    destroy(LoNode, mAllocator);
					LoNode = 0;
                    return;
                }
				LoPosition = P_Select;
				SaPosition = P_Select; // DDT : ??? 
				SaNode = LoNode;
            }
            else
            {
                SetError("cak_x_decl_cursor 2");
            }
			break;

        case cak_x_order:               //   5
            ActNode->SetOrderByStart(NodeNo);
			LoPosition = P_OrderBy;
            break;
        case cak_x_for_update:          //   6
			SetNotYetImplemented("cak_x_for_update");
   //         ActNode->SetUpdateColumnsStart(NodeNo);
			//LoPosition = P_UpdateColumns;
            break;
        case cak_x_group_by:            //   7
            ActNode->SetGroupByStart(NodeNo);
			LoPosition = P_GroupBy;
            break;
        case cak_x_having:              //   8
            ActNode->SetHavingStart(NodeNo);
			LoPosition = P_Having;
            break;
        case cak_x_sort_spec:           //   9
            break;

        case cak_x_search_condition:    //  10
            ActNode->SetSearchConditionStart(NodeNo);
			LoPosition = P_Where;
            break;

        case cak_x_for_reuse:           //  11
            ActNode->SetForReuse();
            break;
        case cak_x_subquery:            //  12
            mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
            break;

        case cak_x_start_union:         //  13  ( lo_level ist a63/14/15/16/17/18/19 )
            break;

        case cak_x_union:               //  14  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Union, NodeNo, false );
			SaPosition = LoPosition = P_Union;
			break;
        case cak_x_union_all:           //  15  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Union, NodeNo, true );
			SaPosition = LoPosition = P_Union;
            break;

        case cak_x_except:              //  16  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Difference, NodeNo, false );
			SaPosition = LoPosition = P_Difference;
			break;
        case cak_x_except_all:          //  17  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Difference, NodeNo, true );
			SaPosition = LoPosition = P_Difference;
			break;

        case cak_x_intersect:           //  18  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Intersect, NodeNo, false );
			SaPosition = LoPosition = P_Intersect;
			break;
        case cak_x_intersect_all:       //  19  ( lo_level/sa_level sind die beiden select's )
			SaNode = LoNode = AddSetNode( ActNode, Position, P_Intersect, NodeNo, true );
			SaPosition = LoPosition = P_Intersect;
			break;

        case cak_x_res_delete:          //  20
            break;
        case cak_x_res_no_delete:       //  21
            break;
        //case cak_x_for_fetch_only:      //  23
        //    break;

        case cak_x_recursive_select:    //  24
            SetNotYetImplemented("cak_x_recursive_select");
            break;

        //case cak_x_new_columnname_list: //  26
        //    break;

        case cak_x_one_val_subquery:    //  27 Subselect indicator (n_symb = cak_x_select or cak_x_where or cak_x_having jenachdem, wo was subselect steht,
                                        //  n_pos gibt des Konten des zugehrigen Subselects an
            mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
            break;

        case cak_x_arith_with_subquery: //  28
            mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
            break;

        //case cak_x_corr_subquery:       //  29
        //    break;
        //case cak_x_one_val_corr_sub:    //  30
        //    break;    

        default:
            SetNotYetImplemented("a63");
            break;
        }
        break;

    case a63query_spec:
        LoNode = AddSelectNode(ActNode, Position, NodeNo, ComplexViewReplacement, LoSubQueryType);
		LoPosition = P_Select;
        break;

    case a64:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
		case cak_x_or_operator:		// 1:
            break;
		case cak_x_and_operator:	// 2:
            break;
        case cak_x_predicate:       // 3: Predicate - n_symbol contains what it is (=,>,<,...)
			if ( mSyntaxTree[NodeNo].n_symb == s_equal || mSyntaxTree[NodeNo].n_symb == s_unequal )
			{
				LoSubQueryType = cak_x_one_val_subquery;
			}
            break;
        case cak_x_between_pred:	// 6:
            break;
        case cak_x_in_pred:         // 7: 
            break;
        case cak_x_in_pred_list:    // 8:
            break;

        case cak_x_left_outer_join:     //  12 - n_symbol contains what it is (=,>,<,...)
            ActNode->SetOuterJoin(JT_LeftOuter);
			if ( ComplexViewReplacement )
				mContext.SQLManContext().a_outer_join = true;
            break;
        case cak_x_right_outer_join:    //  13 - n_symbol contains what it is (=,>,<,...)
            ActNode->SetOuterJoin(JT_RightOuter);
			if ( ComplexViewReplacement )
				mContext.SQLManContext().a_outer_join = true;
            break;
        case cak_x_full_outer_join:     //  14 - n_symbol contains what it is (=,>,<,...)
            ActNode->SetOuterJoin(JT_FullOuter);
			if ( ComplexViewReplacement )
				mContext.SQLManContext().a_outer_join = true;
            break;

      //cak_x_escape              =  15;
      //cak_x_truth_value         =  16;
		case cak_x_value_expression:	// 17 - VAL EXPR
			break;

		default:
//            SetError(); // DDT ???
            break;
        }
        break;

    case a641:
        break;

    case a66:
        switch ( mSyntaxTree[NodeNo].n_subproc() )
        {
        case cak_x_given_sequence:      //   1
			ActNode->SetGivenSequence();
            break;
        case cak_x_table_spec:          //   2  (lo_level of a63/3)
			LoNode = AddTableNode( ActNode, Position, NodeNo );
			break;

        case cak_x_select_in_from_part: //   3  (lo_level of a63/3)
            mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
            break;

		default:
            SetNotYetImplemented("a66");
            break;
        }
        break;


    case a67:
        //switch ( mSyntaxTree[NodeNo].n_subproc() )
        //{
        //mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_Syntax);
        //CreateQueryNodes( low,  ActNode, Position, ComplexViewReplacement );
        //CreateQueryNodes( same, ActNode, Position, ComplexViewReplacement );
		//}
		break;

    case a73:
        //switch ( mSyntaxTree[NodeNo].n_subproc() )
        //{
        //case cak_x_describe: // 11:
        //    break;
        //}
        SetNotYetImplemented("a73");
        break;

    case a80:           // Hint node
        //mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_Syntax);
        mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
        break;

    case a90:           // Initiallink
        break;

    case a92fromsel:    // FROMSEL indicator (sa_level of a63/1)
        mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_NeedAdjustment);
        SetNeedOptimization();
        break;          // subproc+pos indicates the treenode which contains the subselect

    default:
        SetNotYetImplemented("n_proc");
        break;
    }

	if ( IsOk() )
	{
        mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_Syntax);
		if ( low )
	        CreateQueryNodes( low,  LoNode, LoPosition, ComplexViewReplacement, LoSubQueryType );
		if ( same )
	        CreateQueryNodes( same, SaNode, SaPosition, ComplexViewReplacement, SaSubQueryType );
	}
}

Query_Node* Query_QueryGraph::AddSelectNode( Query_Node* ActNode, Query_Position Position, SyntaxTreeNodeNo NodeNo, bool ComplexViewReplacement, SAPDB_Int2 SubQueryType )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_QueryGraph::AddSelectNode", QueryRewrite_Trace, 5);

    Query_Node* NewNode = new(mAllocator) Query_SelectNode(mContext, *this, Position, NodeNo, false); 
    if ( NewNode )
    {
		if ( ComplexViewReplacement && SubSelectNodeMustBeAdded(Position) )
		{
			SyntaxTreeNodeNo N = NodeNo;
			if ( Position == P_FromPart )
				N = SyntaxTree()[NodeNo].n_refback;

			if ( !AddSubSelectNode(N, Position, SubQueryType) )
			{
				SetError("AddSelectNode 3");
		        destroy(NewNode, mAllocator);
				NewNode = 0;
				return 0;
			}
		}

		mSyntaxTree[NodeNo].n_querycheck.addElement(qcn_Syntax);
        if ( !ActNode->AddQuantifier(NewNode, DT_Preserve, Position, DetermineQuantifierType(Position)) )
        {
			SetError("AddSelectNode 2");
			destroy(NewNode, mAllocator);
			NewNode = 0;
		}
    }
	else
	{
	    SetError("AddSelectNode 1");
	}
    return NewNode;
}

Query_Node* Query_QueryGraph::AddSetNode( Query_Node* ActNode, Query_Position Position, Query_Position SetType, SyntaxTreeNodeNo NodeNo, bool All )
{
	Query_Node* NewNode = 0;
	switch ( SetType )
	{
	case P_Union:	
		NewNode = new(mAllocator) Query_UnionNode		(mContext, *this, Position, NodeNo, All); 
		break;
	case P_Difference:
		NewNode = new(mAllocator) Query_DifferenceNode(mContext, *this, Position, NodeNo, All); 
		break;
	case P_Intersect:
		NewNode = new(mAllocator) Query_IntersectNode	(mContext, *this, Position, NodeNo, All); 
		break;
	default:
		return 0;
	}

	if ( NewNode )
    {
        if ( !ActNode->AddQuantifier(NewNode, DT_Preserve, Position, DetermineQuantifierType(Position)) )
        {
            SetError("AddSetNode 2");
            destroy(NewNode, mAllocator);
			NewNode = 0;
        }
    }
    else
    {
        SetError("AddSetNode 1");
    }
	return NewNode;
}

Query_Node* Query_QueryGraph::AddTableNode( Query_Node* ActNode, Query_Position Position, SyntaxTreeNodeNo NodeNo )
{
	Query_Node* NewNode = new(mAllocator) Query_TableNode(mContext, *this, Position, NodeNo); 
    if ( NewNode )
    {
        if ( !ActNode->AddQuantifier(NewNode, DT_Preserve, Position, QT_F) )
        {
            SetError("AddTableNode 2");
            destroy(NewNode, mAllocator);
            NewNode = 0;
        }
    }
    else
    {
        SetError("AddTableNode 1");
    }
	return NewNode;
}

// ------------------------------------------------------------------------------------------------

bool Query_QueryGraph::AddSubSelectNode( SyntaxTreeNodeNo StartNode, Query_Position Position, SAPDB_Int2 SubProc )
{
	SAPDBERR_ASSERT_ARGUMENT( 
		SubProc == cak_x_subquery ||
		SubProc == cak_x_one_val_subquery ||
		SubProc == cak_x_arith_with_subquery );

    // Add 
	//		a92fromsel		for the new subselect	if Position == P_FromPart
	//      a63/n select							if Position == P_SelectList
    //		a63/n where								if Position == P_Where/P_WhereAll
	//		a63/n having							if Position == P_Having/P_HavinAll

	// n=27 for a one_val_subquery else n=12 !

	//
	// -> see ak60put_new_sub_select_node - but that routine wouldn't work for our purpose ;-((

	SyntaxTreeNodeNo ToBeInsert = SyntaxTree().GetFreeNode();
    if ( ToBeInsert > 0 )
    {
        SAPDB_Int2 Pos = -1; 
        SyntaxTreeNodeNo FatherSelNode = cak_qrewrite_ref_init;
        SyntaxTreeNodeNo SelNode = cak_qrewrite_ref_init;

/**/	SyntaxTreeNodeNo Node = SyntaxTree()[StartNode].n_refback;
        while ( Node > 0 )
        {
            if ( ( SyntaxTree()[Node].n_proc == a63 && 
                  (SyntaxTree()[Node].n_subproc() == cak_x_mass_select ||
				   SyntaxTree()[Node].n_subproc() == cak_x_decl_cursor || 
                   SyntaxTree()[Node].n_subproc() == cak_x_start_union)) ||
                 (SyntaxTree()[Node].n_proc == a63query_spec) )
            {
//				Pos++;
                if ( FatherSelNode == cak_qrewrite_ref_init )
				{
					SyntaxTreeNodeNo R = SyntaxTree()[Node].n_refback;
					if ( (SyntaxTree()[Node].n_proc == a63query_spec) &&
						 (SyntaxTree()[R   ].n_proc == a66)           &&
						 (SyntaxTree()[R   ].n_subproc() == cak_x_select_in_from_part) )
						FatherSelNode = SyntaxTree()[Node].n_refback;
					else
	                    FatherSelNode = Node;
				}
                SelNode = Node;
            }
            if ( SyntaxTree()[Node].n_proc == a63 )
                if ( (SyntaxTree()[Node].n_subproc() == cak_x_union)      ||
                     (SyntaxTree()[Node].n_subproc() == cak_x_union_all)  ||
                     (SyntaxTree()[Node].n_subproc() == cak_x_except)     ||
                     (SyntaxTree()[Node].n_subproc() == cak_x_except_all) ||
                     (SyntaxTree()[Node].n_subproc() == cak_x_intersect)  ||
                     (SyntaxTree()[Node].n_subproc() == cak_x_intersect_all) )
                {
                    break;
                }

            Node = SyntaxTree()[Node].n_refback;
        }

		SyntaxTree()[ToBeInsert].n_querycheck.addElement(qcn_Syntax);
		if ( Position == P_FromPart )
		{
			SyntaxTree()[ToBeInsert].n_proc.becomes(a92fromsel);
			SyntaxTree()[ToBeInsert].n_subproc()    = StartNode;
			SyntaxTree()[ToBeInsert].n_pos()        = StartNode;
		}
		else
		{
			SyntaxTree()[ToBeInsert].n_proc.becomes(a63);
			SyntaxTree()[ToBeInsert].n_subproc() = SubProc;
			SyntaxTree()[ToBeInsert].n_pos()     = StartNode;
		}
		switch ( Position )
		{
		case P_FromPart:
			SyntaxTree()[ToBeInsert].n_symb.becomes(s_unknown);
			break;
		case P_SelectList:
			SyntaxTree()[ToBeInsert].n_symb.becomes(s_select);
			break;
		case P_Where:
		case P_WhereAll:
			SyntaxTree()[ToBeInsert].n_symb.becomes(s_where);
			break;
		case P_Having:
		case P_HavingAll:
			SyntaxTree()[ToBeInsert].n_symb.becomes(s_having);
			break;
		default:
			SetError("AddSubSelectNode");
			return false;
			break;
		}

        return SyntaxTree().AddSubSelectNode(SelNode, ToBeInsert, FatherSelNode, Pos);
    }
    return false;
}

// ------------------------------------------------------------------------------------------------

bool Query_QueryGraph::SCVCreateInternalPacket(SQLMan_Context& ViewContext, SAPDB_Int4 ViewStmtLength)
{
    a542internal_packet (ViewContext, /*release_internal_packet*/false, ViewStmtLength);
    if ( ViewContext.IsOk() ) 
    {
        ViewContext.a_cmd_part->sp1p_buf_len() = ViewStmtLength;
        ViewContext.a_cmd_segm->sp1c_prepare() = ViewContext.a_initial_segment_header.sp1c_prepare();
        ViewContext.a_cmd_segm->sp1c_producer().becomes(sp1pr_complex_view_handling);
        ViewContext.a_cmd_segment_header       = ViewContext.a_cmd_segm->sp1s_segm_header();
        return true;
    }
    return false;
}

bool Query_QueryGraph::SCVReleaseInternalPacket(SQLMan_Context& ViewContext)
{
    a542pop_packet(ViewContext);	// release the internal packet AND restore the previous one
    return ViewContext.IsOk(); 
}


void* Query_QueryGraph::SCVSaveContextStmtPart( SQLMan_Context& ViewContext )
{
	int Size = ViewContext.GetStatementPartSize();
    void* SavedPart = ViewContext.GetAllocator().Allocate(Size);
    if ( SavedPart )
	    SAPDB_MemCopyNoCheck (SavedPart, ViewContext.GetStatementPartPtr(), Size);
	return SavedPart;
}

bool Query_QueryGraph::SCVRestoreContextStmtPart( SQLMan_Context& ViewContext, void* SavedPart )
{
    if ( SavedPart )
    {
        SAPDB_MemCopyNoCheck (ViewContext.GetStatementPartPtr(), SavedPart, ViewContext.GetStatementPartSize());
        ViewContext.GetAllocator().Deallocate(SavedPart);
		return true;
    }
	return false;
}

bool Query_QueryGraph::SCVParseViewStatement(SQLMan_Context& ViewContext, Catalog_ViewDescriptor& ViewDescr, bool IsComplexView)
{
	// Save old stmtpart of the context
	void* SavedContextStmtPart = SCVSaveContextStmtPart(ViewContext);
	if ( SavedContextStmtPart )
	{
		// Parse View Definition...
		ViewContext.a_ex_kind.becomes(only_syntax);
		ViewContext.a_sqlmode.becomes(ViewDescr.GetSqlMode());
		if ( IsComplexView )
			ViewContext.a_dt_format.becomes(ViewDescr.GetDateTimeFormat());
		a35_asql_statement(ViewContext);

		//  Restore old values...
		SQLMan_SyntaxTreeMembers OrgSTM = ViewContext.GetSyntaxTreeMembers();
		SCVRestoreContextStmtPart(ViewContext, SavedContextStmtPart);
		ViewContext.SetSyntaxTreeMembers(OrgSTM);

	    return ViewContext.IsOk(); 
	}
	return false;
}

bool Query_QueryGraph::SCVCheckIfViewCanBeSubstituted(Query_SyntaxTree& NewSyntaxTree, SyntaxTreeNodeNo& ViewNameNode, SyntaxTreeNodeNo& ViewSelectNode, SyntaxTreeNodeNo& ColNameNodes )
{
    // So far either number of columns given by a16/3 must match number of columns given by a60/1 or there is no a16/3 node.

    SyntaxTreeNodeNo Node = NewSyntaxTree[0].n_lo_level;
    if ( (NewSyntaxTree[Node].n_proc == a16) && (NewSyntaxTree[Node].n_subproc() == cak_i_view) )
    {
        ViewNameNode = Node = NewSyntaxTree[Node].n_lo_level;    // points to the Symb tabl witch gives the viewname
		if ( (NewSyntaxTree[Node].n_proc == no_proc) &&
		     (NewSyntaxTree[Node].n_symb == s_authid) )
			ViewNameNode = Node = NewSyntaxTree[Node].n_sa_level;

        Node = NewSyntaxTree[Node].n_lo_level;    
        if ( (NewSyntaxTree[Node].n_proc == a16) && 
            (NewSyntaxTree[Node].n_subproc() == cak_x_column_list) )
        {
			SyntaxTreeNodeNo Help = Node;
            Node = NewSyntaxTree[Help].n_lo_level;
            NewSyntaxTree[Help].n_lo_level = 0;

			ColNameNodes = NewSyntaxTree[Help].n_sa_level;
			NewSyntaxTree[Help].n_sa_level = 0;
        }
        // Node now points to a631/-2 or a32/cak_x_start_union
        // set ViewName Node as alias for the "subquery"
        ViewSelectNode = Node;
        NewSyntaxTree[ViewSelectNode].n_sa_level = ViewNameNode;
        NewSyntaxTree[ViewNameNode  ].n_proc.becomes(no_proc);
        NewSyntaxTree[ViewNameNode  ].n_subproc() = 0;
        NewSyntaxTree[ViewNameNode  ].n_symb.becomes(s_identifier);
        NewSyntaxTree[ViewNameNode  ].n_sa_level = 0;
        NewSyntaxTree[ViewNameNode  ].n_lo_level = 0;
        NewSyntaxTree[ViewNameNode  ].n_refback  = ViewSelectNode;

		return true;
    }

	SetError("CheckIfViewCanBeSubstituted");
	NewSyntaxTree.PrintToTrace();
	return false;
}

bool Query_QueryGraph::SCVLinkSyntaxTrees(SyntaxTreeNodeNo FromTabNode, SyntaxTreeNodeNo ViewSyntaxTreeNode, SyntaxTreeNodeNo ViewNameNode, bool ReferencedViaSynonym)
{
    SyntaxTreeNodeNo Node  = SyntaxTree()[FromTabNode].n_lo_level;
    SyntaxTreeNodeNo ToBeRemoved = Node;
    
    while ( Node > 0 )
    {
		if ( ReferencedViaSynonym )
		{
			if ( (SyntaxTree()[Node].n_proc == no_proc) &&
				(SyntaxTree()[Node].n_symb == s_tablename) ) // we have to use the synonym name
			{
				SyntaxTree()[ViewNameNode].n_pos()    = SyntaxTree()[Node].n_pos();
				SyntaxTree()[ViewNameNode].n_length() = SyntaxTree()[Node].n_length();
			}
		}
        if ( (SyntaxTree()[Node].n_proc == no_proc) &&
             (SyntaxTree()[Node].n_symb == s_identifier) ) // there already was an renaming
        {
            SyntaxTree()[ViewNameNode].n_pos()    = SyntaxTree()[Node].n_pos();
            SyntaxTree()[ViewNameNode].n_length() = SyntaxTree()[Node].n_length();
        }
        Node = SyntaxTree()[Node].n_sa_level;
    }

    SyntaxTree().SetUnusedNode(ToBeRemoved, true);
    SyntaxTree()[FromTabNode        ].n_lo_level = ViewSyntaxTreeNode;
    SyntaxTree()[ViewSyntaxTreeNode ].n_refback  = FromTabNode;

    // Change a63/cak_x_mass_select to a63query_spec/-2 if necessary
    if ( (SyntaxTree()[ViewSyntaxTreeNode].n_proc == a63) &&
        (SyntaxTree()[ViewSyntaxTreeNode].n_subproc() == cak_x_mass_select) )
    {
        SyntaxTree()[ViewSyntaxTreeNode].n_proc.becomes(a63query_spec);
        SyntaxTree()[ViewSyntaxTreeNode].n_subproc()  = -2;
    }

    // Change type ComplexView.SyntaxNode from a66/cak_x_table_spec to a66/cak_x_select_in_from_part
    SyntaxTree()[FromTabNode].n_subproc() = cak_x_select_in_from_part;

    return true;
}

bool Query_QueryGraph::SubstituteComplexView( Query_ComplexViewInfo& ComplexView, SyntaxTreeNodeNo& ColNameNodes, bool IsComplexView )
{
    //SetError("Cannot substitute ComplexView");
    //return false;

    // Note: This method is called _before_ CreateNodes !!!
    //       Therefore anything has to be done on Syntax(AP)Tree-Level only !!!
    ///

    SAPDBTRACE_METHOD_DEBUG ("Query_QueryGraph::SubstituteComplexView", QueryRewrite_Trace, 5);
    SAPDBERR_ASSERT_ARGUMENT( ComplexView.SyntaxNode != 0 );

    SQLMan_Context& ViewContext = mContext.SQLManContext();

	// Create ViewDescriptor 
	Catalog_ObjectId ViewID = ComplexView.TabID;
	Catalog_ViewDescriptor ViewDescr(ViewContext, ViewID);
	if ( ViewDescr )
	{
		SAPDB_Int4 ViewStmtLength = ViewDescr.GetDefinitionLength(false);
		if ( SCVCreateInternalPacket(ViewContext, ViewStmtLength) )				// Create InternalPacket for parsing the view
		{
			void* Dest = (void*)ViewContext.a_cmd_part->sp1p_buf();			
			if ( ViewDescr.CopyDefinition (false, true, Dest, ViewStmtLength) )	// Copy the ViewStatement into the InternalPacket
			{
				// Save old syntax tree and create a new one
				SQLMan_SyntaxTreeMembers OrgSTM = ViewContext.GetSyntaxTreeMembers();
				ViewContext.a_ap_tree = 0; // to insure NewSyntaxTree works...
				ViewContext.NewSyntaxTree(2048); // DDT : should be smaller but due to old pascal coding
												 //       we have to insure that no reallocate occurs...

				if ( SCVParseViewStatement(ViewContext, ViewDescr, IsComplexView) )	// Parse ViewStatement
				{
					SQLMan_SyntaxTreeMembers ViewSTM = ViewContext.GetSyntaxTreeMembers();

					// Check if replacement is possible
					Query_SyntaxTree ViewSyntaxTree(ViewContext);
					SyntaxTreeNodeNo ViewSelectNode, ViewNameNode;
					ViewSyntaxTree.PrintToTrace();
					if ( SCVCheckIfViewCanBeSubstituted(ViewSyntaxTree, ViewNameNode, ViewSelectNode, ColNameNodes) )
					{
						// Length of the original Statement
						SAPDB_Int4 PosOffset = mContext.GetSQLStatementSize();

						// Copy SyntaxTree into own SyntaxTree array...
						ViewContext.SetSyntaxTreeMembers(OrgSTM);
						SyntaxTreeNodeNo ViewSyntaxTreeNode = SyntaxTree().DoubleTreePart((tak_ap_node*)ViewSTM.SyntaxTreePtr, ViewSelectNode, PosOffset );
						if ( ColNameNodes > 0 )
							ColNameNodes = SyntaxTree().DoubleTreePart((tak_ap_node*)ViewSTM.SyntaxTreePtr, ColNameNodes, PosOffset );
						ViewContext.GetSyntaxTreeMembers(OrgSTM); // Save the modified syntax tree as original one			
						if ( ViewSyntaxTreeNode > 0 )
						{

							// Append View Statement to SQLStatement...
							void* ToAdd             = ViewContext.GetSQLStatementPtr();
							SAPDB_Int4  AddLength   = ViewContext.GetSQLStatementSize();
							if ( mContext.AppendToSQLStatement(ToAdd, AddLength) )
							{
								ViewContext.SetSyntaxTreeMembers(OrgSTM);
#ifdef SAPDB_SLOW
								Dest = (void*)ViewContext.GetSQLStatementPtr();	// For debugging...
#endif
								// Link both SyntaxTrees
								if ( SCVLinkSyntaxTrees(ComplexView.SyntaxNode, ViewSyntaxTreeNode, ViewNameNode, ComplexView.ReferencedViaSynonym) )   
								{
									// Set backwards references for the new syntaxtree part
									SyntaxTree().SetRefBack(ComplexView.SyntaxNode);

                                    if ( AddSubSelectNode(ComplexView.SyntaxNode, P_FromPart, cak_x_subquery) )    // Add FromSelectNode
									{
										// Delete the new syntax tree and restore the modified new one...
										ViewContext.SetSyntaxTreeMembers(ViewSTM);
										ViewContext.DeleteSyntaxTree();
										ViewContext.SetSyntaxTreeMembers(OrgSTM);
										//Release the packet that we don't need any longer
										SCVReleaseInternalPacket(ViewContext);	
										ComplexView.SyntaxNode = 0;
										ViewContext.a_from_select = true;
										return true;
									} // AddFromSelectNode
								} // LinkSyntaxTrees
							} // AppendSQLStatement
						} // DuplicateTreePart
						ViewContext.SetSyntaxTreeMembers(ViewSTM);
					} // Check if view can be substituted
				} // Parse ViewStatement
				// Restore old syntax tree and statement part of the context
				ViewContext.DeleteSyntaxTree();
				ViewContext.SetSyntaxTreeMembers(OrgSTM);
			} // if ( ViewDescr.CopyDefinition )
			SCVReleaseInternalPacket(ViewContext); //Release the packet that we don't need any longer
		} // Create Internal Packet
	} // if ( ViewDescr )

	SetError("SubsituteComplexView");
    mContext.SQLManContext().ResetReturnCode();
    return false;
}

// ------------------------------------------------------------------------------------------------

bool Query_QueryGraph::AddExpressionRefName( SyntaxTreeNodeNo Node, SAPDB_Int2& ExprCount )
{
	if ( mSyntaxTree[Node].n_symb != s_rowno )
	{
		SyntaxTreeNodeNo RefNode = mSyntaxTree.GetFreeNode();

		if ( RefNode > 0 )
		{
			tak00_columninfo akColInfo;
			a61_new_expr_no(akColInfo, ExprCount);
			SAPDB_Int4 Pos = mContext.AppendToSQLStatement((void*)akColInfo.ccolumnn(), akColInfo.ccolumnn_len());
			if ( Pos > 0 )
			{
				// fill the new node
				mSyntaxTree[RefNode].n_querycheck.addElement(qcn_Syntax);
				mSyntaxTree[RefNode].n_proc.becomes(no_proc);
				mSyntaxTree[RefNode].n_subproc() = 0;
				mSyntaxTree[RefNode].n_symb.becomes(s_reference_name);
				mSyntaxTree[RefNode].n_pos() = Pos + 1; // +1 due to the pascal coding...
				mSyntaxTree[RefNode].n_length() = akColInfo.ccolumnn_len();

				// link nodes
				SyntaxTreeNodeNo PrevNode = mSyntaxTree[Node].n_refback;
				mSyntaxTree[PrevNode].n_lo_level = RefNode;
				mSyntaxTree[RefNode ].n_sa_level = Node;
				mSyntaxTree[RefNode ].n_refback  = PrevNode;
				mSyntaxTree[Node    ].n_refback  = RefNode;

				return true;
			}
			mSyntaxTree.SetUnusedNode(RefNode);
		}
		return false;
	}
	return true;
}

// ------------------------------------------------------------------------------------------------

bool Query_QueryGraph::NegotiateChildParameters( SyntaxTreeNodeNo Node )
{
    if ( (mSyntaxTree[Node].n_proc == no_proc) &&
         ( (mSyntaxTree[Node].n_symb == s_and) ||
           (mSyntaxTree[Node].n_symb == s_and) ) )
    {
        SyntaxTreeNodeNo H = mSyntaxTree[Node].n_lo_level;
        SyntaxTreeNodeNo Not;
        if ( (Not = mSyntaxTree.GetFreeNode()) > 0 )
        {
			mSyntaxTree[Not ].n_querycheck.addElement(qcn_Syntax);
            mSyntaxTree[Not ].n_proc.becomes(no_proc);
            mSyntaxTree[Not ].n_subproc() = 0;
            mSyntaxTree[Not ].n_symb.becomes(s_not);

            mSyntaxTree[Node].n_lo_level = Not;
            mSyntaxTree[Not ].n_refback  = Node;
            mSyntaxTree[Not ].n_sa_level = H;
            mSyntaxTree[H   ].n_refback  = Not;
            
            H = mSyntaxTree[H].n_sa_level;
            if ( mSyntaxTree[H].n_symb == s_authid )
                H = mSyntaxTree[H].n_sa_level;
            if ( mSyntaxTree[H].n_symb == s_tablename )
                H = mSyntaxTree[H].n_sa_level;

            if ( (Not = mSyntaxTree.GetFreeNode()) > 0 )
            {
				mSyntaxTree[Not ].n_querycheck.addElement(qcn_Syntax);
                mSyntaxTree[Not ].n_proc.becomes(no_proc);
                mSyntaxTree[Not ].n_subproc() = 0;
                mSyntaxTree[Not ].n_symb.becomes(s_not);

                mSyntaxTree[Node].n_lo_level = Not;
                mSyntaxTree[Not ].n_refback  = Node;
                mSyntaxTree[Not ].n_sa_level = H;
                mSyntaxTree[H   ].n_refback  = Not;

                return true;
            }
        }
    }
    return false;
}


bool Query_QueryGraph::NormalizeSearchConditions( SyntaxTreeNodeNo StartNode )
{
    return true; // do nothing so far...

    //SyntaxTreeNodeNo Next = mSyntaxTree[StartNode].n_lo_level;

    //if ( mSyntaxTree[StartNode].n_proc == no_proc )
    //{
    //    switch ( mSyntaxTree[StartNode].n_symb )
    //    {
    //    case s_not:
    //        if ( mSyntaxTree[Next].n_proc == no_proc )
    //        {
    //            switch ( mSyntaxTree[Next].n_symb )
    //            {
    //            case s_not:                 // NOT NOT A        -> A
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                StartNode = mSyntaxTree[Next].n_refback;
    //                mSyntaxTree.UnlinkNode(Next, true);
    //                Next      = mSyntaxTree[StartNode].n_lo_level;
    //                break;

    //            case s_or:                // NOT (A OR B)     -> NOT(A) AND NOT(B)
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_and);
    //                NegotiateChildParameters(Next);
    //                break;

    //    //        case s_and:               // NOT (A AND B)    -> NOT(A) OR NOT(B) but we don't want this

    //            case s_equal:             // NOT equal        -> unequal
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_unequal);
    //                break;
    //            case s_unequal:           // NOT unequal      -> equal
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_equal);
    //                break;

    //            case s_greater:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_less_or_eq);
    //                break;
    //            case s_less_or_eq:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_greater);
    //                break;

    //            case s_less:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_greater_or_eq);
    //                break;
    //            case s_greater_or_eq:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_less);
    //                break;

    //            case s_is_null:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_not_null);
    //                break;
    //            case s_is_not_null:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_null);
    //                break;

    //            case s_is_true:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_not_true); // DDT: s_is_false ??
    //                break;
    //            case s_is_not_true:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_true);
    //                break;

    //            case s_is_false:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_not_false); // DDT: s_is_true ??
    //                break;
    //            case s_is_not_false:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_is_false);
    //                break;

    //            case s_like:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_notlike);
    //                break;
    //            case s_notlike:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_like);
    //                break;

    //            case s_sounds:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_notsounds);
    //                break;
    //            case s_notsounds:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_sounds);
    //                break;

    //            case s_in:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_notin);
    //                break;
    //            case s_notin:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_in);
    //                break;
    //        
    //            case s_between:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_notbetween);
    //                break;
    //            case s_notbetween:
    //                mSyntaxTree.UnlinkNode(StartNode, true);
    //                mSyntaxTree[Next].n_symb.becomes(s_between);
    //                break;
    //            }
    //        }
    //        break;

    //        //case s_is_not_false:
    //        //    mSyntaxTree[Next].n_symb.becomes(s_is_true);
    //        //    break;
    //        //case s_is_not_true:
    //        //    mSyntaxTree[Next].n_symb.becomes(s_is_false);
    //        //    break;
    //    }
    //}

    //if ( Next > 0 )
    //    NormalizeSearchConditions(Next);
    //
    //Next = mSyntaxTree[StartNode].n_sa_level;
    //while ( Next > 0 )
    //{
    //    NormalizeSearchConditions(Next);
    //    Next = mSyntaxTree[Next].n_sa_level;
    //}
    //return true;
}

// ------------------------------------------------------------------------------------------------