/*!***************************************************************************

  module      : Query_Graph.hpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#ifndef QUERY_GRAPH_H
#define QUERY_GRAPH_H

// ------------------------------------------------------------------------------------------------

#include "SAPDB/SQLManager/SQLMan_Types.hpp"
#include "SAPDB/SQLManager/Catalog/Catalog_Types.hpp"

#include "SAPDB/SQLManager/Query/Query_Types.hpp"
#include "SAPDB/SQLManager/Query/Query_ManipulationContext.hpp"
#include "SAPDB/SQLManager/Query/Query_SyntaxTree.hpp"
#include "SAPDB/SQLManager/Query/Query_Node.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeRoot.hpp"

#include "SAPDB/SAPDBCommon/SAPDB_Types.hpp"

#include "hak07.h"

// ------------------------------------------------------------------------------------------------

class Catalog_ViewDescriptor;
class SAPDBMem_IRawAllocator;

// ------------------------------------------------------------------------------------------------

struct Query_ComplexViewInfo
{
    SyntaxTreeNodeNo    SyntaxNode;
//    Query_Identifier    SchemaName;
//    Query_Identifier    TableName;
    Catalog_ObjectId    TabID;
	bool                ReferencedViaSynonym;
};

class Query_QueryGraph
{

public:

	Query_QueryGraph( Query_ManipulationContext& Context )
		: mContext(Context)
		, mAllocator(Context.GetAllocator())
		, mSyntaxTree(Context.SQLManContext())
		, mProgress(PG_Init)
		, mRootNode(0)
		, mCurrID(0)
		, mInitialSQLStatement(true)
		, mNeedOptimization(false)
		, mVTrace()
		, mError(false)
		, mNotYetImplemented(false)
		, mChanges(false)
		, mViewReplaced(false)
	{
		PrintToTrace();
	}

	inline ~Query_QueryGraph( void )
    {
        if ( mRootNode )
            destroy(mRootNode, mAllocator);
    }

    inline bool Changes( void )
    {
        return ( mChanges || mSyntaxTree.Changes() );
    }

	inline bool ViewReplaced( void )
	{
		return mViewReplaced;
	}

	inline void SetViewReplaced( void )
	{
		mViewReplaced = true;
	}

    bool Initialize( bool& NeedOptimization );

	inline void SetNeedOptimization( void ) { mNeedOptimization = true; }

    inline bool IsOk( void )
    {
        return !mError && !mNotYetImplemented;
    }

	inline void SetSyntaxError( SAPDB_Int2 BasisError, SAPDB_Int4 Code )
	{
		a07_b_put_error (mContext.SQLManContext(), BasisError, Code);

        if (QueryRewrite_Trace.TracesLevel(2))
            mVTrace << "Graph - SemanticError: " << mContext.SQLManContext().a_returncode << FlushLine;        
        mError = true;
        SetProgress(PG_Error, /*TraceGraph=*/ false);
	}

	inline void SetSemanticError( SAPDB_Int2 BasisError, SAPDB_Int4 Code )
	{
		a07_b_put_error (mContext.SQLManContext(), BasisError, Code);
			
        if (QueryRewrite_Trace.TracesLevel(2))
            mVTrace << "Graph - SemanticError: " << mContext.SQLManContext().a_returncode << FlushLine;        
		mError = true;
        SetProgress(PG_Error, /*TraceGraph=*/ false);
	}

    inline void SetError( char* ErrText, bool TraceGraph=true )
    {
        if (QueryRewrite_Trace.TracesLevel(2))
            mVTrace << "Graph - Error: " << ErrText << FlushLine;        
        mError = true;
        SetProgress(PG_Error, TraceGraph);
    }

    inline void SetNotYetImplemented( char* ErrText )
    {
        if (QueryRewrite_Trace.TracesLevel(2))
            mVTrace << "Graph - Not yet implemented: " << ErrText << FlushLine;        
        mNotYetImplemented = true;
    }

    inline void SetProgress ( Query_Progress P, bool TraceGraph=true )
    {
        if ( mProgress != P )
        {
            mProgress = P;
			if ( TraceGraph )
				PrintToTrace();
        }
    }

    void PrintToTrace( void );

    void PrintQueryToTrace( void );

    inline SAPDBMem_IRawAllocator& Allocator( void )
    {
        return mAllocator; 
    }

    inline Query_SyntaxTree& SyntaxTree( void )
    {
        return mSyntaxTree;
    }

    SQLMan_Identifier GetIdentifier( SyntaxTreeNodeNo NodeNo );
    SQLMan_Identifier GetSchemaName( SyntaxTreeNodeNo& NodeNo );

    void CreateQueryNodes( SyntaxTreeNodeNo NodeNo, Query_Node* ActNode, Query_Position Position, bool ComplexViewReplacement, SAPDB_Int2 SubQueryType );

    bool SubstituteComplexView( Query_ComplexViewInfo& ComplexView, SyntaxTreeNodeNo& ColNameNodes, bool IsComplexView );

	bool AddExpressionRefName( SyntaxTreeNodeNo Node, SAPDB_Int2& ExprCount );

	inline void SetNodeID( Query_Node* B )
    {
        B->SetNodeID(mCurrID++);
    }

private:

    Query_Node* AddSelectNode( Query_Node* ActNode, Query_Position Position, SyntaxTreeNodeNo NodeNo, bool ComplexViewReplacement, SAPDB_Int2 SubQueryType );
	Query_Node* AddSetNode( Query_Node* ActNode, Query_Position Position, Query_Position SetType, SyntaxTreeNodeNo NodeNo, bool All );
	Query_Node* AddTableNode( Query_Node* ActNode, Query_Position Position, SyntaxTreeNodeNo NodeNo );

	bool AddSubSelectNode( SyntaxTreeNodeNo StartNode, Query_Position Position, SAPDB_Int2 SubQueryType );
	inline bool SubSelectNodeMustBeAdded( Query_Position Position );

    inline Query_QuantifierType DetermineQuantifierType( Query_Position Position );

    // Functions used by SubstituteComplexView
    bool SCVCreateInternalPacket( SQLMan_Context& ViewContext, SAPDB_Int4 ViewStmtLength );
    bool SCVReleaseInternalPacket( SQLMan_Context& ViewContext );
	void* SCVSaveContextStmtPart( SQLMan_Context& ViewContext );
	bool SCVRestoreContextStmtPart( SQLMan_Context& ViewContext, void* SavedPart );
    bool SCVParseViewStatement( SQLMan_Context& ViewContext, Catalog_ViewDescriptor& ViewDescr, bool IsComplexView );
    bool SCVCheckIfViewCanBeSubstituted( Query_SyntaxTree& NewSyntaxTree, SyntaxTreeNodeNo& ViewNameNode, SyntaxTreeNodeNo& ViewSelectNode, SyntaxTreeNodeNo& ColNameNodes );
    bool SCVLinkSyntaxTrees( SyntaxTreeNodeNo FromTabNode, SyntaxTreeNodeNo ViewSyntaxTreeNode, SyntaxTreeNodeNo ViewNameNode, bool ReferencedViaSynonym );

    bool NormalizeSearchConditions( SyntaxTreeNodeNo StartNode );
    bool NegotiateChildParameters( SyntaxTreeNodeNo Node );

    Query_ManipulationContext&              mContext;
    SAPDBMem_IRawAllocator&                 mAllocator;

    Query_SyntaxTree                        mSyntaxTree;
    Query_Progress                          mProgress;

    Query_RootNode*                     mRootNode;

    Kernel_VTrace                           mVTrace;

    SAPDB_UInt2                             mCurrID;

    bool                              mInitialSQLStatement;
    bool                              mNeedOptimization;
    /// mError is set to TRUE if the AP-Tree is damaged in some way that requires reparsing of the initial sql statement !!!
    bool                              mError;
    bool                              mNotYetImplemented;
    bool                              mChanges;
	bool							  mViewReplaced;

};

// ------------------------------------------------------------------------------------------------

inline bool Query_QueryGraph::SubSelectNodeMustBeAdded( Query_Position Position )
{
	if ( //Position == P_SelectList ||
		 Position == P_FromPart  ||
		 Position == P_Where ||
		 Position == P_WhereAll ||
		 Position == P_Having ||
		 Position == P_HavingAll 
		)
		return true;
	return false;
}

inline Query_QuantifierType Query_QueryGraph::DetermineQuantifierType( Query_Position Position )
{
    switch ( Position ) 
    {
    case P_FromPart: 
        return QT_F;
        break;
    case P_Where: 
    case P_Having: 
        return QT_E;
        break;
    case P_WhereAll: 
    case P_HavingAll: 
	case P_Start: // DDT : ???
        return QT_A;
        break;
    case P_SelectList:
        return QT_L;
        break;
    default:
        return QT_Unknown;
        break;
    }
    return QT_Unknown;
}

// ------------------------------------------------------------------------------------------------



// ------------------------------------------------------------------------------------------------

#endif