/*!***************************************************************************

  module      : Query_ManipulationContext.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Query/Query_ManipulationContext.hpp"
#include "SAPDB/SQLManager/Query/Query_SyntaxTree.hpp"

#include "hak542.h"
#include "hak99.h"

#include "SAPDB/SQLManager/SharedSQL/SharedSQL_Handles.hpp"

// ------------------------------------------------------------------------------------------------

Query_ManipulationContext::Query_ManipulationContext( SQLMan_Context& SQLManContext, SharedSQL_PrepareHandle* prepareHandle )
    : mSQLManContext(SQLManContext)
	, mPrepareHandle(prepareHandle)
    , mVTrace()
    , mOldAPTreePtr(0)
    , mOldScvIndex(0)
    , mOldMaxNode(0)
    , mOldAPMaxNode(0)
    , mSQLStatementBuffer(0)
    , mSQLStatementBufferMaxSize(0)
    , mSQLStatementBufferUsedSize(0)
    , mSQLStatementBufferChanges(false)
    , mDecision(false)
    , mError(false)
    , mChanges(false)
{
    SAPDBERR_ASSERT_STATE(  SQLManContext.a_ap_tree == SQLManContext.a_session_ap_tree );
    CreateNewSyntaxTreeAndStatementBuffer();
}

Query_ManipulationContext::~Query_ManipulationContext( void )
{
//    SAPDBERR_ASSERT_STATE( mDecision );

    if ( !mDecision )
        RestoreOldSyntaxTreeAndStatementBuffer();

	mSQLManContext.a_queryrewrite_modified = mChanges;
}

SAPDBMem_IRawAllocator& Query_ManipulationContext::GetAllocator( void ) const
{
    return mSQLManContext.GetAllocator();
}

void* Query_ManipulationContext::GetSQLStatementPtr( void ) const
{
    if ( mSQLStatementBuffer )
        return mSQLStatementBuffer;
    else
        return mSQLManContext.GetSQLStatementPtr();
}

int Query_ManipulationContext::GetSQLStatementSize( void ) const
{
    if ( mSQLStatementBuffer )
        return mSQLStatementBufferUsedSize;
    else
        return mSQLManContext.GetSQLStatementSize();
}

bool Query_ManipulationContext::AppendBlankToSQLStatement( void )
{
    if ( g01unicode )
        return (AppendToSQLStatement((char*)csp_unicode_blank, 2)>=0);
    else
        return (AppendToSQLStatement((char*)csp_ascii_blank, 1)>=0);
}

SAPDB_Int4 Query_ManipulationContext::AppendToSQLStatement( char* ToAdd )
{
    SAPDB_Int4 Len = strlen(ToAdd);
    return AppendToSQLStatement(ToAdd, Len);
}

SAPDB_Int4 Query_ManipulationContext::AppendToSQLStatement( void* ToAdd, SAPDB_Int4 AddSize )
{
    SAPDBERR_ASSERT_ARGUMENT( ToAdd && (AddSize>0) );

    SAPDB_Int4 RequiredSize = mSQLStatementBufferUsedSize + AddSize;
    if ( RequiredSize > mSQLStatementBufferMaxSize )
    {
        if ( ! ExtendSQLStatementBuffer(RequiredSize) )
            return -1;
    }
    SAPDB_Int4 pos = mSQLStatementBufferUsedSize;
    SAPDB_Byte* Position = mSQLStatementBuffer + pos;
    SAPDB_MemCopyNoCheck(Position, ToAdd, AddSize);
    mSQLStatementBufferUsedSize += AddSize;
    mSQLStatementBufferChanges = true;
    return pos;
}

bool Query_ManipulationContext::ExtendSQLStatementBuffer( SAPDB_Int4 RequiredSize )
{
    SAPDBERR_ASSERT_ARGUMENT( RequiredSize>0 );

    SAPDB_Int2 BlockSize    = 8192;
    SAPDB_Int4 Blocks       = (RequiredSize/BlockSize) 
                            + (RequiredSize%BlockSize?1:0);
    SAPDB_Int4 NewSize      = Blocks * BlockSize; 

    if ( NewSize > mSQLStatementBufferMaxSize )
    {
        SAPDB_Byte* NewBuffer = (SAPDB_Byte*) mSQLManContext.GetAllocator().Reallocate(NewSize, mSQLStatementBuffer);
        if ( NewBuffer )
        {
            mSQLStatementBuffer         = NewBuffer;
            mSQLStatementBufferMaxSize  = NewSize;
        }
    }
    return ( mSQLStatementBuffer!=0 );
}

bool Query_ManipulationContext::Changes( void ) const
{
    return ( mSQLStatementBufferChanges );
}

void Query_ManipulationContext::CreateNewSyntaxTreeAndStatementBuffer( void )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_ManipulationContext::CreateNewSyntaxTreeAndStatementBuffer", QueryRewrite_Trace, 5);

#if defined(SAPDB_SLOW)
    if ( QueryRewrite_Trace.TracesLevel(1) )
    {
        mVTrace << "Query_ManipulationContext : Original SyntaxTree" << FlushLine;
        a99print_nodes(*mSQLManContext.a_ap_tree, mSQLManContext.a_scv_index, mSQLManContext.a_first_hint_node);
    }
#endif

    // Save values for the old syntax tree
    mOldAPTreePtr       = (tak_ap_node*) mSQLManContext.a_ap_tree;
    mOldScvIndex        = mSQLManContext.a_scv_index;
    mOldMaxNode         = mSQLManContext.a_maxnode;
    mOldAPMaxNode       = mSQLManContext.a_ap_max_node;
    mOldFirstHintNode   = mSQLManContext.a_first_hint_node;

    // Create new syntax tree
	mSQLManContext.a_ap_tree			= 0; // to insure NewSyntaxTree works...
    mSQLManContext.NewSyntaxTree(256);
    mSQLManContext.a_scv_index          = 0;
    mSQLManContext.a_maxnode            = 0;
    mSQLManContext.a_ap_max_node        = 0;
    mSQLManContext.a_first_hint_node    = csp_minint2;

    Query_SyntaxTree ActSyntaxTree = Query_SyntaxTree(mSQLManContext);

    // Copy old syntax tree
    ActSyntaxTree[0] = mOldAPTreePtr[0];
    ActSyntaxTree[0].n_lo_level =
        ActSyntaxTree.DoubleTreePart(mOldAPTreePtr, mOldAPTreePtr[0].n_lo_level);
    if ( mOldFirstHintNode != csp_minint2 )
    {
        SyntaxTreeNodeNo HintNodes = ActSyntaxTree.DoubleTreePart(mOldAPTreePtr, mOldFirstHintNode);
        mSQLManContext.a_first_hint_node = HintNodes;
    }

    // New StatementBuffer
    void* ToAdd             = mSQLManContext.GetSQLStatementPtr();
    SAPDB_Int4  AddLength   = mSQLManContext.GetSQLStatementSize();
    if ( AppendToSQLStatement(ToAdd, AddLength) < 0 )
        mError = true;
    mSQLStatementBufferChanges = false;
    mChanges = true;
    mSQLManContext.a_queryrewrite_modified = mChanges;
}

bool Query_ManipulationContext::RestoreOldSyntaxTreeAndStatementBuffer( void )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_ManipulationContext::RestoreOldSyntaxTreeAndStatementBuffer", QueryRewrite_Trace, 5);

    // Restore old syntax tree
    mSQLManContext.DeleteSyntaxTree();
    mSQLManContext.a_ap_tree            = (tak_ap_max_tree_ptr) mOldAPTreePtr;
    mSQLManContext.a_scv_index          = mOldScvIndex;
    mSQLManContext.a_maxnode            = mOldMaxNode;
    mSQLManContext.a_ap_max_node        = mOldAPMaxNode;
    mSQLManContext.a_first_hint_node    = mOldFirstHintNode;

    // Restore old StatementBuffer
    // ... there is nothing to do but to delete the new one
    if ( mSQLStatementBuffer )
        mSQLManContext.GetAllocator().Deallocate(mSQLStatementBuffer);

    if ( QueryRewrite_Trace.TracesLevel(1) )
    {
        mVTrace << "Query_ManipulationContext : Restore old context" << FlushLine;
#if defined(SAPDB_SLOW)
        if ( QueryRewrite_Trace.TracesLevel(2) )
            a99print_nodes(*mSQLManContext.a_ap_tree, mSQLManContext.a_scv_index, mSQLManContext.a_first_hint_node);
#endif
    }

	// in the end we have to clear the tableinfo added within Query_TableNode::Semantcis
	if ( mPrepareHandle )
		mPrepareHandle->ClearTables();
	
	mChanges = false;
	mSQLManContext.a_queryrewrite_modified = mChanges;
//	mSQLManContext.a_queryrewrite_done.becomes(No_Rewrite);
    if ( QueryRewrite_Trace.TracesLevel(2) )
	{
		// MessageListe ausgeben... !
	}
	mSQLManContext.SetTransInfoWarning(warn9_stmt_optimized,		/*on=*/false);
	mSQLManContext.SetTransInfoWarning(warn10_views_substituted,	/*on=*/false);
#if defined(SAPDB_SLOW)
	mSQLManContext.SetTransInfoWarning(warn15_user_defined_code,    /*on=*/true );
#endif
	mSQLManContext.ResetReturnCode();
    return true;
}

bool Query_ManipulationContext::KeepNewSyntaxTreeAndStatementBuffer( void )
{
    // Keep new syntax tree
    // ... there is nothing to do here

    // Keep new statement buffer if necessary

	// DDT : CREATE ALWAYS A NEW PACKET DURING FIRST TEST PHASE	
//    if ( mSQLStatementBufferChanges )
    {
        // ... create a new internal packet and copy the statement buffer

		SAPDB_Int4 AddSize = 0;
		if ( mSQLManContext.a_return_segm->sp1r_function_code() == csp1_explain_fc )
			AddSize = (mSQLManContext.a_cmd_part->sp1p_buf_size() - mSQLManContext.a_cmd_part->sp1p_buf_len());

        a542internal_packet (mSQLManContext, /*release_internal_packet*/false, mSQLStatementBufferUsedSize+AddSize);
        if ( mSQLManContext.IsOk() ) // DDT
        {
            void* Dest = mSQLManContext.a_cmd_part->sp1p_buf();
            SAPDB_MemCopyNoCheck(Dest, mSQLStatementBuffer, mSQLStatementBufferUsedSize);
            mSQLManContext.a_cmd_part->sp1p_buf_len() = mSQLStatementBufferUsedSize;
            mSQLManContext.a_cmd_segm->sp1c_prepare() = mSQLManContext.a_initial_segment_header.sp1c_prepare();

//			if ( mSQLManContext.a_cmd_segment_header.sp1c_producer() == sp1pr_user_cmd )
	            mSQLManContext.a_cmd_segm->sp1c_producer().becomes(sp1pr_queryrewrite);
//			else
//	            mSQLManContext.a_cmd_segm->sp1c_producer().becomes(mSQLManContext.a_cmd_segment_header.sp1c_producer());
    
			mSQLManContext.a_cmd_segment_header       = mSQLManContext.a_cmd_segm->sp1s_segm_header();
            mSQLManContext.a_cmd_segment_header.sp1c_producer().becomes(sp1pr_queryrewrite);
        }
        else
        {
            mSQLManContext.ResetReturnCode();
            mError = true;
            RestoreOldSyntaxTreeAndStatementBuffer();
        }
    }
    if ( mSQLStatementBuffer )	// we have to delete the tempary buffer in any case
        mSQLManContext.GetAllocator().Deallocate(mSQLStatementBuffer);

#if defined(SAPDB_SLOW)
    if ( QueryRewrite_Trace.TracesLevel(1) )
    {
        mVTrace << "Query_ManipulationContext : Keep new Context" << FlushLine;
        if ( QueryRewrite_Trace.TracesLevel(1) )
            a99print_nodes(*mSQLManContext.a_ap_tree, mSQLManContext.a_scv_index, mSQLManContext.a_first_hint_node);
    }
#endif

	mSQLManContext.ResetReturnCode();
    return true;
}


