/*!***************************************************************************

  module      : Query_Node.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Query/Query_Node.hpp"
#include "SAPDB/SQLManager/Query/Query_NodeSelect.hpp"
#include "SAPDB/SQLManager/Query/Query_Graph.hpp"
#include "SAPDB/SQLManager/Query/Query_Quantifier.hpp"

// ------------------------------------------------------------------------------------------------

Query_Node::Query_Node( Query_ManipulationContext& Context, Query_QueryGraph& QueryGraph, 
                        Query_Position Position, SyntaxTreeNodeNo APNode,
                        bool HeadDistinct, Query_DistinctType BodyDistinctType ) 
    : mContext(Context)
    , mAllocator(Context.GetAllocator())
    , mQueryGraph(QueryGraph) 
    , mAPNode(APNode)
    , mID(0)
    , mFQuantifierCnt(0)
    , mWQuantifierCnt(0)
    , mPosition(Position)
    , mFatherNode(0)
    , mFatherQuantifier(0)
    , mQuantifier(Context.GetAllocator())
    , mHeadDistinct(HeadDistinct)
    , mBodyDistinctType(BodyDistinctType)
    , mSemanticsDone(false)
	, mUnidentifiedColumnsInChildNodes(false)
{
    mQuantifier.Reserve(5);
    mQueryGraph.SetNodeID(this);
}

// copy constructor
Query_Node::Query_Node( const Query_Node& N, Query_Node* Father, Query_Quantifier* Q, Query_NodeMapper* mpMap )
    : mContext(N.mContext) 
    , mAllocator(N.mAllocator)
    , mQueryGraph(N.mQueryGraph) 
    , mAPNode(0)                                    //
    , mID(0)                                        //
    , mFQuantifierCnt(0)                            // will be set in AddQuantifier
    , mWQuantifierCnt(0)                            // will be set in AddQuantifier
    , mPosition(N.mPosition)
    , mFatherNode(0)                            //
    , mFatherQuantifier(0)                             //
    , mQuantifier(N.mAllocator)  //
    , mHeadDistinct(N.mHeadDistinct)
    , mBodyDistinctType(N.mBodyDistinctType)
    , mSemanticsDone(N.mSemanticsDone)
	, mUnidentifiedColumnsInChildNodes(N.mUnidentifiedColumnsInChildNodes)
{
    // Create an own id
    mQueryGraph.SetNodeID(this);

    // Map the mAPNode
    mQueryGraph.SyntaxTree().MapNode(mAPNode, mpMap);

    // Set the linkage to topnode and topquantifier
    mFatherNode = Father;
    mFatherQuantifier	= Q;

    // Copy the quantifiers
    SAPDB_Int2 m = N.mQuantifier.GetSize();
    mQuantifier.Reserve(m);
    for ( SAPDB_Int2 i=0; i<m; m++ )
        AddQuantifier((mQuantifier[i])->CreateDuplicate(this, mpMap));
}

Query_Node::~Query_Node( void )
{
    unsigned qc = mQuantifier.GetSize();
    for ( unsigned i = 0; i < qc; i++ )
    {
        if ( mQuantifier[i] != 0 )
            destroy(mQuantifier[i], mAllocator);
        mQuantifier[i] = 0;
    }
}

// ------------------------------------------------------------------------------------------------

bool Query_Node::IllegalUseOfMethod( char* MethodName )
{
	mQueryGraph.SetError(MethodName, /*TraceGraph*/ false);
	return false;
}

// ------------------------------------------------------------------------------------------------

void Query_Node::PrintToTrace( Kernel_VTrace& VTrace )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::PrintToTrace", QueryRewrite_Trace, 8);

    if ( QueryRewrite_Trace.TracesLevel(2) )
    {
        VTrace << "+-------------------------------" << FlushLine;
        VTrace << "NodeID     : " << mID << FlushLine;
        VTrace << "FatherID   : " << (mFatherNode  ? mFatherNode->NodeID()  : cak_qrewrite_ref_init) << FlushLine;
        for ( unsigned j=0; j < mQuantifier.GetSize(); j++ )
            VTrace << "ChildNodeID: " << mQuantifier[j]->NodeRef()->NodeID() << FlushLine;
        VTrace << "APNode     : " << mAPNode << FlushLine;

        switch ( mPosition )
        {
        case P_Start:
            VTrace << "Position   : Start" << FlushLine;
            break;
        case P_Union:
            VTrace << "Position   : Union" << FlushLine;
            break;
        case P_Difference:
            VTrace << "Position   : Difference" << FlushLine;
            break;
        case P_Intersect:
            VTrace << "Position   : Intersect" << FlushLine;
            break;
        case P_Select:
            VTrace << "Position   : Select" << FlushLine;
            break;
        case P_SelectList:
            VTrace << "Position   : SelectList" << FlushLine;
            break;
        case P_FromPart:
            VTrace << "Position   : FromPart" << FlushLine;
            break;
        case P_Where:
            VTrace << "Position   : Where" << FlushLine;
            break;
        case P_WhereAll:
            VTrace << "Position   : WhereAll" << FlushLine;
            break;

        case P_Having:
            VTrace << "Position   : Having" << FlushLine;
            break;
        case P_OrderBy:
            VTrace << "Position   : OrderBy" << FlushLine;
            break;
        case P_GroupBy:
            VTrace << "Position   : GroupBy" << FlushLine;
            break;
        case P_Unknown:
            VTrace << "Position   : Unkown ???" << FlushLine;
            break;
        default:
            break;
        }
        //if ( mFatherQuantifier )
        //{
        //    VTrace << "Output Cols: " << mFatherQuantifier->OutputColumns() << FlushLine;
        //}
    }
}

void Query_Node::PrintChildNodesToTrace( Kernel_VTrace& VTrace )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::PrintChildNodesToTrace", QueryRewrite_Trace, 8);

    SAPDB_Int2 n = mQuantifier.GetSize();
    for ( SAPDB_Int2 i=0; i < n; i++ )
        mQuantifier[i]->NodeRef()->PrintToTrace(VTrace);
}

// ------------------------------------------------------------------------------------------------

Query_Node* Query_Node::GetFirstChildNode( SAPDB_Int2& ChildCount )
{
	ChildCount = mQuantifier.GetSize();
	if ( ChildCount > 0 )
		return mQuantifier[0]->NodeRef();
	else
		return 0;
}

// ------------------------------------------------------------------------------------------------

bool Query_Node::Semantics( bool UniqueOutputColumnNames, bool ComplexViewReplacement, bool& UnidentifiedColumns )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::Semantics", QueryRewrite_Trace, 6);

    SAPDB_Int2 n = mQuantifier.GetSize();
    for ( SAPDB_Int2 i=0; (i<n) && mQueryGraph.IsOk(); i++ )
    {
        Query_Quantifier* Q =   mQuantifier[i];
        bool Unique         =   (Q->QuantifierType()==QT_F) || 
                                UniqueOutputColumnNames;
		bool Unidentified = false;
        if ( !Q->NodeRef()->Semantics(Unique, ComplexViewReplacement, Unidentified) ) 
            return false;
		mUnidentifiedColumnsInChildNodes |= Unidentified;
    }
	UnidentifiedColumns |= mUnidentifiedColumnsInChildNodes;
    return mQueryGraph.IsOk();
}

bool Query_Node::Semantics_Correlation( Query_SelectNode* TopNode, bool& UnidentifiedColumns )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::Semantics_Correlation", QueryRewrite_Trace, 6);

	if ( mUnidentifiedColumnsInChildNodes )
	{
		SAPDB_Int2 n = mQuantifier.GetSize();
		bool Unidentified = false;
		for ( SAPDB_Int2 i=0; (i<n) && mQueryGraph.IsOk(); i++ )
		{
			Query_Quantifier* Q =   mQuantifier[i];
			if ( !Q->NodeRef()->Semantics_Correlation(TopNode, Unidentified) ) 
				return false;
		}
		if ( !Unidentified )
			mUnidentifiedColumnsInChildNodes = false;
		else
			UnidentifiedColumns |= true;

	    return mQueryGraph.IsOk();
	}
	return true;
}

bool Query_Node::ReplaceComplexViews( void )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::ReplaceComplexViews", QueryRewrite_Trace, 6);

    SAPDB_Int2 n = mQuantifier.GetSize();
    for ( SAPDB_Int2 i=0; (i<n) && mQueryGraph.IsOk(); i++ )
    {
        if ( !mQuantifier[i]->ReplaceComplexViews() ) 
            return false;
    }
    return mQueryGraph.IsOk();
}

// ------------------------------------------------------------------------------------------------

bool Query_Node::GetQuantifierNoFromSyntaxNode( SyntaxTreeNodeNo Node, SAPDB_Int2& QNo )
{
    SAPDB_Int2 n = mQuantifier.GetSize();
    for ( QNo=0; QNo<n; QNo++ )
    {
        if ( mQuantifier[QNo]->NodeRef()->SyntaxNode() == Node ) 
            return true;
    }
	return false;
}

// ------------------------------------------------------------------------------------------------

bool Query_Node::AddQuantifier ( Query_Quantifier* Q )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::AddQuantifier", QueryRewrite_Trace, 5);

    Q->NodeRef()->FatherNode() = this;
    mQuantifier.InsertEnd(Q);

    if ( Q->QuantifierType() == QT_F )
        ++mFQuantifierCnt; 

    if ( Q->NodeRef()->Position() == P_Where || Q->NodeRef()->Position() == P_WhereAll )
        ++mWQuantifierCnt; 

    return true;
}

bool Query_Node::AddQuantifier ( Query_Node* Node, Query_DistinctType DistinctType, Query_Position Position, Query_QuantifierType QuantifierType )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_Node::AddQuantifier", QueryRewrite_Trace, 5);
    SAPDBERR_ASSERT_ARGUMENT( Node != 0 );

    Query_Quantifier* Quantifier = new(mAllocator) Query_Quantifier(Node, DistinctType, QuantifierType);
    Node->FatherNode() = this;
    if ( Quantifier )
    {
        Node->FatherQuantifier() = Quantifier;
        mQuantifier.InsertEnd(Quantifier);

        if ( QuantifierType == QT_F )
            ++mFQuantifierCnt; 

        if ( Position == P_Where ||Position == P_WhereAll )
            ++mWQuantifierCnt; 
        return true;
    }
    mQueryGraph.SetError("AddQuantifier");
    return false;
}

void Query_Node::DiscardDisabledQuantifiers( void )
{
    SAPDB_Int2 n = mQuantifier.GetSize();
    for ( SAPDB_Int2 i=n-1; i>=0; i-- )
    {
        if ( mQuantifier[i]->Disabled() )
        {
			destroy(mQuantifier[i], mAllocator);
			mQuantifier[i] = 0;
            mQuantifier.Delete(i);
        }
    }
}

// ------------------------------------------------------------------------------------------------
