/*!***************************************************************************

  module      : Query_NodeTable.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Query/Query_NodeTable.hpp"
#include "SAPDB/SQLManager/Query/Query_Graph.hpp"
#include "SAPDB/SQLManager/Query/Query_Quantifier.hpp"

#include "SAPDB/SQLManager/SharedSQL/SharedSQL_Handles.hpp"

// ------------------------------------------------------------------------------------------------

#define QRW_REPLACE_ALL_KIND_OF_VIEWS true

// ------------------------------------------------------------------------------------------------

Query_TableNode::Query_TableNode( Query_ManipulationContext& Context, Query_QueryGraph& QueryGraph, Query_Position Position, SyntaxTreeNodeNo APNode )
    : Query_Node(Context, QueryGraph, Position, APNode, false, DT_Unknown)
	, mSchemaName()
	, mTableName()
    , mReferenceName()
    , mTableDescriptor()
    , mOutputColumnCnt(0)
    , mTableType(TT_Unknown)
    , mExplicitSchema(false)
	, mSelectPrivForAllColumns(false)
	, mReferencedViaSynonym(false)
{
}

// copy constructor
Query_TableNode::Query_TableNode( const Query_TableNode& N, Query_Node* Father, Query_Quantifier* Q, Query_NodeMapper* mpMap )
    : Query_Node(N, Father, Q, mpMap) 
{
}

Query_Node* Query_TableNode::CreateDuplicate( Query_Node* Father, Query_Quantifier* Q, Query_NodeMapper* mpMap )
{
    return new(mAllocator) Query_TableNode(*this, Father, Q, mpMap);
}

// ------------------------------------------------------------------------------------------------

void Query_TableNode::PrintToTrace( Kernel_VTrace& VTrace )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_TableNode::PrintToTrace", QueryRewrite_Trace, 8);

    Query_Node::PrintToTrace(VTrace);
    if ( QueryRewrite_Trace.TracesLevel(2) )
    {
        VTrace << "NodeType  : TableAccess" << FlushLine;
        if ( mSemanticsDone )
        {
            switch ( TableType() )
            {
            case TT_BaseTable:
                VTrace << "Tabletype : BaseTable" << FlushLine;
                break;
            case TT_Cursor:
                VTrace << "Tabletype : Cursor" << FlushLine;
                break;
            case TT_Synonym:
                VTrace << "Tabletype : Synonym" << FlushLine;
                break;
            case TT_View:
                VTrace << "TableType : View" << FlushLine;
                break;
            case TT_JoinView:
                VTrace << "TableType : JoinView" << FlushLine;
                break;
            case TT_ComplexView:
                VTrace << "TableType : ComplexView" << FlushLine;
                break;
            case TT_Unknown:
                VTrace << "TableType : Unknown ???" << FlushLine;
                break;
            }
            VTrace << "SchemaName  : " << mTableDescriptor.GetSchemaName() << FlushLine; 
            VTrace << "TableName   : " << mTableDescriptor.GetTableName() << FlushLine;
            VTrace << "Output Cols : " << mOutputColumnCnt << FlushLine; 
        }
    }
    if ( QueryRewrite_Trace.TracesLevel(1) && !QueryRewrite_Trace.TracesLevel(2) )
    {
        if ( mSemanticsDone )
            VTrace << "[" << mID << "] " // << mTableDescriptor.GetSchemaName() << "." 
                                        << mTableDescriptor.GetTableName() << FlushLine;
        else
            VTrace << "[" << mID << "] 'Table'" <<  FlushLine;
    }
}

// ------------------------------------------------------------------------------------------------

bool Query_TableNode::Semantics( bool UniqueOutputColumnNames, bool ComplexViewReplacement, bool& UnidentifiedColumns )
{
    SAPDBTRACE_METHOD_DEBUG ("Query_TableNode::Semantics", QueryRewrite_Trace, 6);

    if ( !mQueryGraph.IsOk() )
        return false;

    if ( !mTableDescriptor )
    {
        Query_SyntaxTree& SyntaxTree = mQueryGraph.SyntaxTree();
        SyntaxTree[mAPNode].n_querycheck.addElement(qcn_Semantic);

        SyntaxTreeNodeNo Node  = SyntaxTree[mAPNode].n_lo_level;

        if ( SyntaxTree[Node].n_proc == no_proc && SyntaxTree[Node].n_symb == s_authid )
        {
            mExplicitSchema = true;
            mSchemaName = mQueryGraph.GetSchemaName(Node);
            mTableName  = mQueryGraph.GetIdentifier(Node);
            mTableDescriptor = Catalog_TableDescriptor(mContext.SQLManContext(), mSchemaName, mTableName, /*ConsiderResultTables=*/ true, /* ConsiderSYSDD */ true);
        }
        else
        {
            mExplicitSchema = false;
            mSchemaName = SQLMan_Identifier();
            mTableName  = mQueryGraph.GetIdentifier(Node);
            mTableDescriptor = Catalog_TableDescriptor(mContext.SQLManContext(), mSchemaName, mTableName, /*ConsiderResultTables=*/ true, /* ConsiderSYSDD */ true);
        }
        if ( mTableDescriptor )
        {
			mReferencedViaSynonym = mTableDescriptor.ReferencedViaSynonym();
			if ( (Node = SyntaxTree[Node].n_sa_level ) > 0 )
			{
				if ( SyntaxTree[Node].n_proc == no_proc && SyntaxTree[Node].n_symb == s_identifier )
					mReferenceName = mQueryGraph.GetIdentifier(Node);
			}

            mTableType = TT_BaseTable;
            if ( mTableDescriptor.IsSynonym() )
                mTableType = TT_Synonym;
            if ( mTableDescriptor.IsResultTable() )
                mTableType = TT_Cursor;
            if ( mTableDescriptor.IsView() )
                mTableType = TT_View;
            if ( mTableDescriptor.IsJoinView() )
                mTableType = TT_JoinView;
            if ( mTableDescriptor.IsComplexView() )
                mTableType = TT_ComplexView;

            if ( mTableDescriptor.HasAllSelectPrivileges() )
				mSelectPrivForAllColumns = true;
			else
				return false; // DDT : unable to handle this case so far...

            mOutputColumnCnt = mTableDescriptor.GetColumnCountForStarResolving(); 
			//                 ^ this will do the correct counting even if there are no rights for all columns

            if ( mOutputColumnCnt == 0 )
            {
                // DDT : No sufficient privileges...
                return false;
            }

			SyntaxTree[mAPNode].n_querycheck.addElement(qcn_RightsChecked);

			// if we are not within a complex view replacement and we are parsing for sharedsql we have to
			// add information for SharedSQL
			SQLMan_Context& context = mContext.SQLManContext();
	        SharedSQL_PrepareHandle* pPrepareHandle = mContext.GetPrepareHandle();
			if ( !ComplexViewReplacement && pPrepareHandle )
			{
				SharedSQL_Table SSQLTable;
				if ( mExplicitSchema )
					SSQLTable = SharedSQL_Table(mTableDescriptor.GetSchemaId(), mTableDescriptor.GetTableName());
				else
					SSQLTable = SharedSQL_Table(SharedSQL_Schema(), mTableDescriptor.GetTableName());

				pPrepareHandle->PutTable(SSQLTable, context.GetErrorList());
			}
			//

            mSemanticsDone = true;
			return true;
        }
        mQueryGraph.SetError("NoValidTableDescriptor");
        return false;
    }
    return true;
}

bool Query_TableNode::SelectPrivilegesForAllColumns( void )
{
	return mSelectPrivForAllColumns;
}

bool Query_TableNode::HasSysKey( void )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

	return mTableDescriptor.HasSysKey();
}


// ------------------------------------------------------------------------------------------------

bool Query_TableNode::ReplaceComplexViews( void )
{
	if ( (mTableType != TT_ComplexView)  
#if defined(QRW_REPLACE_ALL_KIND_OF_VIEWS)
		 && (mTableType != TT_JoinView)  
		 && (mTableType != TT_View) 
#endif
		)
        return true;

    // replace 'this' !

    Query_ComplexViewInfo VInfo;
    VInfo.SyntaxNode			= SyntaxNode();
    VInfo.TabID					= TabID();
	VInfo.ReferencedViaSynonym	= mReferencedViaSynonym;
//    if ( TableName(VInfo.SchemaName, VInfo.TableName) )
    {
//		Query_NodeType NodeType = NT_Unknown;
		SyntaxTreeNodeNo ColNameNodes = cak_qrewrite_ref_init;
        if ( mQueryGraph.SubstituteComplexView(VInfo, ColNameNodes, mTableType==TT_ComplexView) )
		{
			mContext.SQLManContext().SetTransInfoWarning(warn10_views_substituted, /*on=*/true);
			mQueryGraph.SetNeedOptimization();

			// Call CreateQueryNodes for the new tree part
			SyntaxTreeNodeNo A631Node = mQueryGraph.SyntaxTree()[mAPNode].n_lo_level;

			Query_RootNode HelpRoot(this);
			mQueryGraph.CreateQueryNodes( A631Node, &HelpRoot, P_Start, /*ComplexViewReplacement=*/true, /*SubQueryType*/cak_x_subquery );

			SAPDB_Int2 ChildCount = 0;
			Query_Node* NewNode = HelpRoot.GetFirstChildNode(ChildCount);

			if ( mQueryGraph.IsOk() && ChildCount==1 )
			{
				// clear reference of HelpRoot to NewNode
				NewNode->FatherQuantifier()->SetNodeRef(0); 

				// link NewNode with my FatherNode correctly
				NewNode->FatherNode()         = this->FatherNode();
				NewNode->FatherQuantifier()	= this->FatherQuantifier();
				NewNode->FatherQuantifier()->SetNodeRef(NewNode);	
				NewNode->SetNodeID(NodeID()); // we wan't the same ID as the Node that is replaced...

				// Call semantics for the new tree part
				// we don't need to check for unique output columns as the view must be defined that way!
				bool UnidentifiedColumns = false;
				if ( NewNode->Semantics(/*UniqueOutputColumns=*/false, /*ComplexViewReplacement=*/true, UnidentifiedColumns) )
				{		
					if ( UnidentifiedColumns )
						mQueryGraph.SetError("Unidentified Columns in substituted View");

					if ( ColNameNodes > 0 )
					{
						// DDT : asterix must be replaced if there is a volumn list with the view name
						// e.g.  create view v(A,B,C) as select t1.*, t2.* from s1.t1, t2
						//                     ^              ^
						if ( !NewNode->CreateColRefNamesAndResolveAsterix(ColNameNodes) )
							mQueryGraph.SetError("ReplaceComplexViews"); 
					}
					mQueryGraph.SetViewReplaced();

					// Continue replacing complex views 
					NewNode->ReplaceComplexViews();
				}
				
				bool rc = mQueryGraph.IsOk();	// we cannot access mQueryGraph after destroying this anymore...
				Query_TableNode* T = this;
				destroy(T, mAllocator);  
				return rc;
			}
		}
    }
	mQueryGraph.SetError("ReplaceComplexViews"); 
	return false;
}

bool Query_TableNode::IsColumnOfTypeLong( SAPDB_Int2 ColumnIndex )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

#if defined(QRW_REPLACE_ALL_KIND_OF_VIEWS)
		if ( (mTableType != TT_ComplexView) &&  
		     ((mTableType == TT_JoinView) || (mTableType == TT_View)) )
			 --ColumnIndex;
#endif

	Catalog_TableDescriptor::ColumnIterator ColIt = mTableDescriptor.GetColumn(ColumnIndex);
	if ( ColIt )
	{
		return (*ColIt).IsLongDataType();
	}
	return false;
}

// ------------------------------------------------------------------------------------------------


SAPDB_Int2 Query_TableNode::OutputColumns( void )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );
    return mOutputColumnCnt;
}

bool Query_TableNode::GetOutputColumn( SAPDB_Int2 i, SQLMan_Identifier& ColName )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    // DDT : too slow !!!
    Catalog_TableDescriptor::ColumnIteratorForStarResolving iter = mTableDescriptor.BeginStarResolving();
    SAPDB_Int2 k=1;
    while ( k++<i && iter != mTableDescriptor.EndStarResolving() ) { ++iter; } 
    if ( iter != mTableDescriptor.EndStarResolving() )
    {
        (*iter).GetColumnName(ColName);
        return true;
    }
    return false;
}

bool Query_TableNode::GetOutputColumnIndex( SAPDB_Int2 i, SAPDB_Int2& ColumnIndex )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    // DDT : too slow !!!
    Catalog_TableDescriptor::ColumnIteratorForStarResolving iter = mTableDescriptor.BeginStarResolving();
    SAPDB_Int2 k=1;
    while ( k++<i && iter != mTableDescriptor.EndStarResolving() ) { ++iter; } 
    if ( iter != mTableDescriptor.EndStarResolving() )
    {
        ColumnIndex = (*iter).ExtColNo();
#if defined(QRW_REPLACE_ALL_KIND_OF_VIEWS)
		if ( (mTableType != TT_ComplexView) &&  
		     ((mTableType == TT_JoinView) || (mTableType == TT_View)) )
			 ++ColumnIndex;
#endif
        return true;
    }
    return false;
}

bool Query_TableNode::CheckForColumn( SQLMan_Identifier Column, SAPDB_Int2& QuantifierSyntaxNode, SAPDB_Int2& ColumnIndex )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    if ( Catalog_TableDescriptor::ColumnIterator ColIt = mTableDescriptor.FindColumn(Column) )
    {
        QuantifierSyntaxNode = mAPNode;
		if ( !mTableDescriptor.HasSelectPrivilege(*ColIt) )	// DDT : this is correct for select ONLY !
			return false;
        ColumnIndex = (*ColIt).ExtColNo();  // DDT: Or RecColNo ? Must match with a061exist_columnid !
#if defined(QRW_REPLACE_ALL_KIND_OF_VIEWS)
		if ( (mTableType != TT_ComplexView) &&  
		     ((mTableType == TT_JoinView) || (mTableType == TT_View)) )
			 ++ColumnIndex;
#endif
        return true;
    }
    return false;
}

bool Query_TableNode::CheckForColumn( SQLMan_Identifier Schema, SQLMan_Identifier Table, SQLMan_Identifier Column, SAPDB_Int2& QuantifierSyntaxNode, SAPDB_Int2& ColumnIndex )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    if ( Table == SQLMan_Identifier() )
        return CheckForColumn(Column, QuantifierSyntaxNode, ColumnIndex);
    if ( CheckQuantifierName(Schema, Table) )
        return CheckForColumn(Column, QuantifierSyntaxNode, ColumnIndex);
    return false;
}

bool Query_TableNode::CheckQuantifierName( SQLMan_Identifier Schema, SQLMan_Identifier Table )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    if ( mReferenceName == SQLMan_Identifier() )
	{
//		mExplicitSchema
		if ( Schema != SQLMan_Identifier() )
	        return mTableDescriptor.IsEqualTable(Schema, Table);
		return ( Table == mTableDescriptor.GetTableName() );
	}
    if ( (Schema==SQLMan_Identifier()) && (Table==mReferenceName) )
        return true;
    return false;
}

void Query_TableNode::GetIdentificationNames( SQLMan_Identifier& Schema, SQLMan_Identifier& Table )
{
    SAPDBERR_ASSERT_STATE( mSemanticsDone );

    if ( mReferenceName != SQLMan_Identifier() )
    {
        Schema = SQLMan_Identifier();
        Table  = mReferenceName;
    }
    else
    {
        mTableDescriptor.GetSchemaName(Schema);
        mTableDescriptor.GetTableName(Table);
    }
}

bool Query_TableNode::HasReferenceName( void )
{
	return ( mReferenceName != SQLMan_Identifier() );
}

// ------------------------------------------------------------------------------------------------
