/*!***************************************************************************

  module      : Query_Quantifier.cpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#include "SAPDB/SQLManager/Query/Query_Quantifier.hpp"
#include "SAPDB/SQLManager/Query/Query_Node.hpp"

// ------------------------------------------------------------------------------------------------

Query_Quantifier::Query_Quantifier( void )
    : mRef(0)
    , mDistinctType(DT_Unknown)
    , mQuantifierType(QT_Unknown)
    , mDisabled(false)
{
}

Query_Quantifier::Query_Quantifier( Query_Node* Ref, Query_DistinctType DistinctType, Query_QuantifierType QuantifierType )
    : mRef(Ref)
    , mDistinctType(DistinctType)
    , mQuantifierType(QuantifierType)
	, mDisabled(false)
{
    mRef->FatherQuantifier() = this;
}

Query_Quantifier::Query_Quantifier( const Query_Quantifier& Q, Query_Node* Father, Query_NodeMapper* mpMap )
    : mRef(0)
    , mDistinctType(Q.mDistinctType)
    , mQuantifierType(Q.mQuantifierType)
	, mDisabled(Q.mDisabled)
{
    mRef = mRef->CreateDuplicate(Father, this, mpMap);
}

Query_Quantifier::~Query_Quantifier( void )
{
	if ( mRef )
	{
		SAPDBMem_IRawAllocator& Alloc = mRef->GetAllocator();
		destroy(mRef, Alloc);
	}
}

bool Query_Quantifier::ReplaceComplexViews( void )
{
    return mRef->ReplaceComplexViews();
}

Query_Quantifier* Query_Quantifier::CreateDuplicate( Query_Node* Father, Query_NodeMapper* mpMap )
{
    return new(Father->GetAllocator()) Query_Quantifier(*this, Father, mpMap);
}

bool Query_Quantifier::OneTupleCondition( void )
{
    return false; // DDT
} 

bool Query_Quantifier::NoDuplicateCondition( void )
{
    return false; // DDT
}

SAPDB_UInt2 Query_Quantifier::NodeID( void )
{
    return mRef->NodeID();
}

Query_NodeType Query_Quantifier::NodeType( void )
{
    return mRef->GetNodeType();
}

bool Query_Quantifier::NodeTypeSelect( void )
{
    return (mRef->GetNodeType() == NT_Select);
}

bool Query_Quantifier::NodeTypeTable( void )
{ 
	return (mRef->GetNodeType()==NT_TableAccess); 
}

bool Query_Quantifier::NodeTypeSelectOrSet( void )
{
    Query_NodeType T = mRef->GetNodeType();
    return (( T == NT_Select )    ||
            ( T == NT_Union )     ||
            ( T == NT_Intersect ) ||
            ( T == NT_Difference ));
}
bool Query_Quantifier::NodeTypeSet( void )
{
    Query_NodeType T = mRef->GetNodeType();
    return (( T == NT_Union )     ||
            ( T == NT_Intersect ) ||
            ( T == NT_Difference ));
}

bool Query_Quantifier::FromTableSpec( void )
{
	return ( mRef->Position() == P_FromPart );
}

bool Query_Quantifier::HasSysKey( void )
{
	return mRef->HasSysKey();
}

Query_JoinType Query_Quantifier::JoinType( void )
{
    return mRef->GetJoinType();
}

bool Query_Quantifier::OuterJoin( void )
{
    return mRef->OuterJoin();
}

bool Query_Quantifier::Aggregates( void )
{
    return mRef->Aggregates();
}

SAPDB_Int2 Query_Quantifier::OutputColumns( void )
{ 
    return mRef->OutputColumns(); 
}

SAPDB_Int2 Query_Quantifier::FunctionOutputColumns( void )
{ 
    return mRef->FunctionOutputColumns(); 
}

void Query_Quantifier::AdjustColumnIndexForPullDown( Container_Vector<Query_Column>*& Columns )
{
	if ( NodeTypeSelectOrSet() && FromTableSpec() )
	{
		SAPDB_Int2 n = Columns->GetSize();
		for ( SAPDB_Int2 i=0; i<n; i++ )
			(*Columns)[i].ChangeColumnIndexDelta(-1);	// Adjustment for the syskey
	}
}

void Query_Quantifier::AdjustColumnIndexForPullUp( Container_Vector<Query_Column>*& Columns )
{
	if ( NodeTypeSelectOrSet() && FromTableSpec() )
	{
		SAPDB_Int2 n = Columns->GetSize();
		for ( SAPDB_Int2 i=0; i<n; i++ )
			(*Columns)[i].ChangeColumnIndexDelta(+1);	// Adjustment for the syskey
	}
}

bool Query_Quantifier::PullDownSearchCond( Container_Vector<Query_Column>*& Columns, SyntaxTreeNodeNo SearchCond )
{
	AdjustColumnIndexForPullDown(Columns);
	return mRef->PullDownSearchCond(Columns, SearchCond);
}

bool Query_Quantifier::GetOutputColumn( SAPDB_Int2 i, SQLMan_Identifier& ColName )
{
    return mRef->GetOutputColumn(i, ColName);
}

bool Query_Quantifier::GetOutputColumnIndex( SAPDB_Int2 i, SAPDB_Int2& ColumnIndex )
{
	bool rc = mRef->GetOutputColumnIndex(i, ColumnIndex);
	if ( rc && NodeTypeSelectOrSet() && FromTableSpec() )
		++ColumnIndex;	// adjust columnindex as there will be a syskey
	return rc;
}

bool Query_Quantifier::CheckForColumn( SQLMan_Identifier Column, SAPDB_Int2& QuantifierSyntaxNode, SAPDB_Int2& ColumnIndex )
{
    bool rc = mRef->CheckForColumn(Column, QuantifierSyntaxNode, ColumnIndex);
	if ( rc && FromTableSpec() ) 
	{	// adjust columnindex as there will be a syskey
		if ( NodeTypeSelectOrSet() )
			++ColumnIndex;	
		if ( NodeTypeTable() && mRef->HasSysKey() )
			--ColumnIndex;
	}
	return rc;
}

bool Query_Quantifier::CheckForColumn( SQLMan_Identifier Schema, SQLMan_Identifier Table, SQLMan_Identifier Column, SAPDB_Int2& QuantifierSyntaxNode, SAPDB_Int2& ColumnIndex )
{
    bool rc = mRef->CheckForColumn(Schema, Table, Column, QuantifierSyntaxNode, ColumnIndex);
	if ( rc && NodeTypeSelectOrSet() && FromTableSpec() )
		++ColumnIndex;	// adjust columnindex as there will be a syskey
	return rc;
}

bool Query_Quantifier::CheckQuantifierName( SQLMan_Identifier Schema, SQLMan_Identifier Table)
{
    return mRef->CheckQuantifierName(Schema, Table);
}

void Query_Quantifier::GetIdentificationNames(SQLMan_Identifier& Schema, SQLMan_Identifier& Table)
{
    return mRef->GetIdentificationNames(Schema, Table);
}

bool Query_Quantifier::HasReferenceName( void )
{
	return mRef->HasReferenceName();
}

bool Query_Quantifier::IqualIdentificationNames(Query_Quantifier*& Q)
{
    SQLMan_Identifier S1, S2, T1, T2;
    GetIdentificationNames(S1, T1);
    Q->GetIdentificationNames(S2, T2);

    if ( (S1==S2) && (T1==T2) )
        return true;
    return false;
}

bool Query_Quantifier::IsCorrelated( void )
{
    return mRef->IsCorrelated();
}

bool Query_Quantifier::SelectPrivilegesForAllColumns( void )
{
	return mRef->SelectPrivilegesForAllColumns();
}

// ------------------------------------------------------------------------------------------------
