/*!***************************************************************************

  module      : Query_SyntaxTree.hpp

  -------------------------------------------------------------------------

  responsible : DirkT

  special area: SQL Manager / Query
  description : 

  last changed: 2004-11-23  16:00
  first created:2004-04-29  10:00

  -------------------------------------------------------------------------

  copyright:    Copyright (c) 2000-2004 SAP AG




    ========== licence begin  GPL
    Copyright (c) 2000-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/

#ifndef QUERY_SYNTAXTREE_H
#define QUERY_SYNTAXTREE_H

// ------------------------------------------------------------------------------------------------

#include "SAPDB/SQLManager/SQLMan_Types.hpp"
#include "SAPDB/SQLManager/Query/Query_Types.hpp"

#include "SAPDB/SAPDBCommon/SAPDB_MemCopyMove.hpp"
#include "SAPDB/SAPDBCommon/SAPDB_Types.hpp"
#include "SAPDB/SAPDBCommon/ErrorsAndMessages/SAPDBErr_Assertions.hpp"
#include "SAPDB/SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"
#include "SAPDB/SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"

#include "SAPDB/Container/Container_Vector.hpp"

#include "vak001.h"

// ------------------------------------------------------------------------------------------------

class SubQueryMapper;
class SQLMan_Context;

// ------------------------------------------------------------------------------------------------

class NodeMap
{
public:

    NodeMap( void ) 
        : SrcNode(cak_qrewrite_ref_init)
        , DestNode(cak_qrewrite_ref_init)
    {}

    NodeMap( SyntaxTreeNodeNo SNode, SyntaxTreeNodeNo DNode ) 
        : SrcNode(SNode)
        , DestNode(DNode)
    {}

    SyntaxTreeNodeNo SrcNode;
    SyntaxTreeNodeNo DestNode;
};

class Query_NodeMapper
{
public:
    Query_NodeMapper( SAPDBMem_IRawAllocator& Alloc, SAPDB_Int2 InitSize = 10 )
        : mMapVector(Alloc)
    {
        mMapVector.Reserve(InitSize);
    }

    ~Query_NodeMapper( void )
    {
        mMapVector.Delete();
    }

    inline SAPDB_Int2 GetSize( void )
    {
        return mMapVector.GetSize();
    }

    inline void AddMappingNode( SyntaxTreeNodeNo SrcNode, SyntaxTreeNodeNo DestNode )
    {
        NodeMap M(SrcNode, DestNode);
        mMapVector.InsertEnd(M);
    }

    inline SyntaxTreeNodeNo GetDestNodeFromPos( SAPDB_Int2 i )
    {
        return mMapVector[i].DestNode;
    }

    inline SyntaxTreeNodeNo GetDestNodeFromSrcNode( SyntaxTreeNodeNo SrcNode )
    {
        SyntaxTreeNodeNo n = mMapVector.GetSize();
        for (SAPDB_Int2 i=0; i<n; i++)
        {
            if ( mMapVector[i].SrcNode == SrcNode )
                return mMapVector[i].DestNode;
        }
        return cak_qrewrite_ref_init;    
    }

private:
    Container_Vector<NodeMap>   mMapVector; 
};

// ------------------------------------------------------------------------------------------------

class Query_SyntaxTreeNode : public tak_ap_node
{
public:

};

class Query_SyntaxTree 
{
public:
    Query_SyntaxTree( SQLMan_Context& Context );

    Query_SyntaxTree& operator= ( const Query_SyntaxTree& T );

    void PrintToTrace( void );

    // -------------------

    inline /*Query_SyntaxTreeNode&*/ tak_ap_node& operator[](SAPDB_Int2 n)
    {
        return *(mAPTreePtr + n);
    }

    inline const /*Query_SyntaxTreeNode&*/ tak_ap_node& operator[](SAPDB_Int2 n) const
    {
        return *(mAPTreePtr + n);
    }

    inline tak_ap_node* GetAPTree( void )
    {
        return mAPTreePtr;
    }

    // -------------------

    inline bool& Changes( void )
    {
        return mChanges;
    }

    inline bool IsOk( void )
    {
        return !mError;
    }

    void SetError( void )
    {
        mError = true;
    }

    // -------------------

    inline bool NeedsMapping( SyntaxTreeNodeNo Node ) const
    {
        return ( mAPTreePtr[Node].n_querycheck.includes(qcn_NeedAdjustment) || 
                 (mAPTreePtr[Node].n_proc == a92fromsel ) ||
                 (mAPTreePtr[Node].n_proc == a66 && mAPTreePtr[Node].n_subproc() == cak_x_select_in_from_part ) ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_mass_select )         ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_subquery )            ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_one_val_subquery )    ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_arith_with_subquery ) ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_start_union )			||
                 (mAPTreePtr[Node].n_proc == a63query_spec ) );
				// this statement must match with Query_SyntaxTree::MapSubSelectNodes
				// all nodes listed in MapSubSelectNodes must occur here too ! but not vice versa !
    }

    inline bool WithOutPositionOffset( SyntaxTreeNodeNo Node ) const
    {
        return ( mAPTreePtr[Node].n_querycheck.includes(qcn_FixedPos)       ||
			     mAPTreePtr[Node].n_querycheck.includes(qcn_NeedAdjustment) || 
                 (mAPTreePtr[Node].n_proc == a92fromsel ) ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_subquery )            ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_one_val_subquery )    ||
                 (mAPTreePtr[Node].n_proc == a63 && mAPTreePtr[Node].n_subproc() == cak_x_arith_with_subquery ) ||
				 (mAPTreePtr[Node].n_proc == a80 && mAPTreePtr[Node].n_subproc() == cak_x_select_hint ) );
				// this statement must match with Query_SyntaxTree::MapSubSelectNodes
    }

    bool MapNode( SyntaxTreeNodeNo& Node, Query_NodeMapper* mpMap )
    {
        if ( NeedsMapping(Node) )
        {
            SyntaxTreeNodeNo Dest = mpMapper->GetDestNodeFromSrcNode(mAPTreePtr[Node].n_pos());
            if ( Dest != cak_qrewrite_ref_init )
            {
                Node = Dest;
                return true;
            }
        }
        return false;
    }

    // -------------------

    void SetRefBack( SyntaxTreeNodeNo StartNode );

    bool GetFirstTreeNode( SyntaxTreeNodeNo& Node );
    bool GetNextTreeNode( SyntaxTreeNodeNo& Node );
    
    bool GetNextArrayNode( SyntaxTreeNodeNo& Node );
    bool GetFirstArrayNode( SyntaxTreeNodeNo& Node );

    // -------------------

    SyntaxTreeNodeNo GetFreeNode( bool ForcePutNode=false );

    inline SAPDB_Int2 GetFreeNodeCount( void )
    {
        return mFreeNodeCount;
    }

    void SetUnusedNode( SyntaxTreeNodeNo NodeNo, bool Recursive=false );

    // -------------------

    SyntaxTreeNodeNo CheckNodes( void );

    // -------------------

    bool SubtreeContainsNode( SyntaxTreeNodeNo SubtreeStart, SyntaxTreeNodeNo Node );

    // -------------------

    inline SyntaxTreeNodeNo DoubleTreePart( SyntaxTreeNodeNo StartNode, Query_NodeMapper* mpMap=0  )
    {
        SAPDBTRACE_METHOD_DEBUG ("Query_SyntaxTree::DoubleTreePart", QueryRewrite_Trace, 7);
        return DoubleTreePart(mAPTreePtr, StartNode, 0, mpMap);
    }

    // DDT: Duplicate the part of the ap_tree given by Context / StartNo 
    // the duplicate will be within the OWN nodearray !!!
    SyntaxTreeNodeNo DoubleTreePart( tak_ap_node* SourceSyntaxTree, SyntaxTreeNodeNo StartNode, SAPDB_Int2 PosOffset=0, Query_NodeMapper* mpMap=0 );

    // -------------------

    void UnlinkNode( SyntaxTreeNodeNo NodeNo, bool SetUnused );

    void SubstituteNodeBySubtree( SyntaxTreeNodeNo OldNodeNo, SyntaxTreeNodeNo StartSubtreeNodeNo );

    void SetSucessorOnSameLevel( SyntaxTreeNodeNo NodeNo, SyntaxTreeNodeNo NewSuccessor );

    void SetSucessorOnLowLevel( SyntaxTreeNodeNo NodeNo, SyntaxTreeNodeNo NewSuccessor );

    // -------------------

    bool AddSubSelectNode( SyntaxTreeNodeNo SelNode, SyntaxTreeNodeNo SubSelNode, SyntaxTreeNodeNo FatherOfSubSelNode, SAPDB_Int2 Pos );
    bool RemoveSubSelectNode( SyntaxTreeNodeNo FromSelNodeNo );
    SyntaxTreeNodeNo FindSubSelectNode( SyntaxTreeNodeNo Node );

    bool ChangeColumnNameToColumnId( SyntaxTreeNodeNo Node, SyntaxTreeNodeNo FromTabNode, SAPDB_Int2 ColumnIndex );
//    bool ChangeOrderColumnNameToRefName( SyntaxTreeNodeNo Node, SAPDB_Int4& Pos, SAPDB_Int2& Len );

protected:

    bool CheckSingleNode( SyntaxTreeNodeNo NodeNo );

    SyntaxTreeNodeNo DoubleTreePartInternal( tak_ap_node* SourceSyntaxTree, SyntaxTreeNodeNo StartNode, SAPDB_Int2 PosOffset=0 );

    void MapSubSelectNodes( void );

private:

    SQLMan_Context&     mContext;
    tak_ap_node*        mAPTreePtr;
    Query_NodeMapper*   mpMapper;
    SyntaxTreeNodeNo    mFirstUnusedNode;
    SyntaxTreeNodeNo    mFreeNodeCount;
    bool          mChanges;
    bool          mError;
};

// ------------------------------------------------------------------------------------------------

#endif