/*****************************************************************************/
/*!
  @file         SQLMan_SessionIterator.hpp
  @author       MartinKi
  @ingroup

  @brief        Header file for class SQLMan_SessionIterator.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2005 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

#ifndef SQLMAN_SESSIONITERATOR_HPP
#define SQLMAN_SESSIONITERATOR_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SAPDBCommon/SAPDB_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class SQLMan_Context;

/*!
 * @class  SQLMan_SessionIterator
 * @author MartinKi
 *
 * Provides an iterator over all SQL sessions.
 *
 * @warning There can only be one iterator active at any time and
 * while the iterator is active no session can be created or released!
 * It is therfore important to either insure that the iterator either
 * is cycled through quickly, or has a lifetime that is limited to the
 * time it is actually used, or that its Reset()-method is called as
 * soon as the lock it is holding can be released.
 */
class SQLMan_SessionIterator
{
private:
    //no copy-construction or assignment
    SQLMan_SessionIterator(const SQLMan_SessionIterator&);
    SQLMan_SessionIterator operator=(SQLMan_SessionIterator&);

    SAPDB_Bool m_isFirst;
public:
    /*!
     * @brief Constructor. Does not lock the session list.
     */
    SQLMan_SessionIterator();

    /*!
     * @brief Destructor. Releases the lock on the session list if it
     *        should be still set.
     */
    ~SQLMan_SessionIterator();

    /*!
     * @brief Retrieves the next session. The first call of this method
     * will lock the session list. The lock is held until all sessions
     * have been retrieved by further calls to Next(), or until Reset()
     * is called.
     */
    SAPDB_Bool Next(SQLMan_Context*& context);

    /*!
     * @brief Resets the iterator to the first element in the session
     *        list and releases the lock on the session list.
     */
    void Reset();
};

/**************************************************************************/

inline SQLMan_SessionIterator::SQLMan_SessionIterator()
    : m_isFirst(true)
{}

/**************************************************************************/

#endif // SQLMAN_SESSIONITERATOR_HPP
