/*****************************************************************************/
/*!
 @file    SQLMan_UpdateStatistics.cpp
 @author  MartinKi

 @brief Implementation of class SQLMan_UpdateStatistics.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (c) 2003-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SQLManager/SQLMan_UpdateStatistics.hpp"

#include "SQLManager/Catalog/Catalog_TableDescriptor.hpp"
#include "SQLManager/Catalog/Catalog_TableStatisticsDescriptor.hpp"
#include "SQLManager/Catalog/Catalog_ColumnDescriptor.hpp"
#include "SQLManager/Tables/SQLTab_SysUpdStatLog.hpp"
#include "SQLManager/SQLMan_Context.hpp"
#include "SQLManager/SQLMan_Tracing.hpp"
#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"
#include "SAPDBCommon/Messages/SDBMsg_SQLMan.h"

#include "DataAccess/Data_PrimTreeStatisticsCoordinator.hpp"

#include "FileDirectory/FileDir_FileNo.hpp"
#include "FileDirectory/FileDir_Table.hpp"
#include "FileDirectory/FileDir_IDirectory.hpp"

#include "KernelCommon/ParameterNames/KernelParam_DataAccess.hpp"

#include "SAPDBCommon/Fields/SAPDBFields_FieldMap.hpp"
#include "SAPDBCommon/Fields/SAPDBFields_AttributedField.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Topic.hpp"

#include "RunTime/Configuration/RTEConf_ParameterAccessKernelInterface.hpp"
#include "RunTime/System/RTESys_Time.h"
#include "RunTime/Tasking/RTETask_ITask.hpp"
#include "RunTime/RTE_Message.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"

#include "hgg06.h" // gg06SessionGetInt4
#include "heo52.h" // vtimestamp

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/
/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/// default of 0 means 'use parallel update statistics and determine
/// number of servers automatically'
SAPDB_Int4 SQLMan_UpdateStatistics::m_noParallelServers = 0;

/**************************************************************************/

void SQLMan_UpdateStatistics::InitKernelParameters(
    Msg_List& errMsg)
{
	SAPDBTRACE_ROUTINE_DEBUG (
        "SQLMan_UpdateStatistics::InitKernelParameters", SQLMan_Trace, 1);

    SAPDB_Int8 paramVal = 0;
    if ( RTEConf_ParameterAccess::Instance()->GetInteger(
             KERNELPARAM_UPDATESTAT_PARALLEL_SERVERS,
             paramVal,
             errMsg) ) {
        if ( -1 > paramVal ) {
            paramVal = -1;
        }
    }
    m_noParallelServers = static_cast<SAPDB_Int>(paramVal);
}

/**************************************************************************/

SQLMan_UpdateStatistics::SQLMan_UpdateStatistics(
    SQLMan_Context&            context,
    const tsp00_TermId&        termId,
    const SAPDB_Int4           sampleRows,
    const SAPDB_Int4           samplePercentage)
    : m_context( context ),
      m_sampleRowCount( sampleRows ),
      m_samplePercentage( samplePercentage )
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::SQLMan_UpdateStatistics", SQLMan_Trace, 1);

    SAPDB_MemCopyNoCheck( m_termId, &termId, sizeof( m_termId ) );
}

/**************************************************************************/

SAPDB_Bool SQLMan_UpdateStatistics::Initialize(Msg_List& errMsg)
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::Initialize", SQLMan_Trace, 1);

    vtimestamp( *reinterpret_cast<tsp00_Timestamp*>( &m_startTimeStamp ) );

    if ( m_requestFile.GetFileNo().IsValid() ) {
        m_requestFile.Destroy( m_context.TransContext(), errMsg );
    }

    return m_requestFile.Create( m_context.TransContext(), errMsg ).IsValid();
}

/**************************************************************************/

SQLMan_UpdateStatistics::~SQLMan_UpdateStatistics()
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::SQLMan_UpdateStatistics", SQLMan_Trace, 1);

    Msg_List errMsg;
    m_requestFile.Destroy( m_context.TransContext(), errMsg );
}

/**************************************************************************/

SQLMan_UpdateStatistics::ReturnCode
SQLMan_UpdateStatistics::AddTableStatRequest(
    const FileDir_FileNo&   tableFileNo,
    const SAPDB_Int4        noOfPagesToSample,
    const SAPDB_Int4        columnCount,
    const SAPDB_Int2        columnNumbers[],
    const tgg00_StEntryAddr pColumnDescriptions,
    Msg_List&               errMsg)
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::AddTableStatRequest", SQLMan_Trace, 1);

    SAPDBERR_ASSERT_STATE( m_requestFile.GetFileNo().IsValid() );
    if ( m_requestFile.GetFileNo().IsInvalid() ) {
        errMsg = Msg_List(
            Msg_List::Error,
            SDBMSG_SQLMAN_NOT_INITIALIZED );
        return SystemError;
    }

    Data_UpdStatRequestFile::ReturnCode retCode =
        m_requestFile.AddTable(
            m_context.TransContext(),
            tableFileNo,
            noOfPagesToSample,
            columnCount,
            reinterpret_cast<const SAPDB_Int2*>( columnNumbers ),
            pColumnDescriptions, errMsg );

    if ( retCode != Data_UpdStatRequestFile::Okay ) {
        switch ( retCode ) {
            case Data_UpdStatRequestFile::FileNotFound:
                // table has probably been deleted, ignore
                SAPDBTRACE_WRITELN(
                    SQLMan_Trace, 1,
                    "table not found: " << tableFileNo << ", skipping" );
                break;
            default:
                errMsg.AddDetail(
                    SDBMSG_SQLMAN_UPDSTAT_ADDTOREQUESTFILE_FAILED,
                    Msg_Arg( "FILENO",
                             FileDir_FileNoToString( tableFileNo ) ),
                    Msg_Arg( "FILENO1",
                             FileDir_FileNoToString( m_requestFile.GetFileNo() ) )
                    );
                return SystemError;
        }
    }

    return Okay;
}

/**************************************************************************/

// SQLMan_UpdateStatistics::ReturnCode
// SQLMan_UpdateStatistics::UserTableScan(
//     SQLMan_Identifier& schemaName,
//     SQLMan_Identifier& tableName,
//     const SAPDB_Int    tableNameLen)
// {
//	SAPDBTRACE_METHOD_DEBUG (
//         "SQLMan_UpdateStatistics::UserTableScan", SQLMan_Trace, 1);

//     if ( ! m_context.IsOk() ) {
//         return BadContext;
//     }

//     if ( m_requestFile.GetFileNo().IsInvalid() ) {
//         return SystemError;
//     }

//     Catalog_TableDescriptor curTableDesc;
//     if ( (! schemaName.IsBlank())
//          && (tableNameLen == sizeof( tableName ) ) {
//         // first table specified completely
//         curTableDesc = Catalog_TableDescriptor(
//             m_context, schemaName, tableName );
//         if ( ! curTableDesc ) {
//             return TableNotFound;
//         }

//         if ( curTableDesc.IsArchiveTable() ) {
//             return InvalidTableType;
//         }

//         if ( curTableDesc.IsSynonym() ) {
//     Data_UpdStatRequestFile::ReturnCode retCode =
//         m_requestFile.AddTable(
//             trans, tableSurr, columnCount,
//             reinterpret_cast<SAPDB_Int2*>( columnNumbers ),
//             pColumnDescriptions, errMsg );

// }

// CheckTableStatisticsRequired()
//     Catalog_TableDescriptor& tableDesc,
//     Msg_List&                errMsg
// {
//     SQLTab_SysUpdStatRequestTable updStatRequest( m_context );

//     ColumnList colList;
//     if ( updStatRequest.GetOutdatedColumns( tableDesc, colList, errMsg ) )
// }


/**************************************************************************/

SQLMan_UpdateStatistics::ReturnCode
SQLMan_UpdateStatistics::ProcessRequests(
    Msg_List& errMsg)
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::ProcessRequests", SQLMan_Trace, 1);

    SAPDBERR_ASSERT_STATE( m_requestFile.GetFileNo().IsValid() );

    if ( m_requestFile.GetFileNo().IsInvalid() ) {
        errMsg = Msg_List(
            Msg_List::Error,
            SDBMSG_SQLMAN_NOT_INITIALIZED );
        return SystemError;
    }

    Data_PrimTreeStatisticsCoordinator statCoord(
        m_context.TransContext(), m_requestFile );

    Data_PrimTreeStatisticsCoordinator::ReturnCode statRc =
        statCoord.UpdateStatistics( errMsg );

    if ( statRc != Data_PrimTreeStatisticsCoordinator::Okay ) {
        switch ( statRc ) {
            case Data_PrimTreeStatisticsCoordinator::Cancelled:
            {
                // update what we've already got
                ReturnCode updRetCode = this->UpdateStatisticRecords( errMsg );
                return updRetCode == Okay ? Cancelled : updRetCode;
            }
            default:
                return SystemError;
        }
    }

    return this->UpdateStatisticRecords( errMsg );
}

/**************************************************************************/

SQLMan_UpdateStatistics::ReturnCode
SQLMan_UpdateStatistics::UpdateStatisticRecords(
    Msg_List& errMsg)
{
	SAPDBTRACE_METHOD_DEBUG (
        "SQLMan_UpdateStatistics::UpdateStatisticRecords", SQLMan_Trace, 1);

    RTETask_ITask& task =
        *RTETask_ITask::ByTaskID( m_context.TransContext().trTaskId_gg00 );
    const SAPDB_Bool isCancelled = task.IsCancelled();

    if ( isCancelled ) {
        // ignore cancel flag while updating statistics information
        task.ResetCancelled();
    }

    SAPDB_Bool updateStatisticsLogTable = true;

    FileDir_IDirectory& fDir =
        FileDir_IDirectory::SessionInstance( m_context.TransContext() );

    Data_UpdStatRequestFile::Iterator::ReturnCode itRetCode;
    Data_UpdStatRequestFile::Iterator
        it ( m_requestFile.GetIterator( m_context.TransContext() ) );

    SAPDB_Int rc = 0;
    while ( (rc == 0)
            && ( (itRetCode = it.Next( m_context.TransContext(), errMsg ))
                 == Data_UpdStatRequestFile::Iterator::Okay ) ) {

        if ( ! it.IsRequestProcessed() ) {
            SAPDBTRACE_WRITELN(
                SQLMan_Trace, 1,
                "  skipping unprocessed table: "
                << it.GetStatistics( errMsg ).tableFileNo );
            continue;
        }

        const Data_UpdStatRequestFile::TableStatistics&
            stats = it.GetStatistics( errMsg );

        Catalog_TableDescriptor tableDesc( m_context, stats.tableFileNo );
        if ( ! tableDesc ) {
            m_context.ResetReturnCode();
            SAPDBTRACE_WRITELN(
                SQLMan_Trace, 1,
                "table not found in catalog: "
                << stats.tableFileNo << ", skipping" );
            continue;
        }

        Catalog_TableStatisticsDescriptor statisticsDesc;
        const bool c_forUpdate = true;
        if ( ! tableDesc.GetTableStatisticsDescriptor(
                 c_forUpdate, statisticsDesc ) ) {
            // FIXME: get correct error;
            rc = m_context.ReturnCode();
            break;
        }

        FileDir_Table tableFile;
        fDir.GetTableFile(
            m_context.TransContext().trTaskId_gg00,
            stats.tableFileNo, tableFile, errMsg );
        if ( ! tableFile.IsAssigned() ) {
            SAPDBTRACE_WRITELN(
                SQLMan_Trace, 1,
                "fileNo not found in fdir: "
                << stats.tableFileNo << ", skipping" );
            continue;
        }

        statisticsDesc.SetRowCount(
            static_cast<SAPDB_Int4>(
                tableFile.GetEntryCount(
                    m_context.TransContext().trTaskId_gg00) ) );
        statisticsDesc.SetPageCount(
            tableFile.GetLeafPageCount(
                m_context.TransContext().trTaskId_gg00) );

        SAPDB_Int seqNo = 1;
        Data_UpdStatRequestFile::ColumnStatisticsVector::ConstIterator
            colItEnd = stats.columnStatistics.End();
        for ( Data_UpdStatRequestFile::ColumnStatisticsVector::ConstIterator
                  colIt = stats.columnStatistics.Begin();
                  colIt != colItEnd;
              ++colIt, ++seqNo ) {

            statisticsDesc.SetColumnStatistics(
                colIt->columnNumber, colIt->value );
            if ( ! statisticsDesc.Store() ) {
                rc = m_context.ReturnCode();
                break;
            }

            if ( updateStatisticsLogTable ) {
                updateStatisticsLogTable = this->UpdateStatisticsLogTable(
                    tableDesc,
                    colIt->columnNumber,
                    seqNo,
                    it,
                    errMsg );
            }
        }
    }

    if ( isCancelled ) {
        // set cancel flag again
        task.SetCancelled();
        return Cancelled;
    }

    m_context.SetReturnCode( rc );
    return Okay;
}

/*----------------------------------------------------------------*/

SAPDB_Bool
SQLMan_UpdateStatistics::UpdateStatisticsLogTable(
    const Catalog_TableDescriptor&           tableDesc,
    const SAPDB_Int                          columnNumber,
    SAPDB_Int                                seqNo,
    const Data_UpdStatRequestFile::Iterator& it,
    Msg_List&                          errMsg)
{

    SAPDB_Int4      c_nullInt     = 0;
    const SAPDB_Int c_columnCount = 14;

    SQLMan_Identifier c_emptyName;
    SAPDB_MemCopyNoCheck( c_emptyName, bsp_c64, sizeof( SQLMan_Identifier ) );

    RTESys_SQLTimestamp beginTimeStamp;
    RTESys_SQLTimestamp endTimeStamp;

    SAPDBFields_AttributedField curField;
    SQLTab_SysUpdStatLog logTable( m_context );

    SAPDBFields_FieldMap fields( m_context.GetAllocator(), c_columnCount );

    const Data_UpdStatRequestFile::TableStatistics&
        stats = it.GetStatistics( errMsg );

    //
    // add log table entry
    //
    SAPDB_Int colNo = 0;

    // schemaname
    curField.Assign(
        tableDesc.GetSchemaName(), sizeof( tableDesc.GetSchemaName() ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // tablename
    curField.Assign(
        tableDesc.GetTableName(), sizeof( tableDesc.GetTableName() ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // tableid
    curField.Assign(
        const_cast<char*>( stats.tableFileNo.GetCharArray() ),
        sizeof( stats.tableFileNo ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // columnname
    Catalog_TableDescriptor::ColumnIterator columnIt =
        tableDesc.GetColumn( columnNumber );
    if ( columnIt ) {
        curField.Assign(
            (*columnIt).GetColumnName(),
            sizeof( (*columnIt).GetColumnName() ) );
        fields.Fill( colNo++, c_columnCount, curField );
    } else {
        colNo++;
    }

    // indexname
    curField.Assign(
        c_emptyName, sizeof( c_emptyName ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // executed_at
    curField.Assign(
        reinterpret_cast<char*>( m_startTimeStamp ),
        sizeof( m_startTimeStamp ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // implicit
    const char c_No[] = "NO";
    curField.Assign( const_cast<char*>(c_No), sizeof( c_No ) - 1 );
    fields.Fill( colNo++, c_columnCount, curField );

    // sample_pct
    SAPDB_Int4 samplePercent = m_samplePercentage;
    if ( it.GetNoOfPagesToSample() == 0 ) {
        // file was processed completely
        samplePercent = 100;
        curField.Assign( &samplePercent, sizeof( samplePercent ) );
        fields.Fill( colNo++, c_columnCount, curField );
    } else if ( samplePercent > 0 ) {
        curField.Assign( &samplePercent, sizeof( samplePercent ) );
        fields.Fill( colNo++, c_columnCount, curField );
    } else {
        curField.Assign( &c_nullInt, sizeof( c_nullInt ) );
        fields.Fill( colNo++, c_columnCount, curField );
    }

    // sample_row
    SAPDB_Int4 sampleRows = m_sampleRowCount;
    if ( sampleRows > 0 ) {
        curField.Assign( &sampleRows, sizeof( sampleRows ) );
        fields.Fill( colNo++, c_columnCount, curField );
    } else {
        curField.Assign( &c_nullInt, sizeof( c_nullInt ) );
        fields.Fill( colNo++, c_columnCount, curField );
    }

    // execution start
    SAPDB_MemCopyNoCheck(
        beginTimeStamp,
        it.GetBeginTimeStamp( errMsg ),
        sizeof( beginTimeStamp ) );
    curField.Assign(
        beginTimeStamp,
        sizeof( it.GetBeginTimeStamp( errMsg ) ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // execution end
    SAPDB_MemCopyNoCheck(
        endTimeStamp,
        it.GetEndTimeStamp( errMsg ),
        sizeof( endTimeStamp ) );
    curField.Assign(
        endTimeStamp,
        sizeof( it.GetEndTimeStamp( errMsg ) ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // session
    SAPDB_Int4 sessionId = gg06SessionGetInt4(
        m_context.TransContext().trSessionId_gg00 );
    curField.Assign( &sessionId, sizeof( sessionId ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // termid
    curField.Assign( m_termId, sizeof( m_termId ) );
    fields.Fill( colNo++, c_columnCount, curField );

    // seqNo
    curField.Assign( &seqNo, sizeof( seqNo ) );
    fields.Fill( colNo++, c_columnCount, curField );

    const SAPDB_Bool c_updateDuplicates = true;
    const SAPDB_Bool c_withCommit       = true;

    SQLTab_SysUpdStatLog::ReturnCode tabReturnCode =
        logTable.InsertRow(
            fields,
            c_updateDuplicates,
            c_withCommit,
            errMsg );

    // crash on all errors in slow kernel
    SAPDBERR_ASSERT_STATE(
        (tabReturnCode == SQLTab_SysUpdStatLog::Okay)
        || (tabReturnCode == SQLTab_SysUpdStatLog::LockRowFailed) );

    if ( tabReturnCode != SQLTab_SysUpdStatLog::Okay ) {
        if ( tabReturnCode == SQLTab_SysUpdStatLog::LockRowFailed ) {
            errMsg = Msg_List(
                Msg_List::Info,
                SDBMSG_SQLMAN_UPDSTAT_INSERT_LOCK_COLLISION );
            RTE_Message( errMsg );
            errMsg.ClearMessageList();
            // do not try again
            return false;
        }

        // ignore errors in fast kernel
        errMsg.Overrule(
            Msg_List(
                Msg_List::Warning,
                SDBMSG_SQLMAN_UPDATESTAT_IGNORE_INSERT_RETURN,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( stats.tableFileNo ) ),
                Msg_Arg( "RETURNCODE", SAPDB_ToString( tabReturnCode ) )
                ) );
        RTE_Message( errMsg );
        errMsg.ClearMessageList();
    }

    return true;
}
