/*****************************************************************************/
/*!
  @file         SQLMan_UpdateStatistics.hpp
  @author       MartinKi

  @brief        Header file for class SQLMan_UpdateStatistics.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2003 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

#ifndef SQLMAN_UPDATESTATISTICS_HPP
#define SQLMAN_UPDATESTATISTICS_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_UpdStatRequestFile.hpp"

#include "RunTime/System/RTESys_Time.h"

#include "SAPDBCommon/SAPDB_Types.hpp"

#include "ggg00.h"
#include "gsp00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class SQLMan_Context;
class Catalog_TableDescriptor;
class Msg_List;

/*!
 * @class  SQLMan_UpdateStatistics
 * @author MartinKi
 *
 * @brief Implements parts of the 'Update Statistics' commands.
 *
 * This class currently implements only a small subset of all 'Update
 * statistics' variations.
 */
class SQLMan_UpdateStatistics
{
public:
    /// Possible return values.
    enum ReturnCode {
        Okay,        ///< Command completed successfully
        Cancelled,   ///< Command execution was canceled
        SystemError  ///< A system error occured during command execution
    };

    /*!
     * @brief Constructor.
     *
     */
    SQLMan_UpdateStatistics(
        SQLMan_Context&            context,
        const tsp00_TermId&        termId,
        const SAPDB_Int4           sampleRows,
        const SAPDB_Int4           samplePercentage);

    /*!
     * @brief Destructor.
     */
    ~SQLMan_UpdateStatistics();

    /*!
     * @brief Initializes the object.
     *
     * This method can also be called to reset the object. The start
     * timestamp for statistics operations will be set to the time of
     * the function call.
     */
    SAPDB_Bool Initialize(Msg_List& errMsg);

    /*!
     * @brief Adds a compute statistics request for a table to the
     * internal request list.
     *
     * TODO: this method should become private as soon as this class
     * is able to collect the requested tables itself.
     */
    ReturnCode AddTableStatRequest(
        const FileDir_FileNo&   tableFileNo,
        const SAPDB_Int4        noOfPagesToSample,
        const SAPDB_Int4        columnCount,
        const SAPDB_Int2        columnNumbers[],
        const tgg00_StEntryAddr pColumnDescriptions,
        Msg_List&               errMsg);

    /*!
     * @brief Processes the list of requests collected.
     *
     * This method initiates the statistics computation for all
     * requests in the internal request list and also updates the
     * system table SYSUPDSTATLOG.
     */
    ReturnCode ProcessRequests(Msg_List& errMsg);

    /*!
     * @name Kernel parameter/Number of server tasks to be used
     *
     */
    /*@{*/

    /*!
     * @brief Initizalizes maximum number of parallel servers from
     * database parameter.
     */
    static void InitKernelParameters(Msg_List& errMsg);

    /*!
     * @brief Returns true if parallel update statistics is enabled.
     */
    static SAPDB_Bool IsParallel();

    /*! @brief Sets the maximum number of servers to use at a time
     *  during parallel update statistcs.
     */
    static void SetNoOfParallelServers(const SAPDB_Int noOfServers);

    /*! @brief Returns the maximum number of servers to use at a time
     *  during parallel update statistcs.
     */
    static SAPDB_Int GetNoOfParallelServers();

    /*@}*/
private:
    // no copy-construction or assignment
    SQLMan_UpdateStatistics(const SQLMan_UpdateStatistics&);
    SQLMan_UpdateStatistics operator=(const SQLMan_UpdateStatistics&);

    SQLMan_Context&         m_context;
    Data_UpdStatRequestFile m_requestFile;
    tsp00_TermId            m_termId;
    RTESys_SQLTimestamp     m_startTimeStamp;
    SAPDB_Int4              m_sampleRowCount;
    SAPDB_Int4              m_samplePercentage;

    /*!
     * maxmimum number of parallel servers; default of 0 means 'use
     * parallel update statistics and determine number of servers
     * automatically', n > 0: use a maximum of n servers; n = -1: do
     * not use code for parallel update statistics
     */
    static SAPDB_Int4       m_noParallelServers;

    ReturnCode UpdateStatisticRecords(Msg_List& errMsg);

    SAPDB_Bool UpdateStatisticsLogTable(
        const Catalog_TableDescriptor&           tableDesc,
        const SAPDB_Int                          columnNumber,
        SAPDB_Int                                seqNo,
        const Data_UpdStatRequestFile::Iterator& it,
        Msg_List&                                errMsg);
};

/**************************************************************************/

inline SAPDB_Bool SQLMan_UpdateStatistics::IsParallel()
{
    return m_noParallelServers >= 0;
}

/**************************************************************************/

inline void SQLMan_UpdateStatistics::SetNoOfParallelServers(
    const SAPDB_Int noOfServers)
{
    m_noParallelServers = noOfServers;
}

/**************************************************************************/

inline SAPDB_Int SQLMan_UpdateStatistics::GetNoOfParallelServers()
{
    return m_noParallelServers;
}

/**************************************************************************/

#endif // SQLMAN_UPDATESTATISTICS_HPP
