/*****************************************************************************/
/*!
 @file    SQLTab_SysUpdStatLog.cpp
 @author  MartinKi

 @brief Implementation of class SQLTab_SysUpdStatLog.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (c) 2003-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/
#include "SQLManager/SQLMan_Context.hpp"
#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"
#include "SAPDBCommon/Messages/SDBMsg_SQLMan.h"

#include "SQLManager/SQLMan_SQLStatementContext.hpp"
#include "SQLManager/KernelSQL/KSQL_Connection.hpp"
#include "SQLManager/KernelSQL/KSQL_PreparedStatement.hpp"

#include "SAPDB/Messages/Msg_List.hpp"
#include "SAPDBCommon/Fields/SAPDBFields_IFieldSequences.hpp"
#include "SAPDBCommon/Fields/SAPDBFields_AttributedField.hpp"
#include "SAPDBCommon/SAPDB_string.h"
#include "SAPDBCommon/SAPDB_Types.hpp"

#include "SQLManager/Tables/SQLTab_SysUpdStatLog.hpp"

/**************************************************************************/

SQLTab_SysUpdStatLog::SQLTab_SysUpdStatLog(SQLMan_Context& context)
    : m_context( context )
{}

/**************************************************************************/

SQLTab_SysUpdStatLog::ReturnCode
SQLTab_SysUpdStatLog::InsertRow(
    const SAPDBFields_IFieldSequence& values,
    const SAPDB_Bool                  updateDuplicates,
    const SAPDB_Bool                  doCommit,
    Msg_List&                         errMsg)
{

    SQLMan_SQLStatementContext newStatementScope( m_context );
    if ( !m_context.IsOk() )
    {
        errMsg = Msg_List(
            Msg_List::Error,
            SDBMSG_SQLMAN_UPDSTATLOG_NEWCONTEXT_FAILED,
            Msg_Arg( "RETURNCODE",
                     SQLManErr_Interface::GetInstance().GetErrorText(
                         m_context.ReturnCode() ) )
            );
        return SystemError;
    }

    KSQL_Connection con;
    if ( ! con.isConnected() ) {
        errMsg = Msg_List(
            Msg_List::Error,
            SDBMSG_SQLMAN_UPDSTATLOG_CREATESTATEMENT_FAILED,
            Msg_Arg( "RETURNCODE",
                     SQLManErr_Interface::GetInstance().GetErrorText(
                         m_context.ReturnCode() ) )
            );
        return SystemError;
    }

    m_context.SetInDDLTrigger();

    // statement prepare
    KSQL_PreparedStatement stmt = con.createPreparedStatement();

    const int c_maxStmtLen = 500;
    char stmtText[ c_maxStmtLen+1 ] =
        "INSERT SYSDBA.SYSUPDSTATLOG ("
        " schemaname, tablename,"
        " tableid,"
        " columnname,"
        " indexname,"
        " executed_at, implicit, sample_pct, sample_row,"
        " execution_start, execution_end, session,"
        " termid,"
        " seqno)"
        " VALUES"
        " (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

    if ( updateDuplicates ) {
        SAPDB_strncat( stmtText, " UPDATE DUPLICATES", c_maxStmtLen );
    }

    if ( doCommit ) {
        SAPDB_strncat( stmtText, " WITH COMMIT", c_maxStmtLen );
    }

    stmt.prepare( stmtText );

    SAPDB_Int colNo = 0;
    SAPDB_Int index = 0;
    bool      ok    = true;
    SAPDBFields_AttributedField curField;

    ok &= ( values.GetField( index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    stmt.bindIdentifier4Input( ++colNo, curField.GetDataPtr() ); // schemaname

    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    stmt.bindIdentifier4Input( ++colNo, curField.GetDataPtr() ); // tablename


    // tableid
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bind(
            ++colNo, curField.GetDataPtr(), curField.GetLength() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // columnname
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindIdentifier4Input(
            ++colNo, curField.GetDataPtr() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // indexname
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindIdentifier4Input(
            ++colNo,
            reinterpret_cast<const char*>( curField.GetDataPtr() ) );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // executed_at
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindTimestampInput( ++colNo, curField.GetDataPtr() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // implicit
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindAsciiInput(
            ++colNo,
            reinterpret_cast<const char*>( curField.GetDataPtr() ),
            curField.GetLength() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // sample_pct
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindInt4Input(
            ++colNo,
            *reinterpret_cast<const SAPDB_Int4*>( curField.GetDataPtr() ));
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // sample_row
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindInt4Input(
            ++colNo,
            *reinterpret_cast<const SAPDB_Int4*>( curField.GetDataPtr() ));
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // execution_start
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindTimestampInput( ++colNo, curField.GetDataPtr() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // execution_end
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindTimestampInput( ++colNo, curField.GetDataPtr() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // session
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindInt4Input(
            ++colNo,
            *reinterpret_cast<const SAPDB_Int4*>( curField.GetDataPtr() ) );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // termid
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    if ( ok
         && (curField.GetAttribute()
             == SAPDBFields_AttributedField::ValueAttribute)
        ) {
        stmt.bindAsciiInput(
            ++colNo,
            reinterpret_cast<const char*>( curField.GetDataPtr() ),
            curField.GetLength() );
    } else {
        stmt.bindNullValue( ++colNo );
    }

    // seqNo
    ok &= ( values.GetField( ++index, curField, errMsg ) ==
            SAPDBFields_IFieldSequence::fieldFound );
    stmt.bindInt4Input(
        ++colNo,
        *reinterpret_cast<const SAPDB_Int4*>(curField.GetDataPtr()) );

    SAPDBERR_ASSERT_STATE( ok );
    if ( ! ok ) {
        errMsg = Msg_List(
            Msg_List::Error,
            SDBMSG_SQLMAN_UPDSTATLOG_PREPARE_FAILED
            );
        return SystemError;
    }

    // statement execution
    m_sqlReturnCode = stmt.execute();
    if ( m_sqlReturnCode != 0 ) {
        switch ( m_sqlReturnCode ) {
            case -20:
            case 200:
                return DuplicateKey;
            case 500:
                return LockRowFailed;
            default:
                return InsertFailed;
        }
    }

    return Okay;
}

/**************************************************************************/
