/*****************************************************************************/
/*!
  @file         SQLTab_SysUpdStatLog.hpp
  @author       MartinKi
  @ingroup

  @brief        Header file for class SQLTab_SysUpdStatLog.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2003 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

#ifndef SQLTAB_SYSUPDSTATLOG_HPP
#define SQLTAB_SYSUPDSTATLOG_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SQLManager/KernelSQL/KSQL_Statement.hpp"
#include "SAPDBCommon/SAPDB_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

class Msg_List;
class SAPDBFields_IFieldSequence;
class SQLMan_Context;

/*!
 * @class  SQLTab_SysUpdStatLog
 * @author MartinKi
 *
 * This class provides access to the system table SYSUPDSTATLOG.
 */
class SQLTab_SysUpdStatLog
{
public:
    /// Possible return codes.
    enum ReturnCode {
        Okay,
        LockRowFailed,
        DuplicateKey,
        InsertFailed,
        SystemError
    };

    /*!
     * @brief Constructor.
     *
     * Does not create the table but simply provides a handle to access it.
     */
    SQLTab_SysUpdStatLog(SQLMan_Context& context);

    /*!
     * @brief Inserts a new row into the table.
     *
     * @param values           [in] field sequence of values to insert
     * @param updateDuplicates [in] update an existing record if true
     * @param doCommit         [in] commit changes if true
     * @param errMsg           [out] possible error messages
     *
     * @return
     *   - Okay, on success
     *   - DuplicateKey, if <em>updateDuplicates == false</em> and key exists
     *   - InsertFailed, if processing of the internal insert SQL statement
     *     failed; the SQL return code can be obtained by calling
     *     GetReturnCode
     *   - LockRowFailed, if the row could not be inserted because of a
     *     lock collision
     *   - SystemError
     */
    ReturnCode InsertRow(
        const SAPDBFields_IFieldSequence& values,
        const SAPDB_Bool                  updateDuplicates,
        const SAPDB_Bool                  doCommit,
        Msg_List&                         errMsg);

    /*!
     * @brief Returns the return code of the last internal sql command
     * (e.g. insert)
     */
    KSQL_Statement::SQLCode GetReturnCode() const;

private:
    // no copy-construction or assignment
    SQLTab_SysUpdStatLog(const SQLTab_SysUpdStatLog&);
    SQLTab_SysUpdStatLog operator=(const SQLTab_SysUpdStatLog&);

    SQLMan_Context& m_context;
    KSQL_Statement::SQLCode m_sqlReturnCode;
};

/**************************************************************************/

inline KSQL_Statement::SQLCode
SQLTab_SysUpdStatLog::GetReturnCode() const
{
    return m_sqlReturnCode;
}

/**************************************************************************/

#endif // SQLTAB_SYSUPDSTATLOG_HPP
