/*****************************************************************************/
/*!
 @file    SrvTasks_JobCollectTableStatistics.cpp
 @author  MartinKi

 @brief Implementation of class SrvTasks_JobCollectTableStatistics.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (c) 2003-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SQLManager/ErrorHandling/SQLManErr_Interface.hpp"
#include "ServerTasks/SrvTasks_JobCollectTableStatistics.hpp"
#include "ServerTasks/SrvTasks_ITaskForJob.hpp"
#include "ServerTasks/SrvTasks_JobTypes.hpp"
#include "SAPDBCommon/Messages/SDBMsg_SrvTasks.h"

#include "DataAccess/Data_PrimTreeStatisticsCoordinator.hpp"
#include "DataAccess/Data_RandomTreeAccess.hpp"
#include "DataAccess/Data_ScanTreeAccess.hpp"
#include "DataAccess/Data_UpdStatRequestFile.hpp"
#include "DataAccess/Data_PrimTreeStatistic.hpp"
#include "DataAccess/Data_Types.hpp"

#include "Transaction/Trans_Context.hpp"

#include "FileDirectory/FileDir_IDirectory.hpp"
#include "FileDirectory/FileDir_Table.hpp"
#include "FileDirectory/FileDir_FileNo.hpp"

#include "SAPDBCommon/SAPDB_Types.hpp"

#include "hbd30.h"
#include "hbd01.h"
#include "gbd00.h"
#include "gsp00.h"
#include "ggg00.h"

#include "SAPDB/SAPDBCommon/SAPDB_RangeCode.hpp" // Kernel_move_and_fill
#include "Trace/Trace_Entry.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/

/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/

/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/

/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/

/// Helper class to lock/release files on construction/destruction.
class AutoTreeLock
{
private:
    tbd_current_tree& m_current;
    bool              m_lockSet;

public:
    enum LockType {
        LockExclusive,
        LockNonExclusive
    };

    AutoTreeLock(
        tgg00_TransContext& trans,
        tgg00_FileId&       fileId,
        tbd_current_tree&   current,
        LockType            lock)
        : m_current( current )
    {
        if ( trans.trError_gg00 != e_ok ) {
            return;
        }

        bd30GetTree( trans, fileId, m_current,
                 (lock == LockExclusive), m_select );

        m_lockSet = (trans.trError_gg00 == e_ok);
    }

    ~AutoTreeLock()
    {
        this->Release();
    }

    void Release()
    {
        if ( m_lockSet ) {
            bd30ReleaseTree( m_current );
        }
        m_lockSet = false;
    }
};

/*************************************************************************/
/*************************************************************************/

SrvTasks_JobCollectTableStatistics::SrvTasks_JobCollectTableStatistics(
    const Attributes&     jobAttributes,
    const FileDir_FileNo& updStatRequestFileNo,
    const Data_PrimTreeStatisticsCoordinator::TableKey& tableKey,
        const FileDir_FileNo& tableFileNo,
    const FileDir_FileNo& auxFileNo,
    const SAPDB_Int4      noOfPagesToSample,
    SrvTasks_WaitingTask& waitingTask)
    : SrvTasks_Job( jobUpdateStat, jobAttributes, &waitingTask ),
      m_updStatRequestFileNo( updStatRequestFileNo ),
      m_tableKey( tableKey ),
      m_tableFileNo( tableFileNo ),
      m_auxFileNo( auxFileNo ),
      m_noOfPagesToSample( noOfPagesToSample ),
      m_returnCode( ResultOkay )
{}

/*************************************************************************/

const Data_PrimTreeStatisticsCoordinator::TableKey&
SrvTasks_JobCollectTableStatistics::GetTableKey() const
{
    return m_tableKey;
}

/*************************************************************************/

const FileDir_FileNo&
SrvTasks_JobCollectTableStatistics::GetAuxFileNo() const
{
    return m_auxFileNo;
}

/*************************************************************************/

bool
SrvTasks_JobCollectTableStatistics::CreateCopy (
    SAPDBMem_IRawAllocator &allocator,
    Msg_List               &errlist,
    SrvTasks_Job*          &jobCopy ) const
{
    SAPDBTRACE_METHOD_DEBUG (
        "SrvTasks_JobCollectTableStatistics::CreateCopy", DataTree_Trace, 5);

    // This is only called if nobody wants to wait for this job.
    // If nobody is waiting then no output field is needed - and copied

    jobCopy = new (allocator) SrvTasks_JobCollectTableStatistics(
        m_Attributes,
        m_updStatRequestFileNo,
        m_tableKey,
        m_tableFileNo,
        m_auxFileNo,
        m_noOfPagesToSample,
        *m_WaitingTask
        );

    if ( 0 == jobCopy )
    {
        AppendContextInfo (__CONTEXT__, errlist, "This job could not be copied");
        return false;
    }
    jobCopy->SetAttributes(m_Attributes);
    return true;
}

/*************************************************************************/

void SrvTasks_JobCollectTableStatistics::Execute ()
{
    Execute ( m_ServerTaskContext->GetContext().OldTrans() );
}

/*************************************************************************/

void SrvTasks_JobCollectTableStatistics::Execute(
    tgg00_TransContext& trans)
{
    SAPDBTRACE_METHOD_DEBUG(
        "SrvTasks_JobCollectTableStatistics::CollectStatistics",
        DataTree_Trace, 5 );

    /*
     * get request from request file
     */
    Msg_List& errMsg = GetTransContext().ErrorList();
    errMsg.ClearMessageList();
    Data_UpdStatRequestFile requestFile;
    Data_UpdStatRequestFile::ReturnCode reqRetCode =
        requestFile.Open( trans, m_updStatRequestFileNo, errMsg );

    trans.trError_gg00 = e_ok;

    if ( reqRetCode != Data_UpdStatRequestFile::Okay ) {
        errMsg.Overrule (
            Msg_List( Msg_List::Error,
                      SDBMSG_SRVTASKS_TABLESTATISTICS_REQUESTFILE_NOT_FOUND,
                      Msg_Arg( "FILENO",
                               FileDir_FileNoToString( m_updStatRequestFileNo ) )
                ) );
        SAPDBERR_ASSERT_STATE( false );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    FileDir_Table tableFile;
    FileDir_IDirectory& fDir =
        FileDir_IDirectory::Instance( trans.trIndex_gg00 );

    Data_UpdStatRequestFile::Iterator it = requestFile.GetIterator( trans );

    Data_UpdStatRequestFile::Iterator::ReturnCode itRetCode;
    itRetCode = it.Find( trans, m_tableKey, errMsg );
    if ( itRetCode != Data_UpdStatRequestFile::Iterator::Okay  ) {
        errMsg.Overrule(
            Msg_List( Msg_List::Error,
                      SDBMSG_SRVTASKS_TABLESTATISTICS_TABLEKEY_NOT_FOUND,
                      Msg_Arg( "KEYASHEX",
                               m_tableKey.GetKeyPointer(),
                               m_tableKey.GetLength(),
                               m_tableKey.GetLength(),
                               SAPDB_ToStringClass::hex ),
                      Msg_Arg( "FILENO",
                               FileDir_FileNoToString( m_updStatRequestFileNo ) )
                ) );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    Data_UpdStatRequestFile::StatisticsRequestInfo& requestInfo =
        it.GetStatisticsRequest( errMsg );

    /*
     * check if table and aux file exist and lock them
     */
    SAPDBTRACE_WRITELN(
        DataTree_Trace, 5, "processing table: " << requestInfo.tableFileNo );

    SAPDBERR_ASSERT_STATE( requestInfo.tableFileNo == m_tableFileNo );

    fDir.GetTableFile(
        trans.trTaskId_gg00,
        requestInfo.tableFileNo,
        tableFile,
        errMsg );

    if ( ! tableFile.IsAssigned() ) {
        SAPDBTRACE_WRITELN(
            DataTree_Trace, 5, "table not found, skipping");
        m_returnCode = ResultTableFileNotFound;
        return;
    }

    tgg00_FileId tableFileId;
    tableFile.BuildFileId( tableFileId );

    FileDir_Table auxFile;
    fDir.GetTableFile(
        trans.trTaskId_gg00,
        m_auxFileNo,
        auxFile,
        errMsg );
    if ( ! auxFile.IsAssigned() ) {
        errMsg.Overrule(
            Msg_List(
                Msg_List::Error ,
                SDBMSG_SRVTASKS_TABLESTATISTICS_TEMPFILE_NOT_FOUND,
                Msg_Arg("FILENO", FileDir_FileNoToString( m_auxFileNo ) )
                ) );
        SAPDBERR_ASSERT_STATE( false );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }


    tgg00_FileId auxFileId;
    auxFile.BuildFileId( auxFileId );
    // FIXME: the following two elements should be added by BuildFileId:
//     auxFileId.fileType_gg00().addElement( ftsConcurrent_egg00 );
//     auxFileId.fileType_gg00().addElement( ftsShared_egg00 );


    tbd_current_tree current;
    tbd_current_tree tempCurrent;

    AutoTreeLock tableTreeLock(
        trans, tableFileId, current, AutoTreeLock::LockNonExclusive );

    if ( trans.trError_gg00 != e_ok ) {
        SAPDBTRACE_WRITELN(
            DataTree_Trace, 5, "get tree failed for table xxx: trError");

        // table might have been deleted this very moment
        if ( tableFile.GetFileState() == Data_FileDeleted ) {
            SAPDBTRACE_WRITELN(
                DataTree_Trace, 5, "table has been deleted");
            m_returnCode = ResultTableFileNotFound;
            return;
        }

        errMsg.Overrule(
            Msg_List(
                Msg_List::Error ,
                SDBMSG_SRVTASKS_TABLESTATISTICS_GETTREE_FAILED,
                Msg_Arg("FILENO", FileDir_FileNoToString( tableFile.GetFileNo()) ),
                Msg_Arg("RETURNCODE",
                        SQLManErr_Interface::GetInstance().GetBaseErrorText(
                            trans.trError_gg00) )
                ) );
        SAPDBERR_ASSERT_STATE( false );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    AutoTreeLock auxTreeLock(
        trans, auxFileId, tempCurrent, AutoTreeLock::LockExclusive );

    if ( trans.trError_gg00 != e_ok ) {
        errMsg.Overrule(
            Msg_List(
                Msg_List::Error ,
                SDBMSG_SRVTASKS_TABLESTATISTICS_GETTEMP_FAILED,
                Msg_Arg("FILENO", FileDir_FileNoToString( tableFile.GetFileNo()) ),
                Msg_Arg("RETURNCODE",
                        SQLManErr_Interface::GetInstance().GetBaseErrorText(
                            trans.trError_gg00) )
                ) );
        SAPDBERR_ASSERT_STATE( false );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    /*
     * compute statistics
     */
    SAPDB_Int2 columnCount = requestInfo.columnCount;

    SAPDBERR_ASSERT_STATE( columnCount > 0 );
    SAPDBERR_ASSERT_STATE( requestInfo.stackEntries.GetSize() == columnCount );

    tgg00_StEntryAddr pColumnDescription =
        reinterpret_cast<tgg00_StEntryAddr>( &requestInfo.stackEntries[0] );


    Data_UpdStatRequestFile::DistinctValueVector
        distinctVals(
            *reinterpret_cast<SAPDBMem_IRawAllocator*>(
                trans.trAllocator_gg00 ) );

    // prepare distinctVals; we have to initialize columnCount
    // elements using Resize (instead of Reserve), so the vector has
    // the correct size because statistics computation just changes
    // the array contents but does not touch the size
    distinctVals.Clear();
    if ( ! distinctVals.Resize( columnCount ) ) {
        errMsg.Overrule(
            Msg_List(
                Msg_List::Error,
                SDBMSG_SRVTASKS_TABLESTATISTICS_OUT_OF_MEMORY,
                Msg_Arg("ALLOC_SIZE",
                        columnCount
                        * sizeof(Data_UpdStatRequestFile::DistinctValueVector
                                 ::ValueType) )
            ) );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    Data_PageCount          auxLeaves;
    Data_PageCount          auxRecords;
    Data_PrimTreeStatistic  primTreeStatistic( current );
    if ( m_noOfPagesToSample == 0 ) {
        // calculate statistics, don't sample
        Data_SerializableScanTreeAccess scanTreeAccessor( current );

        primTreeStatistic.CalculateColumnStatistic(
            scanTreeAccessor,
            tempCurrent,
            columnCount, 1, columnCount,
            pColumnDescription, auxLeaves, auxRecords,
            *reinterpret_cast<tgg00_ColumnDistinctValues*>(
                distinctVals.Begin() )
            );
    } else {
        const SAPDB_Bool bNoDuplicates = true;

        Data_RandomTreeAccess randomTreeAccessor(
            trans, current, m_noOfPagesToSample, ! bNoDuplicates );

        if( ! randomTreeAccessor.Initialize( m_noOfPagesToSample )) {
            errMsg.Overrule(
                Msg_List(
                    Msg_List::Error,
                    SDBMSG_SRVTASKS_TABLESTATISTICS_INITSAMPLINGITERATOR_FAILED,
                    Msg_Arg("FILENO",
                            FileDir_FileNoToString( tableFile.GetFileNo()) )
                    ) );
            trans.trError_gg00 = e_ok; // to be on the safe side
            m_Result           = error;
            m_returnCode       = ResultSystemError;
            return;
        }

        primTreeStatistic.SampleNewColumnStatistic(
            randomTreeAccessor,
            tempCurrent,
            columnCount, 1, columnCount,
            pColumnDescription, auxLeaves, auxRecords,
            *reinterpret_cast<tgg00_ColumnDistinctValues*>(
                distinctVals.Begin() )
            );
    }
    SAPDBTRACE_WRITELN(
        DataTree_Trace, 5,
        "processing done: trError = "
        << SQLManErr_Interface::GetInstance().GetBaseErrorText(
            trans.trError_gg00) );

    if ( trans.trError_gg00 != e_ok ) {

        if ( trans.trError_gg00 == e_cancelled ) {
            trans.trError_gg00 = e_ok;
            m_Result     = ok;
            m_returnCode = ResultCancelled;
            return;
        }

        if ( trans.trError_gg00 == e_old_fileversion ) {
            trans.trError_gg00 = e_ok;
            m_Result     = ok;
            m_returnCode = ResultOldFileVersion;
            return;
        }

        errMsg.Overrule(
            Msg_List(
                Msg_List::Error ,
                SDBMSG_SRVTASKS_TABLESTATISTICS_COLLECT_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( tableFile.GetFileNo()) ),
                Msg_Arg( "RETURNCODE",
                         SQLManErr_Interface::GetInstance().GetBaseErrorText(
                             trans.trError_gg00) )
                ) );
        trans.trError_gg00 = e_ok;
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }

    tableTreeLock.Release();
    auxTreeLock.Release();

    /*
     * set statistics in request file
     */
    itRetCode = it.SetStatistics(
        trans,
        auxLeaves, auxRecords,
        distinctVals, errMsg );

    if ( itRetCode != Data_UpdStatRequestFile::Iterator::Okay ) {
        errMsg.Overrule(
            Msg_List(
                Msg_List::Error ,
                SDBMSG_SRVTASKS_TABLESTATISTICS_COLLECT_FAILED,
                Msg_Arg( "FILENO",
                         FileDir_FileNoToString( m_updStatRequestFileNo ) ),
                Msg_Arg( "FILENO1",
                         FileDir_FileNoToString( tableFile.GetFileNo()) )
                ) );
        m_Result     = error;
        m_returnCode = ResultSystemError;
        return;
    }
}





//     Data_KeyBoundedPageAccess accessor ( PrimCurrent,
//                                          pStartPrimKey, StartPrimKeyLen,
//                                          pStopPrimKey,  StopPrimKeyLen );
//     Data_PrimTreeStatistic   collector( PrimCurrent );

//     collector.CalculateColumnStatistic(
//         InvCurrent,
//         accessor,
//         columnCount, columnFirstPosition, columnLastPosition,
//         pColumnDescription, leafCount, recordCount,
//         distinctValues );
// }

/*************************************************************************/


/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
