/*****************************************************************************/
/*!
  @file         SrvTasks_JobCollectTableStatistics.hpp
  @author       MartinKi
  @ingroup

  @brief        Header file for class SrvTasks_JobCollectTableStatistics.

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (C) 2003 SAP AG

  This program is free software; you can redisytribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif
*/
/*****************************************************************************/

#ifndef SRVTASKS_JOBCOLLECTTABLESTATISTICS_HPP
#define SRVTASKS_JOBCOLLECTTABLESTATISTICS_HPP

/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "DataAccess/Data_PrimTreeStatisticsCoordinator.hpp"
#include "DataAccess/Data_Types.hpp"
#include "FileDirectory/FileDir_FileNo.hpp"
#include "ServerTasks/SrvTasks_Job.hpp"

#include "Messages/Msg_List.hpp"
#include "RunTime/Tasking/RTETask_Task.h"
#include "SAPDBCommon/SAPDB_Types.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*!
 * @class  SrvTasks_JobCollectTableStatistics
 * @author MartinKi
 *
 * @brief Computes statistics for a given single table.
 *
 * @sa Data_PrimTreeStatisticsCoordinator
 */
class SrvTasks_JobCollectTableStatistics : public SrvTasks_Job
{
public:
    /// Possible return codes.
    enum ReturnCode {
        ResultOkay,
        ResultTableFileNotFound,
        ResultOldFileVersion,
        ResultCancelled,
        ResultSystemError
    };

    /// Constructor.
    SrvTasks_JobCollectTableStatistics(
        const Attributes&     jobAttributes,
        const FileDir_FileNo& updStatRequestFileNo,
        const Data_PrimTreeStatisticsCoordinator::TableKey& tableKey,
        const FileDir_FileNo& tableFileNo,
        const FileDir_FileNo& auxFileNo,
        const SAPDB_Int4      noOfPagesToSample,
        SrvTasks_WaitingTask& waitingTask);

    /*!
     * @brief Returns the table key of the table that is being processed.
     */
    const Data_PrimTreeStatisticsCoordinator::TableKey&
    GetTableKey() const;

    /*!
     * @brief Returns the fileno of the table that is being processed.
     */
    FileDir_FileNo GetTableFileNo() const;

    /*!
     * @brief Returns the fileno of the temporary file that is used to
     * store information during the statistics computation process.
     */
    const FileDir_FileNo& GetAuxFileNo() const;

    /*!
     * @brief Called by server task handling to start statistics computation.
     *
     * @copydoc SrvTasks_Jobs::Execute
     */
    void Execute();

    /*!
     * @brief Used to start statistics computation directly.
     *
     * @warning To check whether the computation was successful,
     * GetReturnCode should be used. GetResult will always return
     * StandardResult::ok as long as no SystemError occured. This
     * implies that GetResult will also return ok if the table could
     * not be processed because it could not be found. GetReturnCode
     * will return ResultTableFileNotFound in this case.
     *
     * @copydoc SrvTasks_Jobs::Execute
     */
    void Execute(tgg00_TransContext& trans);

    /*!
     * @brief Returns the result code of the execution.
     *
     * @return
     *   - ResultOkay, on success
     *   - ResultTableFileNotFound, if the table for which statistics
     *     should be computed could not be found (e.g. it might have
     *     been deleted after having been entered in the request file)
     *   - ResultOldFileVersion, if the job was aborted due to a file
     *     version change on the table processed
     *   - ResultCancelled, if the job was cancelled during execution
     *   - ResultSystemError, if an error occured during execution; in this
     *     case errMsg will be set
     */
    ReturnCode GetReturnCode(Msg_List& errMsg);

    /*!
     * @name Unused methods required by SrvTasks_Job interface
     */
    /*@{*/

    /*!
     * @copydoc SrvTasks_Jobs::CreateCopy
     */
    bool CreateCopy(
        SAPDBMem_IRawAllocator& allocator,
        Msg_List&               errlist,
        SrvTasks_Job*&          jobCopy ) const;

    /*!
     * @copydoc SrvTasks_Jobs::Deallocate
     */
    void Deallocate() {}
    /*@}*/

private:
    FileDir_FileNo m_tableFileNo;
    FileDir_FileNo m_updStatRequestFileNo;
    FileDir_FileNo m_auxFileNo;
    ReturnCode     m_returnCode;
    SAPDB_Int4     m_noOfPagesToSample;
    Data_PrimTreeStatisticsCoordinator::TableKey m_tableKey;

};

/**************************************************************************/

inline FileDir_FileNo
SrvTasks_JobCollectTableStatistics::GetTableFileNo() const
{
    return m_tableFileNo;
}

/**************************************************************************/

inline SrvTasks_JobCollectTableStatistics::ReturnCode
SrvTasks_JobCollectTableStatistics::GetReturnCode(
    Msg_List& errMsg)
{
    this->GetResult( errMsg ); // get message list, mark job done internally
    return m_returnCode;
}

/**************************************************************************/

#endif // SRVTASKS_JOBCOLLECTTABLESTATISTICS_HPP
