/*!
  @file        SrvTasks_Task.cpp
  @author      UweH
  @ingroup     ServerTasks
  @brief       This is the implementation of SrvTasks_Task
  @since       2005-01-04  16:09

\if EMIT_LICENCE
  copyright:    Copyright (c) 2001-2004 SAP AG


    ========== licence begin  GPL
    Copyright (c) 2001-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end
\endif
*/
#include "ServerTasks/SrvTasks_Task.hpp"
#include "ServerTasks/SrvTasks_Job.hpp"
#include "ServerTasks/SrvTasks_WaitingTask.hpp"
#include "ServerTasks/SrvTasks_Coordinator.hpp"
#include "ServerTasks/SrvTasks_Types.hpp"
#include "SAPDBCommon/Tracing/SAPDBTrace_Usage.hpp"
#include "SAPDBCommon/Messages/SDBMsg_SrvTasks.h"
#include "KernelCommon/Kernel_VTrace.hpp"
#include "RunTime/RTE_Crash.hpp"
#include "RunTime/RTE_Message.hpp"
#include "RunTime/Tasking/RTETask_ITask.hpp"
#include "RunTime/Tasking/RTETask_SuspendReason.h"
#include "RunTime/System/RTESys_MemoryBarrier.h"
#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "Transaction/Trans_Context.hpp"
#include "SQLManager/SQLMan_Context.hpp"
#include "heo55k.h" // RTE   : vbegexcl,vendexcl 
#include "hgg08.h"  // g08servertask

//--------------------------------------------------------------------------------------
SrvTasks_Task::SrvTasks_Task (Trans_Context &context)
: m_Context     (context),
  m_pJob        (0),
  m_Index       (SAPDB_MAX_UINT4),
  m_ReleaseSelf (false)
{
    if ( context.Task().Type() != RTETask_ITask::Server)
        RTE_Crash(Msg_List(Msg_List::Error,SDBMSG_SRVTASKS_WRONG_TASK_TYPE,
                           Msg_Arg("TYPE",SAPDB_ToString(context.Task().Type()))));

    if (! SrvTasks_Coordinator::Instance().Register(*this, m_Context.ErrorList()))
        RTE_Crash(m_Context.ErrorList());
}
//--------------------------------------------------------------------------------------
static inline void DestroySQLManContext(Trans_Context &context)
{
    if ( context.OldTrans().trAcvPtr_gg00 != 0 )
    {
        SQLMan_Context* pContext = reinterpret_cast<SQLMan_Context*>(context.OldTrans().trAcvPtr_gg00);
        destroy (pContext, context.Allocator());
        context.OldTrans().trAcvPtr_gg00 = 0;
    }
}
//--------------------------------------------------------------------------------------
void SrvTasks_Task::UnRegisterJob()
{
    vbegexcl(m_Context.Task().ID(),g08servertask);
    m_pJob = 0;
    vendexcl(m_Context.Task().ID(),g08servertask);
}
//--------------------------------------------------------------------------------------
void SrvTasks_Task::Run()
{
    SAPDBTRACE_METHOD_DEBUG ("SrvTasks_Task::Main", SrvTasks_Trace, 5);

    if ( SAPDB_MAX_UINT4 == m_Index )
    {
        m_Context.ErrorList() = Msg_List(Msg_List::Error,	SDBMSG_SRVTASKS_TASK_NOT_PROPERLY_REGISTERED,
                                         Msg_Arg("JOB_TASKID",SAPDB_ToString(m_Context.Task().ID())));
        RTE_Crash(m_Context.ErrorList());
    }
    
    RTE_TaskDisplayName      description;
	RTE_TaskDisplayNameShort shortDescription;
    
    while ( 1 )
    {
        if ( SrvTasks_Trace.TracesLevel(7) )
        {
            Kernel_VTrace trace;
            WriteToTraceIntern(trace,true);
            trace << " SUSPENDED";
        }
        m_Context.Task().ResetCancelled();
        m_Context.Task().Suspend(TSK_VSUSPEND_REASON_NO_WORK);
        m_Context.Task().ResetCancelled();

        if ( m_pJob == 0 )
        {
            m_Context.ErrorList() = Msg_List(Msg_List::Error,SDBMSG_SRVTASKS_TASK_STARTED_WITHOUT_JOB);
            RTE_Crash(m_Context.ErrorList());
        }
        
        m_pJob->GetDescription( description, shortDescription );
                         
        SrvTasks_Coordinator::Instance().SetTaskStarted (m_Context.Task(), m_Index, description, shortDescription );
        
        m_pJob->BeginExecute(*this);

        m_ReleaseSelf = ! m_pJob->GetAttributes().KeepServerTaskReserved;
    
        m_pJob->Execute();  // The jobs execution is starting here.
        
        if ( m_ReleaseSelf )
            SrvTasks_Coordinator::Instance().SetTaskStopped (m_Context.Task(), m_Index);

        if ( m_pJob->GetWaitingTask() != 0 )
        {
            m_pJob->EndExecute(); // after this the control over the job is switched back to the user task
            UnRegisterJob();
        }
        else
        {
            m_pJob->EndExecute(); // nobody tis waiting for this job
            
            // Nobody is waiting so we have to handle any error or message here
            // before destroying all contexts.

            if ( m_pJob->GetResult(m_Context.ErrorList()) == SrvTasks_Job::ok )
            {
                if ( ! m_Context.ErrorList().IsEmpty() )
                    RTE_Message ( m_Context.ErrorList()
                                  +
                                  Msg_List(Msg_List::Error,SDBMSG_SRVTASKS_ERRLIST_NOT_EMPTY) );
            }
            else
            {
                m_pJob->AppendContextInfo(__CONTEXT__,m_Context.ErrorList(),"The jobs result code is not OK.");
                RTE_Message (m_Context.ErrorList());
            }
            m_Context.ErrorList().ClearMessageList();
            destroy(m_pJob,m_Context.Allocator());
            UnRegisterJob();
        }
        if ( m_ReleaseSelf )
        {
            if ( ! SrvTasks_Coordinator::Instance().Release (m_Context.Task(), m_Index) )
            {
                m_Context.ErrorList().AppendNewMessage(Msg_List(Msg_List::Error, SDBMSG_SRVTASKS_RELEASE_TASK_FAILED,
                                                       Msg_Arg("JOB_TASKID", SAPDB_ToString(m_Context.Task().ID())),
                                                       Msg_Arg("INDEX",      SAPDB_ToString(m_Index))));
                RTE_Crash(m_Context.ErrorList());
            }
        }
        DestroySQLManContext(m_Context);
    } // endwhile
}
//--------------------------------------------------------------------------------------
void SrvTasks_Task::Abort(RTETask_ITask &usertask,
                          bool           inRegion)
{
    SAPDBTRACE_METHOD_DEBUG ("SrvTasks_Task::Abort", SrvTasks_Trace, 5);
    m_Context.Task().SetCancelled();
}
//--------------------------------------------------------------------------------------
bool SrvTasks_Task::RegisterJob (RTETask_ITask        & task,
                                 SrvTasks_Job         & job,
                                 Msg_List             & errlist,
                                 SrvTasks_WaitingTask * waitingTask )
{
    SAPDBTRACE_METHOD_DEBUG ("SrvTasks_Task::RegisterJob", SrvTasks_Trace, 5);

    vbegexcl(task.ID(),g08servertask);
    if ( m_pJob != 0 )
    {
        errlist = Msg_List(Msg_List::Error,SDBMSG_SRVTASKS_JOB_ALREADY_EXISTS);
        AppendContextInfo        (__CONTEXT__, errlist, "This Task should execute two jobs.");
        m_pJob->AppendContextInfo(__CONTEXT__, errlist, "This Job is just executed.");
        job.AppendContextInfo    (__CONTEXT__, errlist, "This Job should be executed.");
        Kernel_VTrace trace;
        trace << "ERROR: A servertask should execute two jobs." << FlushLine;
        SrvTasks_Task::WriteToTraceIntern (trace, true);
        trace << FlushLine << "The new jobs is: " << job;
        vendexcl(task.ID(),g08servertask);
        return false;
    }
    
    if ( waitingTask != 0 )
        job.SetWaitingTask(*waitingTask);
        
    if ( job.GetWaitingTask() != 0 )
        m_pJob = &job;
    else
    {
        if ( ! job.CreateCopy (m_Context.Allocator(), errlist, m_pJob) )
        {
            errlist = errlist + Msg_List(Msg_List::Error,	SDBMSG_SRVTASKS_ALLOCATION_FAILED);
            job.AppendContextInfo(__CONTEXT__,errlist,"Job could not be copied.");
            if ( ! SrvTasks_Coordinator::Instance().Release (m_Context.Task(), m_Index) )
            {
                errlist = Msg_List(Msg_List::Error, SDBMSG_SRVTASKS_RELEASE_TASK_FAILED,
                                   Msg_Arg("JOB_TASKID", SAPDB_ToString(m_Context.Task().ID())),
                                   Msg_Arg("INDEX",      SAPDB_ToString(m_Index)) );
                RTE_Crash(errlist);
            }
            return false;
        }
    }
    m_pJob->SetStarted(task);
    if ( SrvTasks_Trace.TracesLevel(7) )
    {
        Kernel_VTrace trace;
        WriteToTraceIntern(trace,true);
        trace << " RESUMED BY T" << task.ID();
    }
    task.ResumeOther(m_Context.Task());
    vendexcl(task.ID(),g08servertask);
    return true;
}
//--------------------------------------------------------------------------------------
bool SrvTasks_Task::IsAborted() const
{
    SAPDBTRACE_WRITELN(SrvTasks_Trace, 6, (m_Context.Task().IsCancelled() ? "CANCELLED" : ""));
    return m_Context.Task().IsCancelled();
}
//--------------------------------------------------------------------------------------
const char * SrvTasks_Task::StateString [] =
{
	"init",
    "nowork",
    "running",
    "waiting"
};
//--------------------------------------------------------------------------------------
void SrvTasks_Task::WriteToTraceIntern (Kernel_VTrace &trace,
                                        bool           withJob) const
{
    trace << "ServerTask #" << m_Index
          << " T" << m_Context.Task().ID()
          << " ";
	if ( withJob )
    {
        if ( m_pJob != 0 )
            trace << *m_pJob;
        else
            trace << ", no job";
    }
    else
        trace << (m_pJob != 0 ? ", job exists" : ", no job");
    trace << ", TaskAddr: 0x" << SAPDB_ToString((const void*)this,SAPDB_ToStringClass::hex);
}
//--------------------------------------------------------------------------------------
SAPDB_OStream & operator<< (SAPDB_OStream &os, const SrvTasks_ITaskForJob &task)
{
    const SrvTasks_Task * const pTask = static_cast<const SrvTasks_Task*>(&task);
    os    << "ServerTask #" << pTask->m_Index
          << " T" << pTask->m_Context.Task().ID()
          << (pTask->m_pJob != 0 ? ", job exists" : ", no job")
          << ", TaskAddr: 0x" << SAPDB_ToString((const void*)&task,SAPDB_ToStringClass::hex)
          << (pTask->m_ReleaseSelf ? ", release self" : "");
     return os;
}
//--------------------------------------------------------------------------------------
void SrvTasks_Task::AppendContextInfo (SAPDB_Char  const * const  filename,
                                       SAPDB_UInt4 const          linenumber,
                                       SAPDBErr_MessageList      &errlist,
                                       SAPDB_Char  const * const  title) const
{
    errlist.AddDetail(SDBMSG_SRVTASKS_SERVER_TASK_INFO,
                      Msg_Arg("SERVER_TASK_ID",SAPDB_ToString(m_Context.Task().ID())),
                      Msg_Arg("SERVER_TASK_INDEX",SAPDB_ToString(m_Index)),
                      Msg_Arg("SERVER_TASK_KIND",(m_ReleaseSelf?"dynamic":"static")),
                      Msg_Arg("DESCRIPTION",(title!=0?title:"")));
    if ( m_pJob != 0 )
        m_pJob->AppendContextInfo(filename,linenumber,errlist,"job of task");
    // void *m_pSessionContext;
}