/****************************************************************************/
/*!

  @file        SysView_FileDirectories.cpp

  -------------------------------------------------------------------------

  @author      MartinKi

  @ingroup     SystemViews
  @brief       This module implements the "FILEDIRECTORIES" view class.

  @see

\if EMIT_LICENCE
  ========== licence begin  GPL
  Copyright (c) 2003-2004 SAP AG

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  ========== licence end
\endif

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "SystemViews/SysView_Defines.hpp"
#include "SystemViews/SysView_ITableObj.hpp"
#include "SystemViews/SysView_FileDirectories.hpp"

#include "SQLManager/SQLMan_SessionIterator.hpp"
#include "SQLManager/SQLMan_Context.hpp"
#include "FileDirectory/FileDir_ISharedDirectory.hpp"
#include "FileDirectory/FileDir_IDirectory.hpp"
#include "Transaction/Trans_Context.hpp"

#include "RunTime/Tasking/RTETask_ITask.hpp"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/



/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS  (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

void SysView_FileDirectories::Create()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::Create", SysView_Trace, 5);

	m_Table->AppendCol (ITOCT_FIXED, SV_TASKID, 10);
	m_Table->AppendCol (ITOCT_CHARBYTE,  SV_MAXUSEDFILEID, 8);
	m_Table->AppendCol (ITOCT_FIXED, SV_ENTRYCOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_CACHEENTRYCOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_HASHTABLESLOTCOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_USEDHASHTABLESLOTCOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_MAXHASHTABLECHAINENTRYCOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_RESIZEHASHTABLECOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_USEDPAGECOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_STATISTICSDELTACOUNT, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_CURRENTSTATISTICSDELTACOUNT, 10);
}

/*---------------------------------------------------------------------------*/

SAPDB_Int SysView_FileDirectories::GetColCount()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::GetColCount", SysView_Trace, 5);

	return SV_CC_FILEDIRECTORIES;
}

/*---------------------------------------------------------------------------*/

SAPDB_Int SysView_FileDirectories::EstimateRows()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::EstimateRows", SysView_Trace, 5);

	return SV_ER_FILEDIRECTORIES;
}

/*---------------------------------------------------------------------------*/

void SysView_FileDirectories::Execute()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::Execute", SysView_Trace, 5);

    m_Table->GetCatalogTable();

    if (m_Context.IsOk())
    {
        this->QuerySharedFileDirectory();
        this->QuerySessionFileDirectories();
    }
}

/*---------------------------------------------------------------------------*/

void SysView_FileDirectories::QuerySharedFileDirectory()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::QuerySharedFileDirectory",
        SysView_Trace, 5);

    FileDir_ISharedDirectory& directory = FileDir_ISharedDirectory::Instance();

    SAPDB_UInt4 i;

    m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
    m_Table->MoveToCol (
        ITOVT_SURROGATE, directory.GetMaxFileNo().GetCharArray(), 0 );
    i = directory.GetEntryCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // cache entry count

    i = directory.GetHashTableSlotCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetUsedHashTableSlotCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetMaxHashTableChainEntryCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetResizeHashTableCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetPageCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetStatisticsDeltaCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

    i = directory.GetCurrentStatisticsDeltaCount();
    m_Table->MoveToCol (ITOVT_UINT4, &i, 0);
}

/*---------------------------------------------------------------------------*/

void SysView_FileDirectories::QuerySessionFileDirectories()
{
    SAPDBTRACE_METHOD_DEBUG (
        "SysView_FileDirectories::QuerySessionFileDirectory",
        SysView_Trace, 5);

    SQLMan_SessionIterator sessionIt;
    SQLMan_Context*        sessionContext;
    while (  sessionIt.Next( sessionContext ) && m_Context.IsOk() ) {
        Trans_Context& sessionTrans =
            Trans_Context::GetContext( sessionContext->TransContext() );

        FileDir_IDirectory* pSessionFDir = sessionTrans.FileDirectory();
        if ( ! pSessionFDir ) {
            // session is just being removed
            continue;
        }

        SAPDB_UInt i;
        RTE_TaskId sessionTaskId = sessionTrans.TaskId();
        m_Table->MoveToCol (ITOVT_INT4, &sessionTaskId, 0);
        m_Table->MoveToCol (
            ITOVT_SURROGATE, pSessionFDir->GetMaxTempFileNo().GetCharArray(), 0);

        i = pSessionFDir->GetEntryCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        i = pSessionFDir->GetCacheEntryCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        i = pSessionFDir->GetHashTableSlotCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        i = pSessionFDir->GetUsedHashTableSlotCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        i = pSessionFDir->GetMaxHashTableChainEntryCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        i = pSessionFDir->GetResizeHashTableCount();
        m_Table->MoveToCol (ITOVT_UINT4, &i, 0);

        m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
        m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
        m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
    }
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
