/****************************************************************************/
/*!

  @file        SysView_Files.cpp

  -------------------------------------------------------------------------

  @author      MartinKi

  @ingroup     SystemViews
  @brief       This module implements the "Files" view class.

  @see

*/
/*-------------------------------------------------------------------------

  copyright:    Copyright (c) 2002-2004 SAP AG


    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end



*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "ggg00.h"
#include "SystemViews/SysView_Defines.hpp"
#include "SystemViews/SysView_ITableObj.hpp"
#include "SystemViews/SysView_Files.hpp"

#include "SQLManager/SQLMan_Context.hpp"
#include "SQLManager/SQLMan_Types.hpp"
#include "SQLManager/SQLMan_SessionIterator.hpp"
#include "SAPDB/SQLManager/Catalog/Catalog_TableDescriptor.hpp"
#include "SAPDB/SQLManager/Catalog/Catalog_IndexDescriptor.hpp"

#include "FileDirectory/FileDir_ShortColumn.hpp"
#include "FileDirectory/FileDir_Index.hpp"
#include "FileDirectory/FileDir_Table.hpp"
#include "FileDirectory/FileDir_Generic.hpp"
#include "FileDirectory/FileDir_Iterator.hpp"
#include "FileDirectory/FileDir_ISharedDirectory.hpp"
#include "FileDirectory/FileDir_IDirectory.hpp"

#include "Transaction/Trans_Context.hpp"

#include "Messages/Msg_List.hpp"
#include "DataAccess/Data_Types.hpp"

#include "FrameControl/FrameCtrl_ICommon.hpp"

#include "SAPDBCommon/SAPDB_string.h"

#include "RunTime/Tasking/RTETask_ITask.hpp"

//#include "hak51.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

/*!
 *    @brief   constant for the maximal number of characters in a tablename
 */
#define TABLENAMELENGTH 32

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS  (PROTOTYPES)                                     *
 *===========================================================================*/

const char c_unknown[] = "UNKNOWN";

/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/

SysView_Files::SysView_Files(
    SQLMan_Context  &context,
    tak40_show_glob &a41v)
    : SysView_SysViewBase (context, a41v),
      m_pageSize (FrameCtrl_ICommon::GetInstance().GetPageSize() / 1024)
{}

/*---------------------------------------------------------------------------*/

void SysView_Files::Create()
{
    SAPDBTRACE_METHOD_DEBUG ("SysView_Files::Create", SysView_Trace, 5);

	m_Table->AppendCol (ITOCT_CHARBYTE, SV_FILEID, 8);
	m_Table->AppendCol (ITOCT_CHARUNICODE, SV_SCHEMANAME, 32);
	m_Table->AppendCol (ITOCT_CHARUNICODE, SV_TABLENAME, TABLENAMELENGTH);
	m_Table->AppendCol (ITOCT_CHARUNICODE, SV_INDEXNAME, 32);
	m_Table->AppendCol (ITOCT_FIXED, SV_TASKID, 10);
	m_Table->AppendCol (ITOCT_FIXED, SV_ROOT, 10);
	m_Table->AppendCol (ITOCT_CHAR, SV_TYPE, 20);
	m_Table->AppendCol (ITOCT_CHAR, SV_FILESTATE, 20);
	m_Table->AppendCol (ITOCT_FIXED, SV_ENTRYCOUNT, 20);
	m_Table->AppendCol (ITOCT_FIXED, SV_TREEINDEXSIZE, 20);
	m_Table->AppendCol (ITOCT_FIXED, SV_TREELEAVESSIZE, 20);
	m_Table->AppendCol (ITOCT_FIXED, SV_LOBSIZE, 20);
}

/*---------------------------------------------------------------------------*/

SAPDB_Int SysView_Files::GetColCount()
{
    SAPDBTRACE_METHOD_DEBUG ("SysView_Files::GetColCount", SysView_Trace, 5);

    return SV_CC_FILES;
}

/*---------------------------------------------------------------------------*/

SAPDB_Int SysView_Files::EstimateRows()
{
    SAPDBTRACE_METHOD_DEBUG ("SysView_Files::EstimateRows", SysView_Trace, 5);

    return SV_ER_FILES;
}

/*---------------------------------------------------------------------------*/

void SysView_Files::Execute()
{
    SAPDBTRACE_METHOD_DEBUG ("SysView_Files::Execute", SysView_Trace, 5);

    const SAPDB_Bool c_outputTaskId = true;

    m_Table->GetCatalogTable();

    if (m_Context.IsOk())
    {
        FileDir_Iterator it( m_Context.GetAllocator() );
        FileDir_IDirectory* fdir =
            &FileDir_ISharedDirectory::IFileDirInstance();

        fdir->GetPrivateFileDirIterator( it, &m_Context.TransContext() );

        this->OutputFdir(
            it,
            ! c_outputTaskId,
            0 );

        // output session-local files
        SQLMan_SessionIterator sessionIt;
        SQLMan_Context*        sessionContext;
        while (  sessionIt.Next( sessionContext ) && m_Context.IsOk() ) {
            Trans_Context& sessionTrans =
                Trans_Context::GetContext( sessionContext->TransContext() );

            FileDir_IDirectory* pSessionFDir = sessionTrans.FileDirectory();
            if ( ! pSessionFDir ) {
                // session is just being removed
                continue;
            }

            RTE_TaskId sessionTaskId = sessionTrans.TaskId();
            pSessionFDir->GetPrivateFileDirIterator( it, &m_Context.TransContext() );
            this->OutputFdir(
                it,
                c_outputTaskId,
                sessionTaskId );
        }
    }
}

/*---------------------------------------------------------------------------*/

void SysView_Files::OutputFdir(
    FileDir_Iterator&   it,
    const SAPDB_Bool    outputSessionNo,
    const RTE_TaskId    outputTaskId)
{
    SAPDBTRACE_METHOD_DEBUG ("SysView_Files::OutputFdir", SysView_Trace, 5);

    Msg_List msgList;
    FileDir_Generic fileEntry;
    RTE_TaskId taskId = m_Context.TransContext().trTaskId_gg00;

    FileDir_IDirectory& fdir = *(it.GetFileDirectory());

    while ( it.GetNext( fileEntry ) ) {
        FileDir_Table tableFile;
        FileDir_Index indexFile;

        m_Table->MoveToCol(
            ITOVT_SURROGATE, fileEntry.GetFileNo().GetCharArray(), 0 );

        switch ( fileEntry.GetFileType() ) {
        case Data_TableFile:
            tableFile.Assign( fileEntry );
            break;
        case Data_ShortColumnFile:
        {
            FileDir_ShortColumn fs;
            fs.Assign( fileEntry );
            fdir.GetTableFile(
                taskId,
                fs.GetParentFileNo(),
                tableFile,
                msgList);
            break;
        }
        case Data_IndexFile:
            indexFile.Assign( fileEntry );
            fdir.GetTableFile(
                taskId,
                indexFile.GetParentFileNo(),
                tableFile,
                msgList);
            break;
        }

        if ( tableFile.IsAssigned() ) {
            Catalog_TableDescriptor
                tableDesc( m_Context, tableFile.GetFileNo() );
            if ( tableDesc ) {
                m_Table->MoveToCol (
                    ITOVT_KNLIDENTIFIER, tableDesc.GetSchemaName(), 0);
                m_Table->MoveToCol (
                    ITOVT_KNLIDENTIFIER, tableDesc.GetTableName(), 0);

                if ( indexFile.IsAssigned() ) {
                    Catalog_IndexDescriptor indexDesc =
                        tableDesc.GetIndexDescriptor( indexFile.GetIndexId() );
                    if ( indexDesc ) {
                        m_Table->MoveToCol (
                            ITOVT_KNLIDENTIFIER, indexDesc.GetName(), 0);
                    } else {
                        m_Context.ResetReturnCode(); // ignore errors
                        m_Table->MoveToCol (
                            ITOVT_CHARPTR, c_unknown, sizeof(c_unknown)-1 );
                    }
                } else {
                    m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
                }
            } else {
                m_Context.ResetReturnCode(); // ignore errors
                if ( tableFile.IsPersistent() ) {
                    // persistent file should have had schema and table name
                    m_Table->MoveToCol (
                        ITOVT_CHARPTR, c_unknown, sizeof(c_unknown)-1 ); // schema
                    m_Table->MoveToCol (
                        ITOVT_CHARPTR, c_unknown, sizeof(c_unknown)-1 ); // tablename
                    m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // indexname
                } else {
                    // for temp files it is ok not to have any catalog
                    // information associated
                    m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
                    m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
                    m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
                }
            }
        } else {
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
        }

        if ( outputSessionNo ) {
            m_Table->MoveToCol (ITOVT_INT4, &outputTaskId, 0);
        } else {
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
        }
        SAPDB_Int4 rootPage = fileEntry.GetRootPage();
        m_Table->MoveToCol (ITOVT_INT4, &rootPage, 0);

        switch ( fileEntry.GetFileType() ) {
        case Data_TableFile:
            if ( fileEntry.GetComment() == 0 ) {
                m_Table->MoveToCol (ITOVT_CHARPTR, "TABLE", 5);
            } else {
                char comment[TABLENAMELENGTH];
                SAPDB_strncpy(
                    comment,
                    fileEntry.GetComment(),
                    TABLENAMELENGTH );
                comment[ TABLENAMELENGTH-1 ] = '\0';
                for ( char* p = comment; *p != '\0'; ++p ) {
                    *p = toupper( *p ); // output needs to be uppercase
                }
                m_Table->MoveToCol (
                    ITOVT_CHARPTR,
                    comment,
                    SAPDB_strlen( comment ) );
            }
            break;
        case Data_IndexFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "INDEX", 5);
            break;
        case Data_ShortColumnFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "SHORT COLUMN FILE", 17);
            break;
        case Data_OmsFixedFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "FIXED OBJECT", 12);
            break;
        case Data_OmsVarFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "VARIABLE OBJECT", 15);
            break;
        case Data_OmsKeyFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "KEYED OBJECT", 12);
            break;
        case Data_OmsKeyPartitionFile:
            m_Table->MoveToCol (ITOVT_CHARPTR, "KEYED OBJECT INDEX", 18);
            break;
        default:
            m_Table->MoveToCol (ITOVT_CHARPTR, "UNKNOWN", 7);
        }

        switch ( fileEntry.GetFileState() ) {
        case Data_FileOk:
            m_Table->MoveToCol (ITOVT_CHARPTR, "OK", 2);
            break;
        case Data_FileDeleted:
            m_Table->MoveToCol (ITOVT_CHARPTR, "DELETED", 7);
            break;
        case Data_FileNotAccessible:
        case Data_FileBad:
            m_Table->MoveToCol (ITOVT_CHARPTR, "BAD", 3);
            break;
        case Data_FileReadOnly:
            m_Table->MoveToCol (ITOVT_CHARPTR, "READ ONLY", 9);
            break;
        default:
            m_Table->MoveToCol (ITOVT_CHARPTR, "UNKNOWN", 7);
        }

        if ( fileEntry.AreCountersInitialized() ) {
            SAPDB_UInt8 entryCount = fileEntry.GetEntryCount( taskId );
            m_Table->MoveToCol (ITOVT_UINT8, &entryCount, 0);

            SAPDB_Int4 leafPageCount = fileEntry.GetLeafPageCount( taskId );
            SAPDB_Int4 kbSize;
            if ( (fileEntry.GetFileType() != Data_OmsFixedFile)
                 && (fileEntry.GetFileType() != Data_OmsVarFile)
                 && (fileEntry.GetFileType() != Data_OmsKeyFile) ) {
                SAPDB_Int4 indexPageCount =
                    fileEntry.GetIndexPageCount( taskId );
                kbSize = indexPageCount * m_pageSize;
                m_Table->MoveToCol (ITOVT_INT4, &kbSize, 0);
            } else {
                m_Table->MoveToCol (ITOVT_NULL, NULL, 0);
            }

            kbSize = leafPageCount * m_pageSize;
            m_Table->MoveToCol (ITOVT_INT4, &kbSize, 0);
        } else {
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // entry count
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // index page size
            m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // leaf page size
        }

        // PTS 1133892 M.Ki
        m_Table->MoveToCol (ITOVT_NULL, NULL, 0); // lob size
    }
}

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
