/*!
    @file    ToolsParsersUtil_IdentifierCheck.hpp
    @author  BerndV
    @brief   Class to analyze parameter values

    \if EMIT_LICENCE

    ========== licence begin  GPL
    Copyright (c) 1998-2005 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end

    \endif
*/

#ifndef _ToolsParsersUtil_IdentifierCheck_HPP_
#define _ToolsParsersUtil_IdentifierCheck_HPP_

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */

/*! @brief Class to analyze parameter values which are "identifiers" */
class ToolsParsersUtil_IdentifierCheck
{
public:
  /*! @brief   checks wether it is a "very simple identifier"

              A "very simple identifier" is a non empty string. It starts with a 
              letter ('A...Z','a...z') or a special character ('#','@','$'). 
              After the first charcater a "very simple identifier" has letters 
              ('A...Z','a...z'), special characters ('#','@','$') or digits ('0...9').

     @param   szString [IN] zero terminated ASCII string

     @return  true if it is a "very simple identifier"
   */
  static bool IsVerySimpleIdentifier
      ( const char * szString );

  /*! @brief   checks wether it is a "simple identifier"

              A "simple identifier" is a non empty string. It starts with a 
              letter ('A...Z','a...z'), a special character ('#','@','$') or 
              a "extended language letter" (like german umlauts). 
              After the first charcater  a "simple identifier" has letters 
              ('A...Z','a...z'), special characters ('#','@','$'),a "extended 
              language letter" (like german umlauts) or digits ('0...9').
              The function uses the same "extended language letters" like the 
              MaxDB kernel.

     @param   szString [IN] zero terminated ASCII string

     @return  true if it is a "simple identifier"
   */
  static bool IsSimpleIdentifier
      ( const char * szString );

  /*! @brief   makes a "very simple identifier" uppercase

              If it is not a "very simple identifier" the function does not
              change the string content.

     @param   szString [INOUT] zero terminated ASCII string

     @return  false if it is not "very simple identifier"
   */
  static bool VerySimpleToupper
      ( char * szString );

  /*! @brief   makes a "very simple identifier" lowercase 

              If it is not a "very simple identifier" the function does not
              change the string content.

     @param   szString [INOUT] zero terminated ASCII string

     @return  false if it is not "very simple identifier"
   */
  static bool VerySimpleTolower
      ( char * szString );

  /*! @brief   makes a "simple identifier" uppercase 

              If it is not a "simple identifier" the function does not
              change the string content.

     @param   szString [INOUT] zero terminated ASCII string

     @return  false if it is not "simple identifier"
   */
  static bool SimpleToupper
      ( char * szString );

  /*! @brief   makes a "simple identifier" lowercase 

              If it is not a "simple identifier" the function does not
              change the string content.

     @param   szString [INOUT] zero terminated ASCII string

     @return  false if it is not "simple identifier"
   */
  static bool SimpleTolower
      ( char * szString );

private:
  static unsigned char oSimpleTolower     [256];
  static unsigned char oSimpleToupper     [256];
  static bool          oSimpleFirst       [256];
  static bool          oSimpleTail        [256];

  static unsigned char oVerySimpleTolower [256];
  static unsigned char oVerySimpleToupper [256];
  static bool          oVerySimpleFirst   [256];
  static bool          oVerySimpleTail    [256];
}; // end class ToolsParsersUtil_IdentifierCheck

#endif // _ToolsParsersUtil_IdentifierCheck_HPP_
