/*!
 * \file    LVCMem_GlobalAnchorRegister.cpp
 * \author  ChristophR
 * \brief   LiveCache global user allocator register.
 */
/*

    ========== licence begin  GPL
    Copyright (c) 2002-2004 SAP AG

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License
    as published by the Free Software Foundation; either version 2
    of the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
    ========== licence end


*/

#include "SAPDBCommon/MemoryManagement/SAPDBMem_NewDestroy.hpp"
#include "liveCache/MemoryManagement/LVCMem_GlobalAnchorRegister.hpp"
#include "liveCache/MemoryManagement/LVCMem_GlobalAnchorItem.hpp"
#include "RunTime/Synchronisation/RTESync_Spinlock.hpp"
#include "hsp77.h"
//#include "RunTime/Synchronisation/RTESync_RWRegion.hpp"

//#include "Oms/OMS_GlobalAnchorObject.hpp"
//#include "Oms/OMS_AllocatorInterface.hpp"

/*---------------------------------------------------------------------------*/

/// Bucket count for allocator lvcGlobalAnchorHash.
enum {
  BUCKET_CNT  = 127
};


/// Hash item for global user allocator list.
class LVCMem_AnchorRegisterItem {
public:
  LVCMem_GlobalAnchorItem    *m_anchorItem;
  SAPDB_Int4                 m_areaId;
  SAPDB_Int4                 m_anchorId;
  LVCMem_AnchorRegisterItem  *m_next;      ///< next item in lvcGlobalAnchorHash bucket
  
  //RTESync_RWRegion          *m_rwRegion;

  LVCMem_AnchorRegisterItem(LVCMem_GlobalAnchorItem *anchorItem, 
     SAPDB_Int4 areaId, SAPDB_Int4 anchorId, LVCMem_AnchorRegisterItem *next)
    : m_anchorItem(anchorItem), m_next(next), m_areaId(areaId), m_anchorId(anchorId)
  {}
};

typedef LVCMem_AnchorRegisterItem *LVCMem_AnchorRegisterItemPtr;

/*---------------------------------------------------------------------------*/

/// Global user allocator hash items.
static LVCMem_AnchorRegisterItemPtr  lvcGlobalAnchorItems[BUCKET_CNT] = {NULL};

/// Global lockId counter;
static SAPDB_Int4 lvcLockIdCounter = 1;

/// Lock for global anchor directory.
static RTESync_Spinlock                     lvcGlobalAnchorLock;

/// Default instance.
LVCMem_GlobalAnchorRegister          LVCMem_GlobalAnchorRegister::m_instance;

/*---------------------------------------------------------------------------*/

LVCMem_GlobalAnchorItem *LVCMem_GlobalAnchorRegister::get(
  SAPDBMem_IRawAllocator &alloc, SAPDBMem_IRawAllocator &blockalloc,
  SAPDB_Int4 areaId, SAPDB_Int4 anchorId)
{
  size_t bucket = (SAPDB_Int4)((((SAPDB_Int8)areaId<<32) | (SAPDB_Int8)anchorId) % BUCKET_CNT);

  RTESync_LockedScope lck(lvcGlobalAnchorLock);

  LVCMem_AnchorRegisterItem *i = lvcGlobalAnchorItems[bucket];
  while (i) {
    if ( i->m_areaId == areaId && i->m_anchorId == anchorId) {
        // return existing anchorItem with the same areaId and identifier
        return i->m_anchorItem;
      }
      i = i->m_next;
  }
  // create new anchor object
  if(lvcLockIdCounter == MAX_INT4_SP00)
    return NULL; //TODO (id reuse)

  char name[41];
  sp77sprintf(name,40,"OMS GSM(%d,%d)",areaId,anchorId);

  // create new global shared memory allocator
  LVCMem_UserAllocator *a = ::new(alloc) 
    LVCMem_UserAllocator(reinterpret_cast<const SAPDB_UTF8*>(name), blockalloc, 32 * 1024, 32 * 1024, 
    SAPDBMem_RawAllocator::FREE_RAW_EXTENDS);
  a->SetBadAllocHandler(SAPDBMem_RawAllocator::THROW_STD_BAD_ALLOC);

  LVCMem_GlobalAnchorItem *anchorItem = ::new(alloc) LVCMem_GlobalAnchorItem(a,lvcLockIdCounter++);
  
  //anchorItem->m_lock = RTESync_CreateRWRegion(lockId++, m_spinlock_pool, alloc);
  //anchorItem->m_lock->enter(false);

  lvcGlobalAnchorItems[bucket] = ::new(alloc) 
    LVCMem_AnchorRegisterItem(anchorItem, areaId, anchorId, lvcGlobalAnchorItems[bucket]);

  return lvcGlobalAnchorItems[bucket]->m_anchorItem;
}

/*---------------------------------------------------------------------------*/
