/*
 * ArchiveTar.xs
 */

#ifndef __cplusplus
#ifndef HAS_BOOL
typedef char bool;
#endif
#endif

#ifndef WIN32
#define UNIX UNIX
#endif

#ifdef UNIX
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#endif

#ifdef WIN32
#include <malloc.h>
#include <string.h>
#endif

#include "zlib.h"

#ifdef _WIN64
typedef __int64 ptr_t;
#else
typedef long ptr_t;
#endif

#ifdef __hpux
#define DONT_DECLARE_STD
#endif

#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#ifdef PERL_OBJECT
#define FILE struct _PerlIO
#endif

/*
 * expect compiler warning 4113 when using Microsoft C
 * in c code generated by xs preprocessor
 */
#if _MSC_VER >= 1200
#pragma warning(disable : 4113)
#endif

/* unix tar typeflag */
#define TAR_AREGTYPE     '\0'
#define TAR_REGTYPE      '0'
#define TAR_LNKTYPE      '1'
#define TAR_SYMTYPE      '2'
#define TAR_CHRTYPE      '3'
#define TAR_BLKTYPE      '4'
#define TAR_DIRTYPE      '5'
#define TAR_FIFOTYPE     '6'

/* gnu tar typeflags */
#define TAR_GNUTYPE      'A'
#define TAR_LONGNAMETYPE 'L'

/* tar magic */
static char TAR_MAGIC[5] = {'u', 's', 't', 'a', 'r'};

/* tar header */
typedef struct {
	char name[100];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[8];
	char typeflag[1];	
	char linkname[100];
	char magic[6];
	char version[2];
	char uname[32];
	char gname[32];
	char devmajor[8];
	char devminor[8];
	char prefix[155];
} tarheader_t;

typedef struct {
	gzFile fp;
	tarheader_t hdr[1];
	int size;
	char *fn;
	int pos;
} tararch_t;

typedef struct {
	tararch_t *arch;
	int size;
	int pos;
} tarfile_t;

/*
 * the BLOCKSIZE read from a file must be 512 bytes,
 * because the tar header and gzip are using 512 byte blocks
 */
#define BLOCKSIZE 512

static int
strnocttoi (char *ptr, int cnt) {
	char c;
	int val;
  
	for (val = 0; cnt > 0; cnt--) {
		c = *ptr++;
		if (c == ' ')
			continue;

		if (c == '\0')
			break;

		val = (val << 3) + (c - '0');
	}

  	return (val);
}

static int
tarrewind (tararch_t *arch) {
	if (arch == 0)
		return -1;

	gzrewind (arch->fp);

	arch->size = 0;
	arch->pos = 0;
	memset (&arch->hdr[0], 0, sizeof (tarheader_t));
	return 0;
}

static int
tarnext (tararch_t* arch) {
	char buff[BLOCKSIZE];
	tarheader_t *hdr;
	int rc, got, tail, pad;

	/* find next block */
	tail = arch->pos % BLOCKSIZE;
	if (tail != 0) {
		got = gzread (arch->fp, buff, BLOCKSIZE - tail);
		if (got < 0) {
			return -1;
		}

		arch->pos += got;
	}

	for (;;) {
		got = gzread (arch->fp, buff, BLOCKSIZE);
		if (got < 0) {
			return -1;
		}

		arch->pos += got;

		if (got != BLOCKSIZE) {
			return -1;
		}

		if (arch->size == 0) {
			/* next header expected */
			hdr = (tarheader_t *) buff;
		
			/* check end of file or null header */
			if ((got == 0)  || (hdr->name[0] == 0)) {
				memset (arch->hdr, 0, sizeof (tarheader_t));
				return 0;
			}

			/* check header magic */
			if (memcmp (hdr->magic, TAR_MAGIC, sizeof (TAR_MAGIC)) != 0) {
				memset (arch->hdr, 0, sizeof (tarheader_t));

				gzclose (arch->fp);
				arch->fp = 0;

				return -1;
			}

			/* keep header information */			
			memcpy (arch->hdr, hdr, sizeof (tarheader_t));
			arch->size = strnocttoi (hdr->size, sizeof (hdr->size));

			return 0;
		} else if (arch->size > BLOCKSIZE) {
			arch->size -= BLOCKSIZE;
		} else {
			arch->size = 0;
		}
	}

	return (-1);
}

static int
tarfind (tararch_t *arch, char *filename) {
	/* is current position */

	if (strcmp (arch->hdr->name, filename) == 0)
		return 0;

	/* search whole archive */
	tarrewind (arch);
	while (tarnext (arch) == 0) {
		/* end of archive */
		if (arch->hdr->name[0] == '\0')
			return -1;

		if (strcmp (arch->hdr->name, filename) == 0)
			return 0;
	}

	return -1;
}

#if 0
static void
hexdump (void *buff, int size) {
	int i;

	for (i = 0; i < size; i++) {
		if (i % 16 == 0)
			printf ("%05d: ", i);

		printf ("%02x ", *(((unsigned char *)buff) + i));
		if (i % 16 == 15)
			printf ("\n");
	}

	printf ("\n");
}
#endif

static int
tarread (tararch_t *arch, char *buff, int want) {
	int got;
 
	if (arch->size == 0)
		return 0;

	if (want > arch->size) {
		got = gzread (arch->fp, buff, arch->size);
		if (got < 0)
			return -1;

		arch->size = 0;
		arch->pos += got;
	} else {
		got = gzread (arch->fp, buff, want);
			if (got < 0)
				return -1;

		arch->size -= want;
		arch->pos += got;
	}

	return (got);
}

static tararch_t *
getarch (HV *self) {
	char *key = "arch";
	SV **psv;

	psv = hv_fetch (self, key, strlen (key), 0);
	if (psv == 0)
		return 0;

	return ((tararch_t *) SvIV (*psv));
}

static tarfile_t *
getfile (HV *self) {
    char *key = "file";
    SV **psv; 

    psv = hv_fetch (self, key, strlen (key), 0);
    if (psv == 0)
            return 0;

    return ((tarfile_t *) SvIV (*psv)); 
}   

/*
 * implementaion of sysseek() is really poor
 * only set to current position is implemented
 */
MODULE = SDB::Common::Archive::Tar PACKAGE = SDB::Common::Archive::Tar

PROTOTYPES: DISABLE

void
rewind (...)
PREINIT:
	tararch_t *arch;
PPCODE:
	if (items != 1)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	arch = getarch ((HV*) SvRV (ST(0)));
	if (arch == 0)
		XSRETURN_UNDEF;

	if (tarrewind (arch) < 0)
		XSRETURN_UNDEF;

	XPUSHs (sv_2mortal (newSViv (0)));
	XSRETURN (1);

void
read (...)
PREINIT:
	tararch_t *arch;
PPCODE:
	if (items != 1)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	arch = getarch ((HV*) SvRV (ST(0)));
	if (arch == 0)
		XSRETURN_UNDEF;

	for (;;) {
		if (tarnext (arch) < 0)
			XSRETURN_UNDEF;

		if (arch->hdr->name[0] == '\0') {
			XSRETURN_UNDEF;
		}

		if (arch->hdr->typeflag[0] == TAR_REGTYPE)
			break;

		if (arch->hdr->typeflag[0] == TAR_AREGTYPE)
			break;
	}

	XPUSHs (sv_2mortal (newSVpv (arch->hdr->name, strlen (arch->hdr->name))));
	XSRETURN (1);

void
new (...)
PREINIT:
	HV *self;
	char *fn;
	tararch_t *arch;
	char *CLASS = 0;
	const char *key = "arch";
	Stat_t sb[1];
	char buff[BLOCKSIZE];
	tarheader_t *hdr;
	int got;
CODE:
	if (items < 1 || items > 2)
		XSRETURN_UNDEF;
	
	if (items == 2 && (strlen (SvPV_nolen (ST(0))) > 0))
		CLASS = SvPV_nolen (ST(0));

	if (CLASS == 0)
		CLASS = "SDB::Common::Archive::Tar";

	fn = (char *) SvPV_nolen (ST (items - 1));
	if (strlen (fn) == 0)
		XSRETURN_UNDEF;

	if (stat (fn, sb) != 0)
		XSRETURN_UNDEF;

	if (S_ISREG (sb->st_mode) == 0)
		XSRETURN_UNDEF;

	arch = (tararch_t *) malloc (sizeof (tararch_t));
	if (arch == 0)
		XSRETURN_UNDEF;

	memset (arch, 0, sizeof (tararch_t));

	arch->fn = (char *) malloc (strlen (fn) + 1);
	if (arch->fn == 0) {
		free (arch);
		XSRETURN_UNDEF;
	}

	strcpy (arch->fn, fn);

	arch->fp = gzopen (fn, "rb");
	if (arch->fp == 0) {
		free (arch->fn);
		free (arch);
		XSRETURN_UNDEF;
	}

	got = gzread (arch->fp, &buff, BLOCKSIZE);
	if (got != BLOCKSIZE) {
		gzclose (arch->fp);
		free (arch->fn);
		free (arch);
		XSRETURN_UNDEF;
	}

	if (gzrewind (arch->fp) < 0) {
		gzclose (arch->fp);
		free (arch->fn);
		free (arch);
		XSRETURN_UNDEF;
	}

	hdr = (tarheader_t *) buff;
	if (memcmp (hdr->magic, TAR_MAGIC, sizeof (TAR_MAGIC)) != 0) {
		gzclose (arch->fp);
		free (arch->fn);
		free (arch);
		XSRETURN_UNDEF;
	}

	self = newHV ();
	ST(0) = sv_bless (newRV ((SV*) self), gv_stashpv (CLASS, 1));
	hv_store (self, key, strlen (key), newSViv ((ptr_t) arch), 0);
	sv_2mortal (ST(0));
	SvREFCNT_dec (self);

void
DESTROY (...)
PREINIT:
	tararch_t *arch;
PPCODE:
	if (items != 1)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	arch = getarch ((HV*) SvRV (ST(0)));
	if (arch == 0)
		XSRETURN_UNDEF;

	if (arch->fp != 0) {
		gzclose (arch->fp);
		arch->fp = 0;
	}

	if (arch->fn != 0) {
		free (arch->fn);
		arch->fn = 0;
	}

	if (arch != 0) 
		free (arch);

	XSRETURN_UNDEF;

void
open (...)
PREINIT:
	SV *sv;
	HV *self;
	tararch_t *arch;
	tarfile_t *file;
	char *filename;
	const char *key = "file";
	const char *CLASS = "SDB::Common::Archive::Tar::File";
PPCODE:
	if (items < 1 || items > 2)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	arch = getarch ((HV*) SvRV (ST(0)));
	if (arch == 0)
		XSRETURN_UNDEF;

	filename = (char *) SvPV_nolen (ST (1));
	if (strlen (filename) == 0)
		XSRETURN_UNDEF;

	if (tarfind (arch, filename) != 0)
		XSRETURN_UNDEF;

	file = (tarfile_t *) malloc (sizeof (tarfile_t));
	if (file == 0)
		XSRETURN_UNDEF;
		
	memset (file, 0, sizeof (tarfile_t));
	file->arch = arch;
	file->size = arch->size;
	file->pos = 0;

	self = newHV ();
	hv_store (self, key, strlen (key), newSViv ((ptr_t) file), 0);
	sv = sv_bless (newRV ((SV*) self), gv_stashpv (CLASS, 1));

	XPUSHs (sv_2mortal (sv));
	XSRETURN (1);

MODULE = SDB::Common::Archive::Tar   PACKAGE = SDB::Common::Archive::Tar::File

PROTOTYPES: DISABLE

void
DESTROY (...)
PREINIT:
	tararch_t *arch;
	tarfile_t *file;
PPCODE:
	if (items != 1)
		XSRETURN (0);

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN (0);

	file = getfile ((HV*) SvRV (ST(0)));
	if (file == 0)
		XSRETURN (0);

	arch = file->arch;
	if (arch == 0)
		XSRETURN (0);

	free (file);
	XSRETURN (0);

void
close (...)
PREINIT:
	HV *self;
	tararch_t *arch;
	tarfile_t *file;
	const char *key = "file";
PPCODE:
	if (items != 1)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	self = (HV*) SvRV (ST(0));
	file = getfile (self);
	if (file == 0)
		XSRETURN_UNDEF;

	arch = file->arch;
	if (arch == 0)
		XSRETURN_UNDEF;

	if (arch->fp == 0) {
		XPUSHs (sv_2mortal (newSViv (0)));
		XSRETURN (1);
	}

	free (file);
	hv_store (self, key, strlen (key), newSViv (0), 0);

	XPUSHs (sv_2mortal (newSViv (0)));
	XSRETURN (1);

void
sysread (...)
PREINIT:
	tarfile_t *file;
	tararch_t *arch;
	int want, got, off;
	char *buff;
PPCODE:
	if (items != 3)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	file = getfile ((HV*) SvRV (ST(0)));
	if (file == 0)
		XSRETURN_UNDEF;

	arch = file->arch;
	if (arch == 0)
		XSRETURN_UNDEF;

	if (SvREADONLY (ST(1)))
		XSRETURN_UNDEF;

	want = SvIV (ST (2));

	buff = (char *) malloc (want);	
	if (buff == 0)
		XSRETURN_UNDEF;

	got = tarread (arch, buff, want);
	if (got < 0) {
		free (buff);
		XSRETURN_UNDEF;
	}

	file->pos += off;

	sv_setpvn (ST(1), (const char *) buff, got);
	free (buff);

	XPUSHs (sv_2mortal (newSViv (got)));
	XSRETURN (1);

void
read (...)
PREINIT:
	SV *sv;
	tarfile_t *file;
	tararch_t *arch;
	int want, got, off;
	char *buff;
PPCODE:
	if (items != 1)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	file = getfile ((HV*) SvRV (ST(0)));
	if (file == 0)
		XSRETURN_UNDEF;

	arch = file->arch;
	if (arch == 0)
		XSRETURN_UNDEF;

	off = 0;
	buff = 0;
	for (;;) {
		want = off + BLOCKSIZE;
		buff = (char *) realloc (buff, want);
		if (buff == 0)
			XSRETURN_UNDEF;

		got = tarread (arch, buff + off, BLOCKSIZE);
		if (got < 0) {
			free (buff);
			XSRETURN_UNDEF;
		}

		off += got;
		if (got == 0) {
			got = off;
			break;
		}
	}

	sv = newSVpvn ("", 0);
	sv_setpvn (sv, (const char *) buff, got);
	free (buff);

	XPUSHs (sv_2mortal (sv));
	XSRETURN (1);

void
sysseek (...)
PREINIT:
	int rc;
	HV *self;
	tararch_t *arch;
	tarfile_t *file;
	const char *key = "file";
	int off;
	int mode;
PPCODE:
	if (items != 3)
		XSRETURN_UNDEF;

	if (!(sv_isobject (ST(0)) && (SvTYPE (SvRV (ST(0))) == SVt_PVHV)))
		XSRETURN_UNDEF;

	self = (HV*) SvRV (ST(0));
	file = getfile (self);
	if (file == 0)
		XSRETURN_UNDEF;

	off = SvIV (ST(1));
	mode = SvIV (ST(2));

	if (mode == SEEK_CUR && off == 0) {
		XPUSHs (sv_2mortal (newSViv (file->pos)));
		XSRETURN (1);
	}

	if (mode == SEEK_SET && off == file->pos) {
		XPUSHs (sv_2mortal (newSViv (file->pos)));
		XSRETURN (1);
	}

	XPUSHs (sv_2mortal (newSViv (-1)));
	XSRETURN (1);

