#!/usr/bin/perl
#
# GetMagic.pl
# perl source for package SAPDB::Install::GetMagic
#
# $Header: //sapdb/V76/c_00/b_09/sys/src/install/perl/sdbrun/FileMagic.pl#1 $
# $DateTime: 2005/03/21 16:21:00 $
# $Change: 93297 $
#

package SAPDB::Install::FileMagic;

#
# AllowMagic32To64 ()
# returns constant 
#
sub AllowMagic32To64 {
	return 1;
}

#
# AllowMagic32To64 ()
# returns constant 
#
sub AllowMagic64To32 {
	return 2;
}

#
# CompareMagic ()
# 
# compares two file magics for GetMagic ()
# 
# when called in array context, it
# returns (0, undef) 
#     if these two magics are equal 
#     or 
#        if platform is equal and 
#        mode of first argument is 32 bit and 
#        mode of second argument is 64 bit and
#        the flag AllowMagic32To64 is set
#     or
#        if platform is equal and 
#        mode of first argument is 64 bit and 
#        mode of second argument is 32 bit and
#        the flag AllowMagic64To32 is set
#
# returns (-1, errortxet)
#     in any other case
#
# when called in scalar context, it returns only the return code
#
sub CompareMagic {
	my ($magic0, $magic1, $flags) = @_;

	$flags = 0 unless defined ($flags);
	my ($rc, $txt);

	if (IsValidMagic ($magic0) != 0) {
		($rc, $txt) = (-1, "invalid magic");
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	if (IsValidMagic ($magic1) != 0) {
		($rc, $txt) = (-1, "invalid magic");
		return (wantarray () ? ($rc, $txt) : $rc);
	}
	
	if ($magic0 eq $magic1) {
		($rc, $txt) = (0, undef);
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	my ($platform0, $mode0) = GetPlatformFromMagic ($magic0);
	unless (defined $platform0) {
		($rc, $txt) = (-1, "unknown magic");
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	my ($platform1, $mode1) = GetPlatformFromMagic ($magic1);
	unless (defined $platform1) {
		($rc, $txt) = (-1, "unknown magic");
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	unless (($platform0 eq $platform1) ||
	        ($platform0 =~ /aix/ && $platform1 =~ /aix/)) {
		($rc, $txt) = (-1, "different platforms");
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	if (($mode0 == $mode1) && ($platform0 eq $platform1)) {
		($rc, $txt) = (0, undef);
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	#
	# allow update from aix4 64-bit to aix5 64-bit
	#
	if (($mode0 == 64) && ($mode1 == 64) &&
	    ($platform0 eq 'aix4') && ($platform1 eq 'aix5')) {
		($rc, $txt) = (0, undef);
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	#
	# disable aix 4 on aix 5
	#
	if (($mode0 == 64) && ($mode1 == 64) &&
	    ($platform0 eq 'aix5') && ($platform1 eq 'aix4')) {
		($rc, $txt) = (-1, "different platforms");
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	if (($flags & AllowMagic32To64 ()) == AllowMagic32To64 ()) {
		if (($mode0 == 32) && ($mode1 == 64)) {
			($rc, $txt) = (0, undef);
			return (wantarray () ? ($rc, $txt) : $rc);
		}
	} elsif (($flags & AllowMagic64To32 ()) == AllowMagic64To32 ()) {
		if (($mode0 == 64) && ($mode1 == 32)) {
			($rc, $txt) = (0, undef);
			return (wantarray () ? ($rc, $txt) : $rc);
		}
	} elsif (($flags & (AllowMagic64To32 () || AllowMagic32To64 ())) ==
	         (AllowMagic64To32 () || AllowMagic32To64 ())) {
		($rc, $txt) = (0, undef);
		return (wantarray () ? ($rc, $txt) : $rc);
	}

	($rc, $txt) = (-1, "invalid object mode");
	return (wantarray () ? ($rc, $txt) : $rc);
}

#
# IsValidMagic ()
# only for internal use, called by CompareMagic ()
#
sub IsValidMagic {
	my ($magic) = @_;

	return (-1) unless (defined $magic);
	return (-1) if ($magic =~ /cannot get magic/);
	return (-1) if ($magic =~ /unknown magic/);
	return 0;
}

#
# GetPlatformMagic ()
# only for internal use, called by CompareMagic ()
#
sub GetPlatformFromMagic {
	my ($magic) = @_;

	if ($magic =~ /elf 32-bit lsb i\d86/) {
		return ("elf-i386", 32);
	} elsif ($magic =~ /elf 64-bit lsb IA64/) {
		return ("elf-ia64", 64);
	} elsif ($magic =~ /elf 32-bit msb SPARC/) {
		return ("elf-sun", 32);
	} elsif ($magic =~ /elf 64-bit msb SPARC/) {
		return ("elf-sun", 64);
	} elsif ($magic =~ /elf 32-bit msb MIPS/) {
		return ("elf-mips", 32);
	} elsif ($magic =~ /elf 64-bit msb MIPS/) {
		return ("elf-mips", 64);
	} elsif ($magic =~ /RS6000 32-bit/) {
		return ("aix4", 32);
	} elsif ($magic =~ /AIX 4 64-bit/) {
		return ("aix4", 64);
	} elsif ($magic =~ /AIX 5 64-bit/) {
		return ("aix5", 64);
	} elsif ($magic =~ /elf HP-UX ABI 64-bit msb PA-RISC/) {
		return ("hpux", 64);
	} elsif ($magic =~ /HP-UX 32-bit/) {
		return ("hpux", 32);
	} elsif ($magic =~ /HP 9000 Series/) {
		return ("hpux", 32);
	} elsif ($magic =~ /HP PA-RISC/) {
		return ("hpux", 32);
	} elsif ($magic =~ /coff ALPHA.+64-bit/) {
		return ("true64", 64);
	}
	return (undef, undef);
}

