#!/usr/bin/perl
#
# SDBRun.pl 
#

#line 7 SDBRun.pl

package SAPDB::Install::Start;

#
# first of all make the archive loader running
# this is done by overloading require
#

sub BEGIN {
	bootstrap SDB::Common;

	import SDB::Common::Getopt::Long;
	import SDB::Common::Archive;
	import SDB::Common::Require;
}

package SAPDB::Install::Start::Overwrite;

sub BEGIN {
	*CORE::GLOBAL::require = \&SDB::Common::Require::require;
}

package SAPDB::Install::Start;

bootstrap SAPDB::Install::ProcState;
bootstrap SAPDB::Install::System::Win32::Console;

#
# package Start
# parse argument line, load and run requested code
#

package SAPDB::Install::Start;

$DEBUG = 0;

#
# this function will be called from SDBRUN as entry point
#
sub Start {
	local @ARGV = @_;

	# check the environment
	return -1 unless (CheckEnvironment (@ARGV) == 0);

	# set dir if program was called using path
	$SAPDB::Install::Config{'RuntimeDir'} =~ s/\\/\//g; 

	$SAPDB::Install::Config{'RuntimeDir'} =

	# path is '.', that means cwd
	$SAPDB::Install::Config{'RuntimeDir'} eq '.' ?
	$SAPDB::Install::Config{'CallerDir'} :

	# path is './...', that means cwd/...
	$SAPDB::Install::Config{'RuntimeDir'} =~ /^\.\/(\S+)/ ?
	$SAPDB::Install::Config{'CallerDir'}.'/'.$1 :

	# path is not an abs path
	$SAPDB::Install::Config{'RuntimeDir'} !~ /^\/|^[a-z]:/i ?
	$SAPDB::Install::Config{'CallerDir'}.'/'.$SAPDB::Install::Config{'RuntimeDir'} :
	$SAPDB::Install::Config{'RuntimeDir'};

	$SAPDB::Install::Config{'RuntimeDir'} =~ s/\\/\//g; 

	# keep unused arguments in @ARGV
	SDB::Common::Getopt::Long::Configure ('pass_through');

	# dont allow auto abbrev of parameter to avoid
	# '-d DBNAME' will be interpreted as '--debug=DBNAME'
	SDB::Common::Getopt::Long::Configure ('noauto_abbrev');

	#
	# get debug options from command line
	#
	my $debug_opt;
	GetOptions ('debug:s', \$debug_opt);
	my @debug_opts = ();
	foreach (split (' ', $debug_opt)) {
		if ($_ =~ /Start|Exporter|Repository/) {
				${"SAPDB::Install::${_}::DEBUG"} = 1;
				next;
		}
		push @debug_opts, $_;
	}

	#
	# put back unknown debug options
	#
	if ($#debug_opts >= 0) {
		push @ARGV, '--debug', join (" ", @debug_opts);
	}
	
	#
	# find out if we are running from installed sdbrun
	#
	my $installed_sdbrun_used;
	GetOptions ('installed_sdbrun_used+', \$installed_sdbrun_used);

	$installed_sdbrun_used = 0 unless defined $installed_sdbrun_used;
	$SAPDB::Install::Config{'UsedSDBRUN'} =
		$installed_sdbrun_used ? 'installed' : 'kit';

	#
	# run script from filesystem if requested
	#
	my $script;
	GetOptions ('script=s', \$script);
	if (defined $script) {
		unless ($script =~ /^\// || $script =~ /^\w:/) {
			$script = 
			$SAPDB::Install::Config{'CallerDir'}.'/'.$script;
		}
		my $rc = StartScript ($script) if (defined $script);
		return ($rc);
	}
	
	# run code from @inc if requested or use defaults
	my $archive, $incpath, $mainsub, $programname;
	GetOptions ('archive=s', \$archive,
	            'inc=s', \$incpath,
	            'main=s', \$mainsub,
	            'programname:s', \$programname,
		    'silent',\$opt_silent,
		    'upuser=s',\$opt_upuser,
		    'TmpLDPath=s',\$opt_tmpldpath);

	# put temporary LD PATH to Config hash
	if(defined $opt_tmpldpath){
		$SAPDB::Install::Config{'TmpLDPath'} = $opt_tmpldpath;	
	}
	# put unprivileged user to Config hash	
	if(defined $opt_upuser){
		$SAPDB::Install::Config{'UPUser'} = $opt_upuser;	
	}
	
	
	$mainsub = "SAPDB::Install::main" unless (defined $mainsub);
	$programname = "SDBRUN" unless (defined $programname);
	my @inc = defined $incpath ? split (':', $incpath) : ();

	# put ProgramName to Config hash
	$SAPDB::Install::Config{'ProgramName'} = $programname;
	
	if($^O =~ /mswin/i and $programname =~ /SDBINST|sdbuninst/i){
		#
		# MS Windows: SDBINST.exe and sdbuninst.exe are linked as 
		# subsystem Windows now and have no console window at startup
		# automatically
		# 
		# => open a console window if In/Output isn't 
		# redirected and command line option "silent" isn't set 
		#

		if($opt_silent || SAPDB::Install::System::Win32::Console::HasRedirectedStdHandles()){
			$SAPDB::Install::Config{'SilentMode'} = 1;
			if($programname =~ /SDBINST/i){
				
				#
				# push batch option to ARGV to avoid 
				# endless reading from stdin
				#
				
				my $opt_batch = 0;
				foreach my $opt (@ARGV){
						if($opt =~ /^-{1,2}b$|^-{1,2}batch$/){
							$opt_batch = 1;
						}
				}
				push @ARGV,'-b' unless ($opt_batch); 
			}
		} else{
			#
			# open console window
			#

			if(SAPDB::Install::System::Win32::Console::OpenConsole()){
				$SAPDB::Install::Config{'HasOwnConsole'} = 1;
			}
		}
	}

	# use SDBINST.TGZ if no other path exits
	@inc = qw (SDBINST.TGZ) if ($#inc < 0);

	# get absolut filename using UNIX-style slashes
	for (my $i = 0; $i <= $#inc; $i++) {
		$inc[$i] =~ s/\\/\//g;

		next if ($inc[$i] =~ /^\/|^\w:\//);
		$inc[$i] = $SAPDB::Install::Config{'RuntimeDir'}.'/'.$inc[$i];
	}

	return (StartRequireInc ($mainsub, @inc));
}

#
# start a requested script from filesystem via do ();
#
sub StartScript {
	my ($script) = @_;
	return undef unless (defined $script);
	
	$script =~ s/\//\\/g if ($^O =~ /mswin/i);
	unless (-r $script && -f $script) {
		print ("cannot read ".$script."\n");
		print ($!."\n");
		return undef;
	}

	do ($script);
}

#
# start requested sub form @inc
#
sub StartRequireInc {
	my ($sub, @inc) = @_;
	my ($mod) = ($sub =~ /([\w:]+)::\w+$/g);

	local @INC = @inc;
	require $mod;

	if ($DEBUG){
		my $package = $sub;
		$package =~ s/::[^:]*$//;
		print "DEBUG: mainpackage = $package\n";
		my %symbols = %{"${package}::"};
		foreach my $symname (sort keys(%symbols)) {
			*mysym = $symbols{$symname};
			print "DEBUG: found symbol $symname = ".
			      "$symbols{$symname}\n";
			print "DEBUG:\$$symname is defined\n"
				if defined $mysym;
			print "DEBUG:\&$symname is defined\n"
				if defined &mysym;
			print "DEBUG:\%$symname is defined\n"
				if defined %mysym;
			print "DEBUG:\@$symname is defined\n"
				if defined @mysym;	
		}
	}

	unless (defined &{"$sub"}) {
		print "cannot find main sub ".$sub."\n";
		return -1;
	}

	my $rc = &{"$sub"}(@ARGV);
	if(defined $SAPDB::Install::Config{'TmpLDPath'} and 
	     -d $SAPDB::Install::Config{'TmpLDPath'}){
		my $tmp = $SAPDB::Install::Config{'TmpLDPath'};
		opendir(DH,$tmp) or return $rc;
		foreach my $node (grep {not /^[.]{1,2}$/} readdir(DH)){
			unlink ("$tmp/$node");
		}
		closedir(DH);
		rmdir($tmp);
	}
	return $rc;
}

sub CheckEnvironment {
	
	my $redirected_handles = 0;
	my $prog_has_console = 1;
	my $interactive_mode = 1;	
	my ($opt_prog,$opt_silent,$opt_batch);
	if($^O =~ /mswin/i){
		local @ARGV = @_;
		SDB::Common::Getopt::Long::Configure ('pass_through');
		GetOptions('programname:s',\$opt_prog,
					'silent',\$opt_silent,
					'batch',\$opt_batch);
		if($opt_prog =~ /sdbuninst|sdbinst|sdbsetup/i){
			$prog_has_console = 0;
			$redirected_handles = SAPDB::Install::System::Win32::Console::HasRedirectedStdHandles();
		}
		if($opt_batch || $opt_silent || $redirected_handles){
			$interactive_mode = 0;
		}
				
	}

	if (exists ($ENV{'SAPDB_NO_INDEPPATH'})) {
		if( $^O =~ /mswin/i and
		    not $prog_has_console and
			not SAPDB::Install::System::Win32::Console::AttachParentConsole()){
			
			if($interactive_mode){				
				SAPDB::Install::System::Win32::Console::SetErrorWithoutConsole(
					"Cannot run in MaxDB development environment.\n".
					"Please remove SAPDB_NO_INDEPPATH from environment!","$opt_prog Startup Error");
				return -1;
			}
		}
		print STDERR "cannot run in MaxDB development environment\n";
		print STDERR "Please remove SAPDB_NO_INDEPPATH from environment!\n";
		return -1;
	}

	my $ps = SAPDB::Install::ProcState->new ();
	my $pid = $$;
	my @visited = ();

	for (;;) {
		my $ppid = $ps->GetPPid ($pid);
		return 0 unless (defined $ppid);

		#
		# window reuses the pid even when the pid is still a ppid of
		# a running process, so we could run into an endless loop
		#
		foreach my $id (@visited) {
			return 0 if ($id == $pid);
		}
		push @visited, $pid;

		#
		# do not allow to be started from dbmsrv
		#
		my $cmd = $ps->GetCmd ($pid);
		if ($cmd =~ /^dbmsrv/i) {
			$ps->Terminate ($pid);
			print "an unexpected error occured\n";
			return -1;
		}

		#
		# some systems define ppid = 0 for pid = 0
		# that means the parent of 'idle' is 'idle' itself,
		# this can cause an endless loop when reading the process table
		#
		return 0 if ($pid == 0);

		$pid = $ppid;
	}
	
	return 0;
}
