/*
 * untgz.c
 * extraction tool used in make of sdbsetup
 */

#ifdef _WIN32
#define WIN32
#else
#define UNIX UNIX
#endif

#ifdef UNIX
#include <errno.h>
#include <fcntl.h>
#include <grp.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <utime.h>
#include <unistd.h>
#endif

#ifdef WIN32
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <time.h>
#include <sys/utime.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>

#include "zlib.h"

#ifdef WIN32
typedef struct _stat stat_t;
typedef struct _utimbuf utimbuf_t;
#define sys_utime _utime
#define stat(a,b) _stat(a, b)
#endif

#ifdef UNIX
typedef struct stat stat_t;
typedef struct utimbuf utimbuf_t;
#define sys_utime utime
#endif

#ifdef _WIN64
typedef __int64 ptr_t;
#else
typedef long ptr_t;
#endif

/*
 * alloca specific defines
 */
#ifdef _AIX
#define NEED_PRAGMA_ALLOCA NEED_PRAGMA_ALLOCA
#endif
#if defined UNIX && !defined _AIX && !defined SNI
#define NEED_ALLOCA_H NEED_ALLOCA_H
#endif

#ifdef NEED_ALLOCA_H
#include <alloca.h>
#endif

#ifdef NEED_PRAGMA_ALLOCA
#pragma alloca
#endif

#ifdef WIN32

#ifndef alloca
#define alloca(size) _alloca(size)
#endif

#ifndef snprintf
#define snprintf _snprintf
#endif

#ifndef chown
#define chown(file,uid,gid)
#endif

#endif

/* 
 * tar typeflag types
 */
#define TAR_AREGTYPE '\0'
#define TAR_REGTYPE  '0'
#define TAR_LNKTYPE  '1'
#define TAR_SYMTYPE  '2'
#define TAR_CHRTYPE  '3'
#define TAR_BLKTYPE  '4'
#define TAR_DIRTYPE  '5'
#define TAR_FIFOTYPE '6'

/*
 * gnu tar typeflags
 */
#define TAR_GNUTYPE      'A'
#define TAR_LONGNAMETYPE 'L'

/*
 * tar magic
 */
static char TAR_MAGIC[5] = {'u', 's', 't', 'a', 'r'};
static char NO_MAGIC[5] = {0, 0, 0, 0, 0};

/*
 * this is known as the tar header
 */
typedef struct {
	char name[100];
	char mode[8];
	char uid[8];
	char gid[8];
	char size[12];
	char mtime[12];
	char chksum[8];
	char typeflag[1];	
	char linkname[100];
	char magic[6];
	char version[2];
	char uname[32];
	char gname[32];
	char devmajor[8];
	char devminor[8];
	char prefix[155];
} tarhdr_t;

typedef struct {
	gzFile fptgz;
	tarhdr_t hdr[1];
	FILE *fp;
	int size;
	char *name;
	char *longname;
	char *archname;
#ifdef UNIX
	uid_t uid;
	gid_t gid;
#endif
} tarobj_t;

/*
 * a lot of prototypes
 */
static int tar (gzFile, tarobj_t *);
static int obj_init (tarobj_t *);
static int obj_sethdr (tarobj_t *, tarhdr_t *);
static int obj_clearhdr (tarobj_t *);
static int obj_begin (tarobj_t *);
static int obj_final (tarobj_t *);
static int obj_block (tarobj_t *, char *, int);
static int obj_free (tarobj_t *);
static int mkdir_recursiv (tarobj_t *, char *, int);
static FILE *fopen_recursiv (tarobj_t *, char *, int, int *);
static int strnocttoi (char *, int);
static int sprintf_hdrmode (char *, tarhdr_t *);
static int sprintf_hdrmtime (char *, tarhdr_t *);
static int sprintf_hdrstat (char *, tarhdr_t *, char *);

/*
 * this BLOCKSIZE read from an file must be 512 bytes
 * because the tar header and gzip are using 512 byte
 * blocks
 */
#define BLOCKSIZE 512

/*
 * these obj_...() functions are handling any action on
 * opened tar archives, every open archive is using a
 * session structure called an 'obj'
 */

/*
 * initialize a new session structure
 */
static int
obj_init (tarobj_t *obj) {
	memset (obj, 0, sizeof (tarobj_t));
	return 0;
}

/*
 * save tar header of present file form archive to session structure
 */
static int
obj_sethdr (tarobj_t *obj, tarhdr_t *hdr) {
	memcpy (obj->hdr, hdr, sizeof (tarhdr_t));
	return 0;
}

static int
obj_clearhdr (tarobj_t *obj) {
	memset (obj->hdr, 0, sizeof (tarhdr_t));
	return 0;
}

/*
 * handle first action when finding next file in archive
 */
static int
obj_begin (tarobj_t *obj) {
	int dirmode;
	int err;
	int rc;
	char *name;

	name = (obj->longname != 0) ? obj->longname : obj->hdr->name;

	obj->name = (char *) malloc (strlen (name) + 1);
	if (obj->name == 0)
		return -1;

	strcpy (obj->name, name);

	switch (*obj->hdr->typeflag) {
	case TAR_AREGTYPE:
	case TAR_REGTYPE:
		dirmode = strnocttoi (obj->hdr->mode, sizeof (obj->hdr->mode));
		dirmode &= 0777;
		if (dirmode & 0600)
			dirmode |= 0100;
		if (dirmode &  060)
			dirmode |=  010;
		if (dirmode &   06)
			dirmode |=   01;

		obj->fp = fopen_recursiv (obj, obj->name, dirmode, &err);
		if (err !=  0)
			return -1;

		break;
	case TAR_DIRTYPE:
		dirmode = strnocttoi (obj->hdr->mode, sizeof (obj->hdr->mode));
		rc = mkdir_recursiv (obj, obj->name, dirmode);
		if (rc < 0)
			return -1;

#ifdef UNIX
		rc = chmod (obj->name, dirmode);
		if (rc != 0)
			return -1;
#endif

		break;
	}
	return 0;
}

/*
 * handle last action before going to next file entry of archive
 */
static int
obj_final (tarobj_t* obj) {
	utimbuf_t buff[1];
	int mode;

	switch (*obj->hdr->typeflag) {
	case TAR_AREGTYPE:
	case TAR_REGTYPE:
		fclose (obj->fp);
		obj->fp = 0;
		buff->modtime = buff->actime = 
		strnocttoi (obj->hdr->mtime, sizeof (obj->hdr->mtime));
		sys_utime (obj->name, buff);
		chown (obj->name, obj->uid, obj->gid);

		mode = 
		strnocttoi (obj->hdr->mode, sizeof (obj->hdr->mode));
		mode &= 06777;
		chmod (obj->name, mode);
		break;
	default:
		chown (obj->name, obj->uid, obj->gid);
		break;
	}


	if (obj->longname != 0) {
		free (obj->longname);
		obj->longname = 0;
	}

	if (obj->name != 0) {
		free (obj->name);
		obj->name = 0;
	}

	return 0;
}


/*
 * handle action when reading data between two headers from archive 
 */
static int
obj_block (tarobj_t *obj, char *ptr, int len) {
	int rc;

	rc = fwrite (ptr, 1, len, obj->fp);
	return (rc);
}

/*
 * helper functions used by tar extractor
 */

/*
 * helper funtion for creating directories
 */
static int
create_subdir (tarobj_t *obj, char *name, int mode) {
	stat_t statbuf;

	if (strlen (name) <= 0)
		return 0;
	
	/* check if directory already exists */
	if (stat (name, &statbuf) == 0) {
		if (statbuf.st_mode & S_IFDIR) {
			return 0;
		}
	}
		
	/* create it */
	if (mkdir (name, mode) != 0) {
		return -1;
	}

	return 0;
}

static int
mkdir_recursiv (tarobj_t *obj, char *name, int mode) {
	char *buff;
	char *ptr;
	char save;
	int len;

	len = strlen (name);
	if (len <= 0)
		return -1;

	buff = (char *) alloca (len + 1);
	strcpy (buff, name);

	/* create working buffer and remove trailing slashes */
	for (;;) {
		len = strlen (buff);
#ifdef UNIX
		/*
		 * we do not need to create root directory
		 */
		if (len == 1 && *buff == '/')
			return 0;
#endif
		if (len <= 0)
			return -1;

		switch (buff[len - 1]) {
		case '/':
		case '\\':
			buff[len - 1] = '\0';
			continue;
		}
		break;
	}
	
	/* try to create any subdirectory */
	for (ptr = buff;; ptr++) {
		if (*ptr != '/' && *ptr != '\\' && *ptr != '\0')
			continue;

		save = *ptr;
		*ptr = 0;
		if (create_subdir (obj, buff, mode))
			return (-1);

		*ptr = save;

		if (save == '\0')
			return 0;
	}
	return 0;
}

/*
 * helper funtion for creating regular files
 */
static FILE *
fopen_recursiv (
tarobj_t *obj, char *name, int dirmode, int *err) {
	char *buff;
	char *ptr;
	FILE *fp;
	utimbuf_t utime_buff[1];
	int file_exitsts;

	buff = (char *) alloca (strlen (name) + 1);
	strcpy (buff, name);

	ptr = strrchr (buff, '/');
	if (ptr == 0)
		ptr = strrchr (name, '\\');

	if (ptr != 0) {
		*ptr = '\0';
		if (mkdir_recursiv (obj, buff, dirmode) != 0) {
			*err = -1;
			return 0;
		}
	}

	/* real open for write */
	fp = fopen (name, "w+b");
	if (fp == 0) {
		/* may file file is in use or no write permission */
		unlink (name);
		fp = fopen (name, "w+b");
	}

	*err = (fp == 0) ? -1 : 0;
	return (fp);
}

/*
 * helper funtion for converting octal numbers
 * from tar header to integer numbers
 */
static int
strnocttoi (char *ptr, int count) {
	int val;
	char c;
  
	for (val = 0; count > 0; count--) {
		c = *ptr++;
		if (c == ' ')
			continue;

		if (c == '\0')
			break;

		val = (val << 3)  + (c - '0');
	}
  	return val;
}

/*
 * helper funtions for makeing tar header human readable
 */

static int
get_hdrmode (tarhdr_t *hdr) {
	int mode;
	mode = strnocttoi (hdr->mode, sizeof (hdr->mode));
	
	return (mode);
}

static time_t
get_hdrtime (tarhdr_t *hdr) {
	time_t mtime;
	mtime = strnocttoi (hdr->mtime, sizeof (hdr->mtime));

	return (mtime);
}

static int
get_hdrsize (tarhdr_t *hdr) {
	int size;

	size = strnocttoi (hdr->size, sizeof (hdr->size));
	return (size);
}

static int
sprintf_hdrmode (char *ptr, tarhdr_t *hdr) {
	const char *all = "-rwxrwxrwx";
	int mode, type, mask;

	mode = strnocttoi (hdr->mode, sizeof (hdr->mode));
	type =  hdr->typeflag[0];

	strcpy (ptr, all);	
	mode &= 07777;
	switch (type) {
	case TAR_SYMTYPE:
		*ptr = 'l';
		break;
	case TAR_CHRTYPE:
		*ptr = 'c';
		break;
	case TAR_BLKTYPE:
		*ptr = 'b';
		break;
	case TAR_DIRTYPE: 
		*ptr = 'd';
		break;
	}

	ptr++;
	for (mask = 0400; mask != 0; mask >>= 1) {
		if (!(mode & mask))
			*ptr = '-';

		if
		((mask == 0100) && (mode & 04000) ||
		 (mask ==  010) && (mode & 02000))
			*ptr = 's';

		ptr++;
	}

	*ptr = '\0';
	return 0;
}

static int
sprintf_hdrmtime (char *ptr, tarhdr_t *hdr) {
	time_t mtime;
	struct tm *p_tm;

	mtime = get_hdrtime (hdr);
	p_tm = localtime (&mtime);
	sprintf (ptr, "%4d-%02d-%02d %02d:%02d:%02d",
	p_tm->tm_year + 1900, p_tm->tm_mon + 1, p_tm->tm_mday,
	p_tm->tm_hour, p_tm->tm_min, p_tm->tm_sec);

	return 0;
}

static int
sprintf_hdrsize (char *ptr, tarhdr_t *hdr) {
	sprintf (ptr, "%12d", get_hdrsize (hdr));
	return 0;
}

static int
sprintf_hdrname (char *text, tarhdr_t *hdr) {
	char *ptr;
	sprintf (text, "%s", hdr->name);
	
	/* remove trailing spaces */
	ptr = text + strlen (text);
	while (ptr > text && *ptr == ' ')
		ptr--;

	ptr++;
	*ptr = '\0';
	return 0;
}

static int
sprintf_hdrstat (char *ptr, tarhdr_t *hdr, char *longname) {
	sprintf_hdrmode (ptr, hdr);
	strcat (ptr, " ");
	sprintf_hdrsize (ptr + strlen (ptr), hdr);
	strcat (ptr, " ");
	sprintf_hdrmtime (ptr + strlen (ptr), hdr);
	strcat (ptr, " ");

	if (longname == 0)
		sprintf_hdrname (ptr + strlen (ptr), hdr);
	else
		strcat (ptr + strlen (ptr), longname);

	return 0;
}

/*
 * open tgz archive
 */
static int
tar_open (tarobj_t *obj, char *filename) {
	int rc, i;
	stat_t statbuf;

	/* avoid reuse of opened object */
	if (obj->fptgz != 0)
		return -1;

	memset (obj, 0, sizeof (tarobj_t));

	rc = stat (filename, &statbuf);
	if (rc != 0)
		return -1;

	obj->fptgz = gzopen (filename, "rb");
	if (obj->fptgz == 0)
		return -1;

	obj->archname = (char *) malloc (strlen (filename) + 1);
	strcpy (obj->archname, filename);
	return 0;
}

/*
 * close archive and free session structure
 */
static int
tar_close (tarobj_t *obj) {
	if (obj->archname != 0)
		free (obj->archname);

	if (obj->name != 0)
		free (obj->name);

	if (obj->longname != 0)
		free (obj->longname);

	if (obj->fptgz != 0)
		gzclose (obj->fptgz);

	return 0;
}

#ifdef UNIX
static int
tar_setowner (tarobj_t *obj, uid_t uid, gid_t gid) {
	if (obj == 0)
		return -1;

	obj->uid = uid;
	obj->gid = gid;

	return 0;
}
#endif

/*
 * move read position to next tar header in archive
 */
static int
tar_next (tarobj_t* obj) {
	char buff[BLOCKSIZE];
	tarhdr_t *hdr;
	int got;
	int rc;

	/* default return values */
	rc = -1;

	for (;;) {
		/* nothing to read */
		if (obj->fptgz == 0)
			rc = -1;

		got = gzread (obj->fptgz, &buff, BLOCKSIZE);
		if (got < 0) {
			int errcode;
			char *errmsg;

			errmsg = (char *) gzerror (obj->fptgz, &errcode);
			if (errcode == Z_STREAM_ERROR) {
				if (obj->fp != 0) {
					fclose (obj->fp);
					obj->fp = 0;
				}

				if (obj->name != 0) {
					free (obj->name);
					obj->name = 0;
				} 

				obj->size = 0;
				obj_clearhdr (obj);
				rc = 0;
				break;
			}

			gzclose (obj->fptgz);
			obj->fptgz = 0;
			rc = 0;
			break;
		}

		if (got != BLOCKSIZE)
			break;

		if (obj->size == 0) {
			hdr = (tarhdr_t *) buff;
		
			if ((got == 0) || (hdr->name[0] == 0)) { 
				/* end of file */

				if (obj->fp != 0) {
					fclose (obj->fp);
					obj->fp = 0;
				}

				if (obj->name != 0) {
					free (obj->name);
					obj->name = 0;
				} 

				obj_clearhdr (obj);
				rc = 0;
				break;
			}

			/*
			 * for some reason WxPerl archive has no magic ?!
			 * this is a boolean expression, you are not expected
			 * to understand this
			 */
			if (
			memcmp (hdr->magic, TAR_MAGIC, sizeof (TAR_MAGIC)) &&
			memcmp (hdr->magic, NO_MAGIC, sizeof (NO_MAGIC))
			) {
				int i;

				gzclose (obj->fptgz);
				obj->fptgz = 0;
				rc = -1;
				break;
			}

			obj_sethdr (obj, hdr);
			obj->size = strnocttoi (hdr->size, sizeof (hdr->size));

			if (hdr->typeflag[0] == TAR_LONGNAMETYPE) {
				if (obj->longname != 0)
					free (obj->longname);

				obj->longname = malloc (obj->size + 1);
				if (obj->longname == 0) {
					rc = -1;
					break;
				}

				obj->longname[0] = 0;
			}

			if (hdr->typeflag[0] >= TAR_GNUTYPE)
				continue;

			rc = 0;
			break;

		} else if (obj->size > BLOCKSIZE) {
			if (obj->hdr->typeflag[0] == TAR_LONGNAMETYPE) {
				int offset;

				offset = strlen (obj->longname);

				memcpy (obj->longname, buff, BLOCKSIZE);
				obj->longname[offset + BLOCKSIZE] = '\0';
			}

			obj->size -= BLOCKSIZE;
		} else {
			if (obj->hdr->typeflag[0] == TAR_LONGNAMETYPE) {
				int offset;

				offset = strlen (obj->longname);

				memcpy (obj->longname, buff, obj->size);
				obj->longname[offset + obj->size] = '\0';
			} else {
				if (obj->longname != 0) {
					free (obj->longname);
					obj->longname = 0;
				}
			}

			obj->size = 0;
		}
	}

	return (rc);
}

/*
 * extract present file from archive to filesystem
 */
static int
tar_extract (tarobj_t* obj) {
	char buff[BLOCKSIZE];
	tarhdr_t *hdr;
	int got;
	int rc;
	
	/* 
	 * is archive open
	 */
	if (obj->fptgz == 0)
		return -1;

	rc = obj_begin (obj);
	if (rc < 0) {
		fprintf (stderr, "cannot open or create %s\n", obj->hdr->name);
		return -1;
	}

	if (obj->size == 0) {
		/* diretory or empty file */
		obj_final (obj);
		return 0;
	}

	rc = -1;
	for (;;) {
		int retval;
		got = gzread (obj->fptgz, &buff, BLOCKSIZE);
		if (got != BLOCKSIZE) {
			if (rc != 0)
				fprintf (stderr, "cannot read from archive\n");

			break;
		}

		if (obj->size > BLOCKSIZE) {
			retval = obj_block (obj, buff, BLOCKSIZE);
			if (retval < 0) {
				fprintf (stderr, "cannot write to file\n");
				return -1;
			}

			obj->size -= BLOCKSIZE;
		} else {
			retval = obj_block (obj, buff, obj->size);
			if (retval < 0) {
				fprintf (stderr, "cannot write to file\n");
				return -1;
			}

			obj_final (obj);
			obj->size = 0;
			rc = 0;
			break;
		}
	}

	return (rc);
}

static int
extract (char *archivename) {
	int rc;
	tarobj_t obj[1];

	memset (obj, 0, sizeof (tarobj_t));
	rc = tar_open (obj, archivename);
	if (rc != 0) {
		fprintf (stderr, "cannot open archive %s\n", archivename);
		return 1;
	}

#ifdef UNIX
	tar_setowner (obj, getuid (), getgid ());
#endif

	for (;;) {
		char buff[1024];

		rc = tar_next (obj);
		if (rc != 0)
			break;

		/* end of archive */
		if ((obj->hdr->name[0] == '\0'))
			break;

		sprintf_hdrstat (buff, obj->hdr, 0);
		printf ("%s\n", buff);

		rc = tar_extract (obj);
		if (rc != 0) {
			fprintf (stderr, "cannot extract %s\n", obj->hdr->name);
			break;
		}
	}

	tar_close (obj);
	return (rc == 0 ? 0 : 1);
}

int
main (int argc, char **argv) {
	char *archivename;

	if (argc != 2) {
		fprintf (stderr, "usage: untgz archiv");
		return -1;
	}

	archivename = argv[1];
	if (archivename == 0 || strlen (archivename) == 0) {
		fprintf (stderr, "usage: untgz archiv");
		return -1;
	}

	return (extract (archivename));
}
